import traceback
import json
from functools import wraps
from time import time
from datetime import datetime, timedelta
from urllib.parse import urlencode, parse_qsl
from itertools import groupby

import falcon
from inv5.logger import log
from inv5.responses import responses, errors
from inv5.resources.resources import swagger, ResourcesBaseClass, DEFAULT_SECURITY, ATTRIBUTE_ERROR_RE
from marshmallow import ValidationError
from pytz import UTC
from sqlalchemy.exc import InternalError, IntegrityError, DataError
from sqlalchemy import cast, case, literal, inspect, and_, or_, alias, func, cast, Integer, DateTime, select, text
from sqlalchemy.orm import aliased

from inv5 import model
from inv5.resources.resource import OperationErrorResponse, List, Create, Resource, CustomGetAction

from inv5.report_scheme import CdrReportGroppedRequestScheme, CdrReportDetailInnerScheme, \
    CdrReportGroppedResponseScheme, DidReportGroppedRequestScheme, DidReportInnerScheme, \
    DidReportGroppedResponseScheme

REPORT_DEBUG = False


# _model_class = model.CdrReportDetail
# _columns = _model_class.get_fields()
# _groupable = [fn for fn in _columns if
#                  getattr(_model_class, fn).type.__class__.__name__ in ('String', 'DateTime') or '_id' in fn or fn in (
#                      'orig_jur_type', 'release_cause', 'term_jur_type')]
# _aggregable = [fn for fn in _columns if fn not in _groupable]

class SuccessReportResponse(responses.BaseResponse):
    description = 'Successfully executed'
    scheme = CdrReportGroppedResponseScheme

    def get_response_body(self):
        return self.data


_grp = ('orig_jur_type', 'release_cause', 'term_jur_type', 'dst_lata', 'dst_ocn', 'src_lata', 'dst_ocn', 'lrn_lata',
        'lrn_ocn',)


class Report(List):
    entity_plural = 'Report items'
    scheme_class = CdrReportGroppedRequestScheme
    scheme_class_get = CdrReportGroppedResponseScheme
    scheme_class_get_inner = CdrReportDetailInnerScheme
    model_class = model.CdrReportDetail
    columns = model_class.get_fields()

    groupable = [fn for fn in columns if
                 getattr(model.CdrReportDetail, fn).type.__class__.__name__ in (
                     'String', 'DateTime', 'Text') or '_id' in fn or fn in _grp]
    aggregable = list(set([fn for fn in columns if
                           getattr(model.CdrReportDetail, fn).type.__class__.__name__ not in (
                               'String', 'DateTime', 'Text') and '_id' not in fn and fn not in _grp] +
                          ['nrf_calls', 'asr', 'qsr', 'ingress_npr', 'egress_npr']))

    def get_spec(self):
        sfields_names, search_fields_list = self.get_fields_from_scheme(self.scheme_class)
        qfields_names, query_fields_list = self.get_query_fields_from_scheme(self.scheme_class)
        fields_names = [f.name for f in inspect(self.model_class).columns]
        spec = swagger.specify.get_spec(
            method='get', description='Gets {}'.format(self.entity_plural.lower()),
            path_parameters=self.path_parameters,
            query_parameters=search_fields_list + [
                {"name": "from", "description": "printing range. Default: print from start 0 position",
                 "type": "integer"},
                {"name": "count", "description": "printing range. Default: print first 1000 items",
                 "type": "integer"},
                {"name": "fields",
                 "description": "aggregated fields, available values:{}".format(self.aggregable)},
                {"name": "group",
                 "description": "groupping fields, available values:{}".format(self.groupable)},

            ] +
                             [dict(name=i, type='string') for i in self.groupable],
            responses=(
                          SuccessReportResponse(),
                          responses.AttributeNotExitsErrorResponse()
                      ) + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )
        spec['get']['produces'] = ['application/json', 'text/csv', 'text/xml']
        return spec

    def run(self, **kwargs):
        class Req:
            def __init__(self, req_user):
                self.query_string = None
                self.context = {'user': req_user}

            pass

        if 'user' in kwargs:
            user = kwargs['user']
            del kwargs['user']
        else:
            user = model.User.get(1)
        req = Req(user)
        req.query_string = urlencode(kwargs)
        self.req = req
        objects_list, total, page, per_page, fields = self.get_objects_list(req, self.model_class, **kwargs)
        # items = [i for i in objects_list]
        items = self.scheme_class_get_inner().dump(objects_list, many=True).data
        return items

    def send_request(self, params):
        m = dict(calls='ingress_total_calls',
                 ingress_billed_time='ingress_bill_time',
                 ingress_cost='ingress_call_cost',
                 ingress_billed_min='ingress_bill_time',
                 code='ingress_code',
                 non_zero_calls='not_zero_calls',
                 egress_cost='egress_total_cost',
                 egress_billed_time='egress_bill_time',
                 )
        args = {k: v for k, v in params.items() if k not in m}
        upd = {m[k]: v for k, v in params.items() if k in m}
        args.update(upd)

        class Responce:
            def json(self):
                return self.data

        items = self.run(**args)
        resp = Responce()
        resp.data = dict(data=items)
        return resp

    def _on_get(self, req, resp, **kwargs):

        from .helpers import check_permission

        if (not hasattr(self, 'no_auth_needed') or self.no_auth_needed == False) and not check_permission(self, req,
                                                                                                          'list'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return
        try:
            import time
            _start = time.time()

            objects_list, total, _from, count, fields = self.get_objects_list(req, self.model_class, **kwargs)
            # items = [i for i in objects_list]
            items = self.scheme_class_get_inner().dump(objects_list, many=True).data
            data = {
                'success': True,
                'data': items,
                'total': total, 'from': _from, 'count': len(items)
            }
            self.set_response(resp, SuccessReportResponse(data=data, scheme=self.scheme_class_get))

            _end = time.time()
            log.debug('report execution time {}'.format(_end - _start))
            # resp.body = json.dumps(data)
            # resp.status = falcon.HTTP_200

        # except AttributeError as e:
        #     self.set_response(
        #         resp, responses.AttributeNotExitsErrorResponse(
        #             data=ATTRIBUTE_ERROR_RE.search(str(e)).group(1)
        #         )
        #     )
        except (InternalError, IntegrityError, DataError) as e:
            import traceback
            model.get_db().session.rollback()
            log.error(traceback.format_exc())
            self.set_response(resp, OperationErrorResponse(e))
        except Exception as e:
            import traceback
            log.error(traceback.format_exc())
            self.set_response(resp, OperationErrorResponse(e))

    def time_field(self, cdr):
        return cdr.c.report_time

    def get_daily_cdr(self, cdr_date):
        cdr = model.cdr_report_detail(cdr_date)
        return cdr

    def get_fields_for_list(self, parsed_qs, **kwargs):
        fields = parsed_qs.get('fields', None)
        if fields:
            fields = fields.split(',')
            for f in fields:
                # if f not in self.get_all_fields_from_scheme(self.scheme_class_get)[0]:
                if f not in self.aggregable:
                    raise Exception('fields {} not valid!'.format(f))
        return fields

    def get_objects_list(self, req, model_class, **kwargs):
        parsed_qs = dict(parse_qsl(req.query_string))
        if REPORT_DEBUG:
            log.debug('report parsed parameters:{}'.format(parsed_qs))

        try:
            _from = int(parsed_qs.get('from'))
        except (ValueError, TypeError):
            _from = 0

        try:
            count = int(parsed_qs.get('count', 1000))
        except ValueError:
            count = 1000

        ordering = {}

        fields = self.get_fields_for_list(parsed_qs, **kwargs)

        filtering = self.get_filtering_for_list(parsed_qs, **kwargs)
        if fields:
            filtering['fields'] = ','.join(fields)

        filtering, query = \
            self.modify_query_from_filtering_for_list(filtering, **kwargs)

        if ordering:
            ordering, query = \
                self.modify_query_from_ordering_for_list(ordering, query, **kwargs)
        if query is None:
            return [], 0, _from, count, fields
        return model_class.get_objects_list(
            query=query,
            filtering=filtering,
            ordering=ordering,
            paging={'from': _from, 'till': _from + count}
        ) + (_from, count, fields)

    def on_filter_ingress_id(self, v, cdr):
        # return text("ingress_id in ({})".format(v))
        return cdr.c.ingress_id.in_([int(i) for i in v.split(',')])

    def on_filter_egress_id(self, v, cdr):
        # return text("egress_id in ({})".format(v))
        return cdr.c.egress_id.in_([int(i) for i in v.split(',')])

    def on_filter_non_zero_calls(self, v, cdr):
        # return text("egress_id in ({})".format(v))
        if v == '1':
            return cdr.c.ingress_not_zero_calls > 0

    def on_filter_scc_below_sec(self, v, cdr):
        return cdr.c.duration < v

    def on_filter_is_hide_unauthorized(self, v, cdr):
        if v is None:
            return None
        if v:
            return cdr.c.release_cause != 3
        else:
            return cdr.c.release_cause == 3

    def on_select_nrf_calls(self, cdr):
        return cdr.c.lrn_calls

    def on_select_asr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.not_zero_calls) / (1.0 * cdr.c.ingress_total_calls))

    def on_select_qsr(self, cdr):
        return 100.00 * ((cdr.c.not_zero_calls + cdr.c.ingress_busy_calls + cdr.c.ingress_cancel_calls) /
                         cdr.c.ingress_total_calls)

    def on_select_ingress_npr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.npr_count) / (1.0 * cdr.c.ingress_total_calls))

    def on_select_egress_npr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.npr_count) / (1.0 * cdr.c.egress_total_calls))

    def modify_query_from_filtering_for_list(self, filtering, **kwargs):

        def similar_fields(field, field_list):
            sf = field.split('_')
            sim = []
            for f in sf:
                for ef in field_list:
                    if f in ef:
                        sim.append(ef)
            if sim:
                return 'similar fields:{}'.format(set(sim))
            return ''

        import sys
        sys.setrecursionlimit(10000)
        cls = self.model_class
        columns = cls.get_fields()
        # columns = [f.name for f in inspect(self.model_class).columns]
        req_query = dict(parse_qsl(self.req.query_string))
        _from = filtering.pop('from', None)
        count = filtering.pop('count', None)

        sort = filtering.pop('sort', None)
        order = filtering.pop('order', '')

        sort_order = None
        if sort:
            _sort = sort.split(',')
            for s in _sort:
                if s not in columns:
                    raise ValidationError('sort field missing {}'.format(s))
            _order = order.split(',')
            for o in _order:
                if o not in ('', 'asc', 'desc'):
                    raise ValidationError('sort order invalid value {}, must be asc or desc (can be empty)'.format(o))
            if len(_sort) != len(_order):
                if sort != '':
                    raise ValidationError('sort length must be equal order length')
                else:
                    _order = ['' for s in _sort]
            sort_order = list(zip(_sort, _order))

        start_time = filtering.pop('start_time', None)
        end_time = filtering.pop('end_time', str(int(datetime.now(UTC).timestamp())))
        try:
            start_time = int(start_time)
            if start_time < 0:
                start_time = int(datetime.now(UTC).timestamp()) + start_time
            start_date_time = datetime.utcfromtimestamp(start_time).replace(tzinfo=UTC)
        except (ValueError, TypeError):
            raise ValidationError('wrong start time {}'.format(start_time))
        try:
            end_time = int(end_time)
            end_date_time = datetime.utcfromtimestamp(end_time).replace(tzinfo=UTC)
        except (ValueError, TypeError):
            raise ValidationError('wrong end time')
        if start_time >= end_time:
            raise ValidationError('start time greater than end_time!')
        groupable = self.groupable
        aggregable = self.aggregable
        tz = str(filtering.pop('tz', '0'))
        if REPORT_DEBUG:
            log.debug('tz={}'.format(tz))
        non_zero_calls = filtering.pop('non_zero_calls', None)
        if REPORT_DEBUG:
            log.debug('non_zero_calls={}'.format(non_zero_calls))
        human_readable = filtering.pop('human_readable', None)
        if REPORT_DEBUG:
            log.debug('human_readable={}'.format(human_readable))

        method = filtering.pop('method', 'total')
        step = filtering.pop('step', 'all')
        groups = filtering.pop('group', None)
        fields = filtering.pop('fields', None)
        field = filtering.pop('field', None)
        fields = fields or field
        if fields:
            fields = fields.split(',')
        if not fields:
            fields = aggregable.copy()
        for fk in filtering.keys():
            if fk not in columns:
                if not hasattr(self, 'on_filter_{}'.format(fk)):
                    raise ValidationError('{}:wrong filter field. {}'.format(fk, similar_fields(fk, columns)))
            if fk not in fields and not hasattr(self, 'on_filter_{}'.format(fk)):
                fields.append(fk)

        if groups is not None:
            groups = groups.split(',')
        else:
            groups = []
        for fk in groups:
            if fk not in groupable:
                raise ValidationError('{}:wrong group field. {}'.format(fk, similar_fields(fk, groupable)))
            if fk not in fields:
                fields.append(fk)

        fields.append('report_time')
        fields = list(set(fields))

        user = self.get_user(self.req)

        if user and user.user_type in ('client', 'agent', 'vendor'):
            if 'ingress_id' in filtering and filtering['ingress_id']:
                ids = set(filtering['ingress_id'].split(','))
                alw = set([str(i) for i in user.allowed_ingress_id])
                filtering['ingress_id'] = ','.join(list(ids.intersection(alw)))
            else:
                filtering['ingress_id'] = ','.join([str(i) for i in user.allowed_ingress_id])
            if filtering['ingress_id'] == '':
                if user.allowed_egress_id:
                    alw = set([str(i) for i in user.allowed_egress_id])
                    if 'egress_id' in filtering and filtering['egress_id']:
                        ids = set(filtering['egress_id'].split(','))
                        filtering['egress_id'] = ','.join(list(ids.intersection(alw)))
                    else:
                        filtering['egress_id'] = ','.join(list(alw))
                    del filtering['ingress_id']
                else:
                    filtering['ingress_id'] = '-1'
            if REPORT_DEBUG:
                log.debug('result {} filtering={}'.format(user.user_type, filtering))

        start_day = start_date_time.date()
        database_start_day = model.cdr_reports_start_date()
        if database_start_day is None:
            raise ValidationError('database has no cdr report data!')
        if database_start_day > start_day:
            start_day = database_start_day
        end_day = end_date_time.date()
        if end_day < start_day:
            end_day = start_day

        days = (end_day - start_day).days
        union_query = None
        union_list = []
        for day_idx in range(days + 1):

            day_cdr = start_day + timedelta(days=day_idx)
            if hasattr(self, 'ya_query'):
                q = self.ya_query(fields, day_cdr, day_cdr, filtering)
                if q:
                    union_list.append(q)
            cdr_date = "%s%02d%02d" % (day_cdr.year, day_cdr.month, day_cdr.day)
            cdr = self.get_daily_cdr(cdr_date)
            if cdr is None:
                log.warning('missing table {}'.format(cdr_date))
                continue
            _select = []
            for k in fields:
                keys = [x.key for x in _select]
                if k in keys:  # already here
                    continue
                # k=k.replace('non_zero_calls','not_zero_calls')
                if k in ('asr', 'qsr', 'ingress_npr', 'egress_npr'):
                    keys = [x.key for x in _select]
                    for df in ['ingress_busy_calls', 'ingress_cancel_calls', 'not_zero_calls',
                               'ingress_total_calls', 'egress_total_calls', 'npr_count']:
                        if not df in keys:
                            _select.append(getattr(cdr.c, df).label(df))
                    continue
                if hasattr(self, 'on_select_' + k):
                    _f = getattr(self, 'on_select_' + k)(cdr)
                    if _f is not None:
                        _select.append(_f.label(k))
                    continue
                elif hasattr(cdr.c, k):
                    _select.append(getattr(cdr.c, k).label(k))
                else:
                    raise ValidationError('{}:no such field to select.{}'.format(k, similar_fields(k, columns)))
                    # _select.append(getattr(cdr.c, k))
            # _select = tuple(_select)
            _filter = []
            if day_idx == 0:
                _filter = [self.time_field(cdr) >= start_date_time]
                if day_idx == days:
                    _filter = [self.time_field(cdr) >= start_date_time, self.time_field(cdr) <= end_date_time]
            else:
                if day_idx == days:
                    _filter = [self.time_field(cdr) <= end_date_time]
            for k, v in filtering.items():
                if hasattr(self, 'on_filter_' + k):
                    _f = getattr(self, 'on_filter_' + k)(v, cdr)
                    if _f is not None:
                        _filter.append(_f)
                    continue
                if not k in columns:
                    raise ValidationError('{}:no such field to filter.{}'.format(k, similar_fields(k, columns)))
                if hasattr(cdr.c, k):
                    if '*' in v:
                        _filter.append(getattr(cdr.c, k).like(v.replace('*', '%')))
                    else:
                        _filter.append(getattr(cdr.c, k) == v)

            q = model.get_db().session.query(*_select)
            if _filter:
                q = q.filter(and_(*_filter))
            union_list.append(q)
        # if hasattr(self,'ya_query'):
        #     q=self.ya_query(fields,start_day,end_day,filtering)
        #     if q:
        #         union_list.append(q)
        if union_list:
            union_query = union_list[0].union_all(*(union_list[1:]))
        if union_query:
            query_tab = alias(union_query.subquery(), '_all_cdr_')
            self.model_class._tab = query_tab
            _select = []
            q_tab = query_tab

            def _aggr(cdr_table, field_name, agg_method=None, need_label=True):
                fd = {'total': func.sum, 'min': func.min, 'max': func.max, 'avrg': func.avg, }

                if agg_method and (field_name in aggregable or hasattr(self, 'on_select_' + field_name)):
                    if agg_method not in fd:
                        raise ValidationError(
                            '{}:no such method.  Allowed values :{}'.format(k, similar_fields(agg_method, fd.keys())))
                    f = fd[agg_method]
                    tf = f(getattr(cdr_table.c, field_name))
                    if need_label:
                        return tf.label(field_name)
                    else:
                        return tf
                else:
                    tf = getattr(cdr_table.c, field_name)
                    if need_label:
                        return tf.label(field_name)
                    else:
                        return tf

            for k in fields:
                if k == 'asr':
                    _select.append(func.round(
                        100.0 * _aggr(q_tab, 'not_zero_calls', 'total') / _aggr(q_tab, 'ingress_total_calls', 'total'),
                        2).label('asr'))
                    continue
                if k == 'qsr':
                    _select.append(func.round(100.0 * (
                            _aggr(q_tab, 'not_zero_calls', 'total') + _aggr(q_tab, 'ingress_busy_calls',
                                                                            'total') + _aggr(q_tab,
                                                                                             'ingress_cancel_calls',
                                                                                             'total')) / _aggr(
                        q_tab, 'ingress_total_calls', 'total'), 2).label('qsr'))
                    continue
                if k == 'ingress_npr':
                    _select.append(func.round(100.0 * (_aggr(q_tab, 'npr_count', 'total')) / _aggr(
                        q_tab, 'ingress_total_calls', 'total'), 2).label('ingress_npr'))
                    continue
                if k == 'egress_npr':
                    _select.append(func.round(100.0 * (_aggr(q_tab, 'npr_count', 'total')) / _aggr(
                        q_tab, 'egress_total_calls', 'total'), 2).label('egress_npr'))
                    continue
                if hasattr(q_tab.c, k):
                    if not hasattr(self, 'on_select_' + k) and k not in aggregable:
                        continue
                    _select.append(_aggr(q_tab, k, method))
            # _select = tuple(_select)
            _group = []
            new_time_field = None
            my_step = 0
            if step:
                step_dict = {'hour': 60 * 60,
                             'day': 60 * 60 * 24,
                             'week': 60 * 60 * 24 * 7,
                             'year': 60 * 60 * 24 * 365,
                             'month': 60 * 60 * 24 * 30,
                             'all': 0}
                if step in step_dict:
                    my_step = step_dict[step]
                else:
                    my_step = int(step)
            if my_step:
                # .op('AT TIME ZONE')(tz)
                # new_time_field = (my_step * (cast(func.extract('epoch', self.time_field(q_tab)), Integer) / my_step))
                if step in step_dict:
                    new_time_field = cast(func.extract('epoch', func.date_trunc(step, self.time_field(q_tab))), Integer)
                else:
                    new_time_field = (
                            my_step * (cast(func.extract('epoch', self.time_field(q_tab)), Integer) / my_step))
                _select.append(new_time_field.label('report_time'))
                _group.append(new_time_field)
            else:
                _group.append(cast(func.extract('millennium', self.time_field(q_tab)), Integer))

            if groups:
                _group = _group + list(map(lambda k: getattr(q_tab.c, k), groups))
                _select = _select + list(map(lambda k: getattr(q_tab.c, k), groups))

            self.modify_select_from_groupping(_select, groups, q_tab)

            query = model.get_db().session.query(*_select)

            if _group:
                query = query.group_by(*tuple(_group))
            if sort_order is None:
                if new_time_field is not None:
                    query = query.order_by(new_time_field)
            else:
                order_by_expr = tuple(
                    map(lambda i: _aggr(q_tab, i[0], method, False).desc() if i[1] == 'desc' else _aggr(q_tab, i[0],
                                                                                                        method, False),
                        sort_order))
                query = query.order_by(*order_by_expr)
            # return filtering, query
            if REPORT_DEBUG:
                log.debug(model.query_to_sting(query))

            if 'release_cause_from_protocol_stack' in groups:
                query = query.filter(text('release_cause_from_protocol_stack IS NOT null'))

            for i in fields:
                if not i in ['mrc', 'nrc', 'did', 'subcharge']:
                    query = query.filter(text('{field_name} IS NOT null'.format(field_name=i)))
                    break

            return {}, query
        else:
            return None, None
        pass

    def modify_select_from_groupping(self, _select, groups, q_tab):
        if 'ingress_id' in groups:
            ing = aliased(model.Resource)
            _select.append(
                select([ing.alias]).where(ing.resource_id == q_tab.c.ingress_id).as_scalar().label('ingress_name'))
        if 'egress_id' in groups:
            ing = aliased(model.Resource)
            _select.append(
                select([ing.alias]).where(ing.resource_id == q_tab.c.egress_id).as_scalar().label('egress_name'))
        if 'ingress_client_id' in groups:
            ing = aliased(model.Client)
            _select.append(
                select([ing.name]).where(ing.client_id == q_tab.c.ingress_client_id).as_scalar().label(
                    'ingress_client_name'))
        if 'egress_client_id' in groups:
            ing = aliased(model.Client)
            _select.append(
                select([ing.name]).where(ing.client_id == q_tab.c.egress_client_id).as_scalar().label(
                    'egress_client_name'))
        if 'release_cause' in groups:
            cls = aliased(model.GlobalRouteError)
            _select.append(
                select([cls.error_description]).where(cls.error_code == q_tab.c.release_cause).as_scalar().label(
                    'release_cause_name'))


class ReportM(List):
    entity_plural = 'Report items'
    scheme_class = CdrReportGroppedRequestScheme
    scheme_class_get = CdrReportGroppedResponseScheme
    scheme_class_get_inner = CdrReportDetailInnerScheme
    model_class = model.CdrReportDetail
    columns = model_class.get_fields()

    groupable = [fn for fn in columns if
                 getattr(model.CdrReportDetail, fn).type.__class__.__name__ in (
                     'String', 'DateTime', 'Text') or '_id' in fn or fn in _grp]
    aggregable = list(set([fn for fn in columns if
                           getattr(model.CdrReportDetail, fn).type.__class__.__name__ not in (
                               'String', 'DateTime', 'Text') and '_id' not in fn and fn not in _grp] +
                          ['nrf_calls', 'asr', 'qsr', 'ingress_npr', 'egress_npr']))

    def get_spec(self):
        sfields_names, search_fields_list = self.get_fields_from_scheme(self.scheme_class)
        qfields_names, query_fields_list = self.get_query_fields_from_scheme(self.scheme_class)
        fields_names = [f.name for f in inspect(self.model_class).columns]
        spec = swagger.specify.get_spec(
            method='get', description='Gets {}'.format(self.entity_plural.lower()),
            path_parameters=self.path_parameters,
            query_parameters=search_fields_list + [
                {"name": "from", "description": "printing range. Default: print from start 0 position",
                 "type": "integer"},
                {"name": "count", "description": "printing range. Default: print first 1000 items",
                 "type": "integer"},
                {"name": "fields",
                 "description": "aggregated fields, available values:{}".format(self.aggregable)},
                {"name": "group",
                 "description": "groupping fields, available values:{}".format(self.groupable)},

            ] +
                             [dict(name=i, type='string') for i in self.groupable],
            responses=(
                          SuccessReportResponse(),
                          responses.AttributeNotExitsErrorResponse()
                      ) + self.get_additional_responses(method='get'),
            security=self.get_security(method='get')
        )
        spec['get']['produces'] = ['application/json', 'text/csv', 'text/xml']
        return spec

    def run(self, **kwargs):
        class Req:
            def __init__(self, req_user):
                self.query_string = None
                self.context = {'user': req_user}

            pass

        if 'user' in kwargs:
            user = kwargs['user']
            del kwargs['user']
        else:
            user = model.User.get(1)
        req = Req(user)
        req.query_string = urlencode(kwargs)
        self.req = req
        objects_list, total, page, per_page, fields = self.get_objects_list(req, self.model_class, **kwargs)
        # items = [i for i in objects_list]
        items = self.scheme_class_get_inner().dump(objects_list, many=True).data
        return items

    def send_request(self, params):
        m = dict(calls='ingress_total_calls',
                 ingress_billed_time='ingress_bill_time',
                 ingress_cost='ingress_call_cost',
                 ingress_billed_min='ingress_bill_time',
                 code='ingress_code',
                 non_zero_calls='not_zero_calls',
                 egress_cost='egress_total_cost',
                 egress_billed_time='egress_bill_time',
                 )
        args = {k: v for k, v in params.items() if k not in m}
        upd = {m[k]: v for k, v in params.items() if k in m}
        args.update(upd)

        class Responce:
            def json(self):
                return self.data

        items = self.run(**args)
        resp = Responce()
        resp.data = dict(data=items)
        return resp

    def _on_get(self, req, resp, **kwargs):

        from .helpers import check_permission

        if (not hasattr(self, 'no_auth_needed') or self.no_auth_needed == False) and not check_permission(self, req,
                                                                                                          'list'):
            self.set_response(
                resp, responses.ForbiddenErrorResponse(data=errors.AuthErrors.Forbidden)
            )
            return
        try:
            import time
            _start = time.time()

            objects_list, total, _from, count, fields = self.get_objects_list(req, self.model_class, **kwargs)
            # items = [i for i in objects_list]
            items = objects_list #self.scheme_class_get_inner().dump(objects_list, many=True).data
            data = {
                'success': True,
                'data': items,
                'total': total, 'from': _from, 'count': len(items)
            }
            self.set_response(resp, SuccessReportResponse(data=data, scheme=self.scheme_class_get))

            _end = time.time()
            log.debug('report execution time {}'.format(_end - _start))
            # resp.body = json.dumps(data)
            # resp.status = falcon.HTTP_200

        # except AttributeError as e:
        #     self.set_response(
        #         resp, responses.AttributeNotExitsErrorResponse(
        #             data=ATTRIBUTE_ERROR_RE.search(str(e)).group(1)
        #         )
        #     )
        except (InternalError, IntegrityError, DataError) as e:
            import traceback
            model.get_db().session.rollback()
            log.error(traceback.format_exc())
            self.set_response(resp, OperationErrorResponse(e))
        except Exception as e:
            import traceback
            log.error(traceback.format_exc())
            self.set_response(resp, OperationErrorResponse(e))

    def time_field(self, cdr):
        return cdr.c.report_time

    def get_daily_cdr(self, cdr_date):
        cdr = model.cdr_report_detail(cdr_date)
        return cdr

    def get_fields_for_list(self, parsed_qs, **kwargs):
        fields = parsed_qs.get('fields', None)
        if fields:
            fields = fields.split(',')
            for f in fields:
                # if f not in self.get_all_fields_from_scheme(self.scheme_class_get)[0]:
                if f not in self.aggregable:
                    raise Exception('fields {} not valid!'.format(f))
        return fields

    def get_objects_list(self, req, model_class, **kwargs):
        parsed_qs = dict(parse_qsl(req.query_string))
        if REPORT_DEBUG:
            log.debug('report parsed parameters:{}'.format(parsed_qs))

        try:
            _from = int(parsed_qs.get('from'))
        except (ValueError, TypeError):
            _from = 0

        try:
            count = int(parsed_qs.get('count', 1000))
        except ValueError:
            count = 1000

        ordering = {}

        fields = self.get_fields_for_list(parsed_qs, **kwargs)

        filtering = self.get_filtering_for_list(parsed_qs, **kwargs)
        if fields:
            filtering['fields'] = ','.join(fields)

        filtering, query = \
            self.modify_query_from_filtering_for_list(filtering, **kwargs)

        if ordering:
            ordering, query = \
                self.modify_query_from_ordering_for_list(ordering, query, **kwargs)
        if query is None:
            return [], 0, _from, count, fields

        return (query, len(query), _from, count, fields)

    def on_filter_ingress_id(self, v, cdr):
        # return text("ingress_id in ({})".format(v))
        return cdr.c.ingress_id.in_([int(i) for i in v.split(',')])

    def on_filter_egress_id(self, v, cdr):
        # return text("egress_id in ({})".format(v))
        return cdr.c.egress_id.in_([int(i) for i in v.split(',')])

    def on_filter_non_zero_calls(self, v, cdr):
        # return text("egress_id in ({})".format(v))
        if v == '1':
            return cdr.c.ingress_not_zero_calls > 0

    def on_filter_scc_below_sec(self, v, cdr):
        return cdr.c.duration < v

    def on_filter_is_hide_unauthorized(self, v, cdr):
        if v is None:
            return None
        if v:
            return cdr.c.release_cause != 3
        else:
            return cdr.c.release_cause == 3

    def on_select_nrf_calls(self, cdr):
        return cdr.c.lrn_calls

    def on_select_asr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.not_zero_calls) / (1.0 * cdr.c.ingress_total_calls))

    def on_select_qsr(self, cdr):
        return 100.00 * ((cdr.c.not_zero_calls + cdr.c.ingress_busy_calls + cdr.c.ingress_cancel_calls) /
                         cdr.c.ingress_total_calls)

    def on_select_ingress_npr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.npr_count) / (1.0 * cdr.c.ingress_total_calls))

    def on_select_egress_npr(self, cdr):
        return 100.00 * ((1.0 * cdr.c.npr_count) / (1.0 * cdr.c.egress_total_calls))

    def modify_query_from_filtering_for_list(self, filtering, **kwargs):

        def similar_fields(field, field_list):
            sf = field.split('_')
            sim = []
            for f in sf:
                for ef in field_list:
                    if f in ef:
                        sim.append(ef)
            if sim:
                return 'similar fields:{}'.format(set(sim))
            return ''

        import sys
        sys.setrecursionlimit(10000)
        cls = self.model_class
        columns = cls.get_fields()
        # columns = [f.name for f in inspect(self.model_class).columns]
        req_query = dict(parse_qsl(self.req.query_string))
        _from = filtering.pop('from', None)
        count = filtering.pop('count', None)

        sort = filtering.pop('sort', None)
        order = filtering.pop('order', '')

        sort_order = None
        if sort:
            _sort = sort.split(',')
            for s in _sort:
                if s not in columns:
                    raise ValidationError('sort field missing {}'.format(s))
            _order = order.split(',')
            for o in _order:
                if o not in ('', 'asc', 'desc'):
                    raise ValidationError('sort order invalid value {}, must be asc or desc (can be empty)'.format(o))
            if len(_sort) != len(_order):
                if sort != '':
                    raise ValidationError('sort length must be equal order length')
                else:
                    _order = ['' for s in _sort]
            sort_order = list(zip(_sort, _order))

        start_time = filtering.pop('start_time', None)
        end_time = filtering.pop('end_time', str(int(datetime.now(UTC).timestamp())))
        try:
            start_time = int(start_time)
            if start_time < 0:
                start_time = int(datetime.now(UTC).timestamp()) + start_time
            start_date_time = datetime.utcfromtimestamp(start_time).replace(tzinfo=UTC)
        except (ValueError, TypeError):
            raise ValidationError('wrong start time {}'.format(start_time))
        try:
            end_time = int(end_time)
            end_date_time = datetime.utcfromtimestamp(end_time).replace(tzinfo=UTC)
        except (ValueError, TypeError):
            raise ValidationError('wrong end time')
        if start_time >= end_time:
            raise ValidationError('start time greater than end_time!')
        groupable = self.groupable
        aggregable = self.aggregable
        tz = str(filtering.pop('tz', '0'))
        if REPORT_DEBUG:
            log.debug('tz={}'.format(tz))
        non_zero_calls = filtering.pop('non_zero_calls', None)
        if REPORT_DEBUG:
            log.debug('non_zero_calls={}'.format(non_zero_calls))
        human_readable = filtering.pop('human_readable', None)
        if REPORT_DEBUG:
            log.debug('human_readable={}'.format(human_readable))

        method = filtering.pop('method', 'total')
        step = filtering.pop('step', 'all')
        groups = filtering.pop('group', None)
        fields = filtering.pop('fields', None)
        field = filtering.pop('field', None)
        fields = fields or field
        if fields:
            fields = fields.split(',')
        if not fields:
            fields = aggregable.copy()
        for fk in filtering.keys():
            if fk not in columns:
                if not hasattr(self, 'on_filter_{}'.format(fk)):
                    raise ValidationError('{}:wrong filter field. {}'.format(fk, similar_fields(fk, columns)))
            if fk not in fields and not hasattr(self, 'on_filter_{}'.format(fk)):
                fields.append(fk)

        if groups is not None:
            groups = groups.split(',')
        else:
            groups = []
        for fk in groups:
            if fk not in groupable:
                raise ValidationError('{}:wrong group field. {}'.format(fk, similar_fields(fk, groupable)))
            if fk not in fields:
                fields.append(fk)

        fields.append('report_time')
        fields = list(set(fields))

        user = self.get_user(self.req)

        if user and user.user_type in ('client', 'agent', 'vendor'):
            if 'ingress_id' in filtering and filtering['ingress_id']:
                ids = set(filtering['ingress_id'].split(','))
                alw = set([str(i) for i in user.allowed_ingress_id])
                filtering['ingress_id'] = ','.join(list(ids.intersection(alw)))
            else:
                filtering['ingress_id'] = ','.join([str(i) for i in user.allowed_ingress_id])
            if filtering['ingress_id'] == '':
                if user.allowed_egress_id:
                    alw = set([str(i) for i in user.allowed_egress_id])
                    if 'egress_id' in filtering and filtering['egress_id']:
                        ids = set(filtering['egress_id'].split(','))
                        filtering['egress_id'] = ','.join(list(ids.intersection(alw)))
                    else:
                        filtering['egress_id'] = ','.join(list(alw))
                    del filtering['ingress_id']
                else:
                    filtering['ingress_id'] = '-1'
            if REPORT_DEBUG:
                log.debug('result {} filtering={}'.format(user.user_type, filtering))

        start_day = start_date_time.date()
        database_start_day = model.cdr_reports_start_date()
        if database_start_day is None:
            raise ValidationError('database has no cdr report data!')
        if database_start_day > start_day:
            start_day = database_start_day
        end_day = end_date_time.date()
        if end_day < start_day:
            end_day = start_day

        days = (end_day - start_day).days
        union_query = None
        union_list = []
        union_dict = {}

        def _aggr(cdr_table, field_name, agg_method=None, need_label=True):
            fd = {'total': func.sum, 'min': func.min, 'max': func.max, 'avrg': func.avg, }

            if agg_method and (field_name in aggregable or hasattr(self, 'on_select_' + field_name)):
                if agg_method not in fd:
                    raise ValidationError(
                        '{}:no such method.  Allowed values :{}'.format(k, similar_fields(agg_method, fd.keys())))
                f = fd[agg_method]
                if hasattr(self, 'on_select_' + field_name):
                    tf = f(getattr(self, 'on_select_' + k)(cdr))
                elif hasattr(cdr_table.c, field_name):
                    tf = f(getattr(cdr_table.c, field_name))
                else:
                    tf = literal(0.0)
                if need_label:
                    return tf.label(field_name)
                else:
                    return tf
            else:
                tf = getattr(cdr_table.c, field_name)
                if need_label:
                    return tf.label(field_name)
                else:
                    return tf

        _gkeys = []
        _agr = []
        for day_idx in range(days + 1):
            day_cdr = start_day + timedelta(days=day_idx)
            cdr_date = "%s%02d%02d" % (day_cdr.year, day_cdr.month, day_cdr.day)
            cdr = self.get_daily_cdr(cdr_date)
            if cdr is None:
                log.warning('missing table {}'.format(cdr_date))
                continue
            _select = []
            for k in fields:
                keys = [x.key for x in _select]
                if k in keys:  # already here
                    continue
                # k=k.replace('non_zero_calls','not_zero_calls')
                if k in ('asr', 'qsr', 'ingress_npr', 'egress_npr'):
                    keys = [x.key for x in _select]
                    for df in ['ingress_busy_calls', 'ingress_cancel_calls', 'not_zero_calls',
                               'ingress_total_calls', 'egress_total_calls', 'npr_count']:
                        if not df in keys:
                            _select.append(_aggr(cdr.c, df, method))
                    continue
                if hasattr(self, 'on_select_' + k):
                    _f = getattr(self, 'on_select_' + k)(cdr)
                    if _f is not None:
                        _select.append(_aggr(cdr, k, method))
                    continue
                elif hasattr(cdr.c, k):
                    if not hasattr(self, 'on_select_' + k) and k not in aggregable:
                        continue
                    _select.append(_aggr(cdr, k, method))
                else:
                    raise ValidationError('{}:no such field to select.{}'.format(k, similar_fields(k, columns)))
                    # _select.append(getattr(cdr.c, k))
            # _select = tuple(_select)
            _filter = []
            if day_idx == 0:
                _filter = [self.time_field(cdr) >= start_date_time]
                if day_idx == days:
                    _filter = [self.time_field(cdr) >= start_date_time, self.time_field(cdr) <= end_date_time]
            else:
                if day_idx == days:
                    _filter = [self.time_field(cdr) <= end_date_time]
            for k, v in filtering.items():
                if hasattr(self, 'on_filter_' + k):
                    _f = getattr(self, 'on_filter_' + k)(v, cdr)
                    if _f is not None:
                        _filter.append(_f)
                    continue
                if not k in columns:
                    raise ValidationError('{}:no such field to filter.{}'.format(k, similar_fields(k, columns)))
                if hasattr(cdr.c, k):
                    if '*' in v:
                        _filter.append(getattr(cdr.c, k).like(v.replace('*', '%')))
                    else:
                        _filter.append(getattr(cdr.c, k) == v)
            _group = []
            new_time_field = None
            my_step = 0
            if step:
                step_dict = {'hour': 60 * 60,
                             'day': 60 * 60 * 24,
                             'week': 60 * 60 * 24 * 7,
                             'year': 60 * 60 * 24 * 365,
                             'month': 60 * 60 * 24 * 30,
                             'all': 0}
                if step in step_dict:
                    my_step = step_dict[step]
                else:
                    my_step = int(step)
            if my_step:
                if step in step_dict:
                    new_time_field = cast(func.extract('epoch', func.date_trunc(step, self.time_field(cdr))),
                                          Integer).label('report_time')
                else:
                    new_time_field = (
                            my_step * (cast(func.extract('epoch', self.time_field(cdr)), Integer) / my_step)).label(
                        'report_time')
                _select.append(new_time_field.label('report_time'))
                _group.append(new_time_field)
            # else:
            #     _group.append(cast(func.extract('millennium', self.time_field(cdr)), Integer))

            if groups:
                _group = _group + list(map(lambda k: getattr(cdr.c, k), groups))
                _select = _select + list(map(lambda k: getattr(cdr.c, k), groups))

            self.modify_select_from_groupping(_select, groups, cdr)

            q = model.get_db().session.query(*_select)
            if _filter:
                q = q.filter(and_(*_filter))
            if _group:
                q = q.group_by(*_group)
                _gkeys = [i.key for i in _group]
                if 'ingress_id' in _gkeys:
                    _gkeys.append('ingress_name')
                if 'egress_id' in _gkeys:
                    _gkeys.append('egress_name')
                if 'ingress_client_id' in _gkeys:
                    _gkeys.append('ingress_client_name')
                if 'egress_client_id' in _gkeys:
                    _gkeys.append('egress_client_name')
                if 'release_cause' in _gkeys:
                    _gkeys.append('release_cause_name')
            _agr = [x.key for x in _select if x.key in aggregable]
            union_list = union_list + q.all()

        # post groupping and filtering
        res = []
        if union_list:
            print(union_list)
            def get0(k):
                if getattr(cls, k, None) and getattr(cls, k).type.python_type.__name__ == 'int':
                    return 0
                return ''

            def gkey(x):
                return tuple(get0(k) if getattr(x, k) is None else getattr(x, k) for k in _gkeys)

            def keydict(x):
                return dict(zip(_gkeys, x))

            def agrdict(x):
                return dict(zip(_agr, x))

            def agrdict0():
                return {k: 0 for k in _agr}

            union_list.sort(key=gkey)
            for key, grp in groupby(union_list, gkey):
                row = {**keydict(key), **agrdict0()}
                for thing in grp:
                    for k in _agr:
                        if not getattr(thing, k, 0) is None:
                            row[k] += getattr(thing, k, 0)
                    #print(key, thing)
                #print()
                from decimal import Decimal
                for k in fields:
                    if k == 'asr':
                        row['asr'] = round(
                            Decimal(100.0) * row['not_zero_calls'] / row['ingress_total_calls'])
                    if k == 'qsr':
                        row['qsr'] = round(
                            Decimal(100.0) * (row['not_zero_calls'] + row['ingress_busy_calls'] + row[
                                'ingress_cancel_calls']) / row['ingress_total_calls'])
                    if k == 'ingress_npr':
                        row['ingress_npr'] = round(
                            Decimal(100.0) * row['npr_count'] / row['ingress_total_calls'])
                    if k == 'egress_npr':
                        row['egress_npr'] = round(
                            Decimal(100.0) * row['npr_count'] / row['ingress_total_calls'])
                for k in list(row.keys()):
                    if k not in fields and k not in _gkeys:
                        del row[k]
                res.append(row)
            #print(res)

            return {}, res
        else:
            return None, None
        pass

    def modify_select_from_groupping(self, _select, groups, q_tab):
        if 'ingress_id' in groups:
            ing = aliased(model.Resource)
            _select.append(
                select([ing.alias]).where(ing.resource_id == q_tab.c.ingress_id).as_scalar().label('ingress_name'))
        if 'egress_id' in groups:
            ing = aliased(model.Resource)
            _select.append(
                select([ing.alias]).where(ing.resource_id == q_tab.c.egress_id).as_scalar().label('egress_name'))
        if 'ingress_client_id' in groups:
            ing = aliased(model.Client)
            _select.append(
                select([ing.name]).where(ing.client_id == q_tab.c.ingress_client_id).as_scalar().label(
                    'ingress_client_name'))
        if 'egress_client_id' in groups:
            ing = aliased(model.Client)
            _select.append(
                select([ing.name]).where(ing.client_id == q_tab.c.egress_client_id).as_scalar().label(
                    'egress_client_name'))
        if 'release_cause' in groups:
            cls = aliased(model.GlobalRouteError)
            _select.append(
                select([cls.error_description]).where(cls.error_code == q_tab.c.release_cause).as_scalar().label(
                    'release_cause_name'))


class DidReport(Report):
    entity_plural = 'Did Report items'
    scheme_class = DidReportGroppedRequestScheme
    scheme_class_get = DidReportGroppedResponseScheme
    scheme_class_get_inner = DidReportInnerScheme
    model_class = model.DidReport
    columns = model_class.get_fields()

    groupable = [fn for fn in columns if
                 getattr(model.DidReport, fn).type.__class__.__name__ in (
                     'String', 'DateTime', 'Text') or '_id' in fn or fn in _grp]
    aggregable = [fn for fn in columns if
                  getattr(model.DidReport, fn).type.__class__.__name__ not in (
                      'String', 'DateTime', 'Text') and '_id' not in fn and fn not in _grp]

    def on_filter_is_hide_unauthorized(self, v, cdr):
        rep = aliased(model.DidBillingRel)
        if v is None:
            return None
        if v == 'true':
            q = select([func.count(rep.did) > 0]).where(rep.did == cdr.c.did).as_scalar()
            return q
        else:
            q = select([func.count(rep.did) == 0]).where(rep.did == cdr.c.did).as_scalar()
            return q

    def get_daily_cdr(self, cdr_date):
        cdr = model.did_report(cdr_date)
        return cdr

    def ya_query(self, fields, start_date, end_date, filtering):
        # def query_did_charge(start_date, end_date, client, vendor, did):
        vendor = filtering.get('ingress_client_id', '')
        vendor_res = filtering.get('egress_id', '')
        client = filtering.get('egress_client_id', '')
        client_res = filtering.get('egress_id', '')
        did = filtering.get('did', '')

        params = ';'.join(
            [str(start_date), str(end_date), str(client), str(vendor), str(client_res), str(vendor_res), str(did)])
        clsd = model.DidChargeDetail
        rel = model.DidBillingRel  # v6
        # rel = DidBillingBrief
        pln = model.DidBillingPlan
        rep = DidReport
        Resource = model.Resource
        dids = None
        if did:
            dids = [i for i in did.split(',')]

        trunks = None
        client_id = None
        if client:
            client_id = [int(i) for i in client.split(',')]
            trunks = [r.resource_id for r in Resource.filter(Resource.client_id.in_(client_id)).all()]
        if client_res:
            client_res_id = [int(i) for i in client_res.split(',')]
            trunks = [r for r in client_res_id]
        vendor_trunks = None
        if vendor:
            vendor_id = [int(i) for i in vendor.split(',')]
            vendor_trunks = [r.resource_id for r in Resource.filter(Resource.client_id.in_(vendor_id)).all()]
        if vendor_res:
            vendor_res_id = [int(i) for i in vendor_res.split(',')]
            trunks = [r for r in vendor_res_id]

        exists = clsd.filter(clsd.params == params).first()
        if exists is None:
            ret = clsd.session().execute(clsd.__table__.insert() \
                .from_select(
                ['params', 'did_number', 'client_billing_plan_id', 'client_trunk_id', 'vendor_trunk_id', 'nrc', 'mrc',
                 'pay_type',
                 'start_date', 'end_date'],
                select([literal(params), rel.did, rel.buy_billing_plan_id, rel.client_res_id, rel.vendor_res_id,
                        case([(rel.start_date >= start_date, pln.did_price)], else_=0.0),
                        pln.monthly_charge,
                        pln.pay_type,
                        func.greatest(cast(rel.start_date, DateTime), start_date),
                        func.least(cast(rel.end_date, DateTime), end_date),
                        ]
                       ). \
                    where(and_(or_(rel.end_date.is_(None), rel.end_date > start_date),
                               rel.start_date <= end_date,
                               rel.client_res_id.in_(trunks) if trunks else True,
                               rel.vendor_res_id.in_(vendor_trunks) if vendor_trunks else True,
                               rel.did.in_(dids) if dids else True,
                               pln.id == rel.buy_billing_plan_id))
            )
            )
            # print(ret)
            q = clsd.__table__.update().values(mrc_number=
                                               case([(clsd.pay_type == 0,
                                                      func.extract('isoyear', clsd.end_date) * 52 +
                                                      func.extract('week', clsd.end_date) - func.extract('week',
                                                                                                         clsd.start_date)
                                                      - func.extract('isoyear', clsd.start_date) * 52
                                                      ),
                                                     (clsd.pay_type == 1, func.extract('month',
                                                                                       clsd.end_date + timedelta(
                                                                                           days=1)) - func.extract(
                                                         'month', clsd.start_date)
                                                      + func.extract('year', clsd.end_date + timedelta(
                                                         days=1)) * 12 - func.extract('year', clsd.start_date) * 12)
                                                     ],
                                                    else_=clsd.end_date - clsd.start_date
                                                    )
                                               ).where(clsd.params == params)
            # print(str(q))
            ret = clsd.session().execute(q)
            # print(ret)
            clsd.session().commit()
            # if clsd.filter(clsd.params == params).first() is None:
            #     clsd(params=params,did='').save()

        _select = []
        for k in fields:
            if hasattr(clsd, k):
                if k == 'mrc':
                    _select.append(getattr(clsd, 'did_mrc').label(k))
                else:
                    _select.append(getattr(clsd, k).label(k))
            else:
                if k == 'report_time':
                    _select.append(literal(end_date).label(k))
                else:
                    _select.append(literal(None).label(k))
        return clsd.session().query(*_select).filter(clsd.params == params)


class HostBasedReport(Report):
    entity_plural = 'Host Based Report items'
    scheme_class = DidReportGroppedRequestScheme
    scheme_class_get = DidReportGroppedResponseScheme
    scheme_class_get_inner = DidReportInnerScheme
    model_class = model.HostBasedReport
    columns = model_class.get_fields()

    groupable = [fn for fn in columns if
                 getattr(model.HostBasedReport, fn).type.__class__.__name__ in (
                     'String', 'DateTime', 'Text') or '_id' in fn or fn in _grp]
    aggregable = [fn for fn in columns if
                  getattr(model.HostBasedReport, fn).type.__class__.__name__ not in (
                      'String', 'DateTime', 'Text') and '_id' not in fn and fn not in _grp]

    def get_daily_cdr(self, cdr_date):
        cdr = model.host_based_report(cdr_date)
        return cdr

    def modify_select_from_groupping(self, _select, groups, q_tab):
        super().modify_select_from_groupping(_select, groups, q_tab)
        if 'ingress_client_id' in groups and 'ingress_ip' in groups:
            ipc = aliased(model.ResourceIp)
            ing = aliased(model.Resource)
            _select.append(
                select([func.max(ing.alias)]).where(and_(ing.resource_id == ipc.resource_id,
                                                         ipc.ip == q_tab.c.ingress_ip,
                                                         ing.client_id == q_tab.c.ingress_client_id,
                                                         )
                                                    ).as_scalar().label('ingress_name'))
        if 'egress_client_id' in groups and 'egress_ip' in groups:
            ipc = aliased(model.ResourceIp)
            ing = aliased(model.Resource)
            _select.append(
                select([func.max(ing.alias)]).where(and_(ing.resource_id == ipc.resource_id,
                                                         ipc.ip == q_tab.c.egress_ip,
                                                         ing.client_id == q_tab.c.egress_client_id,
                                                         )
                                                    ).as_scalar().label('egress_name'))


class StatisticException(Exception):
    pass


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        print('func:%r args:[%r, %r] took: %2.4f sec' % (f.__name__, args, kw, te - ts))
        return result

    return wrap


class StatisticAPI(object):
    @classmethod
    @timing
    def run(cls, **params):
        return Report().run(**params)


class StatisticAPI2(object):
    @classmethod
    @timing
    def run(cls, **params):
        return ReportM().run(**params)


if __name__ == '__main__':
    from inv5.model import init_db

    from inv5 import settings

    init_db()  # conn_string='postgresql://postgres:password@127.0.0.1:5434/softswitch4v5')

    client = model.Client.get(558)
    api = StatisticAPI()
    api2 = StatisticAPI2()
    start_time = int(datetime(2021, 4, 5,tzinfo=UTC).timestamp())
    end_time = int(datetime(2021, 4, 5, 23, 59, 59,tzinfo=UTC).timestamp())
    res = api.run(start_time=start_time, end_time=end_time,
                  field='not_zero_calls,ingress_bill_time,ingress_call_cost',
                  non_zero_calls='1',
                  ingress_id=client._trunks_term_ids,
                  #group='ingress_id',
                  step='day'
                  )
    print(res)
    res = api2.run(start_time=start_time, end_time=end_time,
                   field='not_zero_calls,ingress_bill_time,ingress_call_cost',
                   non_zero_calls='1',
                   ingress_id=client._trunks_term_ids,
                   #group='ingress_id',
                   step='day'
                   )
    print(res)
#{'start_time': 1617224400, 'end_time': 1619816399, 'field': 'not_zero_calls,ingress_bill_time,ingress_call_cost',
# 'non_zero_calls': '1', 'ingress_id': '1185,1184,842'}] took: 3.1291 sec
