from marshmallow import validate
from marshmallow.fields import Str,Int, Bool,DateTime, Float, Decimal, Dict ,Nested, ValidationError
from inv5.schemes.basemodel_scheme import BaseModelScheme
from inv5.schemes.scheme_choice import Choice
from datetime import datetime, timedelta
from dateutil.parser import parse as parse_datetime
from pytz import UTC
from inv5 import model

FQDN_REGEXP = "^((?=[a-z0-9-]{1,63}\.)(xn--)?[a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,63}$"
IP_REGEXP_DOMAIN = "^([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])" \
                   "(\.([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]{0,61}[a-zA-Z0-9]))*$"
# IP_REGEXP = "^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)|localhost$"
IP_REGEXP = "^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)|localhost$"
DATE_REGEXP = r'^(0?[1-9]|1[012])/(0?[1-9]|[1-2][0-9]|3[01])/[12][09][0-9][0-9]$|' \
              r'^[12][09][0-9][0-9]-(0?[1-9]|1[012])-(0?[1-9]|[1-2][0-9]|3[01])$|' \
              r'^(0?[1-9]|1[012])-[12][09][0-9][0-9]-(0?[1-9]|[1-2][0-9]|3[01])$|' \
              r'^(0?[1-9]|1[012])/[12][09][0-9][0-9]/(0?[1-9]|[1-2][0-9]|3[01])$|' \
              r'^[12][09][0-9][0-9]/(0?[1-9]|1[012])/(0?[1-9]|[1-2][0-9]|3[01])$'
TIME_REGEXP = r'(24:00|2[0-3]:[0-5][0-9]|[0-1][0-9]:[0-5][0-9])'
TIMEZONE_REGEXP = r'^(?:Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])$'
PHONE_REGEXP = r'^\+?[0-9]+$'
DIGITS_REGEXP = r'^[0-9]+$'
PREFIX_REGEXP = r'^[0-9#\+]*$'
NAME_REGEXP = "^(\w[ _\-\.]?)*\w$"
USERNAME_REGEXP = "^(\w[ _\-\.\@]?)*\w$"


def valid_date(d):
    try:
        if len(str(d)) != 10:
            raise ValidationError('Invalid date "{}" must be in format "YYYY-MM-DD"'.format(d))
        parse_datetime(str(d))
        return True
    except ValueError:
        raise ValidationError('Invalid date {}'.format(d))

# validate any model field
def _valid(cls_name, field, value, msg=''):
    cls = model.__dict__[cls_name]
    if type(value) == type('') or (not getattr(value, 'urn', None) is None and 'uuid' in getattr(value, 'urn', None)):
        value = "'" + str(value) + "'"
    if cls.filter(model.text('{}={}'.format(field, value))).count() == 0:
        if msg:
            raise ValidationError(msg)
        else:
            raise ValidationError('{} {} is invalid (not exists)'.format(field, value))


def _emails_validate(value):
    from email.utils import parseaddr
    lst = value.split(';')
    for addr in lst:
        if addr == '':
            continue
        name, email = parseaddr(addr)
        validate.Email(error='invalid emails')(email)


class Emails(Str):
    def __init__(self, *args, **kwargs):
        Str.__init__(self, *args, **kwargs)
        # Insert validation into self.validators so that multiple errors can be
        # stored.
        self.validators.insert(0, _emails_validate)

    def _validated(self, value):
        """Format the value or raise a :exc:`ValidationError` if an error occurs."""
        if value is None:
            return None
        _emails_validate(value)
        return value
# region--- Invoice---
class InvoiceScheme(BaseModelScheme):
    invoice_id = Int()
    invoice_number = Str()
    invoice_start = DateTime()
    invoice_end = DateTime()
    invoice_time = DateTime()
    state = Choice()
    client_id = Int()
    client_name = Str()
    company_name = Str()
    amount = Decimal()
    paid = Bool()
    due_date = Str()
    url = Str()
    data_url = Str()
    status = Choice()
    create_type = Choice()
    last_invoice_for = Str()
    payment_term_name = Str()
    invoice_period = Str()
    invoice_credit_amount = Float()
    unpaid_amount = Float()
    pay_amount = Float()
    invoice_num = Str()

    class Meta:
        model = model.Invoice
        fields = (
            'invoice_id', 'bill_period_start', 'bill_period_end', 'invoice_number', 'invoice_time', 'state',
            'client_id', 'client_name', 'company_name', 'amount', 'paid', 'due_date', 'url', 'status', 'create_type',
            'last_invoice_for', 'payment_term_name', 'invoice_period', 'usage_detail_fields', 'invoice_credit_amount',
            'unpaid_amount', 'invoice_num', 'company_type','pay_amount'
        )

class InvoiceGetScheme(InvoiceScheme):
    company_type = Str(validate=validate.OneOf(('termination',  'origination', 'both termination and origination')))
    invoice_payment = Float()
    client_credit = Float()
    client_offset = Float()
    credit_note = Float()
    invoice_period = Str()
    class Meta:
        model = model.Invoice
        fields = (
            'invoice_id', 'invoice_start', 'invoice_end', 'invoice_period','invoice_number', 'invoice_time', 'state',
            'client_id', 'client_name', 'company_name', 'amount', 'paid', 'due_date', 'url', 'status', 'create_type',
            'last_invoice_for', 'payment_term_name', 'invoice_period', 'usage_detail_fields', 'invoice_credit_amount',
            'unpaid_amount', 'invoice_num', 'company_type','data_url','invoice_payment','client_credit','client_offset',
            'credit_note',
        )
        search_fields = ('invoice_id', 'invoice_number', 'state', 'client_id', 'client_name', 'company_name',
                         'paid', 'status', 'create_type', 'payment_term_name', 'invoice_period', 'company_type',
                         'invoice_period'
                         )
        query_fields = (
            'start_date_gt', 'start_date_lt', 'end_date_gt', 'end_date_lt', 'invoice_time_gt', 'invoice_time_lt',
            'due_date_gt', 'due_date_lt', 'invoice_id_in', 'last_invoice_for_gt', 'last_invoice_for_lt', 'amount_gt',
            'invoice_start_gt', 'invoice_start_lt', 'invoice_end_gt', 'invoice_end_lt',)
# 'invoice_period_from_gt','invoice_period_from_lt','invoice_period_to_gt','invoice_period_to_lt')


class InvoiceFileGetScheme(InvoiceGetScheme):
    class Meta:
        model = model.Invoice
        fields = ('invoice_id',)

class InvoiceDataSchemeGet(InvoiceScheme):
    data = Dict(attribute='json_content')
    class Meta:
        model = model.InvoiceSummary
        fields = ('data',)
class InvoiceEmailScheme(BaseModelScheme):
    email = Emails()
    class Meta:
        model = model.Invoice
        fields = ('email',)

#endregion
# region +++InvoiceSummary+++
class InvoiceSummaryScheme(BaseModelScheme):
    client_name = Str(validate=[validate.Length(max=500)])
    invoice_number = Int()
    invoice_num = Str(validate=[validate.Length(max=256)])
    json_content = Dict()
    client_id = Int()
    invoice_generation_time = DateTime()
    invoice_start_time = DateTime()
    invoice_end_time = DateTime()
    task_id = Int()
    payment_id = Int()
    current_balance = Float()
    previous_balance = Float()
    payments_and_credits = Float()
    balance_forward = Float()
    long_distance_charges = Float()
    taxes = Float()
    scc_calls = Float()
    port_charges = Float()
    scc_amount = Float()
    current_charges = Float()
    total_calls = Int()
    total_minutes = Float()
    total_amount = Float()
    unlimited_credit = Bool()
    invoice_type = Choice()
    did_mrc = Float()
    did_nrc = Float()
    create_time = DateTime()
    task = Nested('InvoiceTaskScheme', many=False)
    client = Nested('ClientScheme', many=False)

    class Meta:
        model = model.InvoiceSummary
        fields = (
            'invoice_num', 'json_content', 'client_id', 'invoice_generation_time', 'invoice_start_time',
            'invoice_end_time',
            'task_id', 'payment_id', 'current_balance', 'previous_balance', 'payments_and_credits', 'balance_forward',
            'long_distance_charges', 'taxes', 'scc_calls', 'port_charges', 'scc_amount', 'current_charges',
            'total_calls',
            'total_minutes', 'total_amount', 'unlimited_credit', 'invoice_type', 'did_mrc', 'did_nrc', 'create_time',)


class InvoiceSummarySchemeGet(InvoiceSummaryScheme):
    class Meta:
        model = model.InvoiceSummary
        fields = (
            'invoice_num', 'json_content', 'client_id', 'invoice_generation_time', 'invoice_start_time',
            'invoice_end_time',
            'task_id', 'payment_id', 'current_balance', 'previous_balance', 'payments_and_credits', 'balance_forward',
            'long_distance_charges', 'taxes', 'scc_calls', 'port_charges', 'scc_amount', 'current_charges',
            'total_calls',
            'total_minutes', 'total_amount', 'unlimited_credit', 'invoice_type', 'did_mrc', 'did_nrc', 'create_time',)
        search_fields = (
            'client_name', 'invoice_number', 'invoice_num', 'json_content', 'client_id', 'task_id', 'payment_id',
            'total_calls', 'unlimited_credit', 'invoice_type', 'task', 'did_invoice_charge_detail',
            'did_invoice_charge_breakdown', 'did_invoice_number_fee_detail', 'did_mrc_info', 'did_setup_fee',
            'did_port_usage_info', 'invoice_code_name_detail', 'invoice_country_detail', 'invoice_trunk_detail',
            'invoice_trunk_prefix_detail', 'invoice_daily_detail', 'client',)
        query_fields = (
            'invoice_generation_time_gt', 'invoice_generation_time_lt', 'invoice_start_time_gt',
            'invoice_start_time_lt',
            'invoice_end_time_gt', 'invoice_end_time_lt', 'current_balance_gt', 'current_balance_lt',
            'previous_balance_gt',
            'previous_balance_lt', 'payments_and_credits_gt', 'payments_and_credits_lt', 'balance_forward_gt',
            'balance_forward_lt', 'long_distance_charges_gt', 'long_distance_charges_lt', 'taxes_gt', 'taxes_lt',
            'scc_calls_gt', 'scc_calls_lt', 'port_charges_gt', 'port_charges_lt', 'scc_amount_gt', 'scc_amount_lt',
            'current_charges_gt', 'current_charges_lt', 'total_minutes_gt', 'total_minutes_lt', 'total_amount_gt',
            'total_amount_lt', 'did_mrc_gt', 'did_mrc_lt', 'did_nrc_gt', 'did_nrc_lt', 'create_time_gt',
            'create_time_lt',)


class InvoiceSummarySchemeGet(InvoiceSummaryScheme):
    class InvoiceSummarySchemeModify(InvoiceSummaryScheme):
        pass


class InvoiceSummarySchemeModify(InvoiceSummaryScheme):
    pass


class InvoiceSummaryDataSchemeGet(InvoiceSummaryScheme):
    data = Dict(attribute='json_content')

    class Meta:
        model = model.InvoiceSummary
        fields = ('data',)
        search_fields = (
            'client_name', 'invoice_number', 'invoice_num', 'json_content', 'client_id', 'task_id', 'payment_id',
            'total_calls', 'unlimited_credit', 'invoice_type', 'task', 'did_invoice_charge_detail',
            'did_invoice_charge_breakdown', 'did_invoice_number_fee_detail', 'did_mrc_info', 'did_setup_fee',
            'did_port_usage_info', 'invoice_code_name_detail', 'invoice_country_detail', 'invoice_trunk_detail',
            'invoice_trunk_prefix_detail', 'invoice_daily_detail', 'client',)
        query_fields = (
            'invoice_generation_time_gt', 'invoice_generation_time_lt', 'invoice_start_time_gt',
            'invoice_start_time_lt',
            'invoice_end_time_gt', 'invoice_end_time_lt', 'current_balance_gt', 'current_balance_lt',
            'previous_balance_gt',
            'previous_balance_lt', 'payments_and_credits_gt', 'payments_and_credits_lt', 'balance_forward_gt',
            'balance_forward_lt', 'long_distance_charges_gt', 'long_distance_charges_lt', 'taxes_gt', 'taxes_lt',
            'scc_calls_gt', 'scc_calls_lt', 'port_charges_gt', 'port_charges_lt', 'scc_amount_gt', 'scc_amount_lt',
            'current_charges_gt', 'current_charges_lt', 'total_minutes_gt', 'total_minutes_lt', 'total_amount_gt',
            'total_amount_lt', 'did_mrc_gt', 'did_mrc_lt', 'did_nrc_gt', 'did_nrc_lt', 'create_time_gt',
            'create_time_lt',)


# endregion ---InvoiceSummary---
# region +++InvoiceTask+++
class InvoiceTaskScheme(BaseModelScheme):
    client_name = Str(validate=[validate.Length(max=500)])
    client_type = Choice()
    id = Int()
    client_id = Int(validate=lambda value: _valid('Client', 'client_id', value))
    is_auto = Bool()
    created_by_user_id = Int()
    created_on = DateTime()
    bill_period_start = Str(validate=valid_date)
    bill_period_end = Str(validate=valid_date)
    show_acc_summary = Bool()
    show_summary_of_charges = Bool()
    show_payment_applied = Bool()
    show_rec_charge = Bool()
    show_non_rec_charge = Bool()
    show_usage_charge = Bool()
    show_trunk_detail = Bool()
    show_daily_summary = Bool()
    show_jd_daily_summary = Bool()
    show_jd_trunk_detail = Bool()
    show_prefix_summary = Bool()
    show_code_name_summary = Bool()
    show_country_summary = Bool()
    show_inbound_summary = Bool()
    show_top_100tn_summary = Bool()
    show_did_charge_breakdown = Bool()
    show_did_charge_summary = Bool()
    enable_email = Bool()
    enable_email_with_cdr = Bool()
    include_tax = Bool()
    include_scc_charges = Bool()
    payment_dur_date = Str(validate=valid_date)
    invoice_zone = Str(validate=[validate.Length(max=20)])
    status = Choice()
    progress = Str(validate=[validate.Length(max=1024)])
    proc_start_time = DateTime()
    proc_end_time = DateTime()
    summary = Nested('InvoiceSummaryScheme', many=False)

    class Meta:
        model = model.InvoiceTask
        fields = ('client_id', 'bill_period_start', 'bill_period_end', 'show_acc_summary', 'show_summary_of_charges',
                  'show_payment_applied',
                  'show_rec_charge', 'show_non_rec_charge', 'show_usage_charge', 'show_trunk_detail',
                  'show_daily_summary', 'show_jd_daily_summary', 'show_jd_trunk_detail', 'show_prefix_summary',
                  'show_code_name_summary', 'show_country_summary', 'show_inbound_summary', 'show_top_100tn_summary',
                  'enable_email', 'enable_email_with_cdr', 'include_tax', 'include_scc_charges', 'payment_dur_date',
                  'invoice_zone', 'show_did_charge_breakdown', 'show_did_charge_summary')


class InvoiceTaskSchemeGet(InvoiceTaskScheme):
    invoice_number = Int()
    invoice_num = Str()

    class Meta:
        model = model.InvoiceTask
        fields = ('id','client_id', 'is_auto',
                  'created_on', 'bill_period_start', 'bill_period_end',
                   'status', 'progress','proc_start_time', 'proc_end_time', 'invoice_number', 'invoice_num','client_name')
        search_fields = (
            'id','client_id', 'is_auto',  'created_on', 'bill_period_start', 'bill_period_end',
            'status', 'progress','proc_start_time', 'proc_end_time', 'invoice_number', 'invoice_num','client_name')
        query_fields = (
            'created_on_gt', 'created_on_lt', 'proc_end_time_gt',
            'proc_end_time_lt',)


class InvoiceTaskSchemeModify(InvoiceTaskScheme):
    pass

# endregion ---InvoiceTask---
