import os
from sqlalchemy import (Column, event, distinct, desc, or_, not_, and_, text, PrimaryKeyConstraint, inspect,
                        Index, UniqueConstraint, literal, ForeignKey)
from . import model, settings, scheme
from .scheme import DATE_REGEXP
import falcon
import mimetypes
from urllib.parse import parse_qsl
from .resources.resource import DEFAULT_SECURITY, Create, Resource, List, CustomAction, CustomGetAction, \
    CustomPostAction
from . import responses
from . import swagger
from .logger import log
from datetime import datetime, timedelta, date
from dateutil.parser import parse as parse_datetime
from pytz import UTC
from sqlalchemy import and_, or_, inspect, alias, text, func, select, case, cast
from sqlalchemy import Integer
import json


def datetime_handler(self, x):
    if isinstance(x, datetime):
        return x.isoformat()
    if isinstance(x, date):
        return x.isoformat()
    raise TypeError("Unknown type")


json.JSONEncoder.default = datetime_handler


# region ---InvoiceTask---

class InvoiceTaskCreate(Create):
    scheme_class = scheme.InvoiceTaskScheme
    model_class = model.InvoiceTask
    entity = 'InvoiceTask'
    path_parameters = ()
    no_auth_needed = True

    # security = (DEFAULT_SECURITY)
    # restrict = ()

    def before_create(self, obj, **kwargs):
        # user = self.get_user(self.req)
        # obj.created_by_user_id = user.user_id
        conf = model.SystemParameter.query().first()
        if obj.client_id:
            clscnf = model.ClientInvoiceSetting
            invoice_setting = clscnf.filter(clscnf.client_id == obj.client_id).first()
            if invoice_setting:
                conf = invoice_setting.setting
        if conf and conf.overlap_invoice_protection:
            cls = model.Invoice
            q = cls.filter(and_(cls.client_id == obj.client_id, cls.invoice_start >= obj.bill_period_start,
                                cls.invoice_start <= obj.bill_period_start, cls.state != -1)).first()
            if q:
                raise Exception(
                    'Overlap invoice {} start {} end {}'.format(q.invoice_id, q.invoice_start, q.invoice_end))
        obj.create_time = datetime.now(UTC)
        obj.status = 'Initial'
        obj.is_auto = False
        return obj

    def after_create(self, object_id, req, resp, **kwargs):
        from kombu.exceptions import OperationalError
        from inv5.tasks.invoice import do_invoice_task
        try:
            do_invoice_task.delay(object_id)
            log.debug('InvoiceTaskCreate scheduled success')
        except OperationalError as e:
            log.warning('InvoiceTaskCreate celery worker not run:{}'.format(e))
            do_invoice_task(object_id)
            log.debug('InvoiceTaskCreate finished success')


class InvoiceTaskResource(Resource):
    model_class = model.InvoiceTask
    scheme_class = scheme.InvoiceTaskScheme
    scheme_class_get = scheme.InvoiceTaskSchemeGet
    scheme_class_modify = scheme.InvoiceTaskSchemeModify
    entity = 'InvoiceTask'
    id_field = 'id'
    has_modify_operation = False
    has_delete_operation = False
    no_auth_needed = True


class InvoiceTaskList(List):
    scheme_class = scheme.InvoiceTaskSchemeGet
    model_class = model.InvoiceTask
    entity_plural = 'InvoiceTasks'
    no_auth_needed = True

    def on_filter_client_type(self, q, val, kwargs):
        cls = self.model_class
        if val == 'carrier':
            return q.filter(cls.client_type.is_(None))
        else:
            return q.filter(cls.client_type == val)


# endregion ---InvoiceTask---

class InvoiceFileGet(CustomGetAction):
    model_class = model.Invoice
    scheme_class = scheme.InvoiceGetScheme
    scheme_class_get = scheme.InvoiceFileGetScheme
    has_update_operation = False
    has_delete_operation = False
    entity = 'Invoice pdf file'
    no_auth_needed = True
    path_parameters = ({'name': 'invoice_number', 'description': 'invoice number'},)

    def get_spec(self):
        spec = swagger.specify.get_spec(
            method='get', description='Gets {}'.format(self.entity.lower()),
            path_parameters=self.get_path_parameters(),
            responses=(responses.SuccessResponseJustOk(), responses.ObjectNotFoundErrorResponse(),
                       responses.OperationErrorResponse()),
            security=self.get_security(method='get', action='info')
        )
        spec['get']['produces'] = ['application/pdf', 'text/html', 'application/json']
        return spec

    def on_get(self, req, resp, invoice_number):
        try:
            # if invoice_number.isnumeric():
            #     invoice_n = int(invoice_number)
            # else:
            #     cls = model.InvoiceSummary
            #     obj = cls.filter(cls.invoice_num==invoice_number).first()
            #     if obj:
            #         invoice_n = int(obj.invoice_number)
            #     else:
            #         self.set_response(resp, responses.ObjectNotFoundErrorResponse(
            #             data={'result': 'Invoice {} not found'.format(invoice_number)}))
            #         return
            cls = self.model_class
            obj = None
            if invoice_number.isnumeric():
                obj = cls.filter(cls.invoice_number == invoice_number).first()
            else:
                obj = cls.filter(cls.invoice_num == invoice_number).first()
            if obj:
                if obj.pdf_path:
                    file_path = '{}/{}'.format(settings.FILES['pdf_path'], obj.pdf_path.replace('invoice/',''))
                    if os.path.isfile(file_path):
                        file = open(file_path, 'rb')
                        # resp.data = file.read()
                        resp.body = file.read()
                        resp.content_type = mimetypes.guess_type(obj.pdf_path)[0]
                        filename = obj.pdf_path.split('/')[-1]
                        resp.append_header('Content-Disposition', 'attachment; filename="{}"'.format(filename))
                        resp.status = falcon.HTTP_200
                    else:
                        self.set_response(resp, responses.OperationErrorResponse(
                            data=responses.errors.Error(code=404, message='Invoice {} pdf file missing'.format(
                                invoice_number),
                                                        reason='file deleted from storage')
                        ))
                else:
                    self.set_response(resp, responses.OperationErrorResponse(
                        data=responses.errors.Error(code=404,
                                                    message='Invoice {} pdf file not generated'.format(invoice_number),
                                                    reason='')
                    ))
            else:
                self.set_response(resp, responses.ObjectNotFoundErrorResponse(
                    data={'result': 'Invoice {} not found'.format(invoice_number)}))

        except Exception as e:
            self.set_response(resp, responses.OperationErrorResponse(e))


class InvoiceList(List):
    scheme_class = scheme.InvoiceGetScheme
    model_class = model.Invoice
    entity_plural = 'Invoices'
    no_auth_needed = True

    # path_parameters=( {'name': 'client_id', 'description': 'Parent carrier'}, )

    # def get_filtering_for_list(self, parsed_qs, **kwargs):
    #    ret = super(DnlList, self).get_filtering_for_list(parsed_qs, **kwargs)
    #    ret['client_id'] = kwargs['client_id']
    #    return ret
    def on_filter_invoice_time_lt(self, query, value, kwargs):
        import re
        cls = self.model_class
        if re.match(DATE_REGEXP, value):
            value = value + ' 23:59:59'
        query = query.filter(cls.invoice_time <= value)
        return query

    def modify_query_from_filtering_for_list(self, filtering, **kwargs):
        cls = self.model_class
        filt, ret = super(InvoiceList, self).modify_query_from_filtering_for_list(filtering, **kwargs)
        # ret = ret.filter(model.user_id_filter(cls, self.get_user(self.req).user_id))
        # if 'did' in self.req.url:
        #     filt['company_type'] = 'origination'
        #     if not 'create_type' in self.req.url:
        #         filt['create_type']='manual'
        # else:
        #     filt['company_type'] = 'termination'
        return filt, ret

    def modify_query_from_ordering_for_list(self, ordering, query, **kwargs):
        model_class = self.model_class

        def ToInt(numb):
            try:
                return cast(func.nullif(func.regexp_replace(numb, "\\D", "", "g"), ""), Integer)
            except:
                return text("")

        if 'by' in ordering and ordering['by'] == 'invoice_number':
            model_class = self.model_class
            if ordering['dir'] == 'desc':
                query = query.order_by(ToInt(model.Invoice.invoice_number).desc())
            else:
                query = query.order_by(ToInt(model.Invoice.invoice_number))

            ordering = {}

        return ordering, query


class InvoiceResource(Resource):
    model_class = model.Invoice
    scheme_class = scheme.InvoiceScheme
    scheme_class_get = scheme.InvoiceGetScheme
    entity = 'Invoice'
    id_field = 'invoice_id'
    has_update_by = True
    has_update_operation = False
    has_delete_operation = False
    unique_field = 'invoice_number'
    no_auth_needed = True


class InvoiceDataGet(CustomGetAction):
    model_class = model.InvoiceSummary
    scheme_class = scheme.InvoiceDataSchemeGet
    path_parameters = ({'name': 'invoice_number', 'description': 'invoice md5 number'},)
    description = 'Get json invoice data'
    no_auth_needed = True

    def on_get(self, req, resp, invoice_number):
        cls = self.model_class
        obj = None
        if invoice_number.isnumeric():
            obj = cls.filter(cls.invoice_number == invoice_number).first()
        else:
            obj = cls.filter(cls.invoice_num == invoice_number).first()
        if obj:
            resp.data = json.dumps(obj.json_content).encode()
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_200
        else:
            resp.data = str.encode(json.dumps({'success': False, 'message': 'Not found'}))
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_404


class InvoiceAutoTest(CustomGetAction):
    model_class = model.Client
    scheme_class = scheme.InvoiceDataSchemeGet
    path_parameters = ({'name': 'client_id', 'description': 'client id'},

                       )
    query_parameters = (
        {'name': 'date', 'description': 'test date'},
    )
    description = 'Check future invoice will generated'
    no_auth_needed = True

    def apply(self, obj, req, resp, **kwargs):
        if obj:
            params = dict(parse_qsl(req.query_string))
            if 'date' in params:
                now = parse_datetime(params['date'])
                now.replace(tzinfo=UTC)
            else:
                now = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
            inv_data = []

            def invoice_generate(client_id, invoice_start_time, invoice_end_time):
                inv_data.append(dict(start_time=invoice_start_time, end_time=invoice_end_time))
                return None, 'success'

            cnt, msg = obj.invoice_check_and_generate(now, invoice_generate, False)
            last_invoice_date = obj.create_time
            if obj.last_invoice_history:
                last_invoice_date = obj.last_invoice_history.last_invoice_for

            data = dict(will_generate=cnt == 1, generated_period=inv_data, messages=msg,
                        invoice_cycle=obj.payment_term.cycle,
                        invoice_cycle_days=obj.payment_term.more_days_arr,
                        last_invoice_date=last_invoice_date,
                        invoice_next_date=obj.invoice_next_date,
                        invoice_next_date_of_test=obj.payment_term.next_date_for_now(last_invoice_date, now.date())
                        )
            resp.data = json.dumps(data).encode()
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_200
        else:
            resp.data = str.encode(json.dumps({'success': False, 'message': 'Not found'}))
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_404


class InvoiceAutoRestart(CustomGetAction):
    model_class = model.Client
    scheme_class = scheme.InvoiceDataSchemeGet
    path_parameters = ({'name': 'client_id', 'description': 'client id'},

                       )
    query_parameters = (
        {'name': 'date', 'description': 'restart date'},
    )
    description = 'invoice  on given date will generated'
    no_auth_needed = True

    def apply(self, obj, req, resp, **kwargs):
        if obj:
            params = dict(parse_qsl(req.query_string))
            if 'date' in params:
                now = parse_datetime(params['date'])
                now.replace(tzinfo=UTC)
            else:
                now = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
            inv_data = []

            def invoice_generate(client_id, invoice_start_time, invoice_end_time):
                inv_data.append(dict(start_time=invoice_start_time, end_time=invoice_end_time))
                return None, 'success'

            cnt, msg = obj.invoice_check_and_generate(now, invoice_generate, False)
            last_invoice_date = obj.create_time
            if obj.last_invoice_history:
                last_invoice_date = obj.last_invoice_history.last_invoice_for

            data = dict(will_generate=cnt == 1, generated_period=inv_data, messages=msg,
                        invoice_cycle=obj.payment_term.cycle,
                        invoice_cycle_days=obj.payment_term.more_days_arr,
                        last_invoice_date=last_invoice_date,
                        invoice_next_date=obj.invoice_next_date,
                        invoice_next_date_of_test=obj.payment_term.next_date_for_now(last_invoice_date, now.date())
                        )
            #schedule restart
            try:
                from inv5.tasks.invoice import do_auto_invoice_restart_task
                from kombu.exceptions import OperationalError
                try:
                    do_auto_invoice_restart_task.delay(now,obj.client_id)
                except OperationalError:
                    log.debug('no celery, run online')
                    do_auto_invoice_restart_task(now,obj.client_id)
            except Exception as e:
                data = dict(will_generate=False,messages=str(e))

            resp.data = json.dumps(data).encode()
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_200
        else:
            resp.data = str.encode(json.dumps({'success': False, 'message': 'Not found'}))
            resp.content_type = 'application/json'
            resp.status = falcon.HTTP_404

class InvoiceSend(CustomPostAction):
    model_class = model.Invoice
    scheme_class = scheme.InvoiceEmailScheme
    path_parameters = ({'name': 'invoice_number', 'description': 'invoice to send'},
                       )
    body_parameters = ('Email list to send',scheme.InvoiceEmailScheme)
    description = 'Email invoice to client with current settings'
    no_auth_needed = True

    #def apply(self, obj, req, resp, invoice_number):
    def on_post(self, req, resp, invoice_number):
        from kombu.exceptions import OperationalError
        from inv5.tasks.invoice import do_invoice_send
        cls = self.model_class
        if invoice_number.isnumeric():
            obj = cls.filter(cls.invoice_number == invoice_number).first()
        else:
            obj = cls.filter(cls.invoice_num == invoice_number).first()
        email = req.data.get('email', None)
        log.debug('InvoiceSend request {}'.format(req.data))
        if obj:
            try:
                email=req.data.get('email',None)
                do_invoice_send.delay(invoice_number,email)
                log.debug('InvoiceSend scheduled success')
                self.set_response(resp, responses.SuccessResponseObjectInfo(
                    data={'message': 'Invoice {} scheduled for send'.format(invoice_number)}))
            except OperationalError as e:
                log.warning('InvoiceSend celery worker not run:{}'.format(e))
                do_invoice_send(invoice_number,email)
                log.debug('InvoiceSend finished success')
                self.set_response(resp, responses.SuccessResponseObjectInfo(
                    data={'message': 'Invoice {} was sent'.format(invoice_number)}))
            except Exception as e:
                self.set_response(resp, responses.OperationErrorResponse(e))
                return True
        else:
            self.set_response(resp, responses.ObjectNotFoundErrorResponse(
                data={'result': 'Invoice {} not found'.format(invoice_number)}))



