import calendar
import random
import traceback
from datetime import date, datetime, timedelta
from decimal import Decimal

import bcrypt
import jwt
import requests
from dateutil.parser import parse as parse_datetime
from marshmallow import ValidationError
from pytz import UTC
from sqlalchemy import (Column, or_, and_, text, PrimaryKeyConstraint, Index, UniqueConstraint)
from sqlalchemy import (
    Integer, SmallInteger, Float, Text, String, DateTime, Date, Time, Boolean, ForeignKey, BigInteger, Table, Sequence,
    JSON
)
from sqlalchemy import (cast)
from sqlalchemy.dialects import postgresql
from sqlalchemy.exc import OperationalError, InternalError
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import foreign, relationship, synonym, column_property, aliased
from sqlalchemy.sql import func, select, literal_column, case
from urllib3.util import parse_url

from . import settings
from .const import COUNTRY, FIELDS, UIFIELDS
from .db import BaseModel, get_db, errors
# from falcon_rest.db import Column
from .db.fields import Numeric, ChoiceType, PrefixRange, Ip4
from .logger import log
from .mail import MailSender, EmailLog

default_billing_decimal = None


def get_default_billing_decimal():
    global default_billing_decimal
    if default_billing_decimal is None:
        config = SystemParameter.query().session.query(SystemParameter.default_billing_decimal).first()
        default_billing_decimal = config.default_billing_decimal
        if default_billing_decimal is None:
            default_billing_decimal = 2
    return default_billing_decimal


def get_time_zone_hour(time_zone):
    if time_zone is None:
        return 0
    if '(' in time_zone:
        time_zone = time_zone.replace('(', '').replace(')', '')
    if ':' in time_zone:
        time_zone = time_zone.replace(':', '')
    if time_zone[0] not in ('+','-'):
        time_zone = '+'+time_zone
    sign = time_zone[0]
    hour = int(time_zone[1:3])
    mins = 0.0
    if len(time_zone)>3:
        if time_zone[3:].isdigit():
            mins = int(time_zone[3:])
            mins = (1 / 60) * mins
    hour = hour + mins
    if sign == "+":
        hour = hour
    else:
        hour = -1 * hour
    return hour

def get_time_zone(str_time_zone):
    from datetime import timezone
    h=get_time_zone_hour(str_time_zone)
    return timezone(timedelta(hours=h))


def dec_prn(v, d=None):
    from decimal import Decimal
    if d is None:
        d = get_default_billing_decimal()
    if d is None or d not in range(0, 16):
        d = 2
    if type(v) == type(float):
        v = Decimal(v)
    if isinstance(v, Decimal):
        if v < 0:
            fmt = '({0:.' + str(d) + 'f})'
            return fmt.format(abs(v))
        fmt = '{0:.' + str(d) + 'f}'
        return fmt.format(v)
    return str(v)


def query_to_sting(q):
    try:
        return str(q.statement.compile(dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}))
    except:
        return str(q)


def rev(d):
    return dict((v, k) for k, v in d.items())


def init_db(conn_string=settings.DB_CONN_STRING, db_label='default', declarative_base=BaseModel):
    from .db import _DB, _databases
    from .settings import DB_CONN_STRING, DB_POOLSIZE
    import sqlalchemy
    conn_string = DB_CONN_STRING
    log.info('DB [{}]: Initializing DB, label {}'.format(conn_string, db_label))
    print('DB [{}]: Initializing DB, label {}'.format(conn_string, db_label))
    _databases[db_label] = _DB(declarative_base, db_label)
    _databases[db_label].engine = sqlalchemy.create_engine(conn_string, pool_size=int(DB_POOLSIZE))
    _databases[db_label].create_session()
    return _databases[db_label]


# noinspection PyClassHasNoInit,PyAbstractClass
class AuthUser:  # AuthUserBase):
    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def can_restore(self, object_revision):
        raise NotImplementedError('Method "can_restore" must be implemented in any derived class')

    def get_id(self):
        raise NotImplementedError('Method "get_id" must be implemented in any derived class')


class User(BaseModel, AuthUser):
    __tablename__ = 'users'
    USER_TYPE_CHOICES = [
        (1, 'admin'),
        (3, 'client'),
        (2, 'agent'),
        (None, 'viewer')
    ]
    user_id = Column(Integer, primary_key=True)
    name = Column(String(60), nullable=False, unique=True)
    password = Column(String(60), nullable=False)
    role_id = Column(ForeignKey('role.role_id', ondelete='SET NULL'), index=True)
    reseller_id = Column(Integer)
    create_time = Column(DateTime(timezone=True), nullable=False, server_default=func.now())
    active = Column(Boolean, nullable=False, default=True)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'), nullable=True, index=True)
    email = Column(String(100))
    # first_name = Column(String(40))
    # last_name = Column(String(40))
    # user_type = Column(Integer, nullable=False)
    user_type = Column(ChoiceType(USER_TYPE_CHOICES), nullable=False)
    create_user_id = Column(Integer)
    last_login_time = Column(DateTime(timezone=True))
    card_id = Column(Integer)
    is_online = Column(Integer, nullable=False, default=2)
    login_ip = Column(String(36))
    default_mod = Column(String(100), default='')
    default_mod2 = Column(String(100), default='')
    last_seen = Column(DateTime(timezone=True))
    report_group = Column(Boolean, nullable=False, default=True)
    outbound_report = Column(Boolean, nullable=False, default=True)
    all_termination = Column(Boolean, nullable=False, default=True)
    show_carrier_trunk_drop_only = Column(Boolean, nullable=False, default=False)
    report_fields = Column(Text)
    # landing_page = Column(Text)
    # cdr_api_token = Column(String(64))
    # cdr_expire = Column(DateTime(timezone=True))
    # avatar_id = Column(Integer)
    # last_notification_time = Column(DateTime(timezone=True))

    # fullname = column_property(first_name + ' ' + last_name)

    login_url = synonym('landing_page')

    # role = relationship('Role', back_populates='users', uselist=False)
    agent = relationship('Agent', primaryjoin='foreign(Agent.user_id) == User.user_id', back_populates='user',
                         uselist=False)
    client = relationship('Client', primaryjoin='foreign(Client.user_id) == User.user_id', back_populates='user',
                          uselist=False, enable_typechecks=False)

    client_limits = relationship('UsersLimit', uselist=True, back_populates='user', single_parent=True,
                                 cascade="all, delete-orphan")

    # auth_ip = relationship('UserAuthIp', uselist=True, back_populates='user', single_parent=True,
    #                        cascade="all, delete-orphan", lazy="dynamic")
    #
    # codedeck_alerts = relationship('UserCodedeckAlerts', uselist=False, back_populates='user',
    #                                cascade="all, delete-orphan")
    #
    # api_key = relationship('UserApiKey', uselist=False, cascade="all, delete-orphan")

    def before_save(self):
        pass
        # if not self.user_id:
        # if self.user_type == 'admin':
        #     if not self.codedeck_alerts:
        #         ar = UserCodedeckAlerts(show_az_codedeck=True, show_us_codedeck=True)
        #         self.codedeck_alerts = ar
        # else:
        #     self.codedeck_alerts = None

    @property
    def is_admin(self):
        return self.user_type == 'admin'

    @property
    def soon_expired_gateways(self):
        return None
        # from inv5.model import VoipGateway
        # if self.is_admin:
        #     return VoipGateway.expired7()
        # else:
        #     return None

    @property
    def is_active(self):
        if self.user_type == 'client' and self.client:
            return self.active and self.client.status
        if self.user_type == 'agent' and self.agent:
            return self.active and self.agent.status
        return self.active

    @property
    def referal_key(self):
        if self.user_type == 'agent' and self.agent:
            return self.agent.referal_key
        return None

    @hybrid_property
    def passwd(self):
        return self.password

    @staticmethod
    def get_hashed_password(plain_text_password):
        if not plain_text_password:
            plain_text_password = str(round(random.random() * 1000000000, 0))
        if isinstance(plain_text_password, type('')):
            plain_text_password = plain_text_password.encode()
        hash = bcrypt.hashpw(
            plain_text_password,
            bcrypt.gensalt()
        )
        if isinstance(hash, type(b'')):
            hash = hash.decode()
        return hash

    @passwd.setter
    def passwd(self, value):
        self.password = self.get_hashed_password(value)

    @classmethod
    def get_user_from_credentials(cls, credentials):
        try:
            user = cls.filter(
                or_(User.email == credentials['email_or_name'], User.name == credentials['email_or_name'])
            ).first()
            if user:  # .is_active:
                return user
            return None
        except OperationalError as e:  # errors.NoResultFound:
            log.error('database error {}'.format(e))
            # cls.query().session.rollback()
            raise e
        except InternalError as e:  # errors.NoResultFound:
            log.error('database error {}'.format(e))
            cls.query().session.rollback()
            raise e

    def get_token_data(self):
        return dict(user_id=self.user_id)

    @classmethod
    def get_user_from_token_data(cls, token_data):
        try:
            user = cls.filter(User.user_id == token_data['user_id']).filter(User.active == True).one()
            if user.is_active:
                return user
            return None
        except:
            cls.query().session.rollback()
            return None

    @classmethod
    def get_user_by_id(cls, user_id):
        try:
            user = cls.filter(User.user_id == user_id).filter(User.active == True).one()
            if user.is_active:
                return user
            return None
        except errors.NoResultFound:  # pragma: no cover
            return None

    def can_login(self, req, resp):
        # resources.BaseResource.set_response(
        #     resp, resources.responses.UnAuthorizedErrorResponse(
        #         data=resources.errors.Error(
        #             code,
        #             msg,
        #             'user not active'
        #         )
        #     )
        # )
        return False

    def can_login_ip(self, req, resp):
        # request_ip = helpers.get_request_ip(req)
        # if request_ip in settings.ALWAYS_ALLOWED_AUTH_IPS:
        #     return True
        # if self.auth_ip.count():
        #     match = self.auth_ip.filter(UserAuthIp.ip == request_ip).first()
        #     if match:
        #         return True
        # else:
        #     return True
        #
        # resources.BaseResource.set_response(
        #     resp, resources.responses.UnAuthorizedErrorResponse(
        #         data=resources.errors.Error(
        #             1003,
        #             'You are not allowed to login from the IP {}'.format(request_ip),
        #             'ip_not_allowed'
        #         )
        #     )
        # )
        return False

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def can_restore(self, object_revision):
        return True

    def get_id(self):
        return self.user_id

    # @hybrid_property
    # def role_name(self):
    #     try:
    #         return self.role.role_name
    #     except:
    #         return ''

    def cdr_token(self):
        return None, None, None

        # from api_dnl.utils.statisticapi2 import StatisticAPI
        api = StatisticAPI()
        ingress_id = None
        egress_id = None

        if self.user_type in ['client', 'agent']:
            if self.allowed_ingress_id:
                ingress_id = ','.join([str(i) for i in self.allowed_ingress_id])
            else:
                if self.allowed_egress_id:
                    egress_id = ','.join([str(i) for i in self.allowed_egress_id])
                else:
                    ingress_id = '-1'
        else:
            if self.allowed_egress_id:
                egress_id = ','.join([str(i) for i in self.allowed_egress_id])

        all_fields = DailyCdrField.query().filter(DailyCdrField.id.notin_([96, 108])).all()
        allowed_fields = None
        if self.user_type in ['client', 'agent']:
            allowed_fields = ','.join([str(f.id) for f in all_fields if f.client_viewable])
        elif self.user_type in ['vendor']:
            allowed_fields = ','.join([str(f.id) for f in all_fields if f.vendor_viewable])
        cdr, exp = None, 0
        try:
            if egress_id:
                if ingress_id:
                    cdr, exp = api.authenticate(user=self.name, allowed_ingress_id=ingress_id,
                                                allowed_egress_id=egress_id, allowed_fields=allowed_fields)
                else:
                    cdr, exp = api.authenticate(user=self.name,
                                                allowed_egress_id=egress_id, allowed_fields=allowed_fields)
            else:
                if ingress_id:
                    cdr, exp = api.authenticate(user=self.name, allowed_ingress_id=ingress_id,
                                                allowed_fields=allowed_fields)
                else:
                    cdr, exp = api.authenticate(user=self.name, allowed_fields=allowed_fields)
        except Exception as e:
            log.error('user get cdr token error {}'.format(str(e)))
            cdr_term, exp = None, None
        if cdr:
            self.cdr_api_token = cdr
            self.cdr_expire = datetime.utcfromtimestamp(exp)
            User.session().add(self)
        return cdr, exp, cdr

    @hybrid_property
    def clients_id(self):

        if self.name in ['admin', '#icx_api']:
            return [l.client_id for l in Client.query().all()]
        else:
            if self.user_type == 'admin':
                return [l.client_id for l in self.client_limits]
            elif self.user_type == 'agent':
                if self.agent and self.agent.carriers:
                    return [l.client_id for l in self.agent.carriers]
            else:  # self.user_type=='client':
                if self.client:
                    return list(set([self.client.client_id] + [l.client_id for l in self.client_limits]))
        return []

    @property
    def allowed_ingress_id(self):

        res = Resource
        if self.user_type == 'admin':
            return None
        if self.clients_id:
            return [r.resource_id for r in res.filter(and_(res.client_id.in_(self.clients_id), res.ingress == True))]
        return [0]

    @property
    def allowed_egress_id(self):
        res = Resource
        if self.user_type in ('admin'):
            return None
        if self.clients_id:
            return [r.resource_id for r in res.filter(and_(res.client_id.in_(self.clients_id), res.egress == True))]
        return None

    @property
    def _client_id(self):
        q = Client.filter(Client.user_id == self.user_id).first()
        if q:
            return q.client_id
        return None

    @property
    def default_billing_decimal(self):

        config = SystemParameter.get(1)
        return config.default_billing_decimal

    @property
    def is_preload(self):

        config = SystemParameter.get(1)
        return config.is_preload

    @property
    def report_count(self):

        config = SystemParameter.get(1)
        return config.report_count

    def check_api_key(self, token):
        pass

    def create_api_key(self):
        pass


class Agent(BaseModel):
    __tablename__ = 'agent'
    METHOD_TYPE_DICT = {0: 'By Profit', 1: 'By Revenue'}
    FREQUENCY_TYPE_DICT = {0: 'daily', 1: 'weekly', 2: 'monthly'}

    agent_id = Column(Integer, primary_key=True)
    agent_name = Column(String(100), unique=True, nullable=False)
    create_on = Column(DateTime(timezone=True), nullable=False, server_default=func.now())
    email = Column(String, nullable=False)
    commission = Column(Float)
    # method_type = Column(Integer, nullable=False)  # 0 - By Profit, 1 - By Revenue
    method_type = Column(ChoiceType(METHOD_TYPE_DICT), nullable=False)
    frequency_type = Column(ChoiceType(FREQUENCY_TYPE_DICT), nullable=False)  # 0 - Daily, 1 - Weekly, 2 - monthly
    user_id = Column(Integer)
    agent_role_id = Column(Integer)
    status = Column(Boolean, default=True)
    update_on = Column(DateTime(timezone=True), nullable=True, onupdate=func.now())
    update_by = Column(String(100), nullable=True)
    edit_permission = Column(Boolean, default=True)
    invoice_setting_id = Column(ForeignKey('invoice_settings.id', ondelete='CASCADE'), nullable=True)
    mail_sender_id = Column(ForeignKey('mail_sender.id', ondelete='CASCADE'), nullable=True)

    carriers = relationship('AgentClient', back_populates='agent', uselist=True, single_parent=True,
                            cascade="all, delete-orphan")
    user = relationship('User', primaryjoin='foreign(Agent.user_id) == User.user_id',
                        back_populates='agent', uselist=False, single_parent=True,
                        cascade="all, delete-orphan"
                        )

    @property
    def ingress_ids(self):
        res = Resource
        return ','.join([str(r.resource_id) for r in res.filter(
            and_(res.client_id.in_([c.client_id for c in self.carriers]), res.ingress == True))])

    @property
    def egress_ids(self):
        res = Resource
        return ','.join([str(r.resource_id) for r in res.filter(
            and_(res.client_id.in_([c.client_id for c in self.carriers]), res.egress == True))])

    # usage_count = column_property
    @hybrid_property
    def user_name(self):
        return self.user.name

    def create_user(self):
        if not self.user:
            self.user = User(name=self.agent_name, passwd='agent')
            self.user.email = self.email
            self.user.user_type = 'agent'
            self.user.role_id = 2
            # self.user.client_id = self.client_id
            return self.user.save()
        else:
            return self.user_id

    @property
    def referal_key(self):
        import jwt
        exp = (datetime.now(UTC) + timedelta(hours=2)).timestamp()
        token_data = {'agent_id': self.agent_id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE, algorithm='HS256').decode('utf-8')
        return token

    @property
    def referal_link(self):
        config = SystemParameter.get(1)
        return '{}/#/referral/{}'.format(config.root_url, self.referal_key)

    @classmethod
    def get_from_referral(cls, token):
        import jwt
        try:
            data = jwt.decode(token, settings.JWT_SIGNATURE)
            return cls.get(data['agent_id'])
        except Exception as e:
            log.debug('cannot get agent form referal-key {} error is {}'.format(token, str(e)))
            return None

    def after_save(self):
        for ac in AgentClient.filter(AgentClient.agent_id == self.agent_id).all():
            ac.client.save()


class AgentClient(BaseModel):
    __tablename__ = 'agent_clients'
    __table_args__ = (UniqueConstraint('agent_id', 'client_id'),)
    METHOD_TYPE_DICT = {0: 'By Profit', 1: 'By Revenue'}

    id = Column(Integer, primary_key=True)
    agent_id = Column(ForeignKey('agent.agent_id', ondelete='CASCADE'), nullable=False)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'), nullable=False)
    # commission = Column(Numeric(5, 2))
    assigned_on = Column(DateTime(timezone=True), server_default=func.now())
    assigned_by = Column(String(100))
    method_type = Column(ChoiceType(METHOD_TYPE_DICT), nullable=False, default=0)

    agent = relationship(Agent, back_populates='carriers', uselist=False)
    client = relationship('Client', back_populates='agents', uselist=False)

    registered_on = synonym('assigned_on')

    agent_name = column_property(
        select([Agent.agent_name]).where(Agent.agent_id == agent_id).correlate_except(Agent))
    commission = column_property(
        select([Agent.commission]).where(Agent.agent_id == agent_id).correlate_except(Agent))

    def _method_type(self):
        return self.client.profit_type


Agent.usage_count = column_property(select([func.coalesce(func.count(AgentClient.client_id), text("0"))]). \
    where(AgentClient.agent_id == Agent.agent_id).correlate_except(AgentClient).label(
    'usage_count'))


class UsersLimit(BaseModel):
    __tablename__ = 'users_limit'

    id = Column(Integer, primary_key=True, server_default=text("nextval('users_limit_id_seq'::regclass)"))
    user_id = Column(ForeignKey('users.user_id', ondelete='CASCADE'))
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'))

    client = relationship('Client', uselist=False, back_populates='client_limits')
    user = relationship('User', uselist=False, back_populates='client_limits')

    @hybrid_property
    def client_name(self):
        return self.client.name


def user_id_filter(cls, user_id):
    return or_(and_(User.user_id == user_id, User.user_type == 'admin'),
               and_(User.user_id == user_id, User.user_type == 'client', cls.client_id == User.client_id),
               cls.client_id.in_(select([UsersLimit.client_id]).where(
                   and_(User.user_id == user_id, User.user_type == 'client',
                        User.user_id == UsersLimit.user_id))),
               cls.client_id.in_(select([AgentClient.client_id]).where(
                   and_(User.user_id == user_id, User.user_type == 'agent', User.user_id == Agent.user_id,
                        Agent.agent_id == AgentClient.agent_id)))
               )


class CurrencyUpdates(BaseModel):
    __tablename__ = 'currency_updates'
    currency_updates_id = Column(Integer, primary_key=True)
    currency_id = Column(ForeignKey('currency.currency_id', ondelete='SET NULL'), nullable=False, index=True)
    rate = Column(Float, nullable=False, default=1)
    currency = relationship('Currency')


class Currency(BaseModel):
    __tablename__ = 'currency'

    currency_id = Column(Integer, primary_key=True)
    code = Column(String(100), nullable=False, unique=True)
    # rate = Column(Float, nullable=False, default=1)#, server_default=func.text('1.0'))
    active = Column(Boolean, nullable=False, default=True)
    update_by = Column(String)
    rates = relationship('CurrencyUpdates', back_populates='currency', uselist=True, single_parent=True,
                         cascade="all, delete-orphan")

    @hybrid_property
    def rate(self):
        try:
            rate = CurrencyUpdates.filter(CurrencyUpdates.currency_id == self.currency_id). \
                filter(CurrencyUpdates.currency_updates_id == CurrencyUpdates.query().
                       session.query(func.max(CurrencyUpdates.currency_updates_id)).filter(
                CurrencyUpdates.currency_id == self.currency_id)).first().rate
            return rate
        except:
            return None

    @rate.setter
    def rate(self, value):
        try:
            if not value is None and not self.currency_id is None:
                r = CurrencyUpdates(currency_id=self.currency_id, rate=value)
                self.rates.append(r)
        except:
            pass


class Code(BaseModel):
    __tablename__ = 'code'
    __table_args__ = (Index('class4_idx_code_code_deck_prefix', 'code_deck_id', 'code'),)

    code_id = Column(Integer, primary_key=True)
    code = Column(PrefixRange, nullable=False, index=True)
    code_deck_id = Column(ForeignKey('code_deck.code_deck_id', ondelete='CASCADE'), index=True)
    city = Column(String(20))
    state = Column(String(20))
    country = Column(String(100))
    name = Column(String(100))
    code_name = synonym('name')

    code_deck = relationship('CodeDeck', back_populates='codes', uselist=False)


class CodeCountry(BaseModel):
    __tablename__ = 'code_country'

    country_code = Column(String(20), primary_key=True)
    country = Column(String(80), unique=True)
    phone_code = Column(PrefixRange())

    @classmethod
    def init(cls):
        if len(cls.query().all()) != len(COUNTRY):
            cls.query().delete()
            for c in COUNTRY:
                cls.session().add(cls(country_code=c[0], country=c[1], phone_code=c[2]))
            cls.session().commit()
            log.info('CodeCountry reinited...')


class CodeDeck(BaseModel):
    __tablename__ = 'code_deck'

    code_deck_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, unique=True)
    update_at = Column(DateTime(timezone=True), onupdate=func.now())
    update_by = Column(String)
    client_id = Column(Integer)  # ,
    # default=0)  # Equal to zero is the default, 0-a / z, 1-us, other values for the code deck dedicated
    codes = relationship(Code, back_populates='code_deck', uselist=True,
                         single_parent=True,
                         cascade="all, delete-orphan")
    update_on = synonym('update_at')

    @staticmethod
    def init():
        if not CodeDeck.get(0):
            r = CodeDeck(code_deck_id=0, name='A/Z', client_id=1)
            r.save()
        else:
            r = CodeDeck.get(0)
            if not (r.name == 'A/Z' and r.client_id == 0):
                r.name = 'A/Z'
                r.client_id = 0
                r.update_by = 'system'
                r.save()

        if not CodeDeck.get(1):
            r = CodeDeck(code_deck_id=1, name='US', client_id=2)
            r.save()
        else:
            r = CodeDeck.get(1)
            if not (r.name == 'US' and r.client_id == 1):
                r.name = 'US'
                r.client_id = 1
                r.update_by = 'system'
                r.save()

    # def code_count(self):return Code.filter(Code.code_deck_id==self.code_deck_id).count()

    def __repr__(self):
        return self.name


CodeDeck.code_count = column_property(
    select([func.count(Code.code_id).label('code_count')]).where(Code.code_deck_id == CodeDeck.code_deck_id). \
        group_by(Code.code_deck_id).correlate_except(Code))


class SystemParameter(BaseModel):
    __tablename__ = 'system_parameter'
    AUTO_RATE_MAIL_SSL = {0: 'no', 1: 'yes'}
    # DEFAULT_US_IJ_RULE = {0: 'A-Z', 1: 'US Non-JD', 2: 'US JD', 3: 'OCN-LATA-JD', 4: 'OCN-LATA-NON-JD'}
    DEFAULT_US_IJ_RULE = {0: 'Inter', 1: 'Intra', 2: 'Max of Inter/Intra'}
    LANDING_PAGE = {0: 'QoS Report', 1: 'Summary Report', 2: 'Orig - Term Report', 3: 'Carrier Management'}
    POSITION = {0: 'bottom', 1: 'top', 2: 'middle'}
    INFOPOSITION = {0: 'top', 1: 'left'}
    CHARGE_TYPE = {1: 'credit total amount', 2: 'create actual received amount'}
    INVOICE_SEND_MODE = {1: 'link', 0: 'attachment'}
    INVOICE_SEND_MODE_CDR = {None: 'none', 1: 'link', 0: 'attachment'}
    QOS_SAMPLE_PERIOD = {0: '15 min', 2: '1 hour', 3: '1 day'}
    LOW_CALL_ATTEMPT_HANDLING = {0: 'none', 1: 'use latest value'}
    REPORT_COUNT = {0: '1 Hour', 1: '24 Hour'}
    sys_id = Column(Integer, primary_key=True)
    # default_avatar_id = Column(Integer)
    # allow_cdr_fields = Column(Text)
    auto_carrier_notification = Column(Boolean)
    auto_delivery_address = Column(String(200))
    auto_delivery_group_by = Column(SmallInteger, default=0)
    auto_delivery_timezone = Column(String(20), default=text("'+00'::character varying"))
    auto_rate_mail_ssl = Column(ChoiceType(AUTO_RATE_MAIL_SSL), default=1)
    auto_rate_pwd = Column(String(100))
    auto_rate_smtp = Column(String(50))
    auto_rate_smtp_port = Column(Integer)
    auto_rate_username = Column(String(100))
    backup_leave_last = Column(String(10))
    backup_period = Column(String(10))
    bar_color = Column(String(20), default=text("'6B9B20'::character varying"))
    base_url = Column(String(512))
    billing_info = synonym('pdf_tpl')
    billing_info_position = synonym('tpl_number')
    billing_info_location = synonym('tpl_number')
    # billing_info = Column(Text)
    # billing_info_location = Column(ChoiceType(POSITION), default=0)
    call_timeout = Column(Integer, default=3600)
    cdrs_deleteafterdays = Column(String(20))
    charge_type = Column(ChoiceType(CHARGE_TYPE))
    # codedeck_last_az_update = Column(DateTime(timezone=True))
    # codedeck_last_us_update = Column(DateTime(timezone=True))
    # codedeck_prompt_az_update = Column(DateTime(timezone=True))
    # codedeck_prompt_us_update = Column(DateTime(timezone=True))
    company_info = Column(String(500))
    company_info_location = Column(ChoiceType(INFOPOSITION), default=0)
    conf_max_duration = Column(Integer)
    conf_number = Column(String(10))
    csv_delimiter = Column(String(5))
    daily_payment_confirmation = Column(Boolean)
    daily_payment_email = Column(Text)
    date_format = Column(String(20))
    datetime_format = Column(String(20))
    de_pin_len = Column(Integer)
    default_billing_decimal = Column(Integer, nullable=False, default=6)
    default_report_decimal = Column(Integer, nullable=False, server_default='2')
    default_code_deck = Column(Integer)
    default_us_ij_rule = Column(ChoiceType(DEFAULT_US_IJ_RULE))
    # doc_url = Column(String(512))
    dr_period = Column(String(10))
    egress_pdd_timeout = Column(Integer, nullable=False, default=6000)
    emailname = Column(String(50))
    emailpassword = Column(String(50))
    emailusername = Column(String(50))
    events_alertszerotime = Column(Boolean)
    events_deleteafterdays = Column(String(10))
    events_notfoundaccount = Column(Boolean)
    events_notfoundtariff = Column(Boolean)
    events_unprofitable = Column(Boolean)
    fail_calls = Column(Integer)
    # favicon_id = Column(Integer)
    finance_email = Column(String(1000))
    forbidden_times = Column(Integer)
    fromemail = Column(String(50))
    ftp_email = Column(String(50))
    ftp_pass = Column(String(20))
    ftp_username = Column(String(50))
    full_cdr_save_days = Column(Integer, default=30)
    inactivity_timeout = Column(Integer, default=30)
    ingress_pdd_timeout = Column(Integer, nullable=False, default=60000)
    invoice_decimal_digits = Column(Integer, default=2)
    invoice_name = Column(String(200))
    invoice_send_mode = Column(ChoiceType(INVOICE_SEND_MODE), default=0)
    # invoice_send_mode_cdr = Column(ChoiceType(INVOICE_SEND_MODE_CDR), default=0)
    invoices_cdr_fields = Column(String(80))
    invoices_delay = Column(String(10))
    invoices_fields = Column(String(50))
    invoices_lastno = Column(String(20))
    # invoices_logo_id = Column(Integer)
    invoices_separate = Column(String(10))
    invoices_tplno = Column(String(20))
    is_hide_unauthorized_ip = Column(SmallInteger, default=0)
    is_preload = Column(Boolean, default=True)
    is_show_mutual_balance = Column(SmallInteger, default=1)
    landing_page = Column(ChoiceType(LANDING_PAGE), default=0)
    login_captcha = Column(Boolean)
    login_fit_image = Column(Boolean)
    # login_image = Column(String)
    login_page_content = Column(Text)
    loginemail = Column(String(12))
    # logo_icon_id = Column(Integer)
    # logo_image_id = Column(Integer)
    logs_deleteafterdays = Column(String(10))
    lowBalance_period = Column(Integer)
    low_call_attempt_handling = Column(ChoiceType(LOW_CALL_ATTEMPT_HANDLING))
    mail_server_from = Column(String(50))
    mailserver_host = Column(String(20))
    minimal_call_attempt_required = Column(Integer)
    msgmonthlyfee = Column(Numeric(20, 3))
    noc_email = Column(String(1000))
    non_zero_cdr_save_days = Column(Integer, default=60)
    notify_carrier = Column(Boolean)
    notify_carrier_cc = Column(Text)
    overlap_invoice_protection = Column(Boolean, default=True)
    payment_content = Column(Text)
    payment_received_confirmation = Column(Boolean)
    payment_setting_subject = Column(String(100))
    paypal_account = Column(String(255))
    # paypal_skey = Column(String(255))
    paypal_test_mode = Column(Boolean)
    paypal_service_charge = Column(Integer, default=0)
    pdf_tpl = Column(Text)
    # portal_display_sip_ip = Column(Boolean, default=True)
    # portal_enable_egress_trunks = Column(Boolean, default=True)
    # portal_enable_ingress_trunks = Column(Boolean, default=True)
    # portal_enable_ip_change = Column(Boolean, default=True)
    # portal_enable_public_products = Column(Boolean, default=True)
    # portal_offset = Column(Boolean, default=True)
    qos_sample_period = Column(ChoiceType(QOS_SAMPLE_PERIOD))
    radius_log_routes = Column(String(10))
    rate_clean_days = Column(Integer)
    rates_deleteafterdays = Column(String(10))
    realm = Column(String)
    report_code_save_days = Column(Integer, default=60)
    report_count = Column(ChoiceType(REPORT_COUNT), default=0)
    report_daily_save_days = Column(Integer, default=180)
    report_hourly_save_days = Column(Integer, default=30)
    require_comment = Column(Integer)
    ring_timeout = Column(Integer, nullable=False, default=60)
    send_cdr_fields = Column(Text)
    signup_content = Column(String(200))
    simple_cdr_save_days = Column(Integer, default=180)
    smtp_secure = Column(Integer, default=0)
    smtphost = Column(String(50))
    smtpport = Column(String(50))
    stats_rotate_delay = Column(String(10))
    stripe_account = Column(String(255))
    stripe_public_account = Column(String(255))
    stripe_service_charge = Column(Integer, default=0)
    switch_alias = Column(String(255))
    switch_ip = Column(String(20))
    switch_port = Column(String(20))
    sys_ani = Column(PrefixRange)
    sys_area = Column(String(50))
    sys_currency = Column(String(20))
    sys_timezone = Column(String(20))
    system_admin_email = Column(String(1000))
    system_rate_mail = Column(String(255))
    themer = Column(Integer, default=0)
    tpl_number = Column(ChoiceType(POSITION), default=0)
    welcome_message = Column(String)
    withdraw_email = Column(String(1000))
    workstation = Column(String)
    yourpay_store_number = Column(String(255))
    cdr_token = Column(String(255))

    # rel_code_deck = relationship(CodeDeck, primaryjoin=foreign(default_code_deck) == CodeDeck.code_deck_id)
    # cmd_debug = Column(String(200))
    # payment_from = Column(Text)
    # payment_subject = Column(Text)
    # payment_from_cc = Column(Text)
    # login_fit_image = Column(Boolean)
    # cdr_token = Column(String(255))
    # cdr_token_alias = Column(String(255))
    # payment section
    cc_email = synonym('notify_carrier_cc')
    # credit_full_payment = synonym(credit_full_payment)
    emails = synonym('finance_email')
    enable_email_notification = synonym('notify_carrier')
    # paypal_account
    paypal_fee = synonym('paypal_service_charge')
    stripe_fee = synonym('stripe_service_charge')
    stripe_publisher_key = synonym('stripe_public_account')
    stripe_secret_key = synonym('stripe_account')
    # --- login page synonyms ---
    image_url = synonym('login_image')
    fit_to_screen = synonym('login_fit_image')
    use_captcha = synonym('login_captcha')
    content = synonym('login_page_content')
    # mail sender
    smtp_host = synonym('mailserver_host')
    smtp_port = synonym('smtpport')
    username = synonym('emailusername')
    password = synonym('emailpassword')
    email = synonym('fromemail')
    # ---
    # ---Invoice settings synonyms---#
    # allow_invoice_overlap = synonym('overlap_invoice_protection')
    # billing_info = synonym('billing_info')#??
    billing_info_position = synonym('billing_info_location')  # ??
    # cdr_fields = synonym('send_cdr_fields')
    # +company_info = synonym()
    company_info_position = synonym('company_info_location')
    decimal = synonym('invoice_decimal_digits')
    invoice_number_format = synonym('invoices_lastno')
    # logo = synonym()
    logo_url = synonym('login_image')
    send_rate_as = synonym('system_rate_mail')
    pre_load_data = synonym('is_preload')

    @classmethod
    def init(cls):
        q = cls.get(1)
        if not q:
            q = cls(sys_id=1)
            q.save()

    @hybrid_property
    def allow_invoice_overlap(self):
        return not bool(self.overlap_invoice_protection)

    @allow_invoice_overlap.setter
    def allow_invoice_overlap(self, value):
        self.overlap_invoice_protection = not bool(value)

    @hybrid_property
    def cdr_fields(self):
        try:
            return self.send_cdr_fields.split(',')
        except:
            return []

    @cdr_fields.setter
    def cdr_fields(self, value):
        try:
            self.send_cdr_fields = ','.join([str(x) for x in value])
        except:
            pass

    @hybrid_property
    def code_deck(self):
        try:
            return self.rel_code_deck.name
        except:
            return ''

    @code_deck.setter
    def code_deck(self, value):
        try:
            self.default_code_deck = CodeDeck.filter(CodeDeck.name == value).one().code_deck_id
        except:
            # self.default_code_deck = 1
            raise ValidationError('bad code deck:{}'.format(value))

    @hybrid_property
    def currency_name(self):
        try:
            return self.currency.code
        except:
            return ''

    @currency_name.setter
    def currency_name(self, value):
        try:
            self.currency_id = Currency.filter(Currency.code == value).one().currency_id
        except:
            pass

    @hybrid_property
    def _logo_url(self):
        try:
            return '{}{}{}/config/export/public/{}'.format(settings.API_SCHEME, settings.API_HOST,
                                                           settings.API_BASE_PATH, self.invoices_logo_id)
        except:
            return None

    @hybrid_property
    def _logo_path(self):
        try:
            return self.logo_url
        except:
            return None

    @property
    def root_url(self):
        url = parse_url(self.base_url)
        return '{}://{}'.format(url.scheme, url.netloc)

    @property
    def ilogo_url(self):
        url = parse_url(self.base_url)
        if url:
            ret = requests.get('{}://{}/homes/getInvoiceLogoUrl'.format(url.scheme, url.netloc), verify=False)
            if ret.status_code in (200, 201, 202):
                return ret.text
            else:
                return '{}://{}/upload/images/ilogo.png'.format(url.scheme, url.netloc)
        return None


# +++ VersionInformation
class VersionInformation(BaseModel):
    __tablename__ = 'version_information'
    __table_args__ = (
        UniqueConstraint('program_name', 'switch_name', name='version_information_program_switch_name_idx'),
    )
    id = Column(Integer, primary_key=True, server_default=text("nextval('version_information_id_seq'::regclass)"))
    program_name = Column(String(50), unique=True)
    major_ver = Column(String(100))
    minor_ver = Column(String(100))
    build_date = Column(String(11))
    start_time = Column(DateTime(True))
    cli_ip = Column(String(30))
    cli_port = Column(Integer())
    last_status_time = Column(String(100))
    status = Column(Integer())
    licensed_cps = Column(Integer())
    licensed_channel = Column(Integer())
    expires = Column(DateTime(True))
    switch_name = Column(String(100))
    serial_number = Column(String(100))


class DailyCdrField(BaseModel):
    __tablename__ = 'daily_cdr_fields'

    id = Column(Integer, primary_key=True)
    type = Column(SmallInteger, default=0)
    field = Column(String, nullable=False)
    label = Column(String)

    db_name = synonym('field')
    display_name = synonym('label')

    @classmethod
    def get_labels(cls, fields):
        labels = []
        for f in fields.split(','):
            l = cls.filter(cls.field == f).first()
            if l:
                labels.append(l.label)
            else:
                labels.append(f.capitalize())
        return ','.join(labels)

    @hybrid_property
    def admin_default_origination(self):
        try:
            return ((int(self.type) & (1 << 5)) != 0)
        except:
            return False

    @admin_default_origination.setter
    def admin_default_origination(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 5)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 5)

    @hybrid_property
    def admin_default(self):
        try:
            return ((int(self.type) & (1 << 4)) != 0)
        except:
            return False

    @admin_default.setter
    def admin_default(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 4)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 4)

    @hybrid_property
    def client_viewable(self):
        try:
            return ((int(self.type) & (1 << 3)) != 0)
        except:
            return False

    @client_viewable.setter
    def client_viewable(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 3)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 3)

    #
    @hybrid_property
    def vendor_viewable(self):
        try:
            return ((int(self.type) & (1 << 2)) != 0)
        except:
            return False

    @vendor_viewable.setter
    def vendor_viewable(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 2)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 2)

    @hybrid_property
    def client_cdr_delivery(self):
        try:
            return ((int(self.type) & (1 << 1)) != 0)
        except:
            return False

    @client_cdr_delivery.setter
    def client_cdr_delivery(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 1)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 1)

    @hybrid_property
    def vendor_cdr_delivery(self):
        try:
            return ((int(self.type) & (1 << 0)) != 0)
        except:
            return False

    @vendor_cdr_delivery.setter
    def vendor_cdr_delivery(self, value):
        if value:
            self.type = (self.type if self.type else 0) | (1 << 0)
        else:
            self.type = (self.type if self.type else 0) & ~(1 << 0)

    @classmethod
    def convert_fields(cls, fields):
        d = {r.field: r.label for r in cls.query().all()}
        ret = [d[f] if f in d else f for f in fields]
        # for f in fields:
        #     if f in d:
        #         ret.append(d[f])
        #     else:
        #         ret.append(f)
        return ret

    @staticmethod
    def init():
        def to_label(old):
            def cap(s):
                if s in ('id', 'Id'):
                    return 'ID'
                else:
                    return s.capitalize()

            return ' '.join([cap(s) for s in old.split('_')])

        # cols=[c.name for c in inspect(CdrReportDetail).columns]
        for k, v in FIELDS.items():
            q = DailyCdrField.get(k)
            if q:
                if q.field != v:
                    q.field = v
                    if UIFIELDS[k] != '':
                        q.label = UIFIELDS[k]
                    else:
                        q.label = to_label(FIELDS[k])
                    try:
                        q.save()
                    except:
                        pass
            else:
                if UIFIELDS[k] != '':
                    label = UIFIELDS[k]
                else:
                    label = to_label(FIELDS[k])
                rec = DailyCdrField(id=k, field=v, label=label)
                try:
                    rec.save()
                except:
                    pass
        for rec in DailyCdrField.query().all():
            if rec.id not in FIELDS:
                rec.delete()


def add_months(sourcedate, months):
    month = sourcedate.month - 1 + months
    year = sourcedate.year + month // 12
    month = month % 12 + 1
    day = min(sourcedate.day, calendar.monthrange(year, month)[1])
    return date(year, month, day)


class PaymentTerm(BaseModel):
    __tablename__ = 'payment_term'
    DAYS_OF_WEEK = {1: "Monday", 2: "Tuesday", 3: "Wednesday", 4: "Thursday", 5: "Friday", 6: "Saturday", 7: "Sunday"}
    TYPE_DICT = {1: 'Every', 2: 'Day of Month', 3: 'Day of Week', 4: 'Some day of month', 5: 'Twice a month'}
    payment_term_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, unique=True)
    type = Column(ChoiceType(TYPE_DICT), nullable=False)
    days = Column(Integer, nullable=False, default=1)
    grace_days = Column(Integer)
    notify_days = Column(Integer)
    more_days = Column(String(100))
    finance_rate = Column(Float)

    grace_period = synonym('grace_days')
    type_name = synonym('type')

    clients = relationship('Client', back_populates='payment_term', uselist=True)

    @property
    def cycle(self):
        try:
            ordinal = lambda n: "%d%s" % (n, "tsnrhtdd"[(n / 10 % 10 != 1) * (n % 10 < 4) * n % 10::4])
            if self.type == 'Day of Month':
                return 'Every {} of the month'.format(ordinal(int(self.days)))
            if self.type == 'Every':
                return 'Every {} day(s)'.format(ordinal(int(self.days)))
            if self.type == 'Day of Week':
                return 'Every {} of the week'.format(self.DAYS_OF_WEEK[self.days])
            if self.type == 'Some day of month':
                return 'At {} days of month'.format(','.join([ordinal(int(d)) for d in self.more_days_arr]))
            if self.type == 'Twice a month':
                return 'Twice in a month (15th and last day)'
        except Exception as e:
            return self.type + ' (not completed)'
        return None

    def due_date(self, d):
        nx = self.next_date(d)
        return nx + timedelta(days=self.grace_days)

    def next_date(self, d):
        now = datetime.now(UTC).date()
        return self.next_date_for_now(d, now)

    def next_date_for_now(self, d, now):
        from datetime import date

        if isinstance(d, type('')):
            d = parse_datetime(d).date()
        elif isinstance(d, datetime):
            d = date(d.year, d.month, d.day)
        elif isinstance(d, date):
            d = date(d.year, d.month, d.day)
        elif d is None:
            d = date(2001, 9, 11)
        else:
            raise Exception("next_date:wrong input data type {}".format(d))

        if not d or d > now:
            d = now
        if self.type == 'Day of Month':
            next = date(d.year, d.month, 1) + timedelta(days=self.days - 1)
            if next <= d:
                next = add_months(next, 1)
            while next < now:
                next = add_months(next, 1)
            return next
        if self.type == 'Every':
            next = d + timedelta(days=int(self.days))
            while next < now:
                next = next + timedelta(days=self.days)
            return next
        if self.type == 'Day of Week':
            days_ahead = (self.days - 1) - d.weekday()
            if days_ahead <= 0:
                days_ahead += 7
            next = d + timedelta(days=days_ahead)
            while next < now:
                next = next + timedelta(days=7)
            return next
        if self.type == 'Some day of month':
            days = sorted([int(i) for i in self.more_days_arr])
            next = date(d.year, d.month, days[0])
            while next <= d or next <= now:
                for day in days:
                    next = date(next.year, next.month, day)
                    if next > d and next > now:
                        return next
                next = date(next.year, next.month, days[0])
                next = add_months(next, 1)
            return next
        if self.type == 'Twice a month':
            days = sorted([1, 15])
            next = date(d.year, d.month, days[0])
            while next < d or next < now:
                for day in days:
                    next = date(next.year, next.month, day)
                    if next > d and next >= now:
                        return next
                next = date(next.year, next.month, days[0])
                next = add_months(next, 1)
            return next

        return None

    @hybrid_property
    def more_days_arr(self):
        try:
            return sorted([int(i) for i in set([str(self.days)] + self.more_days.split(',')) if i != ''])
        except:
            return []

    @more_days_arr.setter
    def more_days_arr(self, value):
        self.days = value[0]
        # raise Exception(value)
        self.more_days = ','.join([str(s) for s in value[1:]])


class TimeProfile(BaseModel):
    __tablename__ = 'time_profile'
    TYPE_DICT = {0: 'all time', 1: 'weekly', 2: 'daily'}
    DAYS_OF_WEEK = {1: "Monday", 2: "Tuesday", 3: "Wednesday", 4: "Thursday", 5: "Friday", 6: "Saturday", 7: "Sunday"}
    time_profile_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, unique=True)
    start_time = Column(Time)
    end_time = Column(Time)
    start_week = Column(ChoiceType(DAYS_OF_WEEK))
    end_week = Column(ChoiceType(DAYS_OF_WEEK))
    type = Column(ChoiceType(TYPE_DICT), nullable=False)  # 0：all time;1：weekly;2：daily
    time_zone = Column(String(10))

    start_day_of_week = synonym('start_week')
    end_day_of_week = synonym('end_week')
    type_name = synonym('type')

    def __repr__(self):
        return str(self.time_profile_id) + ':' + self.name + ' start ' + str(self.start_time) + ' end ' + \
               str(self.end_time) + ' start week ' + str(self.start_week) + ' end week' + str(
            self.end_week) + ' ' + str(self.type)


class DigitTranslation(BaseModel):
    __tablename__ = 'digit_translation'

    translation_id = Column(Integer, primary_key=True)
    translation_name = Column(String(100), nullable=False, unique=True)
    translation = Column(DateTime(timezone=True), nullable=False, default=datetime.now(UTC))
    update_at = synonym('translation')
    digit_map_name = synonym('translation_name')

    items = relationship('TranslationItem', back_populates='translation', uselist=True)
    trunks = relationship('ResourceTranslationRef', back_populates='translation', uselist=True)

    # @hybrid_property
    def digit_map_count(self):
        return TranslationItem.filter(translation_id=self.translation_id).count()

    def __repr__(self):
        return str(self.translation_id) + ':' + self.translation_name


class TranslationItem(BaseModel):
    # Master called conversion
    __tablename__ = 'translation_item'
    METHOD_DICT = {0: 'ignore', 1: 'partial replacement', 2: 'all replace'}
    ref_id = Column(Integer, primary_key=True)
    translation_id = Column(ForeignKey('digit_translation.translation_id', ondelete='CASCADE'), nullable=False,
                            index=True)  # number conversion
    ani = Column(PrefixRange, default='')  # prefix_range caller
    dnis = Column(PrefixRange, default='')  # prefix_range called
    action_ani = Column(String(30), nullable=False)  # after conversion of the caller
    action_dnis = Column(String(30), nullable=False)  # converted after the call
    ani_method = Column(ChoiceType(METHOD_DICT), nullable=False)  # 0 - ignore; 1 - partial replacement; 2 - all replace
    dnis_method = Column(ChoiceType(METHOD_DICT),
                         nullable=False)  # 0 - ignore; 1 - partial replacement; 2 - all replace

    ANI_replace_to = synonym('action_ani')
    DNIS_replace_to = synonym('action_dnis')
    ANI_prefix = synonym('ani')
    DNIS_prefix = synonym('dnis')

    translation = relationship('DigitTranslation', back_populates='items', uselist=False)

    # @hybrid_property
    # def digit_map_name(self): return self.translation.translation_name
    digit_map_name = column_property(select([DigitTranslation.translation_name]). \
                                     where(DigitTranslation.translation_id == translation_id). \
                                     correlate_except(DigitTranslation))


class ResourceTranslationRef(BaseModel):
    __tablename__ = 'resource_translation_ref'
    # different time periods, different master calls are converted
    ref_id = Column(Integer, primary_key=True)
    resource_id = Column(ForeignKey('resource.resource_id'), index=True)  # gateway group - foreign key; docking gateway
    translation_id = Column(ForeignKey('digit_translation.translation_id'),
                            index=True)  # conversion rules - foreign key
    time_profile_id = Column(ForeignKey('time_profile.time_profile_id'), index=True)  # time period - foreign key

    gateway = relationship('Resource', back_populates='digit_maps', uselist=False)
    translation = relationship('DigitTranslation', back_populates='trunks', uselist=False)
    time_profile = relationship('TimeProfile', uselist=False)

    @hybrid_property
    def digit_map_name(self):
        return self.translation.translation_name


class RouteStrategy(BaseModel):
    __tablename__ = 'route_strategy'

    route_strategy_id = Column(Integer, primary_key=True)
    name = Column(String(256), nullable=False, unique=True)
    update_at = Column(DateTime(timezone=True), onupdate=func.now())
    update_by = Column(String)
    is_virtual = Column(Boolean)
    routes = relationship('Route', back_populates='route_strategy', uselist=True,
                          single_parent=True,
                          cascade="all, delete-orphan",
                          )

    route_plan_id = synonym('route_strategy_id')

    @hybrid_property
    def usage_count_(self):
        return self.routes.count()

    @classmethod
    def init(cls):
        q = cls.get(1)
        if not q:
            q = cls(route_strategy_id=1, name='ORIGINATION_ROUTING_PLAN', update_by='system')
            q.save()
        if not q.name == 'ORIGINATION_ROUTING_PLAN':
            q.name = 'ORIGINATION_ROUTING_PLAN'
            q.update_by = 'system'
            q.save()

    @staticmethod
    def get_orig_routing_plan_id():
        RouteStrategy.init()
        return 1


# region RateTable

class RateTable(BaseModel):
    __tablename__ = 'rate_table'
    RATE_TYPE_DICT = {0: 'DNIS', 1: 'LRN', 2: 'LRN BLOCK', 3: 'ANI'}
    JUR_TYPE_DICT = {0: 'A-Z', 1: 'US Non Jurisdictional', 2: 'US Jurisdictional', 3: 'OCN-LATA'}
    rate_table_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, unique=True)
    modify_time = Column(DateTime(True), onupdate=func.now())
    create_time = Column(DateTime(True), nullable=False, server_default=func.now())
    code_deck_id = Column(ForeignKey('code_deck.code_deck_id', ondelete='SET NULL'), index=True, nullable=True)
    currency_id = Column(ForeignKey('currency.currency_id', ondelete='SET NULL'), nullable=False, index=True, default=1)
    rate_type = Column(ChoiceType(RATE_TYPE_DICT), default=0)  # 0-dnis;1-lrn;2-lrn block
    billing_method = synonym('rate_type')
    jurisdiction_country_id = Column(Integer,
                                     index=True)  # Billing area countries, there are values that the rate is calculated according to intra / inter, no value that dnis billing.
    update_at = Column(DateTime(True), onupdate=func.now())
    update_by = Column(String)
    lnp_dipping_rate = Column(Float)
    jurisdiction_prefix = Column(PrefixRange)
    noprefix_min_length = Column(Integer)
    noprefix_max_length = Column(Integer)
    prefix_min_length = Column(Integer)
    prefix_max_length = Column(Integer)
    jur_type = Column(ChoiceType(JUR_TYPE_DICT), default=0)
    rate_type_name = synonym('jur_type')

    origination = Column(Boolean, default=False)
    is_virtual = Column(Boolean)

    require_setup_fee = Column(Boolean, server_default='False')
    default_setup_fee = Column(Float, server_default='0')
    require_local_rate = Column(Boolean, server_default='False')

    code_deck = relationship(CodeDeck)
    currency = relationship(Currency)
    rates = relationship('Rate', back_populates='rate_table', uselist=True,
                         single_parent=True,
                         cascade="all, delete-orphan")
    resources = relationship('Resource', back_populates='rate_table', uselist=True)
    trunks = relationship('ResourcePrefix', uselist=True,
                          primaryjoin='and_(ResourcePrefix.rate_table_id == RateTable.rate_table_id)')

    def usage_count(self):
        return Resource.filter(Resource.rate_table_id == self.rate_table_id).count()

    code_deck_name = column_property(
        select([CodeDeck.name]).where(CodeDeck.code_deck_id == code_deck_id).correlate_except(CodeDeck))
    currency_name = column_property(
        select([Currency.code]).where(Currency.currency_id == currency_id).correlate_except(Currency))

    @property
    def trunks(self):
        if not self.resources:
            return []
        return [r.resource_id for r in self.resources if r.client_id and r.client_type == 'carrier']

    @classmethod
    def init(cls):
        q = cls.filter(cls.name == 'DEFAULT_TESTING').first()
        if not q:
            cls(name='DEFAULT_TESTING').save()

    @property
    def rate_download_link(self):
        exp = datetime.now(UTC) + timedelta(days=settings.JWT_TTL_DAYS)
        token_data = {'rate_table_id': self.rate_table_id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE).decode('utf-8')
        return '{}{}{}/rate/download/{}/rate_{}.csv'.format(settings.API_SCHEME, settings.API_HOST,
                                                            settings.API_BASE_PATH, token, self.rate_table_id)

    @property
    def switch_ip(self):
        ips = []
        for r in self.resources:
            ips = ips + ['{}:{}'.format(i.ip, i.port) for i in r.ip if not i.ip is None]
        return ','.join(ips)


class Rate(BaseModel):
    __tablename__ = 'rate'

    rate_id = Column(Integer, primary_key=True, server_default=text("nextval('rate_rate_id_seq'::regclass)"))
    rate_table_id = Column(ForeignKey('rate_table.rate_table_id', ondelete='CASCADE'), nullable=False, index=True)
    code = Column(PrefixRange, nullable=True, server_default=text("''::prefix_range"), index=True)
    rate = Column(Numeric(30, 10))
    setup_fee = Column(Numeric(30, 10), nullable=False, server_default=text("0"))
    effective_date = Column(DateTime(True), default=datetime.now(UTC), index=True)
    end_date = Column(DateTime(True), index=True)
    min_time = Column(Integer, nullable=False, default=1, server_default=text("0"))
    grace_time = Column(Integer, nullable=False, server_default=text("0"))
    interval = Column(Integer, nullable=False, default=1, server_default=text("1"))
    time_profile_id = Column(ForeignKey('time_profile.time_profile_id', ondelete='CASCADE'), index=True)
    seconds = Column(Integer, default=60, nullable=False, server_default=text("60"))
    code_name = Column(String(100))
    basic_percentages = Column(Float, server_default=text("0"))
    gift_percentages = Column(Float, default=0, server_default=text("0"))
    rate_type = Column(Integer, default=3, nullable=False, server_default=text("3"))
    intra_rate = Column(Numeric(30, 10))
    inter_rate = Column(Numeric(30, 10))
    local_rate = Column(Numeric(30, 10))
    country = Column(String(1000))
    zone = Column(String(10), default='0', server_default=text("0"))
    ocn = Column(String(10))
    lata = Column(String(10))
    create_time = Column(DateTime(True))
    did_type = Column(Integer, index=True)
    __table_args__ = (
        # UniqueConstraint('rate_table_id', 'code', 'effective_date','end_date'),
        UniqueConstraint('rate_table_id', 'code', 'effective_date', name='rate_rate_table_id_code_effective_date_key'),
        Index('class4_idx_rate_table_id_code', code, rate_table_id, postgresql_using='gist', postgresql_ops={
            'code': 'gist_prefix_range_ops',
        })
    )

    id = synonym('rate_id')
    profile = synonym('time_profile_id')
    gmt = synonym('zone')

    rate_table = relationship(RateTable, back_populates='rates', uselist=False)
    is_effective_now = column_property(
        and_(effective_date <= datetime.now(UTC), or_(end_date.is_(None), end_date >= datetime.now(UTC))))
    rate_table_name = column_property(
        select([RateTable.name]).where(RateTable.rate_table_id == rate_table_id).correlate_except(RateTable))


# endregion RateTable


class DynamicRoute(BaseModel):
    __tablename__ = 'dynamic_route'

    ROUTING_RULE_DICT = {6: 'LCR', 1: 'By LCR', 2: 'By ASR', 3: 'By ACD',
                         4: 'by maximum ASR', 5: 'by maximum ACD'}
    LCR_FLAG_DICT = {None: 'not set', 1: '15 minutes', 2: '30 minutes', 3: '1 hours', 4: '1 days'}

    dynamic_route_id = Column(Integer, primary_key=True)
    name = Column(String(100), nullable=False, unique=True)
    routing_rule = Column(ChoiceType(ROUTING_RULE_DICT), nullable=False,
                          default=6)  # Routing rules: 4 - by maximum ASR; 5 - by maximum ACD; 6 - at the lowest cost
    time_profile_id = Column(ForeignKey('time_profile.time_profile_id', ondelete='SET NULL'), index=True)
    update_at = Column(DateTime(timezone=True), onupdate=func.now())
    update_by = Column(String)
    lcr_flag = Column(ChoiceType(LCR_FLAG_DICT), default=1)  # 1-15 minutes; 2-30 minutes; 3-1 hours; 4-1 days
    is_virtual = Column(Boolean)

    lcr_flag_name = synonym('lcr_flag')
    route_rule_name = synonym('routing_rule')
    # rule = synonym('routing_rule')
    last_modified = synonym('update_at')
    modified_by = synonym('update_by')

    qos_cycle = synonym('lcr_flag')

    time_profile = relationship(TimeProfile)
    routes = relationship('Route', back_populates='dynamic_route', uselist=True)

    egress_trunks = relationship('DynamicRouteItem', uselist=True, back_populates='route',
                                 single_parent=True,
                                 cascade="all, delete-orphan"
                                 )
    qos = relationship('DynamicRouteQos', uselist=True, back_populates='dynamic_route',
                       single_parent=True, cascade="all, delete-orphan"
                       )
    pri = relationship('DynamicRoutePri', uselist=True, back_populates='dynamic_route',
                       single_parent=True, cascade="all, delete-orphan"
                       )
    over = relationship('DynamicRouteOverride', uselist=True, back_populates='dynamic_route',
                        single_parent=True, cascade="all, delete-orphan"
                        )
    # @hybrid_property
    # def usage_count(self):return len(self.routes)
    time_profile_name = column_property(
        select([TimeProfile.name]).where(TimeProfile.time_profile_id == time_profile_id).correlate_except(TimeProfile))


class DynamicRouteItem(BaseModel):
    __tablename__ = 'dynamic_route_items'
    __table_args__ = (
        UniqueConstraint('dynamic_route_id', 'resource_id'),
    )
    id = Column(Integer, primary_key=True)
    dynamic_route_id = Column(
        ForeignKey('dynamic_route.dynamic_route_id', ondelete='CASCADE'), nullable=False, index=True
    )
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)

    route = relationship(DynamicRoute, uselist=False, back_populates='egress_trunks')
    trunk = relationship('Resource',
                         primaryjoin='Resource.resource_id==foreign(DynamicRouteItem.resource_id)',
                         back_populates='dynamic_routes',
                         uselist=False)

    @hybrid_property
    def trunk_name(self):
        try:
            return self.trunk.alias
        except:
            return None

    @hybrid_property
    def is_active(self):
        try:
            return self.trunk.is_active
        except:
            return None

    @hybrid_property
    def client_id(self):
        try:
            return self.trunk.client_id
        except:
            return None

    @hybrid_property
    def client_name(self):
        try:
            return self.trunk.client.name
        except:
            return None


DynamicRoute.trunk_count = column_property(select([func.count(DynamicRouteItem.id).label('trunk_count')]). \
                                           where(DynamicRoute.dynamic_route_id == DynamicRouteItem.dynamic_route_id). \
                                           correlate_except(DynamicRouteItem))


class DynamicRouteQos(BaseModel):
    __tablename__ = 'dynamic_route_qos'
    __table_args__ = (
        UniqueConstraint('dynamic_route_id', 'digits'),
    )

    id = Column(Integer, primary_key=True, server_default=text("nextval('dynamic_route_qos_id_seq'::regclass)"))
    dynamic_route_id = Column(ForeignKey('dynamic_route.dynamic_route_id', ondelete='CASCADE'), nullable=False,
                              index=True)
    digits = Column(PrefixRange, nullable=False, server_default=text("''::prefix_range"))
    min_asr = Column(Float)
    max_asr = Column(Float)
    min_abr = Column(Float)
    max_abr = Column(Float)
    min_acd = Column(Float)
    max_acd = Column(Float)
    min_pdd = Column(Integer)
    max_pdd = Column(Integer)
    min_aloc = Column(Float)
    max_aloc = Column(Float)
    limit_price = Column(Float)

    dynamic_route = relationship('DynamicRoute', uselist=False, back_populates='qos')


class DynamicRoutePri(BaseModel):
    __tablename__ = 'dynamic_route_pri'

    id = Column(Integer, primary_key=True, server_default=text("nextval('dynamic_route_pri_id_seq'::regclass)"))
    dynamic_route_id = Column(ForeignKey('dynamic_route.dynamic_route_id', ondelete='CASCADE'), nullable=False,
                              index=True)
    digits = Column(PrefixRange, nullable=False, server_default=text("''::prefix_range"))
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    resource_pri = Column(Integer, nullable=False, server_default=text("0"))

    dynamic_route = relationship('DynamicRoute', uselist=False, back_populates='pri')
    resource = relationship('Resource')


class DynamicRouteOverride(BaseModel):
    __tablename__ = 'dynamic_route_override'

    id = Column(Integer, primary_key=True, server_default=text("nextval('dynamic_route_override_id_seq'::regclass)"))
    dynamic_route_id = Column(ForeignKey('dynamic_route.dynamic_route_id', ondelete='CASCADE'), nullable=False,
                              index=True)
    digits = Column(PrefixRange, nullable=False, server_default=text("''::prefix_range"))
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    percentage = Column(Integer, nullable=False)

    dynamic_route = relationship('DynamicRoute', uselist=False, back_populates='over')
    resource = relationship('Resource')


class Product(BaseModel):
    __tablename__ = 'product'

    DEFINED_BY_DICT = {0: 'Code', 1: 'Code Name', 3: 'Country'}
    ROUTED_BY_DICT = {1: 'LRN', 0: 'DNIS', None: 'None'}
    product_id = Column(Integer, primary_key=True)
    static_route_id = synonym('product_id')
    name = Column(String(100))
    modify_time = Column(DateTime(True), nullable=False, onupdate=func.now(), default=datetime.now(UTC))
    introduction = Column(String(80))
    dynamic_route_id = Column(Integer)
    update_by = Column(String)
    code_type = Column(ChoiceType(DEFINED_BY_DICT), default=0)
    defined_by = synonym('code_type')
    code_deck_id = Column(Integer, ForeignKey('code_deck.code_deck_id', ondelete='SET NULL'))
    route_lrn = Column(ChoiceType(ROUTED_BY_DICT))
    routed_by = synonym('route_lrn')

    code_deck = relationship('CodeDeck')
    items = relationship('ProductItems', uselist=True, back_populates='product', lazy='dynamic')
    update_at = synonym('modify_time')

    # @classmethod
    # def get_model_class_primary_key(cls):
    #    return cls.static_route_id #get_db(cls.db_label).get_model_class_primary_key(cls)
    def route_count(self):
        # return Route.filter(Route.static_route_id == self.product_id).count()
        pass

    def prefix_count(self):
        return ProductItems.filter(ProductItems.product_id == self.product_id).count()

    @hybrid_property
    def code_deck_name(self):
        try:
            return self.code_deck.name
        except:
            return ""

    @classmethod
    def init(cls):
        q = cls.get(1)
        if not q:
            q = Product(product_id=1, name='ORIGINATION_STATIC_ROUTE', update_by='system', code_type='Code',
                        routed_by='DNIS')
            q.save()
        if not q.name == 'ORIGINATION_STATIC_ROUTE' or not q.routed_by == 'DNIS':
            q.name = 'ORIGINATION_STATIC_ROUTE'
            q.update_by = 'system'
            q.code_type = 'Code'
            q.routed_by = 'DNIS'
            q.save()

    @staticmethod
    def get_orig_static_route_id():
        Product.init()
        return 1

    def __repr__(self):
        return self.name


class Route(BaseModel):
    __tablename__ = 'route'
    __table_args__ = (
        UniqueConstraint('route_strategy_id', 'digits', 'ani_prefix'),
        UniqueConstraint('route_type', 'static_route_id', 'dynamic_route_id'),
    )

    ROUTE_TYPE_DICT = {1: 'dynamic routing', 2: 'static routing', 3: 'dynamic routing-static routing',
                       4: 'static routing-dynamic routing'}
    ROUTE_TYPE_FLG_DICT = {
        1: 'Dynamic Routing',
        2: 'Static Routing',
        3: 'Static Routing JD',
        4: 'Dynamic Routing - Static Routing',
        5: 'Dynamic Routing - Static Routing JD',
        6: 'Static Routing - Dynamic routing',
        7: 'Static Routing JD - Dynamic routing'
    }
    route_id = Column(Integer, primary_key=True)
    digits = Column(PrefixRange, default='')
    dynamic_route_id = Column(ForeignKey('dynamic_route.dynamic_route_id', ondelete='SET NULL'), index=True)
    static_route_id = Column(ForeignKey('product.product_id', ondelete='SET NULL'), index=True)
    route_type = Column(ChoiceType(ROUTE_TYPE_DICT),
                        nullable=False)  # Routing: 1 - dynamic routing; 2 - static routing; 3 - dynamic routing and then static routing; 4 - static routing and then dynamic routing
    route_type_name = synonym('route_type_flg')
    route_strategy_id = Column(
        ForeignKey('route_strategy.route_strategy_id', ondelete='CASCADE'), nullable=False, index=True
    )
    lnp = Column(Boolean, default=False)
    lrn_block = Column(Boolean, default=False)
    dnis_only = Column(Boolean, default=True)
    code_deck_type = Column(Integer, default=0)  # 0-icx code deck;1-custom code deck
    update_at = Column(DateTime(True), onupdate=func.now())
    update_by = Column(String)
    intra_static_route_id = Column(Integer)
    inter_static_route_id = Column(Integer)
    jurisdiction_country_id = Column(Integer)
    ani_prefix = Column(PrefixRange)
    ani_min_length = Column(Integer, default=0)
    ani_max_length = Column(Integer, default=32)
    digits_min_length = Column(Integer, default=0)
    digits_max_length = Column(Integer, default=32)
    code_name = Column(String)
    country = Column(String(50))
    route_type_flg = Column(ChoiceType(ROUTE_TYPE_FLG_DICT))

    ANI_prefix = synonym('ani_prefix')
    DNIS_prefix = synonym('digits')
    ANI_min = synonym('ani_min_length')
    ANI_max = synonym('ani_max_length')
    DNIS_min = synonym('digits_min_length')
    DNIS_max = synonym('digits_max_length')
    route_plan_id = synonym('route_strategy_id')

    dynamic_route = relationship(DynamicRoute, back_populates='routes', uselist=False)
    static_route = relationship('Product')
    route_strategy = relationship(RouteStrategy, back_populates='routes', uselist=False)

    @classmethod
    def check_duplicates(cls, ani_prefix, digits, ani_min_length, ani_max_length, digits_min_length, digits_max_length,
                         obj=None):
        q = cls.filter(and_(cls.ani_min_length == ani_min_length, cls.ani_max_length == ani_max_length,
                            cls.digits_min_length == digits_min_length, cls.digits_max_length == digits_max_length))
        if ani_prefix is None or ani_prefix == '':
            q = q.filter(or_(cls.ani_prefix.is_(None), cls.ani_prefix == ''))
        else:
            q = q.filter(cls.ani_prefix == ani_prefix)
        if digits is None or digits == '':
            q = q.filter(or_(cls.digits.is_(None), cls.digits == ''))
        else:
            q = q.filter(cls.digits == digits)
        if obj:
            q = q.filter(cls.route_strategy_id == obj.route_strategy_id)
            if obj.route_id:
                q = q.filter(cls.route_id != obj.route_id)
        item = q.first()
        if item:
            return True
        return False

    @hybrid_property
    def route_types(self):
        return self.ROUTE_TYPE_FLG_DICT

    @classmethod
    def init(cls):
        # RouteStrategy.init()
        # Product.init()
        rs = RouteStrategy.get_orig_routing_plan_id()
        rp = Product.get_orig_static_route_id()
        r = cls.filter(and_(cls.route_strategy_id == rs, cls.static_route_id == rp)).first()
        if not r:
            cls(route_type='static routing', route_strategy_id=rs, static_route_id=rp,
                update_by='system', update_at=datetime.now(UTC), code_deck_type=0).save()

    route_plan_name = column_property(
        select([RouteStrategy.name]).where(RouteStrategy.route_strategy_id == route_strategy_id).correlate_except(
            RouteStrategy))

    DynamicRouteX = DynamicRoute.__table__.alias('dynamic_route_x')
    dynamic_route_name = column_property(
        select([DynamicRouteX.c.name]).where(DynamicRouteX.c.dynamic_route_id == dynamic_route_id).correlate_except(
            DynamicRouteX))

    ProductX = Product.__table__.alias('product_x')
    static_route_name = column_property(
        select([Product.name]).where(Product.product_id == static_route_id).correlate_except(Product))

    intra_static_route_name = column_property(
        select([Product.name]).where(Product.product_id == intra_static_route_id).correlate_except(Product))

    inter_static_route_name = column_property(
        select([Product.name]).where(Product.product_id == inter_static_route_id).correlate_except(Product))


class ProductRoutRateTable(BaseModel):
    __tablename__ = 'product_rout_rate_table'
    # __table_args__ = (UniqueConstraint('rate_table_id', 'rout_id'),)

    STATUS = {0: 'Inactive', 1: 'Active'}
    TYPE = {0: 'public', 1: 'private'}
    MARKETPLACE = {0: 'A-Z', 1: 'US'}
    id = Column(Integer, primary_key=True, server_default=text("nextval('product_rout_rate_table_id_seq'::regclass)"))
    rout_id = Column(ForeignKey('route_strategy.route_strategy_id', ondelete='CASCADE'), nullable=False)
    rate_table_id = Column(ForeignKey('rate_table.rate_table_id', ondelete='CASCADE'), nullable=False)
    product_name = Column(String(256), nullable=False, unique=True)
    agent_id = Column(Integer)
    status = Column(ChoiceType(STATUS), default=1, server_default=text("1"))  # 0:Inactive; 1:Active
    type = Column(ChoiceType(TYPE), default=0, server_default=text("0"))  # 0:Public; 1:Private
    marketplace = Column(ChoiceType(MARKETPLACE), default=0, server_default=text("0"))  # 0:A-Z; 1:US
    par_id = Column(Integer)
    prefix = Column(String)
    description = Column(Text)
    allowed_client = Column(Text)
    update_by = Column(String)
    update_at = Column(DateTime(timezone=True), default=datetime.now(UTC))

    route_plan_id = synonym('rout_id')
    name = synonym('product_name')
    tech_prefix = synonym('prefix')
    update_on = synonym('update_at')

    route_plan = relationship('RouteStrategy')
    rate_table = relationship('RateTable')

    clients_ref = relationship('ProductClientsRef', uselist=True, back_populates='product',
                               single_parent=True, cascade="all, delete-orphan")
    agents_ref = relationship('ProductAgentsRef', uselist=True, back_populates='product',
                              single_parent=True, cascade="all, delete-orphan")

    trunks = relationship('ResourcePrefix', primaryjoin='foreign(ResourcePrefix.product_id) == ProductRoutRateTable.id',
                          uselist=True, back_populates='product')

    @hybrid_property
    def clients(self):
        try:
            clients = set([tr.client_id for tr in self.clients_ref])
            return list(clients)
        except:
            return []

    @clients.setter
    def clients(self, value):
        exc = []
        for tr in self.clients_ref:
            if not tr.client_id in value:
                tr.delete()
            else:
                exc.append(tr.client_id)
        for client_id in value:
            if client_id in exc:
                continue
            try:
                self.clients_ref.append(ProductClientsRef(client_id=client_id))
            except Exception as e:
                pass

    @hybrid_property
    def agents(self):
        try:
            agents = set([tr.agent_id for tr in self.agents_ref])
            return list(agents)
        except:
            return []

    @agents.setter
    def agents(self, value):
        exc = []
        for tr in self.agents_ref:
            if not tr.agent_id in value:
                tr.delete()
            else:
                exc.append(tr.agent_id)
        for agent_id in value:
            if agent_id in exc:
                continue
            try:
                self.agents_ref.append(ProductAgentsRef(agent_id=agent_id))
            except Exception as e:
                pass

    @hybrid_property
    def allowed_clients(self):
        try:
            return self.allowed_client.split(',')
        except:
            return []

    @allowed_clients.setter
    def allowed_clients(self, value):
        try:
            self.allowed_client = ','.join([str(x) for x in value])
        except:
            pass

    @property
    def agent_rate_download_link(self):
        return self.rate_table.rate_download_link

        import jwt
        exp = datetime.now(UTC) + timedelta(days=settings.JWT_TTL_DAYS)
        token_data = {'agent_id': self.agent_id, 'rate_table_id': self.rate_table_id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE).decode('utf-8')
        return '{}/home/agent/download/{}'.format(settings.API_URL, token)

    @property
    def rate_download_link(self):
        return self.rate_table.rate_download_link

        import jwt
        exp = datetime.now(UTC) + timedelta(days=settings.JWT_TTL_DAYS)
        token_data = {'agent_id': self.agent_id, 'rate_table_id': self.rate_table_id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE).decode('utf-8')
        return '{}/home/agent/download/{}'.format(settings.API_URL, token)


class ProductAgentsRef(BaseModel):
    __tablename__ = 'product_agents_ref'

    id = Column(Integer, primary_key=True)  # , server_default=text("nextval('product_agents_ref_id_seq'::regclass)"))
    product_id = Column(ForeignKey('product_rout_rate_table.id', ondelete='CASCADE'))
    agent_id = Column(ForeignKey('agent.agent_id', ondelete='CASCADE'))
    agent = relationship('Agent', )
    product = relationship('ProductRoutRateTable', uselist=False, back_populates='agents_ref')

    name = column_property(
        select([ProductRoutRateTable.name]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))
    tech_prefix = column_property(
        select([ProductRoutRateTable.tech_prefix]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))
    type = column_property(
        select([ProductRoutRateTable.type]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))
    marketplace = column_property(
        select([ProductRoutRateTable.marketplace]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))
    description = column_property(
        select([ProductRoutRateTable.description]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))
    rate_table_id = column_property(
        select([ProductRoutRateTable.rate_table_id]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))

    route_plan_id = column_property(
        select([ProductRoutRateTable.route_plan_id]).where(ProductRoutRateTable.id == product_id).correlate_except(
            ProductRoutRateTable))

    @property
    def agent_rate_download_link(self):
        return self.product.rate_download_link

        import jwt
        exp = datetime.now(UTC) + timedelta(days=settings.JWT_TTL_DAYS)
        token_data = {'agent_id': self.agent_id, 'rate_table_id': self.rate_table_id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE).decode('utf-8')
        return '{}/home/agent/download/{}'.format(settings.API_URL, token)


ProductAgentsRef.agent_name = column_property(
    select([Agent.agent_name]).where(Agent.agent_id == ProductAgentsRef.agent_id).correlate_except(Agent))


class ProductClientsRef(BaseModel):
    __tablename__ = 'product_clients_ref'
    __table_args__ = (
        UniqueConstraint('product_id', 'client_id'),
    )
    id = Column(Integer, primary_key=True)  # , server_default=text("nextval('product_clients_ref_id_seq'::regclass)"))
    product_id = Column(ForeignKey('product_rout_rate_table.id', ondelete='CASCADE'))
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'))
    product = relationship('ProductRoutRateTable', uselist=False, back_populates='clients_ref')
    client = relationship('Client', uselist=True)
    trunks = relationship('Resource', primaryjoin='and_(Resource.client_id==foreign(ProductClientsRef.client_id),'
                                                  'ResourcePrefix.product_id==foreign(ProductClientsRef.product_id),'
                                                  'foreign(ResourcePrefix.resource_id)==Resource.resource_id)',
                          viewonly=True, uselist=True)


class ProductItems(BaseModel):
    __tablename__ = 'product_items'
    __table_args__ = (
        UniqueConstraint('product_id', 'digits', 'code_name'),
    )
    STRATEGY_CHOICES = {
        0: 'By Percentage',
        1: 'Top-Down',
        2: 'Round-Robin'
    }
    item_id = Column(Integer, primary_key=True)
    product_id = Column(Integer, ForeignKey('product.product_id', ondelete='CASCADE'))
    alias = Column(String(100))
    digits = Column(PrefixRange(), nullable=False, server_default=text("''::prefix_range"))
    # strategy = Column( Integer)# 0-;By Percentage;1-Top-Down;2-Round-Robin
    strategy = Column(ChoiceType(STRATEGY_CHOICES), default='Top-Down', nullable=False, server_default=text("1"))
    time_profile_id = Column(Integer, ForeignKey('time_profile.time_profile_id', ondelete='SET NULL'))
    min_len = Column(Integer)
    max_len = Column(Integer)
    min_asr = Column(Float)
    max_asr = Column(Float)
    min_abr = Column(Float)
    max_abr = Column(Float)
    min_acd = Column(Float)
    max_acd = Column(Float)
    min_pdd = Column(Integer)
    max_pdd = Column(Integer)
    min_aloc = Column(Float)
    max_aloc = Column(Float)
    limit_price = Column(Float)
    update_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    update_by = Column(String)
    code_name = Column(String(100))
    # time = Column( Float)
    # flag = Column( SmallInteger, autoincrement=False)
    # record_id = Column( Integer, nullable=False)
    static_route_id = synonym('product_id')
    did = synonym('digits')

    product = relationship('Product', uselist=False, back_populates='items')
    trunks = relationship('ProductItemsResource', uselist=True, back_populates='product',
                          # lazy='joined',
                          single_parent=True,
                          cascade="all, delete-orphan", lazy='dynamic')
    time_profile = relationship('TimeProfile', uselist=False)

    time_profile_name = column_property(
        select([TimeProfile.name]).where(TimeProfile.time_profile_id == time_profile_id).correlate_except(TimeProfile))
    static_route_name = column_property(
        select([Product.name]).where(Product.product_id == product_id).correlate_except(Product))

    @hybrid_property
    def product_name(self):
        return self.product.name

    @classmethod
    def assign_client_did(cls, resource, did, user=None):
        # Route.init()
        static_route_id = Product.get_orig_static_route_id()
        cls.filter(and_(cls.product_id == static_route_id, cls.digits == did)).delete()
        item = cls(product_id=static_route_id, digits=did, strategy='Top-Down', update_by=user)
        res_item = ProductItemsResource(order_type=1)
        res_item.resource = resource
        item.trunks.append(res_item)
        resource.session().add(item)  #
        resource.session().add(res_item)  #


class ProductItemsResource(BaseModel):
    __tablename__ = 'product_items_resource'
    __table_args__ = (UniqueConstraint('item_id', 'resource_id'),)

    id = Column(Integer, primary_key=True, server_default=text("nextval('product_items_resource_id_seq'::regclass)"))
    item_id = Column(ForeignKey('product_items.item_id', ondelete='CASCADE'), nullable=False, index=True)
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    by_percentage = Column(Integer)
    order_id = Column(Integer)
    order_type = Column(Integer)  # 1-buy order;2-sell order

    trunk_id = synonym('resource_id')

    product = relationship('ProductItems', uselist=False, back_populates='trunks')
    resource = relationship('Resource', uselist=False, back_populates='product_items')

    @hybrid_property
    def trunk_name(self):
        return self.resource.alias

    @hybrid_property
    def carrier_name(self):
        return self.resource.client.name


class CarrierGroup(BaseModel):
    __tablename__ = 'carrier_group'

    group_id = Column(Integer, primary_key=True)
    group_name = Column(String, unique=True, nullable=False)

    clients_detail = relationship('Client', back_populates='group', uselist=True)

    def clients_count(self):
        return len(self.clients_detail)

    @hybrid_property
    def clients(self):
        try:
            cli = self.clients_detail
            if cli:
                return [q.client_id for q in cli]
            else:
                return []
        except:
            return None

    @clients.setter
    def clients(self, value):
        for t in self.clients_detail:
            t.group_id = None  # self.clients_rel.remove(t)
        for client_id in value:
            try:

                q = Client.get(client_id)
                q.group_id = self.group_id
                self.clients_detail.append(q)
            except Exception as e:
                pass


class TrunkGroup(BaseModel):
    __tablename__ = 'trunk_group'
    # TRUNK_TYPE_DICT = {1: "class4", 2: "exchange", 3: "product_default", 4: "product_agent"}
    TRUNK_TYPE_DICT = {1: "ingress", 2: "egress"}
    group_id = Column(Integer, primary_key=True)
    group_name = Column(String, nullable=False, unique=True)
    trunk_type = Column(ChoiceType(TRUNK_TYPE_DICT), default=1)

    trunks = relationship('Resource', back_populates='group', uselist=True)

    # --

    def trunk_match(self, q):
        return q and ((q.egress and self.trunk_type == 'egress') or
                      (q.ingress and self.trunk_type == 'ingress'))

    @hybrid_property
    def all_trunks(self):
        try:
            return [q.resource_id for q in self.trunks]
        except:
            return []

    @all_trunks.setter
    def all_trunks(self, value):

        # for t in self.trunks:
        # if self.group_id and t.group_id == self.group_id:
        #    t.group_id = None
        #    t.save()
        # self.trunks.remove(t)
        tr = Resource.filter(Resource.group_id == self.group_id).all()
        for t in tr:
            t.group_id = None
            self.session().add(t)
        for trunk_id in value:
            try:
                q = Resource.filter(Resource.resource_id == trunk_id).first()
                if self.trunk_match(q):
                    # q.group_id = self.group_id
                    # q.save()
                    self.trunks.append(q)
            except:
                pass

    # self.query().session.refresh(self)

    def __repr__(self):
        return self.group_name


class InvoiceSetting(BaseModel):
    __tablename__ = 'invoice_settings'
    INVOICE_FORMAT = {1: 'PDF', 2: 'Excel', 3: 'HTML'}
    POSITION = {0: 'bottom', 1: 'top', 2: 'middle'}
    INFOPOSITION = {0: 'top', 1: 'left'}
    INVOICE_SEND_MODE = {None: 'none', 1: 'link', 0: 'attachment'}
    INVOICE_SEND_MODE_CDR = {None: 'none', 1: 'link', 0: 'attachment'}
    id = Column(BigInteger, primary_key=True)
    invoice_setting_name = Column(String(200))
    invoice_name = Column(String(200))
    tpl_number = Column(ChoiceType(POSITION))
    company_info_location = Column(ChoiceType(INFOPOSITION))
    pdf_tpl = Column(Text)
    company_info = Column(String(500))
    overlap_invoice_protection = Column(Boolean)
    send_cdr_fields = Column(Text)
    invoice_send_mode = Column(ChoiceType(INVOICE_SEND_MODE))
    invoice_decimal_digits = Column(Integer)
    created_at = Column(DateTime(True), server_default=func.now())
    updated_at = Column(DateTime(True))
    updated_by = Column(String(255))
    mail_sender_id = Column(Integer)

    billing_info = synonym('pdf_tpl')
    billing_info_position = synonym('tpl_number')
    billing_info_location = synonym('tpl_number')
    create_by = synonym('updated_by')
    create_on = synonym('created_at')
    update_on = synonym('updated_at')
    invoices_cdr_fields = synonym('send_cdr_fields')

    @hybrid_property
    def allow_invoice_overlap(self):
        return not bool(self.overlap_invoice_protection)

    @property
    def ilogo_url(self):
        conf = SystemParameter.get(1)
        return conf.ilogo_url

    @hybrid_property
    def cdr_fields(self):
        try:
            return self.send_cdr_fields.split(',')
        except:
            return []

    @cdr_fields.setter
    def cdr_fields(self, value):
        try:
            self.send_cdr_fields = ','.join([str(x) for x in value])
        except:
            pass


class ClientInvoiceSetting(BaseModel):
    __tablename__ = 'client_invoice_settings'

    id = Column(Integer, primary_key=True)
    client_id = Column(ForeignKey('client.client_id'), nullable=False)
    invoice_setting_id = Column(ForeignKey('invoice_settings.id'))

    client = relationship('Client', back_populates='invoice_setting')
    setting = relationship('InvoiceSetting', )


class Client(BaseModel):
    __tablename__ = 'client'
    INVOICE_FORMAT = {1: 'PDF', 2: 'Excel', 3: 'HTML'}
    CDR_LIST_FORMAT = {1: 'Excel', 2: 'CSV', 3: 'zip', 4: 'tar.gz'}
    PROFIT_TYPE = {1: 'percentage', 2: 'value'}
    AUTO_INVOICE_TYPE = {0: 'buy', 1: 'sell', 2: 'both'}
    AUTO_DAILY_BALANCE_RECIPIENT = {0: "Partner Billing Contact", 1: "Owner Billing Contact", 2: "Both"}
    SCC_TYPE = {0: 'meeting the short duration defined neighboring', 1: 'that exceed the defined percentage'}
    AUTO_SUMMARY_GROUP_BY = {0: 'By Country', 1: 'By Code Name', 2: 'By Code'}
    CLIENT_TYPE = {0: 'vendor', 1: 'client', None: 'carrier'}
    COMPANY_TYPE = {0: 'termination', 1: 'origination', 2: 'both termination and origination'}
    INVOICE_USE_BALANCE_TYPE = {0: 'actual', 1: 'mutual'}
    BREAKDOWN_BY_RATE_TABLE = {1: 'Breakdown A-Z Rate Table by Destination',
                               2: 'Breakdown US Rate Table by Jurisdiction'}
    RATE_VALUE = {1: 'Actual Value', 2: 'Average Value', None: 'null', 0: 'undefined'}
    BILLING_MODE = {1: 'prepay', 2: 'postpay'}
    DID_BILLING_BY = {0: 'by minute', 1: 'by port'}

    client_id = Column(Integer, primary_key=True)

    name = Column(String(500), unique=True)  # name IS 'name';
    address = Column(String(500))  # address
    allowed_credit = Column(Numeric(30, 10), nullable=False, default=0, server_default='0')  # 'Allow arrears'
    attach_cdrs_list = Column(Boolean)  # whether the attachment bill details
    auto_daily_balance_recipient = Column(ChoiceType(AUTO_DAILY_BALANCE_RECIPIENT),
                                          default=0)  # 0: Partner' s Billing Contact, 1: Owner 's Billing Contact, 2: Both'
    auto_invoice_type = Column(ChoiceType(AUTO_INVOICE_TYPE), nullable=False, default=2,
                               server_default='2')  # 0: buy (client), 1: sell (vendor), 2: both
    auto_invoicing = Column(Boolean, nullable=False, default=True)  # Automatically generate invoice
    auto_send_zone = Column(String, server_default=text("'+00:00'"))
    auto_summary_group_by = Column(ChoiceType(AUTO_SUMMARY_GROUP_BY),
                                   default=0)  # 0: By Country; 1: By Code Name; 2: By Code
    auto_summary_hour = Column(Integer)
    auto_summary_include_cdr = Column(Boolean, nullable=False, default=False)
    auto_summary_not_zero = Column(SmallInteger, default=0)
    auto_summary_period = Column(Integer, default=24)  # 1,2,4,6,8,12,24, -15, -30
    billing_email = Column(String)
    breakdown_by_rate_table = Column(ChoiceType(
        BREAKDOWN_BY_RATE_TABLE))  # 1: Breakdown A-Z Rate Table by Destination, 2: Breakdown US Rate Table by Jurisdiction';
    call_limit = Column(Integer)
    carrier_template_id = Column(Integer, nullable=False,
                                 default=0)  # using the Carrier Template template, defaults to 0, does not use the template
    cdr_format = synonym('cdr_list_format')
    cdr_list_format = Column(ChoiceType(CDR_LIST_FORMAT))  # bill format: 1-Excel, 2-CSV, 3-zip, 4-tar.gz
    client_type = Column(ChoiceType(CLIENT_TYPE))  # did: 0 vendor, 1 client
    company = Column(String)  # company name
    # company_type = Column(ChoiceType(COMPANY_TYPE))  # did: 0 vendor, 1 client
    corporate_contact_email = Column(String(100))
    cps_limit = Column(Integer)
    create_time = Column(DateTime(True), nullable=False, server_default=func.now())  # useless - ready to delete
    currency_id = Column(ForeignKey('currency.currency_id', ondelete='SET NULL'), nullable=False, index=True, default=1)
    # currency_id = Column(Integer)
    daily_balance_already_sent = Column(Boolean, default=False)
    daily_balance_notification = Column(Integer)
    daily_balance_recipient = Column(Integer)
    daily_balance_send_time = Column(Time)
    daily_balance_send_time_zone = Column(String(20))
    daily_cdr_generation = Column(Boolean, default=False)
    daily_cdr_generation_type = Column(SmallInteger, default=0)
    daily_cdr_generation_zone = Column(String, server_default=text("'+00:00'"))
    daily_limit = Column(Integer)
    database_name = Column(String(50))
    decimals_num = Column(Integer)
    details = Column(String(1000))
    did_invoice_include = Column(String(1000))
    email = Column(String(100))  # email

    # enable_auto_invoice = Column(Boolean, default=False)
    # enable_auto_report = Column(Boolean, default=False)
    # enable_client_portal = Column(Boolean, default=False)
    # enable_notification = Column(Boolean, default=False)
    # enable_short_dur_penalty = Column(Boolean, default=False)
    enable_payment_alert = Column(Boolean, default=False)
    enough_balance = Column(Boolean, default=True)  # if the balance is insufficient
    finance_email_cc = Column(String)
    format = synonym('invoice_format')
    group_id = Column(ForeignKey('carrier_group.group_id', ondelete='SET NULL'), index=True)
    hourly_limit = Column(Integer)
    include_available_credit = Column(SmallInteger, default=0)
    include_payment_history = Column(SmallInteger, default=0)
    include_payment_history_days = Column(Integer, default=1)
    invoice_include_payment = Column(Boolean)
    invoice_jurisdictional_detail = Column(Boolean, default=False)
    invoice_past_amount = Column(Numeric(30, 10), default=0)
    invoice_show_details = Column(Boolean, default=False)
    invoice_start_from = Column(Date)
    invoice_use_balance_type = Column(ChoiceType(INVOICE_USE_BALANCE_TYPE), default=0)  # 0-actual, 1-mutual'
    is_active_call = Column(Boolean, default=False)
    is_auto_balance = Column(Boolean, default=False)
    is_auto_summary = Column(Boolean, default=False)
    is_bill_mismatch = Column(Boolean, default=False)
    is_breakdown_by_rate_table = Column(Boolean)  # invoice whether to display Break Down by Rate Table
    is_call_simulation = Column(Boolean, default=False)
    is_capture = Column(Boolean, default=False)
    is_daily_balance_notification = Column(Boolean, default=False)
    is_discon = Column(Boolean, default=False)
    is_egress_sim = Column(Boolean, default=False)
    is_ingress_sim = Column(Boolean, default=False)
    is_invoice_account_summary = Column(Boolean, default=False)
    is_link_cdr = Column(Boolean, default=True)  # can link to download cdr
    is_location = Column(Boolean, default=False)
    is_manage = Column(Boolean, default=False)
    is_orig_term = Column(Boolean, default=False)
    is_panel_accountsummary = Column(Boolean, default=True)  # can access Account Summary
    is_panel_balance = Column(Boolean, default=True)  # can access Balance
    is_panel_cdrslist = Column(Boolean, default=True)  # Can access CDRs List
    is_panel_cid_blocking = Column(Boolean, default=True)
    is_panel_didrequest = Column(Boolean, default=True)  # can access DID Request
    is_panel_invoices = Column(Boolean, default=True)  # can access Invoices
    is_panel_mydid = Column(Boolean, default=False)  # can access My DID
    is_panel_onlinepayment = Column(Boolean, default=True)  # can access Online Payment
    is_panel_paymenthistory = Column(Boolean, default=True)  # can access Payment History
    is_panel_products = Column(Boolean, default=True)  # can access products
    is_panel_ratetable = Column(Boolean, default=True)  # can access RateTable
    is_panel_sippacket = Column(Boolean, default=True)  # Can access SIP PACKET Search
    is_panel_summaryreport = Column(Boolean, default=True)  # can access Summary Report
    is_panel_trunks = Column(Boolean, default=True)  # can access Trunks
    is_panelaccess = Column(Boolean)  # can access the client template - no longer use, ready to delete
    is_qos = Column(Boolean, default=False)
    is_rate_anal = Column(Boolean, default=False)
    is_report = Column(Boolean, default=False)
    is_route = Column(Boolean, default=False)
    is_send_trunk_update = Column(Boolean, default=True)  # whether to send mail when changing its trunk ip
    is_show_by_date = Column(Boolean, default=False)
    is_show_detail_trunk = Column(Boolean, default=False)
    is_show_total_trunk = Column(Boolean, default=False)
    is_spam = Column(Boolean, default=False)
    is_termin = Column(Boolean, default=False)
    is_tools = Column(Boolean, default=False)
    is_trunk = Column(Boolean, default=False)
    is_usage = Column(Boolean, default=False)
    is_vendor_invoice = Column(Boolean, default=False)
    last_autobalance_time = Column(DateTime(True))
    last_autoreport_time = Column(DateTime(True))
    last_cdrdown_time = Column(DateTime(True))
    last_invoiced = Column(DateTime(True), server_default=func.now())  # last invoice generation date
    last_lowbalance_time = Column(DateTime(True))
    login = Column(String(60), unique=True)  # customer template access account, customer login account
    logo = Column(String(100))  # logo
    low_balance_notice = Column(Boolean, nullable=False, default=True)  # whether to provide balance insufficient hints
    low_balance_notification_time_cycle = Column(Integer, default=1)
    low_balance_notification_time_type = Column(Integer, default=0)
    low_balance_number = Column(Integer)
    mail_sended = Column(Integer, default=0)  # 0 - no send, 1 - balance insufficient send
    mode = Column(ChoiceType(BILLING_MODE), nullable=False, default=1)  # PAYMENT MODE: 1 - Prepaid, 2 postpaid
    # mode = Column(Integer, nullable=False, default=1)  # PAYMENT MODE: 1 - Prepaid, 2 postpaid
    noc_email = Column(String(100))
    notify_admin_balance = Column(Numeric(30, 10), default=0)  # prompt system balance
    notify_client_balance = Column(Numeric(30, 10), default=0)  # prompt customer balance
    notify_client_balance_type = Column(Integer, default=0)
    numer_of_days_balance = Column(Integer, default=1)
    offset_balance = Column(Boolean, default=True)
    orig_rate_table_id = Column(Integer, index=True)
    par_id = Column(Integer)
    password = Column(String(60))  # client template access password
    # payment_received_notice = Column(Boolean)

    paypal = Column(String(100))
    phone = Column(String(20))
    profit_margin = Column(Float(53), nullable=False, default=0)  # minimum profit margin
    profit_type = Column(ChoiceType(PROFIT_TYPE), default=1)  # 1 - Percentage, 2 - value
    rate_delivery_email = Column(String(250))
    rate_email = Column(String(100))
    role_id = Column(Integer, index=True)  #
    scc_bellow = Column(Integer, default=0)  # call_duration Call time is less than this value of a short call
    scc_charge = Column(Numeric(30, 10))  # short call amount
    scc_percent = Column(Integer,
                         default=0)  # short call answered * 100 / all greater than this value in the invoice to add a brief call fine
    scc_type = Column(Integer,
                      default=0)  # 0 meeting the short duration defined neighboring, 1 that exceed the defined percentage
    service_charge_id = Column(Integer)
    status = Column(Boolean, nullable=False, default=True)  ##Status: true-active, false-inactive
    tax = Column(Numeric(5, 2))
    tax_id = Column(String(100))
    term_rate_table_id = Column(Integer, index=True)  # 'cost - useless, ready to delete';
    transaction_fee_id = Column(Integer)
    unlimited_credit = Column(Boolean, default=False)
    update_at = Column(DateTime(True), onupdate=func.now())
    update_by = Column(String)
    usage_detail_fields = Column(Text)
    user_id = Column(Integer)
    vendor_payment_term_id = Column(Integer)
    zero_balance_notice = Column(Boolean)
    zero_balance_notice_first_sended = Column(Boolean, default=False)
    zero_balance_notice_last_sent = Column(DateTime(True))
    zero_balance_notice_time = Column(Integer, default=0)

    # invoice setting fields
    payment_term_id = Column(ForeignKey('payment_term.payment_term_id', ondelete='SET NULL'), index=True)
    invoice_format = Column(ChoiceType(INVOICE_FORMAT))  # Generate Invoice Format: 1-PDF, 2-Excel, 3-HTML
    invoice_zone = Column(String(10))
    decimal_place = Column(Integer, default=5)  # number of decimal places, default 5 bits
    rate_value = Column(ChoiceType(RATE_VALUE), default=1)  # rate_value 1 - Actual Value 2 - Average Value
    invoice_zero = Column(Boolean, default=True)  # default does not generate

    include_tax = Column(Boolean, default=False)
    email_invoice = Column(Boolean, default=False)
    is_send_as_link = Column(Boolean, default=False)
    is_show_daily_usage = Column(Boolean, default=False)
    is_short_duration_call_surcharge_detail = Column(Boolean, default=False)

    is_show_code_100 = Column(Boolean, default=False)
    is_show_code_name = Column(Boolean, default=False)
    is_show_country = Column(Boolean, default=False)

    company_type = column_property(case([(client_type.is_(None), 'termination')], else_='origination'))
    # client_invoice_settings_id = Column(ForeignKey('client_invoice_settings.id', ondelete='SET NULL'), index=True)
    client_invoice_settings_id = column_property(select([ClientInvoiceSetting.invoice_setting_id]).where(
        ClientInvoiceSetting.client_id == client_id).correlate_except(ClientInvoiceSetting))
    did_billing_by = Column(ChoiceType(DID_BILLING_BY), server_default='0')
    fee_per_port = Column(Numeric(30, 10))

    is_online = column_property(select([User.is_online == 1]).where(User.user_id == user_id).correlate_except(User))

    client_name = synonym('name')
    auto_send_invoice = synonym('email_invoice')
    daily_usage_group_by = synonym('auto_summary_group_by')
    daily_balance_summary = synonym('is_daily_balance_notification')
    daily_usage_summary = synonym('is_show_daily_usage')
    daily_cdr_delivery = synonym('daily_cdr_generation')
    main_email = synonym('email')
    billing_method = synonym('profit_type')
    username = synonym('login')
    active = synonym('status')
    is_active = synonym('status')
    send_invoice_as_link = synonym('is_send_as_link')
    is_unlimited = synonym('unlimited_credit')
    scc_below = synonym('scc_bellow')
    port_limit = synonym('call_limit')

    payment_term_name = column_property(
        select([PaymentTerm.name]).where(PaymentTerm.payment_term_id == payment_term_id).correlate_except(PaymentTerm))

    currency = relationship(Currency)
    group = relationship(CarrierGroup, back_populates='clients_detail', uselist=False)

    payment_term = relationship(PaymentTerm, back_populates='clients', uselist=False)

    agents = relationship(AgentClient, back_populates='client', uselist=True, single_parent=True,
                          cascade="all, delete-orphan")
    invoices = relationship('Invoice', back_populates='client', uselist=True, single_parent=True,
                            cascade="all, delete-orphan")
    last_invoice_history = relationship('InvoiceHistory',
                                        primaryjoin="foreign(InvoiceHistory.client_id)==Client.client_id",
                                        back_populates='client', uselist=False, single_parent=True,
                                        cascade="all, delete-orphan")

    # payments_history = relationship('PaymentGatewayHistory', uselist=False, back_populates='client')
    #
    # default_ip = relationship('ClientDefaultIp', uselist=True, back_populates='client', single_parent=True,
    #                           cascade="all, delete-orphan")
    # client_taxes = relationship('ClientTaxes', uselist=True, back_populates='client', single_parent=True,
    #                             cascade="all, delete-orphan")

    user = relationship(User, back_populates='client',
                        primaryjoin='foreign(Client.user_id) == User.user_id', uselist=False)
    # foreign(Client.user_id) == User.user_id
    resources = relationship('Resource', uselist=True, back_populates='client')
    resource = relationship('Resource', uselist=False,
                            primaryjoin='and_(Resource.client_id==Client.client_id,foreign(Resource.alias)==Client.name)'
                            , single_parent=True, cascade="all, delete-orphan")

    ip_addresses = relationship('ResourceIp',
                                secondary="join(Resource, ResourceIp, Resource.resource_id == ResourceIp.resource_id)",
                                primaryjoin='Resource.client_id==Client.client_id',
                                secondaryjoin='ResourceIp.resource_id==Resource.resource_id',
                                viewonly=True, uselist=True)
    payment = relationship('ClientPayment', uselist=True, back_populates='client',
                           single_parent=True,
                           cascade="all, delete-orphan")
    c4 = relationship('C4ClientBalance',
                      primaryjoin="Client.client_id == foreign( cast(C4ClientBalance.client_id,Integer) )",
                      viewonly=True, uselist=True)
    low_balance_config = relationship('CarrierLowBalanceConfig', uselist=False, back_populates='carrier',
                                      single_parent=True,
                                      cascade="all, delete-orphan"
                                      )
    c4 = relationship('C4ClientBalance', uselist=False,
                      # back_populates='carrier',
                      primaryjoin="foreign(Client.client_id) == cast(C4ClientBalance.client_id,Integer)"
                      )
    balance_actual = relationship('BalanceHistoryActual', back_populates='carrier', uselist=True, lazy='dynamic',
                                  single_parent=True,
                                  cascade="all, delete-orphan")
    balance_mutual = relationship('BalanceHistory', back_populates='carrier', uselist=True, lazy='dynamic',
                                  single_parent=True,
                                  cascade="all, delete-orphan")
    egress_trunks = relationship('Resource',
                                 primaryjoin='and_(Resource.client_id == Client.client_id,Resource.egress == True)'
                                 )
    ingress_trunks = relationship('Resource',
                                  primaryjoin='and_(Resource.client_id == Client.client_id,Resource.ingress == True)'
                                  )
    origination_trunks = relationship('Resource',
                                      primaryjoin='and_(Resource.client_id == Client.client_id,foreign(Resource.alias)==Client.name,Resource.trunk_type2 == 1)'
                                      )

    invoice_setting = relationship(ClientInvoiceSetting, uselist=False, back_populates='client',
                                   cascade="all, delete-orphan")

    client_limits = relationship('UsersLimit', uselist=True, back_populates='client', cascade="all, delete-orphan")

    # vendor_api = relationship('DidVendor', uselist=False, back_populates='vendor', cascade="all, delete-orphan")

    __is_new = None

    @property
    def invoice_conf(self):
        if self.invoice_setting and self.invoice_setting.setting:
            return self.invoice_setting.setting
        else:
            return SystemParameter.query().first()

    @property
    def invoice_send_mode_cdr(self):
        if self.attach_cdrs_list:
            return 'link'
        return None

    # ++IStatistic implementations
    def _ingress(self):
        return ','.join([str(tr.resource_id) for tr in self.ingress_trunks])

    def _egress(self):
        return ','.join([str(tr.resource_id) for tr in self.egress_trunks])

    @property
    def _trunks_term_ids(self):
        return ','.join([str(tr.resource_id) for tr in self.resources if tr.trunk_type2 == 'Termination Traffic'])

    @property

    def _trunks_term_az_ids(self):
        return ','.join(
            [str(tr.resource_id) for tr in Resource.filter(
                and_(Resource.client_id == self.client_id, ResourcePrefix.resource_id == Resource.resource_id,
                     RateTable.rate_table_id == ResourcePrefix.rate_table_id, RateTable.jur_type==0,
                     Resource.trunk_type2 == 'Termination Traffic'
                     )).all()])

    @property
    def _trunks_term_us_ids(self):
        return ','.join(
            [str(tr.resource_id) for tr in Resource.filter(
                and_(Resource.client_id == self.client_id, ResourcePrefix.resource_id == Resource.resource_id,
                     RateTable.rate_table_id == ResourcePrefix.rate_table_id, RateTable.jur_type.in_([1, 2]),
                     Resource.trunk_type2 == 'Termination Traffic')).all()])

    @property
    def _trunks_orig_ids(self):
        return ','.join([str(tr.resource_id) for tr in self.resources if tr.trunk_type2 == 'DID Traffic'])

    @property
    def ingress_ids(self):
        return self._ingress()

    @property
    def egress_ids(self):
        return self._egress()

    # --IStatistic implementations
    def before_save(self):
        return
        log.debug('Client before save')
        # super(Client,self).before_save()
        # if self.client_id:
        # Resource.filter(Resource.client_id == self.client_id).update({'enough_balance': self.enough_balance})
        #    pass

        if self.mode == 'prepay' and self.client_type != 'vendor':
            # if hasattr(self,'_test_credit' ) and self._test_credit:
            #    self._test_credit = None
            #    raise ValidationError('Cannot set test_credit when mode is prepay!')
            if self.allowed_credit:
                self.allowed_credit = 0.0
            #    raise ValidationError('Cannot set credit limit when mode is prepay!')
            if self.unlimited_credit:
                self.unlimited_credit = False
            #    raise ValidationError('Cannot set unlimited credit when mode is prepay!')
            pass

        if self.unlimited_credit:
            if self.allowed_credit:
                self.allowed_credit = 0.0
        log.debug('Client before save allowed credit {}'.format(self.allowed_credit))
        if self.allowed_credit is None:
            self.allowed_credit = 0.0
        try:
            get_db(self.db_label).session.autoflush = False
            if self.resource:
                log.debug('Client resource before save')
                self.resource.before_save()
        except Exception as e:
            log.error('Client before save error {}'.format(traceback.print_exc()))
            get_db(self.db_label).session.autoflush = True
        # for r in self.resources:
        #    r.before_save()

        self.__is_new = not bool(self.client_id)
        log.debug('Client __is_new {}'.format(self.__is_new))

    def after_save(self):
        return
        log.debug('Client resource after save')
        balance = 0.0

        _p_changed = False
        if not self.user and self.login and self.password:
            self.user = User(name=self.login, passwd=self.password)
            _p_changed = True
        if self.user:
            try:
                if self.user.active != self.status:
                    self.user.active = self.status
                    _p_changed = True
                if self.user.client_id != self.client_id:
                    self.user.client_id = self.client_id
                    _p_changed = True
                if self.user.name != self.login:
                    self.user.name = self.login
                    _p_changed = True
                try:
                    if not bcrypt.checkpw(self.password.encode('utf-8'), self.user.password.encode('utf-8')):
                        self.user.passwd = self.password
                        _p_changed = True
                except:
                    self.user.passwd = self.password
                    _p_changed = True
                if not self.user.email and self.email and len(self.email.split(';')):
                    self.user.email = self.email.split(';')[0]
                    _p_changed = True
                conf = SystemParameter.get(1)
                # if conf and not self.user.landing_page:
                #     self.user.landing_page = conf.base_url
                #     _p_changed = True
                if not self.user.role_id:
                    self.user.role_id = 1
                    _p_changed = True
                if self.user.user_type != 'client':
                    self.user.user_type = 'client'
                if _p_changed:
                    log.debug('Client resource after save user changed!')
                    self.user.save()
            except Exception as e:
                log.error('save client {} user error {}'.format(self.client_id, str(e)))

        if self.__is_new:

            ClientBalanceOperationAction.add_create(self.client_id, Decimal(0.0), self.update_by + '_create_balance')

            # if not C4ClientBalance.filter(cast(C4ClientBalance.client_id,Integer) == self.client_id).first():
            #    C4ClientBalance(client_id=str(self.client_id),balance=balance,ingress_balance=balance).save()
            if not CarrierLowBalanceConfig.filter(CarrierLowBalanceConfig.client_id == self.client_id).first():
                CarrierLowBalanceConfig(client_id=self.client_id).save()
        if hasattr(self, '_test_credit') and self._test_credit:
            balance = self._test_credit
            # ClientBalanceOperationAction.add_create(self.client_id, Decimal(0.0),self.update_by+'_test_credit')
            ClientPayment(client_id=self.client_id, amount=balance, payment_type='prepay payment received',
                          description='test credit', update_by=self.update_by,
                          receiving_time=self.create_time).save()
            del self._test_credit
        try:
            Resource.filter(Resource.client_id == self.client_id).update({'enough_balance': self.enough_balance},
                                                                         synchronize_session='fetch')
            Resource.session().commit()
            pass
        except:
            Resource.session().rollback()
            pass

    # def ingress_count(self):
    #    return Resource.filter(Resource.client_id==self.client_id).filter(Resource.ingress==True).count()
    # def egress_count(self):
    #    return Resource.filter(Resource.client_id==self.client_id).filter(Resource.egress==True).count()

    @hybrid_property
    def auto_summary_not_zero_bool(self):
        try:
            return bool(self.auto_summary_not_zero)
        except:
            return True

    @auto_summary_not_zero_bool.setter
    def auto_summary_not_zero_bool(self, value):
        try:
            self.auto_summary_not_zero = int(value)
        except:
            self.auto_summary_not_zero = 1

    def get_product_trunk(self, product_id):
        p = ProductRoutRateTable.get(product_id)
        if p:
            return Resource.filter(Resource.client_id == self.client_id).filter(
                Resource.product_id == p.id).first()

    def add_product_trunk(self, product_id):
        p = ProductRoutRateTable.get(product_id)
        if p:
            tr = self.get_product_trunk(product_id)
            if not tr:
                trunk = Resource(product_id=p.id, route_strategy_id=p.route_plan_id, rate_table_id=p.rate_table_id,
                                 name=self.name + '_' + p.name, alias=self.name)
                pref = ResourcePrefix(tech_prefix=p.tech_prefix)
                trunk.prefixes.append(pref)
                self.resources.append(trunk)

    def del_product_trunk(self, product_id):
        p = ProductRoutRateTable.get(product_id)
        if p:
            tr = self.get_product_trunk(product_id)
            if tr:
                self.resources.remove(tr)
            tr.delete()

    def balance(self):
        return self.actual_balance

    @hybrid_property
    def _actual_balance(self):
        # d = str(datetime.today().date())
        if self.c4:
            return self.c4.balance
        else:
            return 0.0

        d = self.query().session.query(
            func.max(BalanceHistoryActual.date).label('d')).filter(
            BalanceHistoryActual.client_id == self.client_id).first().d
        try:
            return BalanceHistoryActual.filter(BalanceHistoryActual.client_id == self.client_id).filter(
                BalanceHistoryActual.date == d).first().actual_balance
        except:
            return 0.0

    @hybrid_property
    def credit_limit(self):
        if self.allowed_credit:
            return -self.allowed_credit
        else:
            return 0.0

    @credit_limit.setter
    def credit_limit(self, value):
        if self and value:
            self.allowed_credit = -Decimal(value)
        else:
            self.allowed_credit = 0.0

    @hybrid_property
    def ip_address(self):
        return self.ip_addresses
        ret = []
        try:
            for res in self.resources:
                ret.append(res)
            return ret
        except:
            return ''

    @ip_address.setter
    def ip_address(self, value):
        try:
            if len(self.resources):
                res = self.resources[0]  # TODO: what resource to take?
            else:
                res = Resource(name=self.name, alias=self.name, client_id=self.client_id)
                self.resources.append(res)
                res = self.resources[0]
            for v in value:
                dir = 1
                if self.ingress:
                    dir = 0
                ip = ResourceIp(ip=value.ip, port=v.port, direction=dir)
                res.append(ip)

        except:
            return None

    @hybrid_property
    def billing_mode(self):
        try:
            return self.mode
        except:
            return ''

    @billing_mode.setter
    def billing_mode(self, value):
        try:
            self.mode = value
        except:
            self.mode = 1
            pass

    @hybrid_property
    def test_credit(self):
        return self.test_credit_value
        """
        try:
            s=sum([p.amount for p in self.payment if ( p.payment_type == 'prepay payment received' and p.description == 'test credit') ])
            return s
        except Exception as e:
            log.debug(str(e))
            return 0
        """

    @test_credit.setter
    def test_credit(self, value):
        self._test_credit = value

    def create_user(self):
        if not self.user:
            if not (self.login and self.password):
                return None
            if not self.login:
                self.login = self.name
            if not self.password:
                self.password = self.name
            q = User.filter(User.name == self.login).first()
            if q:
                return None
            self.user = User(name=self.login, passwd=self.password)
            if not self.user.email and self.email and len(self.email.split(';')):
                self.user.email = self.email.split(';')[0]
            self.user.user_type = 'client'
            self.user.role_id = 1
            self.user.client_id = self.client_id
            return self.user.save()
        else:
            return self.user_id

    def apply_mail(self, template, to=None, att=[]):
        # if not to:
        #    to=self.email
        # class Ret:
        #    pass
        # obj=Ret()
        # for k,v in self.__dict__.items():
        #    obj.__dict__['client_'+k]=v
        #    obj.__dict__[k] = v
        return MailSender.apply_mail(self, template, to, att=att)

    def _past_due(self):
        state = aliased(Invoice.state, 'inv_state')
        return Invoice.query().session.query(func.sum(func.coalesce(Invoice.total_amount, 0) -
                                                      func.coalesce(Invoice.pay_amount, 0)).label('total')). \
            filter(and_(Invoice.client_id == self.client_id, Invoice.paid.isnot(True), state == 9))

    def past_due(self):
        class Rec:
            pass

        rec = Rec()

        today = datetime.today()
        p7 = today - timedelta(days=7)
        p15 = today - timedelta(days=15)
        p30 = today - timedelta(days=30)
        i_d_d = Invoice.due_date
        qtotal = self._past_due().filter(i_d_d < today)
        rec.total = qtotal.one().total
        rec.past_7 = self._past_due().filter(and_(i_d_d < today, i_d_d >= p7)).one().total
        rec.past_15 = self._past_due().filter(and_(i_d_d < p7, i_d_d >= p15)).one().total
        rec.past_30 = self._past_due().filter(and_(i_d_d < p15, i_d_d >= p30)).one().total
        rec.past_gt_30 = self._past_due().filter(i_d_d < p30).one().total
        rec.balance = self.balance()
        return rec

    def regenerate_balance(self):
        td = datetime.today()
        yd = datetime.today() - timedelta(hours=24)
        d0 = date(yd.year, yd.month, yd.day)
        d1 = date(td.year, td.month, td.day)

        incoming_q = BalanceHistoryActual.filter("client_id = %d" % self.client_id).filter(
            "date='%s'" % d1).first()
        if incoming_q:
            incoming = incoming_q.actual_ingress_balance
            if incoming is None:
                incoming = 0.0
        else:
            incoming = 0.0
        pays = ClientPayment.filter("client_id = %d" % self.client_id).filter("payment_time>'%s'" % d1).all()
        psent = sum([p.total for p in pays if p.payment_type in ('invoice payment sent', 'payment sent')])
        precv = sum(
            [p.total for p in pays if p.payment_type in ('invoice payment received', 'prepay payment received')])
        drecv = sum([p.total for p in pays if p.payment_type in ('debit note received',)])
        dsent = sum([p.total for p in pays if p.payment_type in ('debit note sent',)])
        crecv = sum([p.total for p in pays if p.payment_type in ('credit note received',)])
        csent = sum([p.total for p in pays if p.payment_type in ('credit note sent',)])

        scc = sum([p.total for p in pays if p.payment_type in ('scc',)])

        sump = sum([p.total for p in pays])
        payment_received = float(sump)
        outcoming = float(incoming) + float(sump)

        q = BalanceHistoryActual.filter("client_id = %d" % self.client_id).filter(
            "date='%s'" % d1).first()
        if not q:
            q = BalanceHistoryActual(client_id=self.client_id, date=d1)
        if q:
            q.actual_balance = outcoming
            q.actual_egress_balance = outcoming
            q.payment_received = precv
            q.payment_sent = psent
            q.debit_received = drecv
            q.debit_sent = dsent
            q.credit_note_received = crecv
            q.credit_note_sent = csent
            q.short_charges = scc
            q.save()

        q = C4ClientBalance.filter(C4ClientBalance.client_id == str(self.client_id)).first()
        if not q:
            q = C4ClientBalance(client_id=self.client_id)
        if q:
            q.balance = outcoming
            q.ingress_balance = incoming
            q.egress_balance = outcoming
            q.save()

        q = BalanceHistory.filter("client_id = %d" % self.client_id).filter(
            "date='%s'" % d1).first()
        if not q:
            q = BalanceHistory(client_id=self.client_id, date=d1)
        if q:
            if q.mutual_ingress_balance:
                outcoming = q.mutual_ingress_balance + sump
            else:
                outcoming = sump
            q.mutual_balance = outcoming
            q.mutual_egress_balance = outcoming
            q.payment_received = precv
            q.payment_sent = psent
            q.debit_received = drecv
            q.debit_sent = dsent
            q.credit_note_received = crecv
            q.credit_note_sent = csent
            q.short_charges = scc
            q.save()

    def payment_terms(self):
        return self.payment_term.name

    @property
    def invoice_cycle(self):
        return self.payment_term.cycle

    @property
    def invoice_next_date(self):
        if self.payment_term:
            if self.last_invoice_history:
                return self.payment_term.next_date(self.last_invoice_history.last_invoice_for)
            else:
                return self.payment_term.next_date(self.create_time)
        return None

    def is_auto_report_time(self, t):

        zone_hours = get_time_zone_hour(self.auto_send_zone)
        t_client_tz = t + timedelta(hours=zone_hours)
        client_hours = t_client_tz.hour
        if self.auto_summary_hour is None:
            return client_hours == 0
        else:
            return self.auto_summary_hour == client_hours

    @property
    def company_name(self):
        return self.company

    @property
    def sys_company_name(self):
        return SystemParameter.get(1).company_info

    @property
    def login_url(self):
        # if self.user:
        #     return self.user.login_url
        # else:
        return SystemParameter.get(1).base_url

    # return SystemParameter.get(1).landing_page

    @property
    def sys_login_url(self):
        return SystemParameter.get(1).base_url

    @hybrid_property
    def usage_fields(self):
        try:
            return self.usage_detail_fields.split(',')
        except:
            return []

    @usage_fields.setter
    def usage_fields(self, value):
        try:
            self.usage_detail_fields = ','.join([str(x) for x in value])
        except:
            pass

    def create_from_template(self, name, template_id):
        # obj = CarrierTemplate.filter(CarrierTemplate.id == template_id).first()
        # if not obj:
        #     raise Exception('Bad template id: {} not found'.format(template_id))
        #     return
        # self.name = name
        # self.carrier_template_id = obj.id
        # self.allowed_credit = obj.allowed_credit if obj.allowed_credit else 0.0
        # self.attach_cdrs_list = obj.attach_cdrs_list
        # self.auto_daily_balance_recipient = obj.auto_daily_balance_recipient
        # self.auto_invoice_type = obj.auto_invoice_type if obj.auto_invoice_type else 'both'
        # self.auto_invoicing = False if not obj.auto_invoicing is None and not obj.auto_invoicing else True
        # self.auto_send_zone = obj.auto_send_zone
        # self.auto_summary_group_by = obj.auto_summary_group_by
        # self.auto_summary_hour = obj.auto_summary_hour
        # self.auto_summary_include_cdr = obj.auto_summary_include_cdr if obj.auto_summary_include_cdr else False
        # self.auto_summary_not_zero = obj.auto_summary_not_zero
        # self.auto_summary_period = obj.auto_summary_period
        # self.billing_mode = obj.billing_mode
        # self.breakdown_by_rate_table = obj.breakdown_by_rate_table
        # self.call_limit = obj.call_limit
        # self.cdr_format = obj.cdr_format
        # self.cdr_list_format = obj.cdr_list_format
        # self.cps_limit = obj.cps_limit
        # # self.create_by = obj.create_by
        # # self.create_on = obj.create_on
        # if obj.currency_id and Currency.get(obj.currency_id):
        #     self.currency_id = obj.currency_id
        # self.daily_balance_notification = obj.daily_balance_notification
        # self.daily_balance_recipient = obj.daily_balance_recipient
        # self.daily_balance_send_time = obj.daily_balance_send_time
        # self.daily_balance_send_time_zone = obj.daily_balance_send_time_zone
        # self.daily_balance_summary = obj.daily_balance_summary
        # self.daily_cdr_delivery = obj.daily_cdr_delivery
        # self.daily_cdr_generation = obj.daily_cdr_generation
        # self.daily_cdr_generation_type = obj.daily_cdr_generation_type
        # self.daily_cdr_generation_zone = obj.daily_cdr_generation_zone
        # self.daily_limit = obj.daily_limit
        # self.daily_usage_group_by = obj.daily_usage_group_by
        # self.daily_usage_summary = obj.daily_usage_summary
        # self.decimal_place = obj.decimal_place
        # # self.email_invoice = obj.email_invoice
        # self.format = obj.format
        # self.hourly_limit = obj.hourly_limit
        # self.include_available_credit = obj.include_available_credit
        # self.include_tax = obj.include_tax
        # self.invoice_format = obj.invoice_format
        # self.invoice_include_payment = obj.invoice_include_payment
        # self.invoice_jurisdictional_detail = obj.invoice_jurisdictional_detail
        # self.invoice_show_details = obj.invoice_show_details
        # self.invoice_start_from = obj.invoice_start_from
        # self.invoice_use_balance_type = obj.invoice_use_balance_type
        # self.invoice_zero = obj.invoice_zero
        # self.invoice_zone = obj.invoice_zone
        # self.is_auto_balance = obj.is_auto_balance
        # self.is_auto_summary = obj.is_auto_summary
        # self.is_breakdown_by_rate_table = obj.is_breakdown_by_rate_table
        # self.is_daily_balance_notification = obj.is_daily_balance_notification
        # self.email_invoice = obj.is_email_invoice
        # self.is_invoice_account_summary = obj.is_invoice_account_summary
        # self.is_send_as_link = obj.is_send_as_link
        # self.is_send_trunk_update = obj.is_send_trunk_update
        # self.is_short_duration_call_surcharge_detail = obj.is_short_duration_call_surcharge_detail
        # self.is_show_by_date = obj.is_show_by_date
        # self.is_show_code_100 = obj.is_show_code_100
        # self.is_show_code_name = obj.is_show_code_name
        # self.is_show_country = obj.is_show_country
        # # self.is_show_daily_trunk = obj.is_show_daily_trunk
        # self.is_show_daily_usage = obj.is_show_daily_usage
        # self.is_show_detail_trunk = obj.is_show_detail_trunk
        # self.is_show_total_trunk = obj.is_show_total_trunk
        # self.last_invoiced = obj.last_invoiced
        # self.mode = obj.mode
        # self.notify_client_balance = obj.notify_client_balance
        # self.notify_client_balance_type = obj.notify_client_balance_type
        # self.numer_of_days_balance = obj.numer_of_days_balance
        # self.payment_received_notice = obj.payment_received_notice
        # if obj.payment_term_id and PaymentTerm.get(obj.payment_term_id):
        #     self.payment_term_id = obj.payment_term_id
        # self.profit_margin = obj.profit_margin
        # self.profit_type = obj.profit_type
        # self.rate_value = obj.rate_value
        # self.scc_bellow = obj.scc_bellow
        # self.scc_below = obj.scc_below
        # self.scc_charge = obj.scc_charge
        # self.scc_percent = obj.scc_percent
        # self.scc_type = 1 if obj.scc_type == 'that exceed the defined percentage' else 0
        # self.send_invoice_as_link = obj.send_invoice_as_link
        # self.tax = obj.tax
        # if obj.test_credit:
        #     self._test_credit = obj.test_credit
        # self.unlimited_credit = obj.unlimited_credit
        # self.update_on = datetime.now(UTC)
        # self.usage_detail_fields = obj.usage_detail_fields
        # c = CarrierLowBalanceConfig()
        # c.create_from_template(obj.low_balance_config)
        # # c.carrier=self
        # self.low_balance_config = c
        pass

    def get_assigned_billing_plan_resource(self, plan):
        # plan=DidBillingPlan.get(billing_plan_id)
        log.debug('get_assigned_billing_plan_resource self={} plan={}'.format(self, plan))
        name = '{}_DID_{}'.format(self.name, plan.name)
        ret = Resource.filter(and_(Resource.client_id == self.client_id, Resource.alias == name)).first()
        if not ret:
            r = self.resource
            if not r:
                if self.resources and len(self.resources):
                    r = self.resources[0]
                else:
                    raise ValidationError(
                        {'client_id': ['client {} does not have default resource!'.format(self.name)]})
            ret = Resource(alias=name, client_id=self.client_id, media_type=r.media_type, capacity=r.capacity,
                           trunk_type2=r.trunk_type2, egress=r.egress, ingress=r.ingress,
                           # rate_table_id=plan.rate_table_id,
                           enough_balance=r.enough_balance, billing_rule=r.billing_rule,
                           route_strategy_id=r.route_strategy_id, t38=True)
            for ip in r.ip:
                i = ResourceIp(ip=ip.ip, port=ip.port, direction=1)
                ret.ip.append(i)
        # ret.save()
        ret.ingress = False  # True
        ret.egress = True  # False
        ret.trunk_type2 = 'DID Traffic'
        ret.update_at = datetime.now(UTC)
        # ret.create_time = datetime.now(UTC)
        # ret.rate_table = plan.rate_table ?? works bad
        ret.rate_table_id = plan.rate_table_id
        ret.client_id = self.client_id
        # self.resources.append(ret)
        self.session().add(ret)
        log.debug('get_assigned_billing_plan_resource ret{}'.format(ret))
        return ret

    def fmt_balance(self):
        return dec_prn(self.c4.balance, 2)

    @property
    def beginning_balance(self):
        b = BalanceHistoryActual
        q = b.filter(
            and_(b.client_id == self.client_id, b.date == (datetime.now(UTC) - timedelta(days=1)).date())).first()
        if q:
            return q.actual_balance
        return None

    @property
    def fmt_current_date(self):
        return str(datetime.now(UTC).date())

    @property
    def fmt_remaining_credit(self):
        beg = self.beginning_balance
        if not beg is None:
            if self.mode == 'postpay':
                rem = (-self.allowed_credit) - abs(beg)
                return dec_prn(rem, 2)
        return '0.00'

    # return dec_prn(self.c4.balance, 2)

    @property
    def notify_balance(self):
        lb = self.low_balance_config
        if not lb:
            return None
        c4 = self.c4
        if not c4:
            return None
        if lb.value_type == 'Actual Balance':
            if self.mode == 'prepay':
                return lb.actual_notify_balance
            if self.mode == 'postpay':
                return self.notify_client_balance

        elif lb.value_type == 'Percentage':
            return self.allowed_credit - lb.percentage_notify_balance * self.allowed_credit / 100.0
        return None

    @property
    def fmt_notify_balance(self):
        if not self.notify_balance is None:
            return dec_prn(self.notify_balance, 2)
        return '-/-'

    @property
    def fmt_payment_terms(self):
        if self.payment_term:
            return self.payment_term.cycle
        return '-/-'

    @property
    def fmt_allow_credit(self):
        return dec_prn(self.credit_limit, 2)

    @property
    def fmt_credit_limit(self):
        return dec_prn(self.credit_limit, 2)

    def fmt_first_name(self):
        try:
            return self.user.first_name
        except:
            return ''

    @property
    def fmt_last_name(self):
        try:
            return self.user.last_name
        except:
            return ''

    #   egress_trunks = relationship("egress_trunk.EgressTrunkModel",

    #                       backref="client_id",secondaryjoin = "egress_trunk.EgressTrunkModel.egress==True")
    #   igress_trunks = relationship("entities.resource.egress_trunk.IngressTrunkModel",

    def last_invoice(self):
        return Invoice.filter(Invoice.client_id == self.client_id). \
            order_by(Invoice.invoice_time.desc()).first()

    def egress_trunks_count(self):
        return len(self.egress_trunks)

    def ingress_trunks_count(self):
        return len(self.ingress_trunks)

    def unpaid_invoices_count(self):
        return Invoice.filter(Invoice.client_id == self.client_id).filter(Invoice.paid == False).count()

    def unread_message_count(self):
        from .mail import EmailLog
        cls = EmailLog
        return cls.filter(and_(cls.type.in_(cls.MESSAGE_TYPES), cls.status != 'fail',
                               cls.is_view != 1, cls.client_id == self.client_id)).count()

    def unread_alert_count(self):
        cls = EmailLog
        return cls.filter(and_(cls.type.in_(cls.ALERT_TYPES), cls.status != 'fail',
                               cls.is_view != 1, cls.client_id == self.client_id)).count()

    @hybrid_property
    def daily_usage_on_non_zero(self):
        return self.auto_summary_not_zero == 1

    @daily_usage_on_non_zero.setter
    def daily_usage_on_non_zero(self, value):
        if value:
            self.auto_summary_not_zero = 1
        else:
            self.auto_summary_not_zero = 0

    @hybrid_property
    def currency_name(self):
        try:
            return self.currency.code
        except:
            return ''

    @currency_name.setter
    def currency_name(self, value):
        try:
            self.currency_id = Currency.filter(Currency.code == value).one().currency_id
        except:
            return None
        return value

    @hybrid_property
    def pt_name(self):
        try:
            return self.payment_term.name
        except:
            return ''

    @pt_name.setter
    def pt_name(self, value):
        try:
            self.payment_term_id = PaymentTerm.filter(PaymentTerm.name == value).one().payment_term_id
        except:
            raise Exception('Bad payment term:' + str(value) + ' not found')

    # pay.save()

    @hybrid_property
    def is_prepay(self):
        return self.mode == 'prepay'

    @is_prepay.setter
    def is_prepay(self, value):
        if value:
            self.mode = 'prepay'
            self.unlimited_credit = False
            self.allowed_credit = None
        else:
            self.mode = 'postpay'

    @hybrid_property
    def fake(self):
        return self.low_balance_config.is_notify

    @fake.setter
    def fake(self, value):
        self.low_balance_config.is_notify = value

    @hybrid_property
    def low_balance_alert(self):
        return self.low_balance_config.is_notify

    @low_balance_alert.setter
    def low_balance_alert(self, value):
        self.low_balance_config.is_notify = value

    # --user-prop
    def last_payment(self):
        return ClientPayment.filter(
            and_(ClientPayment.client_id == self.client_id, ClientPayment.payment_type.in_([3, 4, 5, 6]))). \
            order_by(ClientPayment.date.desc()).first()

    @hybrid_property
    def _mutual_balance(self):
        d = str(datetime.today().date())
        d = self.query().session.query(
            func.max(BalanceHistory.date).label('d')).filter(
            BalanceHistory.client_id == self.client_id).first().d
        try:
            return BalanceHistory.filter(BalanceHistory.client_id == self.client_id).filter(
                BalanceHistory.date == d).first().mutual_balance
        except:
            return 0.0

    @classmethod
    def user_id_filter(cls, user_id):
        return user_id_filter(cls, user_id)

    @hybrid_property
    def _client_invoice_settings_id(self):
        return self.client_invoice_settings_id

    @_client_invoice_settings_id.setter
    def _client_invoice_settings_id(self, value):
        ext = ClientInvoiceSetting.filter(ClientInvoiceSetting.client_id == self.client_id).first()
        if not ext:
            ext = ClientInvoiceSetting(client_id=self.client_id)
        ext.invoice_setting_id = value
        self.session().add(ext)

    def invoice_check_and_generate(self, now, generate_callback, AUTO_INVOICE_DEBUG):
        try:
            now_date = int(now.timestamp())
            time_zone_hour = 0
            hour = now.hour
            if not self.invoice_zone is None:
                time_zone_hour = get_time_zone_hour(self.invoice_zone)

            today = now + timedelta(hours=time_zone_hour)
            # skip other time zone
            if today.hour != 0:  # time_zone_hour != today.hour:
                msg = 'invoice, client {} timezone {} does not match current {}. continue'.format(
                    self.client_id, self.invoice_zone, today.hour)
                log.debug(msg)
                return 0, msg

            if not self.name or not self.payment_term or not self.payment_term.type:
                msg = 'invoice, client {} has not name or payment_term {} or payment_term.type'.format(
                    self.client_id, self.payment_term_id)
                log.debug(msg)
                if AUTO_INVOICE_DEBUG:
                    InvoiceDebug(client_id=self.client_id, text=msg).save()
                return 0, msg

            client_invoice_history = Invoice.filter(
                and_(Invoice.client_id == self.client_id, Invoice.state != 'void')).order_by(
                Invoice.invoice_end.desc()).first()

            if not client_invoice_history:
                msg = 'Invoice Client {} with Id:  {} is type {} is never invoiced before'.format(
                    self.name, self.client_id, self.payment_term.type)
                log.debug(msg)
                if AUTO_INVOICE_DEBUG:
                    InvoiceDebug(client_id=self.client_id, text=msg).save()
                _l = self.create_time
                last_invoiced = datetime(_l.year, _l.month, _l.day)
                InvoiceHistory(last_invoice_for=last_invoiced,
                               next_invoice_date=self.payment_term.next_date_for_now(last_invoiced, today.date()),
                               client_id=self.client_id).save()
            else:
                _l = client_invoice_history.invoice_end + timedelta(seconds=1)
                last_invoiced = datetime(_l.year, _l.month, _l.day)
            if (today.replace(tzinfo=UTC) - last_invoiced.replace(tzinfo=UTC)).days > 90:
                last_invoiced = (today - timedelta(days=90)).replace(day=1)
            invoice_start_time = last_invoiced
            _n = self.payment_term.next_date_for_now(last_invoiced, today.date()) - timedelta(days=1)
            invoice_end_time = datetime(_n.year, _n.month, _n.day, 23, 59, 59)
            msg = 'Client {} id {} invoice term {} type {} cycle {}, next invoice invoice_end_time {}, today is {}'. \
                format(self.name, self.client_id, self.payment_term_name, self.payment_term.type,
                       self.payment_term.cycle, invoice_end_time, today)
            log.debug(msg)
            if AUTO_INVOICE_DEBUG:
                InvoiceDebug(client_id=self.client_id, text=msg).save()

            if invoice_end_time.timestamp() <= today.timestamp():
                inv, warn = generate_callback(self.client_id, invoice_start_time, invoice_end_time)
                if hasattr(inv, 'invoice_number'):
                    msg = 'Client {} id {} SUCCESS invoice_number {} period {} {} {}'.format(
                        self.name, self.client_id, inv.invoice_number, str(invoice_start_time)[0:10],
                        str(invoice_end_time)[0:10], warn)
                    log.debug(msg)
                    if AUTO_INVOICE_DEBUG:
                        InvoiceDebug(client_id=self.client_id, text=msg).save()

                    InvoiceHistory(last_invoice_for=today.date(),
                                   next_invoice_date=self.payment_term.next_date(today.date()),
                                   client_id=self.client_id,
                                   last_invoice_period='{} {}'.format(str(invoice_start_time)[0:10],
                                                                      str(invoice_end_time)[0:10]),
                                   last_invoice_amount=inv.total_amount,
                                   invoice_filename=inv.filename,
                                   next_reminder_timestamp=int(
                                       datetime(*inv.payment_dur_date.timetuple()[:-4]).timestamp()),
                                   reminders_count=1,
                                   start_time=invoice_start_time,
                                   end_time=invoice_end_time
                                   ).save()
                    return 1, warn
                else:
                    msg = 'Client {} id {}  period {} {} {}'.format(
                        self.name, self.client_id, str(invoice_start_time)[0:10],
                        str(invoice_end_time)[0:10], warn)
                    return 1, msg

            msg = 'Client {} id {} SKIP period {} {} does not match payment term, or overlapped.last auto invoice was at {}'.format(
                self.name, self.client_id, str(invoice_start_time)[0:10], str(invoice_end_time)[0:10], last_invoiced
            )
            log.debug(msg)
            if AUTO_INVOICE_DEBUG:
                InvoiceDebug(client_id=self.client_id, text=msg).save()
            return 0, msg
        except Exception as e:
            msg = 'Invoice error:{} {}'.format(e, str(traceback.print_exc()))
            log.debug(msg)
            if AUTO_INVOICE_DEBUG:
                if self.client_id and self.client_name:
                    msg = 'Client {} id {} {}'.format(self.name, self.client_id, msg)
                InvoiceDebug(client_id=self.client_id, text=msg).save()
            return 0, msg
        return 0, 'no messages'


class CarrierLowBalanceConfig(BaseModel):
    __tablename__ = 'client_low_balance_config'
    SEND_TIME_TYPE = {0: 'daily', 1: 'hourly'}
    VALUE_TYPE = {0: 'Actual Balance', 1: 'Percentage'}

    client_id = Column(Integer, ForeignKey('client.client_id', ondelete='CASCADE'), primary_key=True)
    is_notify = Column(Boolean)
    value_type = Column(ChoiceType(VALUE_TYPE))
    send_time_type = Column(ChoiceType(SEND_TIME_TYPE))
    actual_notify_balance = Column(Numeric(30, 2))
    percentage_notify_balance = Column(Numeric(30, 2))
    daily_send_time = Column(Integer)
    duplicate_days = Column(Integer)
    send_to = Column(Integer)
    duplicate_send_days = Column(Integer, nullable=False, default=0)
    last_alert_time = Column(DateTime(timezone=True))
    disable_trunks_days = Column(Integer, default=5)

    carrier = relationship(Client, uselist=False, back_populates='low_balance_config', enable_typechecks=False)
    client = synonym('carrier')

    @hybrid_property
    def carrier_name(self):
        return self.carrier.name

    def create_from_template(self, obj):
        if obj:
            self.is_notify = obj.is_notify
            self.value_type = obj.value_type
            self.send_time_type = obj.send_time_type
            self.actual_notify_balance = obj.actual_notify_balance
            self.percentage_notify_balance = obj.percentage_notify_balance
            self.daily_send_time = obj.daily_send_time
            self.duplicate_days = obj.duplicate_days
            self.send_to = obj.send_to
            self.duplicate_send_days = obj.duplicate_send_days
            self.last_alert_time = obj.last_alert_time
            self.disable_trunks_days = obj.disable_trunks_days


class ResourceExt(BaseModel):
    __tablename__ = 'resource_ext'
    resource_id = Column(Integer, primary_key=True)
    vendor_tech_prefix = Column(PrefixRange)


class Resource(BaseModel):
    __tablename__ = 'resource'

    AGENT_TYPE = {1: 'exchange', 2: 'default agent', 3: 'client agent'}
    # AUTH_TYPE={0:'All',1:'Authorized by Host Only', 2: 'Accept Egress Registration', 3:'Register to Egress Trunk'}
    AUTH_TYPE = {1: 'Authorized by Host Only', 2: 'Authorized by SIP Registration', 3: 'Register to gateway'}
    BILLING_METHOD = {0: 'by minutes', 1: 'by port'}
    BILLING_TYPE = {0: 'DNIS', 1: 'LRN', 2: 'LRN Block All', 3: 'LRN Block Higher Rate Codes Only'}
    BILLING_PORT_TYPE = {0: 'Bill by allowed total port', 1: 'Bill By allowed per DID port',
                         2: 'Bill by max total port'}
    CLI_TYPE = {0: 'white', 1: 'white non cli', 2: 'gray'}
    DTMF_TYPE = {0: 'all', 1: 'info', 2: 'rfc2833', 3: 'inband'}
    JURISDICTION_USE_DNIS = {False: 'LRN', True: 'DNIS'}
    LNP_DIPPING = {1: 'Add Header', 0: 'Not Add Header'}
    MEDIA_TYPE = {2: 'Bypass Media', 0: 'Proxy Media + Transcoding', 1: 'Proxy Media'}
    PASS_THROUGH = {1: 'transparent', 2: 'transparent', 3: 'not pass through the ban'}
    PROFIT_TYPE = {1: 'percentage', 2: 'value'}
    PROTO = {1: 'sip', 2: 'h323', 0: 'all'}
    RATE_ROUNDING = {0: 'Up', 1: 'Down'}
    RES_STRATEGY = {1: 'top-down', 2: 'round-robin'}
    ROUTE_TYPE = {0: 'other', 1: 'intra', 2: 'inter', 3: 'highest'}
    RPID = {0: 'Never', 1: 'Pass Through', 3: 'Always'}
    PAID = {0: 'No', 1: 'Yes'}
    OLI = {0: 'No', 1: 'Yes'}
    PCI = {0: 'No', 1: 'Yes'}
    PRIV = {0: 'No', 1: 'Yes'}
    DIV = {0: 'No', 1: 'Yes'}
    RPID_ID_TYPE = {0: 'None', 1: 'Subscriber', 2: 'User', 3: 'Term', 4: 'Proxy'}
    RPID_PARTY = {0: 'None', 1: 'Calling', 2: 'Called', 3: 'Proxy'}
    RPID_PRIVACY = {0: 'None', 1: 'Full', 2: 'Name', 3: 'Url', 4: 'OFF', 5: 'Ipaddr', 6: 'Proxy'}
    RPID_SCREEN = {0: 'None', 1: 'No', 2: 'Yes', 3: 'Proxy'}
    SERVICE_TYPE = {0: 'Self Service', 1: 'Standand Deck'}
    STATUS = {1: 'untested', 2: 'tested', 3: 'failed', None: 'None'}
    TRUNK_TYPE = {1: 'class4', 2: 'exchange', 3: 'product_default', 4: 'product_agent'}
    TRUNK_TYPE2 = {0: 'Termination Traffic', 1: 'DID Traffic'}

    resource_id = Column(Integer, primary_key=True)
    name = Column(String(100), unique=True)  # , nullable=False)
    ingress = Column(Boolean)
    egress = Column(Boolean)
    active = Column(Boolean, nullable=False, default=True)

    account_id = Column(String(32))
    agent_type = Column(ChoiceType(AGENT_TYPE), default=2)  # 1-exchange; 2-default agent; 3-client agent
    alias = Column(String(100), nullable=False, unique=True, default='')
    amount_per_port = Column(Numeric)  # money per port
    ani_cap_limit = Column(Integer)
    ani_cps_limit = Column(Integer)
    auth_type = Column(ChoiceType(AUTH_TYPE),
                       default=1)  # 1 - Authorized by Host Only 2 - Accept Egress Registration 3 - Register to Egress Trunk###permission field: 0 - ALL; 1 - resource_auth;
    bill_by = Column(Integer,
                     default=4)  # 0 - DNIS, DNIS Billing, do not need to query LRN 1 - LRN, LRN billing, ignoring carrier changes 2 - LRN Block, LRN Billing Number Operator has changed, then declined 3 - LRN Block Higher Rate, LRN rates higher than DNIS rates 4 - Follow Rate Deck, using the above schedule for the rate table
    # rate_profile = Column(Integer, default=0)  # 0 --- false; 1 --- true

    billing_method = Column(ChoiceType(BILLING_METHOD), default=0)  # 0 - by minutes; 1- by port
    billing_port_type = Column(ChoiceType(BILLING_PORT_TYPE), default=0)
    billing_rule = Column(Integer)
    billing_type = Column(ChoiceType(BILLING_TYPE),
                          default=0)  # 0: DNIS; 1: LRN; 2: LRN Block All; 3: LRN Block Higher Rate Codes Only
    block_404_number_time = Column(Integer)
    block_time = Column(Integer)  # loop detection used in seconds to meet the block condition after the block how long
    canada_other = Column(ChoiceType(ROUTE_TYPE), default=0)  # 0 --- other; 1 --- intra; 2 --- inter; 3 --- highest
    canada_route = Column(ChoiceType(ROUTE_TYPE), default=0)  # 0 --- other; 1 --- intra; 2 --- inter; 3 --- highest
    capacity = Column(Integer)  # Allows the number of online calls
    cli_type = Column(ChoiceType(CLI_TYPE))  # 0-white; 1-white non cli; 2-gray
    client_id = Column(ForeignKey('client.client_id', ondelete='SET NULL'), index=True)  # wholesaler
    cost_per_port = Column(Numeric(30, 10))
    counter_time = Column(
        Integer)  # loop detection used, the unit seconds, this time period to receive the number of calls are called the same, then block
    cps_limit = Column(Integer)  # Allows the number of calls per second
    create_time = Column(DateTime(timezone=True), default=datetime.now(UTC))
    delay_bye_limit = Column(Integer)
    delay_bye_second = Column(Integer)
    disable_by_alert = Column(Boolean, default=False)
    display_name = Column(Integer)  # 0 --- false; 1 --- true; whether or not to pass through the caller's display name
    div = Column(ChoiceType(DIV))
    dnis_cap_limit = Column(Integer)
    dnis_cps_limit = Column(Integer)
    dnis_only = Column(Boolean, nullable=False, default=True)
    dtmf_detect = Column(Integer,
                         default=0)  # 0 - auto detect, media_type = transcode; 1 - scheduled detection, media_type = transcode and main and called dtmf are different and have the same speech code
    dtmf_type = Column(ChoiceType(DTMF_TYPE), default=3)  # 0 - all; 1 - info; 2 - rfc2833; 3 - inband
    dummy_trunk = Column(Boolean)  # whether the origination client failover trunk
    egress_bill_after_action = Column(Boolean, default=True)
    enfource_cid = Column(Boolean, default=False)
    enough_balance = Column(Boolean, default=True)  # false whether the balance is insufficient
    group_id = Column(ForeignKey('trunk_group.group_id', ondelete='SET NULL'))
    ignore_early_media = Column(Boolean, default=False)
    ignore_early_nosdp = Column(Integer)
    ignore_ring = Column(Boolean, default=False)
    inband = Column(Integer)  # 0 Does Not Support / Does Not Require; 1 Support / Require
    info = Column(Integer)  # 0 Does Not Support / Does Not Require; 1 Support / Require
    intl_route = Column(ChoiceType(ROUTE_TYPE), default=0)  # 0 --- other; 1 --- intra; 2 --- inter; 3 --- highest
    is_del = Column(Integer, default=0)  # 0: not del; 1: is del
    is_virtual = Column(Boolean)
    jurisdiction_use_dnis = Column(ChoiceType(JURISDICTION_USE_DNIS), default=False)
    last_priority = Column(Integer, default=0)
    lnp = Column(Boolean, nullable=False, default=False)
    lnp_dipping = Column(Boolean, default=False)
    lnp_dipping_rate = Column(Float)
    lrn_block = Column(Boolean, nullable=False, default=False)
    lrn_prefix = Column(Integer, nullable=False, default=1)  # 0 --- false; 1 --- true
    marketplace = Column(Integer, default=0)  # 0: International A-Z; 1: US Domestic
    max_duration = Column(Integer)
    media_timeout = Column(Integer)
    media_type = Column(ChoiceType(MEDIA_TYPE), nullable=False,
                        default=2)  # 0 - Proxy Media + Transcoding; 1 - Proxy Media; 2 - Bypass Media
    number = Column(
        Integer)  # loop detection Used to indicate the number of calls received for the same time in the same time, then block
    oli = Column(ChoiceType(OLI))
    paid = Column(ChoiceType(PAID))
    pass_response_code = Column(Integer,
                                default=0)  # 0 - false, 1 - true, if this option is true, the last sip error returned from the egress is returned to ingress
    pass_through = Column(ChoiceType(
        PASS_THROUGH))  # pass through the rules: 1 - transparent; 2 - not transparent; 3 - not pass through the ban
    pci = Column(ChoiceType(PCI))
    price_per_actual_channel = Column(Float, default=0)
    price_per_max_channel = Column(Float, default=0)
    priority = Column(Integer, default=0)
    priv = Column(ChoiceType(PRIV))
    private = Column(Integer, nullable=False, default=0)
    product_id = Column(Integer)
    profit_margin = Column(Float(53), nullable=False, default=0)  # minimum profit margin
    profit_type = Column(ChoiceType(PROFIT_TYPE), nullable=False, default=1,
                         server_default=text('1'))  # 1 - Percentage; 2 - value
    proto = Column(ChoiceType(PROTO), default=1)  # 1 - sip; 2 - h323; 0 - all
    random_table_id = Column(Integer)
    rate_decimal = Column(Integer, default=6)  # Determines how many bits are the fractional bits of the rate.
    rate_profile = Column(Integer, default=0)
    rate_rounding = Column(ChoiceType(RATE_ROUNDING),
                           default=0)  # rate decimal place method: 0 --- rounding; 1 --- discarded
    rate_table_id = Column(ForeignKey('rate_table.rate_table_id', ondelete='SET NULL'), index=True)
    rate_use_rpid = Column(Boolean, default=False)
    rating_type = Column(Integer, default=0)  # 0: US JD; 1: US Non-JD; 2: OCN-LATA JD; 3: OCN-LATA Non-JD
    re_invite = Column(Integer)  # 0 - false; 1 - true; send re-INVITE to detect whether the other party is active
    re_invite_interval = Column(Integer)  # seconds unit, it is recommended not less than 60 seconds
    redirect = Column(Integer, default=0)  # 0 --- false; 1 --- true
    res_strategy = Column(ChoiceType(RES_STRATEGY))  # Landing Gateway Selection Strategy: 1 - top-down, 2 - round-robin
    resource_block_group_id = Column(Integer)
    resource_template_id = Column(Integer)
    rfc2833 = Column(Integer)  # 0 Does Not Support / Does Not Require; 1 Support / Require
    rfc_2833 = Column(Boolean, nullable=False, default=False)
    rfc_r833_payload = Column(Integer)
    ring_timeout = Column(Integer, nullable=False, default=60)
    route_strategy_id = Column(ForeignKey('route_strategy.route_strategy_id', ondelete='SET NULL'), index=True)
    rpid = Column(ChoiceType(RPID), default=1)
    rpid_id_type = Column(ChoiceType(RPID_ID_TYPE),
                          default=0)  # 0 --- none call out without id_type: 1 --- subscriber; 2 --- user; 3 --- term; 4 --- proxy use caller IDID_type
    rpid_party = Column(ChoiceType(RPID_PARTY),
                        default=0)  # 0 --- none call out without party; 1 --- calling; 2 --- called; 3 --- proxy use caller RPID party
    rpid_privacy = Column(ChoiceType(RPID_PRIVACY),
                          default=0)  # 0 --- none call out without privacy: 1 --- full; 2 --- name; 3 --- url; 4 --- off; 5 --- ipaddr; 6 --- proxy use Calling RPID privacy
    rpid_screen = Column(ChoiceType(RPID_SCREEN),
                         default=0)  # 0 - none without screen; 1 - no; 2 - yes; 3 - proxy use caller RPID screen
    service_type = Column(ChoiceType(SERVICE_TYPE), nullable=False, default=0)  # 0: Self Service; 1: Standand Deck
    status = Column(ChoiceType(STATUS), default=1)  # 1 - untested; 2 - tested; 3 - failed
    switch_profile_id = Column(Integer)
    t38 = Column(Boolean, nullable=False, default=False)
    tdm = Column(Boolean, nullable=False, default=True)
    transaction_fee_id = Column(Integer)
    transnexus = Column(Integer)
    trunk_type = Column(ChoiceType(TRUNK_TYPE))  # 1-class4; 2-exchange; 3-product_default; 4-product_agent
    trunk_type2 = Column(ChoiceType(TRUNK_TYPE2), nullable=False,
                         default=0)  # {0:'Termination Traffic', 1:'DID Traffic'}
    update_at = Column(DateTime(timezone=True), onupdate=func.now())
    update_by = Column(String)
    update_time = Column(DateTime(timezone=True), onupdate=func.now())
    us_other = Column(ChoiceType(ROUTE_TYPE), default=0)  # 0 --- other; 1 --- intra; 2 --- inter; 3 --- highest
    us_route = Column(ChoiceType(ROUTE_TYPE), default=0)  # 0 --- other; 1 --- intra; 2 --- inter; 3 --- highest
    wait_ringtime180 = Column(Integer, default=6000)
    vendor_tech_prefix = Column(PrefixRange)

    rate_profile = Column(Integer)
    priv_value = Column(Integer, server_default=text("0"))
    use_sip404_ani = Column(Boolean, server_default=text("false"))

    __mapper_args__ = {'polymorphic_on': ingress}

    authorization_type = synonym('auth_type')
    bypass_media = synonym('media_type')
    call_limit = synonym('capacity')
    carrier_id = synonym('client_id')
    host_routing_strategy = synonym('res_strategy')
    is_active = synonym('active')
    min_duration = synonym('delay_bye_second')
    min_profit_type = synonym('profit_type')
    billing_by_port = synonym('billing_method')
    min_profit_value = synonym('profit_margin')
    pass_lrn = synonym('lrn_prefix')
    pdd = synonym('wait_ringtime180')
    route_plan_id = synonym('route_strategy_id')
    trunk_id = synonym('resource_id')
    ingress_trunk_name = synonym('alias')
    allowed_ports = synonym('dnis_cap_limit')
    trunk_name = synonym('alias')

    client_name = column_property(select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))
    client_type = column_property(
        select([Client.client_type]).where(Client.client_id == client_id).correlate_except(Client))
    client_mode = column_property(select([Client.mode]).where(Client.client_id == client_id).correlate_except(Client))
    client_status = column_property(
        select([Client.status]).where(Client.client_id == client_id).correlate_except(Client))
    client_allowed_credit = column_property(
        select([Client.allowed_credit]).where(Client.client_id == client_id).correlate_except(Client))
    client_call_limit = column_property(
        select([Client.call_limit]).where(Client.client_id == client_id).correlate_except(Client))
    client_cps_limit = column_property(
        select([Client.cps_limit]).where(Client.client_id == client_id).correlate_except(Client))

    rate_email = column_property(
        select([Client.rate_email]).where(Client.client_id == client_id).correlate_except(Client))
    email = column_property(select([Client.email]).where(Client.client_id == client_id).correlate_except(Client))
    carrier = column_property(select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))
    client_is_online = column_property(
        select([Client.is_online]).where(Client.client_id == client_id).correlate_except(Client))
    carrier_is_active = column_property(
        select([Client.status]).where(Client.client_id == client_id).correlate_except(Client))

    rate_table_name = column_property(
        select([RateTable.name]).where(RateTable.rate_table_id == rate_table_id).correlate_except(RateTable))
    # dynamic_count = column_property(select([func.count(DynamicRouteItem.id)]). \
    #                                 where(DynamicRouteItem.resource_id == resource_id). \
    #                                 correlate_except(DynamicRouteItem)
    #                                 )
    # static_count = column_property(select([func.count(ProductItemsResource.id)]). \
    #                                where(ProductItemsResource.resource_id == resource_id). \
    #                                correlate_except(ProductItemsResource)
    #                                )

    client = relationship(Client, uselist=False, back_populates='resources', enable_typechecks=False)
    group = relationship(TrunkGroup, back_populates='trunks', uselist=False)
    route_strategy = relationship(RouteStrategy)
    product_items = relationship(ProductItemsResource, uselist=True, back_populates='resource',
                                 single_parent=True, cascade="all, delete-orphan")
    ip = relationship('ResourceIp',
                      primaryjoin='and_(ResourceIp.resource_id == Resource.resource_id,ResourceIp.reg_type==0)',
                      uselist=True, back_populates='resource', single_parent=True, cascade="all, delete-orphan")

    reg_user = relationship('ResourceIp',
                            primaryjoin='and_(ResourceIp.resource_id == Resource.resource_id,ResourceIp.reg_type==1)',
                            uselist=True,  # back_populates='resource',
                            single_parent=True, cascade="all, delete-orphan")

    reg_gateway = relationship('ResourceIp',
                               primaryjoin='and_(ResourceIp.resource_id == Resource.resource_id,ResourceIp.reg_type==2)',
                               uselist=True,  # back_populates='resource',
                               single_parent=True, cascade="all, delete-orphan")

    rate_table = relationship('RateTable', back_populates='resources', uselist=False)

    codec_list = relationship('ResourceCodecsRef', uselist=True, back_populates='resource', single_parent=True,
                              cascade="all, delete-orphan", order_by='ResourceCodecsRef.id')
    # vendor_dids = relationship('DidBillingRel',
    #                            primaryjoin='and_(foreign(DidBillingRel.vendor_res_id) == Resource.resource_id)',
    #                            # ,DidBillingRel.end_date.is_(None))',
    #                            uselist=True, back_populates='vendor_res',
    #                            cascade="all, delete-orphan")
    # client_dids = relationship('DidBillingRel',
    #                            primaryjoin='and_(foreign(DidBillingRel.client_res_id) == Resource.resource_id)',
    #                            # DidBillingRel.end_date.is_(None))',
    #                            uselist=True, back_populates='client_res',
    #                            cascade="all, delete-orphan")
    #
    prefixes = relationship('ResourcePrefix', uselist=True, back_populates='resource',
                            single_parent=True, cascade="all, delete-orphan", lazy='dynamic'
                            )
    directions = relationship('ResourceDirection', uselist=True, back_populates='resource',
                              cascade="all, delete-orphan")
    replace_actions = relationship('ResourceReplaceAction', uselist=True, back_populates='resource',
                                   cascade="all, delete-orphan")

    digit_maps = relationship(ResourceTranslationRef, back_populates='gateway', uselist=True, single_parent=True,
                              cascade="all, delete-orphan", enable_typechecks=False)
    # ingress_switch_profiles = relationship('EgressProfile',
    #                                        primaryjoin='and_(foreign(EgressProfile.ingress_id)==Resource.resource_id,Resource.ingress==True)')
    #
    dynamic_routes = relationship('DynamicRouteItem',
                                  primaryjoin='Resource.resource_id==foreign(DynamicRouteItem.resource_id)',
                                  uselist=True, back_populates='trunk', single_parent=True, cascade="all, delete-orphan"
                                  )

    # host = synonym('ip')

    def before_save(self):
        return
        for p in self.prefixes:
            p.before_save()
        if self.rate_decimal is None:
            conf = SystemParameter.get(1)
            if conf:
                self.rate_decimal = conf.default_billing_decimal
        for i in self.ip + self.reg_gateway + self.reg_user:
            if i.direction is None:
                if self.ingress:
                    i.direction = 0
                else:
                    i.direction = 1

    def apply_mail(self, tpl, to=None, att=[]):
        from .mail import MailSender
        return MailSender.apply_mail(self, tpl, to, att)

    @hybrid_property
    def vendor_tech_prefix(self):
        ext = ResourceExt.filter(ResourceExt.resource_id == self.resource_id).first()
        if ext:
            return ext.vendor_tech_prefix
        return None

    @vendor_tech_prefix.setter
    def vendor_tech_prefix(self, value):
        ext = ResourceExt.filter(ResourceExt.resource_id == self.resource_id).first()
        if not ext:
            ext = ResourceExt(resource_id=self.resource_id)
        ext.vendor_tech_prefix = value
        self.session().add(ext)

    @hybrid_property
    def enable_global_404_blocking(self):
        try:
            return bool(self.block_404_number_time)
        except:
            return False

    @enable_global_404_blocking.setter
    def enable_global_404_blocking(self, value):
        self.block_404_number_time = int(value)

    @hybrid_property
    def ignore_early_no_sdp(self):
        try:
            return bool(self.ignore_early_nosdp)
        except:
            return False

    @ignore_early_no_sdp.setter
    def ignore_early_no_sdp(self, value):
        self.ignore_early_nosdp = int(value)

    @hybrid_property
    def prefix(self):
        try:
            return ','.join([p.tech_prefix for p in self.prefixes])
        except:
            return ''

    @prefix.setter
    def prefix(self, value):
        try:
            if value != None:
                if len(self.prefixes) > 0:
                    self.prefixes[0].tech_prefix = value
                else:
                    pr = ResourcePrefix(tech_prefix=value,
                                        route_strategy_id=self.route_strategy_id, rate_table_id=self.rate_table_id)
                    self.prefixes.append(pr)
        except:
            pass

    @hybrid_property
    def static_route_items(self):
        ret = []

        for it in self.product_items:
            ret.append(dict(prefix=it.product.digits, static_route_id=it.product.product_id))
        return ret

    @static_route_items.setter
    def static_route_items(self, value):
        for it in self.product_items:
            it.delete()
        for it in value:
            prefix = it['prefix']
            id = it['static_route_id']
            p = ProductItems.filter(and_(ProductItems.product_id == id, ProductItems.digits == prefix)).first()
            if not p:
                p = ProductItems(product_id=id, digits=prefix)
            tr = ProductItemsResource(resource_id=self.resource_id)
            p.trunks.append(tr)
            self.product_items.append(tr)

    # self.session().add(tr)
    # self.session().add(p)

    # def rate_table_name(self):
    #    return self.rate.name
    @property
    def rate_email(self):
        try:
            return self.client.rate_email
        except:
            return None

    def route_plan_name(self):
        return self.route_strategy.name

    def group_name(self):
        return self.group.name

    def product_name(self):
        try:
            return self.product_rout.product_name
        except:
            return None

    # def carrier(self):
    #    return self.client.name

    def rate_email(self):
        return self.client.rate_email

    def usage_count(self):
        cnt1 = ProductItemsResource.filter(ProductItemsResource.resource_id == self.resource_id).count()
        cnt2 = DynamicRouteItem.filter(DynamicRouteItem.resource_id == self.resource_id).count()
        return cnt1 + cnt2

    def trunk_count(self):
        return Resource.filter(Resource.client_id == self.client_id).count()

    @hybrid_property
    def direction(self):
        if self.ingress:
            return 'ingress'
        else:
            return 'egress'

    @hybrid_property
    def type(self):
        if self.ingress and self.active:
            return 'orig'
        if self.egress and self.active:
            return 'term'
        return None

    @hybrid_property
    def codes(self):
        try:
            return [t.translation.translation_name for t in self.digit_maps]
        except:
            return []

    @codes.setter
    def codes(self, value):
        for t in self.digit_maps:
            self.digit_maps.remove(t)
        for name in value:
            q = DigitTranslation.filter(DigitTranslation.translation_name == name).first()
            if q:
                ref = ResourceTranslationRef(translation_id=q.translation_id)
                self.digit_maps.append(ref)

    @hybrid_property
    def codecs(self):
        try:
            return [c.codec.name for c in self.codec_list]
        except Exception as e:
            return []

    @codecs.setter
    def codecs(self, value):
        # for t in self.codec_list:
        # 	self.codec_list.remove(t)
        self.codec_list[:] = []
        for name in value:
            q = Codec.filter(Codec.name == name).first()
            if q:
                ref = ResourceCodecsRef(codec_id=q.id)
                self.codec_list.append(ref)

    @hybrid_property
    def actions(self):
        try:
            DIRECTION = {0: 'ALL', 1: 'ingress', 2: 'egress'}
            ACT = {'plus prefix': 'Add Prefix', 'minus prefix': 'Delete Prefix', 'plus suffix': 'Add Suffix',
                   'minus suffix': 'Delete Suffix'}
            AF = {'plus prefix': 'new_digits', 'minus prefix': 'num_digits', 'plus suffix': 'new_digits',
                  'minus suffix': 'num_digits'}
            DIR = {'modify the caller': 'ani', 'modify the called': 'dnis'}
            TYPE = {0: 'modify the caller', 1: 'modify the called'}
            ret = []
            ac = None
            if self.directions:
                ac = {'ani': {}, 'dnis': {}}
                for a in self.directions:
                    if not a.action in ['plus prefix', 'minus prefix', 'plus suffix', 'minus suffix']:
                        continue
                    if not a.direction == self.direction:
                        continue
                    if ac[DIR[a.type]] != {}:
                        ret.append(ac)
                        ac = {'ani': {}, 'dnis': {}}

                    ac[DIR[a.type]]['action'] = ACT[a.action]
                    ac[DIR[a.type]][AF[a.action]] = a.digits
                    ac[DIR[a.type]]['match_prefix'] = a.dnis

            if self.replace_actions:
                if not ac:
                    ac = {'ani': {}, 'dnis': {}}
                for a in self.replace_actions:
                    TYPE = {0: 'ani only replace the caller', 1: 'dnis only replace the called',
                            2: 'both replace the calling and called'}
                    if a.type in ['ani only replace the caller', 'both replace the calling and called']:
                        if ac['ani'] != {}:
                            ret.append(ac)
                            ac = {'ani': {}, 'dnis': {}}
                        ac['ani']['action'] = 'Replace'
                        ac['ani']['new_digits'] = a.ani
                        ac['ani']['match_prefix'] = a.ani_prefix
                        ac['ani']['min_length'] = a.ani_min_length
                        ac['ani']['max_length'] = a.ani_max_length

                    if a.type in ['dins only replace the called', 'both replace the calling and called']:
                        if ac['dnis'] != {}:
                            ret.append(ac)
                            ac = {'ani': {}, 'dnis': {}}
                        ac['dnis']['action'] = 'Replace'
                        ac['dnis']['new_digits'] = a.dnis
                        ac['dnis']['match_prefix'] = a.dnis_prefix
                        ac['dnis']['min_length'] = a.dnis_min_length
                        ac['dnis']['max_length'] = a.dnis_max_length
            if ac:
                ret.append(ac)
            return ret
        except Exception as e:
            log.warning('Cannot get actions for did repository {}'.format(str(e)))

    @actions.setter
    def actions(self, value):
        try:
            ACT = {'plus prefix': 'Add Prefix', 'minus prefix': 'Delete Prefix', 'plus suffix': 'Add Suffix',
                   'minus suffix': 'Delete Suffix'}
            ACTT = rev(ACT)
            for d in self.directions:
                d.delete()
            for d in self.replace_actions:
                d.delete()
            # if value:
            #    a=value
            for a in value:
                ani = None
                dnis = None
                for t in ['ani', 'dnis']:
                    if not t in a:
                        continue
                    ac = a[t]
                    digits = None
                    num_digits = None
                    if not 'action' in ac:
                        continue
                    actp = ac['action']
                    if actp in ('Add Prefix', 'Add Suffix') and 'new_digits' in ac:
                        digits = ac.get('new_digits', None)
                    if actp in ('Delete Prefix', 'Delete Suffix') and 'num_digits' in ac:
                        digits = ac.get('num_digits', None)

                    if actp == 'Replace' and t == 'ani' and 'new_digits' in ac:
                        ani = ac.get('new_digits', None)

                    if actp == 'Replace' and t == 'dnis' and 'new_digits' in ac:
                        dnis = ac.get('new_digits', None)

                    rd = 'modify the caller' if t == 'ani' else 'modify the called'
                    if digits or num_digits:
                        self.directions.append(
                            ResourceDirection(action=ACTT[actp], type=rd, digits=digits,
                                              direction=self.direction, dnis=ac.get('match_prefix')))

                if ani or dnis:
                    tp, ani_prefix, ani_min_length, ani_max_length, dnis_prefix, dnis_min_length, dnis_max_length = None, None, None, None, None, None, None
                    if ani:
                        tp = 'ani only replace the caller'
                        aac = a['ani']
                        ani_prefix = aac.get('match_prefix', None)
                        ani_min_length = aac.get('min_length', None)
                        ani_max_length = aac.get('max_length', None)
                    if dnis:
                        tp = 'dnis only replace the caller'
                        aac = a['dnis']
                        dnis_prefix = aac.get('match_prefix', None)
                        dnis_min_length = aac.get('min_length', None)
                        dnis_max_length = aac.get('max_length', None)
                    if ani and dnis:
                        tp = 'both replace the calling and called'
                    if tp:
                        rep = ResourceReplaceAction(type=tp, ani=ani, dnis=dnis,
                                                    ani_prefix=ani_prefix, ani_min_length=ani_min_length,
                                                    ani_max_length=ani_max_length,
                                                    dnis_prefix=dnis_prefix, dnis_min_length=dnis_min_length,
                                                    dnis_max_length=dnis_max_length)
                        self.replace_actions.append(rep)
        except Exception as e:
            log.warning('Cannot get actions for did repository {}'.format(str(e)))

    def assign_vendor_did_prefix(self, did, rate_table):
        log.debug('assign_vendor_did_prefix(self={}, did={}, rate_table={})'.format(self, did, rate_table))
        if self.client and self.client.resource:
            tech_prefix = self.client.resource.vendor_tech_prefix or ''
            ResourcePrefix.filter(
                and_(ResourcePrefix.code == did, ResourcePrefix.resource_id == self.resource_id)).delete()
            did_prefix = ResourcePrefix(code=did, route_strategy_id=RouteStrategy.get_orig_routing_plan_id(),
                                        # tech_prefix=tech_prefix+did
                                        tech_prefix=tech_prefix
                                        )
            did_prefix.rate_table = rate_table
            self.prefixes.append(did_prefix)
            self.enough_balance = True
            return did_prefix
        else:
            raise Exception('Bad vendor configuration!')

    # self.save()

    @property
    def fmt_switch_ip(self):
        ret = ','.join(
            ['{}:{}'.format(s.profile.sip_ip, s.profile.sip_port) for s in self.ingress_switch_profiles if s.profile])
        if ret == '':
            ret = ','.join(
                ['{}:{}'.format(s.sip_ip, s.sip_port) for s in
                 SwitchProfile.filter(SwitchProfile.status == 'connected')])
        return ret

    @property
    def fmt_allowed_ip(self):
        return self.fmt_ip_listing

    @property
    def fmt_ip_listing(self):
        return ','.join(['{}:{}'.format(i.ip, i.port) for i in self.ip if i.ip != None])

    @property
    def fmt_IP_listing(self):
        return self.fmt_ip_listing

    @property
    def fmt_route_info(self):
        return ','.join(['prefix {} product {} rate_table {} '.format(i.tech_prefix,
                                                                      i.product_name if i.product_name else '-//-',
                                                                      i.rate_table_name) for i in self.prefixes])

    @property
    def fmt_trunk_type(self):
        return self.direction

    @property
    def fmt_detail_table(self):
        return """
        <table class="detail_table">
        <tr><td>Alias</td><td>{}</td></tr>
        <tr><td>Direction</td><td>{}</td></tr>
        <tr><td>IP</td><td>{}</td></tr>
        <tr><td>Prefix</td><td>{}</td></tr>
        <tr><td>Capacity/Cps limit</td><td>{}/{}</td></tr>
        <tr><td>Is active </td><td>{}</td></tr>
        </table>
        """.format(self.alias, self.direction, self.fmt_ip_listing, self.fmt_route_info, self.capacity, self.cps_limit,
                   self.is_active)

    @property
    def fmt_carrier_name(self):
        return self.client_name


class RegisterOfRecord(BaseModel):
    __tablename__ = 'register_of_record'
    STATE = {0: 'no', 1: 'yes'}
    id = Column(Integer, primary_key=True)  # , server_default=text("nextval('register_of_record_id_seq'::regclass)"))
    time = Column(BigInteger)
    state = Column(ChoiceType(STATE), default=1)
    username = Column(String(100), index=True)
    direction = Column(Integer)
    ip = Column(Ip4)
    port = Column(Integer)
    expires_time = Column(BigInteger)


class ResourceIp(BaseModel):
    __tablename__ = 'resource_ip'
    # REG_STATUS = {0: 'not registered', 1: 'success', 2: 'error'}
    ADDR_TYPE = {0: 'ip', 1: 'host'}
    REG_STATUS = {0: 'not registered', 1: 'registered', 2: 'register failed'}
    # __table_args__ = (UniqueConstraint('resource_id', 'ip','port'),)
    resource_ip_id = Column(Integer, primary_key=True)
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    ip = Column(String(50))
    port = Column(Integer, default=5060)
    fqdn = Column(String(100))  # domain name
    sip_rpid = Column(String(64))  # opensips
    disable_by_alert = Column(Boolean, default=False)
    priority = Column(Integer, default=0)
    last_priority = Column(Integer, default=0)
    addr_type = Column(ChoiceType(ADDR_TYPE), default=0)
    direction = Column(Integer)  # Mark this ip is incoming or outgoing: 0 - incoming; 1 - outgoing
    username = Column(String(50))
    password = Column(String(50))
    reg_type = Column(SmallInteger, nullable=False,
                      default=0)  # register user type, 0 for none; 1 for register user; 2 for register gateway account
    reg_status = Column(ChoiceType(REG_STATUS), nullable=False,
                        default=0)  # 0 for un-register; 1 for registered; 2 for register failed
    reg_srv_ip = Column(String(100))
    reg_srv_port = Column(Integer, nullable=False, default=5060)
    expires = Column(Integer, nullable=False, default=3600)
    profile_id = Column(Integer, nullable=False, default=0)
    need_register = Column(Boolean, default=False)  # 0:Authorized by IP Only;1:Authorized by SIP Registration
    options_ping_inv = Column(Integer, nullable=True)

    resource = relationship(Resource,
                            primaryjoin='and_(ResourceIp.resource_id == Resource.resource_id,ResourceIp.reg_type==0)',
                            uselist=False, back_populates='ip', enable_typechecks=False)
    # limits = relationship('ResourceIpLimit', back_populates='ip', uselist=True, single_parent=True,
    #                       cascade="all, delete-orphan")
    id = synonym('resource_ip_id')

    # registered = column_property(case([(select([RegisterOfRecord.state]). \
    #                                     where(RegisterOfRecord.username == username).correlate_except(
    #     RegisterOfRecord).as_scalar() == 1, 'yes')], else_='no'))
    registered = column_property(case([(reg_status == 1, 'yes')], else_='no'))

    def before_save(self):
        if self.resource_id:
            res = Resource.get(self.resource_id)
            if res and res.client.client_type != 'vendor':
                prefs = list(set([p.tech_prefix for p in
                                  ResourcePrefix.filter(ResourcePrefix.resource_id == self.resource_id).all()]))
                if res.ingress:
                    sq = ResourceIp.filter(and_(ResourceIp.ip == self.ip, ResourceIp.port == self.port)). \
                        filter(ResourceIp.resource_ip_id != self.resource_ip_id). \
                        outerjoin(Resource, Resource.resource_id == ResourceIp.resource_id). \
                        filter(Resource.ingress == True)
                    if prefs:
                        sq = sq.outerjoin(ResourcePrefix, Resource.resource_id == ResourcePrefix.resource_id). \
                            filter(ResourcePrefix.tech_prefix.in_(prefs))
                    else:
                        sq = sq.outerjoin(ResourcePrefix, Resource.resource_id == ResourcePrefix.resource_id). \
                            filter(ResourcePrefix.tech_prefix.is_(None))
                    q = sq.first()
                    if q:
                        raise ValidationError(
                            'Ip {} port {} is duplicate on ingress trunks with same prefix! (duplicate resource_ip_id {})'.format(
                                self.ip, self.port, q.resource_ip_id))

    def after_save(self):
        pass


Resource.ip_count = column_property(select([func.count(ResourceIp.resource_ip_id)]). \
                                    where(
    and_(Resource.resource_id == ResourceIp.resource_id, ResourceIp.reg_type == 0)). \
                                    correlate_except(ResourceIp)
                                    )


class Codec(BaseModel):
    __tablename__ = 'codecs'

    id = Column(Integer, primary_key=True)
    name = Column(String(40), nullable=False, unique=True)
    detail = Column(String(100))

    trunks = relationship('ResourceCodecsRef', uselist=True, back_populates='codec')

    def __repr__(self):
        return str(self.id) + ' ' + self.name + ' ' + self.detail


class ResourceCodecsRef(BaseModel):
    __tablename__ = 'resource_codecs_ref'

    id = Column(Integer, primary_key=True, server_default=text("nextval('resource_codecs_ref_id_seq'::regclass)"))
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    codec_id = Column(Integer, ForeignKey('codecs.id'), index=True)

    resource = relationship(Resource, uselist=False, back_populates='codec_list')
    codec = relationship(Codec, uselist=False, back_populates='trunks')


# resource.ResourceModel.codec = relationship('resource_codecs_ref.ResourceCodecsRefModel',backref='ResourceModel.resource_id')

class ResourceDirection(BaseModel):
    __tablename__ = 'resource_direction'
    DIRECTION = {0: 'ALL', 1: 'ingress', 2: 'egress'}
    ACTION = {1: 'plus prefix', 2: 'plus suffix', 3: 'minus prefix', 4: 'minus suffix'}
    TYPE = {0: 'modify the caller', 1: 'modify the called'}
    NUMBER_TYPE = {0: 'all', 1: 'greater than', 2: 'equal to', 3: 'less than'}

    direction_id = Column(Integer, primary_key=True,
                          server_default=text("nextval('resource_direction_direction_id_seq'::regclass)"))
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), nullable=False, index=True)
    time_profile_id = Column(ForeignKey('time_profile.time_profile_id', ondelete='SET NULL'), index=True)
    action = Column(ChoiceType(ACTION), nullable=False)
    direction = Column(ChoiceType(DIRECTION), nullable=False)
    digits = Column(String(10), nullable=False)
    dnis = Column(PrefixRange, server_default=text("''::prefix_range"))
    number_length = Column(Integer)
    number_type = Column(ChoiceType(NUMBER_TYPE))
    type = Column(ChoiceType(TYPE), nullable=False)

    resource = relationship('Resource', uselist=False, back_populates='directions')
    time_profile = relationship('TimeProfile')


class ResourcePrefix(BaseModel):
    __tablename__ = 'resource_prefix'
    # __table_args__ = (UniqueConstraint('resource_id', 'code'),)
    __table_args__ = (UniqueConstraint('resource_id', 'tech_prefix', 'code'),)
    id = Column(Integer, primary_key=True)
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), index=True)
    tech_prefix = Column(PrefixRange, nullable=False, default='', index=True)  # Префикс номера соответствия
    route_strategy_id = Column(Integer, index=True)
    rate_table_id = Column(Integer, index=True)
    code = Column(PrefixRange)  # Called number prefix, with tech_prefix not the same, do not need to delete the prefix,
    # select routing_palan and rate_table role. The same ingress + tech_prefix code must be unique.
    code_cps = Column(Integer)
    code_cap = Column(Integer)
    product_id = Column(Integer, index=True)
    resource = relationship(Resource, uselist=False, back_populates='prefixes', enable_typechecks=False)
    rate_table = relationship('RateTable',
                              primaryjoin='foreign(ResourcePrefix.rate_table_id) == RateTable.rate_table_id')
    route_strategy = relationship('RouteStrategy',
                                  primaryjoin='foreign(ResourcePrefix.route_strategy_id) == RouteStrategy.route_strategy_id')
    routing_plan_id = synonym('route_strategy_id')
    trunk_id = synonym('resource_id')
    prefix = synonym('tech_prefix')
    # prefix = column_property(tech_prefix)

    # ips = relationship(ResourceIp,uselist=True,primaryjoin=foreign(resource_id)==ResourceIp.resource_id)
    product = relationship('ProductRoutRateTable',
                           primaryjoin='foreign(ResourcePrefix.product_id) == ProductRoutRateTable.id',
                           uselist=False, back_populates='trunks')

    product_name = column_property(select([ProductRoutRateTable.name]). \
                                   where(ProductRoutRateTable.id == product_id).correlate_except(ProductRoutRateTable))
    rate_table_name = column_property(select([RateTable.name]). \
                                      where(RateTable.rate_table_id == rate_table_id).correlate_except(RateTable))
    code_deck_name = column_property(select([RateTable.code_deck_name]). \
                                     where(RateTable.rate_table_id == rate_table_id).correlate_except(RateTable))
    trunk_name = column_property(select([Resource.alias]). \
                                 where(Resource.resource_id == resource_id).correlate_except(Resource))
    carrier_name = column_property(select([Client.name]). \
                                   where(
        and_(Resource.resource_id == resource_id, Client.client_id == Resource.client_id)).correlate_except(Resource,
                                                                                                            Client))

    rate_email = column_property(select([Client.rate_email]). \
                                 where(
        and_(Resource.resource_id == resource_id, Client.client_id == Resource.client_id)).correlate_except(Resource,
                                                                                                            Client))
    alias = column_property(select([Resource.alias]). \
                            where(Resource.resource_id == resource_id).correlate_except(Resource))

    carrier_id = column_property(select([Resource.client_id]). \
                                 where(Resource.resource_id == resource_id).correlate_except(Resource))
    client_id = column_property(select([Resource.client_id]). \
                                where(Resource.resource_id == resource_id).correlate_except(Resource))

    ingress = column_property(select([Resource.ingress]). \
                              where(Resource.resource_id == resource_id).correlate_except(Resource))

    @hybrid_property
    def ips(self):
        return self.resource.ip

    def before_save(self):
        if self.product_id and self.rate_table_id is None and self.route_strategy_id is None:
            prod = ProductRoutRateTable.get(self.product_id)
            if prod:
                self.rate_table_id = prod.rate_table_id
                self.route_strategy_id = prod.route_plan_id


class ResourceReplaceAction(BaseModel):
    __tablename__ = 'resource_replace_action'
    TYPE = {0: 'ani only replace the caller', 1: 'dnis only replace the called',
            2: 'both replace the calling and called'}
    id = Column(Integer, primary_key=True, server_default=text("nextval('resource_replace_action_id_seq'::regclass)"))
    resource_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'), index=True)
    ani_prefix = Column(String(50))
    ani = Column(String(30))
    ani_min_length = Column(Integer)
    ani_max_length = Column(Integer)
    type = Column(ChoiceType(TYPE))
    dnis_prefix = Column(String(30))
    dnis = Column(String(30))
    dnis_min_length = Column(Integer)
    dnis_max_length = Column(Integer)

    resource = relationship('Resource', uselist=False, back_populates='replace_actions')


class QosResource(BaseModel):
    __tablename__ = 'qos_resource'
    DIRECTION = {0: 'ingress', 1: 'egress'}
    report_time = Column(DateTime(True), nullable=False, index=True,
                         server_default=text("('now'::text)::timestamp(0) with time zone"), primary_key=True)
    res_id = Column(Integer, primary_key=True)
    call = Column(Integer)
    cps = Column(Integer)
    direction = Column(ChoiceType(DIRECTION))
    server_ip = Column(String(50), index=True)
    server_port = Column(Integer, index=True)
    channels = Column(Integer)
    qos_resource_report_time_res_id_idx = Index('qos_resource_report_time_res_id_idx', 'report_time', 'res_id')
    res_name = column_property(
        select([Resource.alias]).where(Resource.resource_id == res_id).correlate_except(Resource))


class VoipGateway(BaseModel):
    __tablename__ = 'voip_gateway'
    __table_args__ = (UniqueConstraint('lan_ip', 'lan_port'),)
    id = Column(Integer, primary_key=True)
    name = Column(String(200), unique=True)
    paid_replace_ip = Column(Integer, default=0)
    lan_ip = Column(String(255))
    lan_port = Column(Integer)
    active_call_ip = Column(String(64))
    active_call_port = Column(Integer)
    sip_capture_ip = Column(String(255))
    sip_capture_port = Column(Integer)
    sip_capture_path = Column(String(255))

    telnet_port = synonym('lan_port')
    telnet_ip = synonym('lan_ip')
    switch_id = synonym('id')

    sip_host = relationship('SwitchProfile', uselist=True, back_populates='voip_gateway',
                            single_parent=True,
                            cascade="all, delete-orphan"
                            )

    # info = relationship('VoipGatewayInfo', uselist=False, single_parent=True, cascade="all, delete-orphan")
    version_info = relationship('VersionInformation',
                                primaryjoin='VersionInformation.switch_name==foreign(VoipGateway.name)')

    @classmethod
    def expired7(cls):
        ol = cls.filter(cls.expire < int((datetime.now() - timedelta(days=7)).timestamp())).all()
        return [dict(name=i.name, lan_ip=i.lan_ip, expire=i.expire) for i in ol]


class SwitchProfile(BaseModel):
    __tablename__ = 'switch_profile'
    __table_args__ = (Index('switch_profile_sip_ip_sip_port_idx',
                            'sip_ip', 'sip_port', unique=True),)
    PROFILE_STATUS = {0: 'disabled', 1: 'connected'}

    id = Column(Integer, primary_key=True)
    voip_gateway_id = Column(ForeignKey('voip_gateway.id', ondelete='CASCADE'), index=True)
    switch_name = Column(String(100))
    profile_name = Column(String(100))
    sip_ip = Column(String(64))
    sip_port = Column(Integer)
    sip_debug = Column(Integer, default=0)
    sip_trace = Column(Boolean, nullable=False, default=False)
    proxy_ip = Column(String(64))
    proxy_port = Column(Integer)
    support_rpid = Column(Integer, nullable=False, default=0)
    support_oli = Column(Integer, nullable=False, default=0)
    support_priv = Column(Integer, nullable=False, default=0)
    support_div = Column(Integer, nullable=False, default=0)
    support_paid = Column(Integer, nullable=False, default=0)
    support_pci = Column(Integer, nullable=False, default=0)
    support_x_lrn = Column(Integer, nullable=False, default=1)
    support_x_header = Column(Integer, nullable=False, default=1)
    sip_capture_ip = Column(String(255))
    sip_capture_port = Column(Integer)
    sip_capture_path = Column(String(255))
    lan_ip = Column(String(255))
    lan_port = Column(Integer)
    profile_status = Column(ChoiceType(PROFILE_STATUS))
    paid_replace_ip = Column(Integer, default=0)
    auth_register = Column(Integer)
    default_register = Column(Integer)
    report_ip = Column(String(100))
    report_port = Column(Integer, default=3300)
    active_call_ip = Column(String(64))
    active_call_port = Column(Integer)
    cps = Column(Integer)
    cap = Column(Integer)
    pcap_token = Column(String(255))
    self_cps = Column(Integer)
    self_cap = Column(Integer)
    last_started_on = Column(DateTime())

    status = synonym('profile_status')
    cli_ip = synonym('sip_ip')
    cli_port = synonym('sip_port')
    switch_id = synonym('voip_gateway_id')
    host_name = synonym('profile_name')

    server_name = column_property(select([VoipGateway.name]). \
                                  where(voip_gateway_id == VoipGateway.switch_id).correlate_except(VoipGateway))

    voip_gateway = relationship(VoipGateway, uselist=False, back_populates='sip_host')

    # egress = relationship('EgressProfile', uselist=False, back_populates='profile')

    def expiration_date(self):
        try:
            return self.pcap_token
        except:
            return ''

    @property
    def limits(self):
        try:
            return self.voip_gateway.limits
        except:
            return {}


class InvoicePayment(BaseModel):
    __tablename__ = 'invoice_payment'

    id = Column(Integer, primary_key=True, server_default=text("nextval('invoice_payment_id_seq'::regclass)"))
    invoice_no = Column(String(50))
    payment_time = Column(DateTime(True))
    payment_amount = Column(Numeric)


class ClientPayment(BaseModel):
    __tablename__ = 'client_payment'
    PAYMENT_TYPE_DICT = {1: 'n/a', 2: 'undefined',
                         3: 'invoice payment sent',
                         4: 'invoice payment received',
                         5: 'prepay payment received',
                         6: 'payment sent',
                         7: 'credit note received',
                         8: 'credit note sent',
                         9: 'reset',
                         10: 'offset',
                         11: 'debit received',
                         12: 'debit sent',
                         13: 'mutual reset',
                         14: 'actual reset',
                         15: 'prepayment'}
    USER_PAYMENT_TYPES = ('invoice payment sent',
                          'invoice payment received',
                          'prepay payment received',
                          'payment sent',
                          'credit note received',
                          'credit note sent',
                          'debit received',
                          'debit sent'
                          )
    METHOD = {None: 'undefined', 0: 'paypal', 1: 'stripe'}

    @hybrid_property
    def total(self):
        if self.payment_type in (
                'invoice payment sent', 'payment sent', 'credit note sent', 'debit received', 'scc cost'):
            return -self.amount
        return self.amount

    client_payment_id = Column(Integer, primary_key=True)
    payment_time = Column(DateTime(True), nullable=False, default=datetime.now(UTC))
    amount = Column(Numeric(20, 10), nullable=False)
    result = Column(Boolean, nullable=False, default=True)
    client_id = Column(ForeignKey('client.client_id', ondelete='SET NULL'), index=True)
    cause = Column(String(20), server_default=text('NULL'))
    description = Column(String(500))  # Set the note for reset balance
    approved = Column(Boolean)
    current_balance = Column(Numeric(30, 10), default=0)
    invoice_number = Column(String(100), index=True)
    payment_type = Column(ChoiceType(PAYMENT_TYPE_DICT), default=1)

    payment_method = Column(ChoiceType(METHOD), server_default=text('NULL'))
    email_sended = Column(Boolean, server_default=text('false'))
    update_by = Column(String)
    receiving_time = Column(DateTime(True))
    egress_amount = Column(Numeric, server_default=text('NULL'))

    payment_type_name = synonym('payment_type')
    transaction_type = synonym('payment_type')
    date = synonym('payment_time')
    paid_on = synonym('receiving_time')
    note = synonym('description')
    balance = synonym('current_balance')

    # inv_list = relationship('Invoice', primaryjoin='foreign(Invoice.invoice_number)==ClientPayment.invoice_number', uselist=True, back_populates='payments')

    invoices = relationship('PaymentInvoice', uselist=True, back_populates='payment')
    client = relationship(Client, uselist=False, back_populates='payment', enable_typechecks=False)

    invoice_unpaid_amount = column_property(egress_amount - amount)

    carrier_name = column_property(
        select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))

    client_name = column_property(
        select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))

    __is_new = None

    @hybrid_property
    def company_name(self):
        return self.client.company_name

    @hybrid_property
    def payment_terms(self):
        return self.client.payment_terms()

    def before_save(self):
        self.__is_new = not bool(self.client_payment_id)
        if self.invoice_number:
            i = Invoice.filter(Invoice.invoice_number == self.invoice_number).one()
            if i.paid:
                raise ValidationError({'invoice_number': ['invoice fully paid. Cannot add payment']})
            if i:
                self.egress_amount = i.total_amount - i.pay_amount
        pass

    def after_save(self):
        try:
            if self.__is_new and self.client_id and self.result:
                # self.client.regenerate_balance()
                ClientBalanceOperationAction.check_and_create(self.client_id, user=self.update_by)
                self.current_balance = Client.get(self.client_id).actual_balance

                if self.payment_type in (
                        'invoice payment received', 'prepay payment received', 'credit note sent', 'debit sent'):
                    ClientBalanceOperationAction(client_id=self.client_id,
                                                 balance=self.amount,
                                                 ingress_balance=self.amount,
                                                 egress_balance=Decimal(0.0),
                                                 action='Increase ingress balance', create_by=self.update_by).save()
                elif self.payment_type in (
                        'invoice payment sent', 'payment sent', 'credit note received', 'debit received'):
                    ClientBalanceOperationAction(client_id=self.client_id,
                                                 balance=self.amount,
                                                 ingress_balance=Decimal(0.0),
                                                 egress_balance=self.amount,
                                                 action='Decrease egress balance', create_by=self.update_by).save()
                """
                if self.payment_type in ('reset'):
                    BalanceDailyResetTask(start_time=self.date, mutual=1, actual=1,client_id=self.client_id,create_by=self.update_by).save()
                    BalanceDailyResetTask(start_time=self.date,reset_balance=1,mutual=1,actual=1,client_id=self.client_id,create_by=self.update_by).save()
                if self.payment_type in ('mutual reset'):
                    BalanceDailyResetTask(start_time=self.date, mutual=1,client_id=self.client_id,create_by=self.update_by).save()
                    BalanceDailyResetTask(start_time=self.date, reset_balance=1, mutual=1,client_id=self.client_id,create_by=self.update_by).save()
                if self.payment_type in ('actual reset'):
                    BalanceDailyResetTask(start_time=self.date, actual=1,client_id=self.client_id,create_by=self.update_by).save()
                    BalanceDailyResetTask(start_time=self.date, reset_balance=1, actual=1,client_id=self.client_id,create_by=self.update_by).save()
                """

                if self.invoice_number:
                    i = Invoice.filter(Invoice.invoice_number == self.invoice_number).first()
                    if i:
                        i.pay_amount = ClientPayment.session().query(
                            func.sum(ClientPayment.amount).label('amount')).filter(
                            ClientPayment.invoice_number == self.invoice_number).one().amount  # ( i.pay_amount if i.pay_amount else Decimal(0.0) ) + self.amount
                        if i.pay_amount >= i.total_amount:
                            i.paid = True
                        i.save()
                        ip = InvoicePayment(payment_amount=self.amount, invoice_no=self.invoice_number,
                                            payment_id=self.client_payment_id)
                        ip.save()
                self.auto_apply_mail()
        except Exception as e:
            log.error('Client payment error: {}'.format(e))
            self.result = False
            get_db().session.rollback()
            pass

    def auto_apply_mail(self):
        # try:
        #     from api_dnl.tasks import payment_apply_mail_task
        #     if self.client.payment_received_notice and not self.email_sended:
        #         if self.payment_type in ('invoice payment received', 'prepay payment received', 'credit note sent',
        #                                  'debit received'):
        #             payment_apply_mail_task.delay(self.client_payment_id, 'payment_received')
        #         if self.payment_type in ('invoice payment sent', 'payment sent', 'credit note received', 'debit sent'):
        #             payment_apply_mail_task.delay(self.client_payment_id, 'payment_sent')
        #
        # except Exception as e:
        #     log.error('Client payment send mail error: {}'.format(e))
        #     pass
        pass

    def delete(self, **kwargs):
        user = kwargs['user'].name if 'user' in kwargs else ''
        op = None
        if self.payment_type in ('invoice payment received', 'prepay payment received',
                                 'credit note received', 'debit sent'):
            op = ClientBalanceOperationAction(client_id=self.client_id,
                                              balance=self.amount,
                                              ingress_balance=self.amount,
                                              egress_balance=Decimal(0.0),
                                              action='Decrease ingress balance', create_by=self.update_by)
        elif self.payment_type in ('invoice payment sent', 'payment sent', 'credit note sent', 'debit received'):
            op = ClientBalanceOperationAction(client_id=self.client_id,
                                              balance=self.amount,
                                              ingress_balance=Decimal(0.0),
                                              egress_balance=self.amount,
                                              action='Increase egress balance', create_by=self.update_by)

        ret = super(ClientPayment, self).delete(**kwargs)
        if op:
            op.save()
        return ret

    def fmt_name(self):
        return self.client.login

    def fmt_amount(self):
        return dec_prn(self.amount, 2)

    def fmt_current_balance(self):
        return dec_prn(self.current_balance)

    def fmt_receiving_time(self):
        return str(self.receiving_time)

    def fmt_note(self):
        return self.note

    # return """<table> <tr><td>note</td><td>{}</td></tr> <tr><td>description</td><td>{}</td></tr>
    # <tr><td>cause</td><td>{}</td></tr> </table>""".format(self.note,self.description,self.cause)

    def apply_mail(self, template, to=None, att=[]):
        return MailSender.apply_mail(self, template, to, att=att)


class C4ClientBalance(BaseModel):
    __tablename__ = 'c4_client_balance'
    id = Column(Integer, primary_key=True, server_default=text("nextval('c4_client_balance_id_seq'::regclass)"))
    client_id = Column(String(100), nullable=False, index=True, unique=True)
    balance = Column(Numeric(30, 6), nullable=False, server_default=text("0.00"))
    ingress_balance = Column(Numeric(30, 6), nullable=False, server_default=text("0.00"))
    egress_balance = Column(Numeric(30, 6), nullable=False, server_default=text("0.00"))
    create_time = Column(DateTime(True), nullable=False,
                         server_default=text("('now'::text)::timestamp(0) with time zone"))

    # carrier = relationship(Client,   primaryjoin="Client.client_id == foreign( cast(C4ClientBalance.client_id,Integer) )")

    @classmethod
    def client_balance(cls, client_id):
        q = cls().filter(text("client_id::integer=%s" % client_id)).first()
        if q:
            return q.balance
        else:
            return 0


Client.actual_balance = column_property(select([C4ClientBalance.balance]). \
                                        where(Client.client_id == cast(C4ClientBalance.client_id, Integer)). \
                                        correlate_except(C4ClientBalance))


class ClientBalance(BaseModel):  # Not used??
    __tablename__ = 'client_balance'

    client_id = Column(String(100), primary_key=True)
    balance = Column(String(100), nullable=False)
    create_time = Column(DateTime(True), nullable=False)
    ingress_balance = Column(String(100))
    egress_balance = Column(String(100))
    bod_balance = Column(String(100), nullable=False)


class ClientBalanceOperationAction(BaseModel):
    __tablename__ = 'client_balance_operation_action'
    ACTION = {0: 'Added a new client balance',
              1: 'Reset client balance, ingress balance, egress balance',
              2: 'Increase ingress balance', 3: 'Decrease ingress balance',
              4: 'Increase egress balance', 5: 'Decrease egress balance'}
    UPDATE_RESULT = {0: 'Initial', 1: 'Update completed', 2: 'Update failed'}

    id = Column(Integer, primary_key=True,
                server_default=text("nextval('client_balance_operation_action_id_seq'::regclass)"))
    client_id = Column(Integer, index=True)
    balance = Column(Numeric(12, 6), nullable=False, server_default=text("0.00"))
    ingress_balance = Column(Numeric(12, 6), nullable=False, server_default=text("0.00"))
    egress_balance = Column(Numeric(12, 6), nullable=False, server_default=text("0.00"))
    action = Column(ChoiceType(ACTION))
    update_result = Column(ChoiceType(UPDATE_RESULT), nullable=False, index=True, default=0, server_default=text("0"))
    update_by = Column(String(100))
    update_time = Column(DateTime(True))
    create_by = Column(String(100))
    create_time = Column(DateTime(True), nullable=False,
                         server_default=text("('now'::text)::timestamp(0) with time zone"))

    def after_save(self):
        # try:
        #     from .tasks import balance_log_task
        #     balance_log_task.delay(self.client_id)
        # except:
        #     pass
        pass

    @staticmethod
    def check_and_create(client_id, user, summ=0.0):
        c4 = C4ClientBalance.filter(C4ClientBalance.client_id == str(client_id)).first()
        if not c4:
            # c4 = C4ClientBalance(client_id = client_id)
            # c4.save()
            id = ClientBalanceOperationAction(client_id=client_id, balance=summ, ingress_balance=summ,
                                              egress_balance=summ,
                                              action='Added a new client balance', create_by=user).save()
            log.debug('ClientBalanceOperationAction Added a new client balance {}'.format(id))
        return c4

    @staticmethod
    def add_create(client_id, summ, user):
        balance = Decimal(summ)
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(0.0)
        id = ClientBalanceOperationAction(client_id=client_id, balance=balance, ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Added a new client balance', create_by=user).save()
        log.debug('ClientBalanceOperationAction Added a new client balance {}'.format(id))

    @staticmethod
    def add_payment(client_id, summ, user):
        balance = Decimal(summ)
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(0.0)
        id = ClientBalanceOperationAction(client_id=client_id,
                                          balance=balance,
                                          ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Increase ingress balance', create_by=user).save()
        log.debug('ClientBalanceOperationAction Increase ingress balance {}'.format(id))

    @staticmethod
    def del_payment(client_id, summ, user):
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(summ)
        balance = Decimal(summ)
        id = ClientBalanceOperationAction(client_id=client_id,
                                          balance=balance,
                                          ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Decrease ingress balance', create_by=user).save()
        log.debug('ClientBalanceOperationAction Decrease ingress balance {}'.format(id))

    @staticmethod
    def add_invoice(client_id, summ, user):
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(summ)
        balance = Decimal(summ)
        id = ClientBalanceOperationAction(client_id=client_id,
                                          balance=balance,
                                          ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Decrease egress balance', create_by=user).save()
        log.debug('ClientBalanceOperationAction Decrease egress balance {}'.format(id))

    @staticmethod
    def del_invoice(client_id, summ, user):
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(summ)
        balance = Decimal(summ)
        id = ClientBalanceOperationAction(client_id=client_id,
                                          balance=balance,
                                          ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Increase egress balance', create_by=user).save()
        log.debug('ClientBalanceOperationAction Increase egress balance {}'.format(id))

    @staticmethod
    def add_reset(client_id, summ, user):
        ingress_balance = Decimal(summ)
        egress_balance = Decimal(summ)
        balance = Decimal(summ)
        id = ClientBalanceOperationAction(client_id=client_id,
                                          balance=balance,
                                          ingress_balance=ingress_balance,
                                          egress_balance=egress_balance,
                                          action='Reset client balance, ingress balance, egress balance',
                                          create_by=user).save()
        log.debug('ClientBalanceOperationAction Reset balance {}'.format(id))


class BalanceHistoryActual(BaseModel):
    __tablename__ = 'balance_history_actual'
    __table_args__ = (UniqueConstraint('date', 'client_id'),)

    id = Column(Integer, primary_key=True, server_default=text("nextval('balance_history_actual_id_seq'::regclass)"))
    date = Column(Date, index=True, nullable=False)
    payment_received = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    credit_note_sent = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    debit_note_sent = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    unbilled_incoming_traffic = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    short_charges = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    payment_sent = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    credit_note_received = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    debit_note_received = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    unbilled_outgoing_traffic = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    actual_balance = Column(Numeric(20, 5), nullable=False)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'), nullable=False)
    actual_ingress_balance = Column(Numeric(20, 5), nullable=False, server_default=text("0"))
    actual_egress_balance = Column(Numeric(20, 5), nullable=False, server_default=text("0"))

    incoming_traffic = synonym('unbilled_incoming_traffic')
    outgoing_traffic = synonym('unbilled_outgoing_traffic')
    credit_sent = synonym('credit_note_sent')
    credit_received = synonym('credit_note_received')
    carrier_id = synonym('client_id')

    carrier = relationship(Client, back_populates='balance_actual', uselist=False)

    client_name = column_property(select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))

    carrier_name = synonym('client_name')


class BalanceHistory(BaseModel):
    __tablename__ = 'balance_history'

    id = Column(Integer, primary_key=True)
    date = Column(Date)
    payment_received = Column(Numeric(20, 5), nullable=False, default=0)
    credit_note_sent = Column(Numeric(20, 5), nullable=False, default=0)
    debit_note_sent = Column(Numeric(20, 5), nullable=False, default=0)
    payment_sent = Column(Numeric(20, 5), nullable=False, default=0)
    credit_note_received = Column(Numeric(20, 5), nullable=False, default=0)
    debit_note_received = Column(Numeric(20, 5), nullable=False, default=0)
    invoice_set = Column(Numeric(20, 5), nullable=False, default=0)
    invoice_received = Column(Numeric(20, 5), nullable=False, default=0)
    mutual_balance = Column(Numeric(20, 5), nullable=False)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'))
    mutual_ingress_balance = Column(Numeric(20, 5), nullable=False, default=0)
    mutual_egress_balance = Column(Numeric(20, 5), nullable=False, default=0)

    invoice_sent = synonym('invoice_set')
    credit_sent = synonym('credit_note_sent')
    credit_received = synonym('credit_note_received')

    carrier = relationship(Client, back_populates='balance_mutual', uselist=False)

    client_name = column_property(select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))
    carrier_name = synonym('client_name')

    @staticmethod
    def today_mutual_balance(client_id):
        def z(d):
            return d if not d is None else Decimal(0.0)

        dt = datetime.now(UTC)
        d = datetime(year=dt.year, month=dt.month, day=dt.day)
        dp = (d - timedelta(days=1))
        # dt0 = datetime(year=dt.year, month=dt.month, day=1)
        dt0 = d
        p = ClientPayment
        payment_received_total = z(p.session().query(func.sum(p.amount).label('payment_received_total')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([4, 5]),
                 p.payment_time.between(dt0, dt))).first().payment_received_total)
        print('payment_received_total {}'.format(payment_received_total))

        payment_sent_total = z(p.session().query(func.sum(p.amount).label('payment_sent_total')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([3, 6]),
                 p.payment_time.between(dt0, dt))).first().payment_sent_total)
        print('payment_sent_total {}'.format(payment_sent_total))

        debit_note_received = z(p.session().query(func.sum(p.amount).label('debit_note_received')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([11]),
                 p.payment_time.between(dt0, dt))).first().debit_note_received)
        print('debit_note_received {}'.format(debit_note_received))

        debit_note_sent = z(p.session().query(func.sum(p.amount).label('debit_note_sent')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([12]),
                 p.payment_time.between(dt0, dt))).first().debit_note_sent)
        print('debit_note_sent {}'.format(debit_note_sent))

        credit_note_received = z(p.session().query(func.sum(p.amount).label('credit_note_received')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([7]),
                 p.payment_time.between(dt0, dt))).first().credit_note_received)
        print('credit_note_received {}'.format(credit_note_received))

        credit_note_sent = z(p.session().query(func.sum(p.amount).label('credit_note_sent')).filter(
            and_(p.client_id == client_id, p.payment_type.in_([8]),
                 p.payment_time.between(dt0, dt))).first().credit_note_sent)
        print('credit_note_sent {}'.format(credit_note_sent))

        p = Invoice
        invoice_received = z(p.session().query(func.sum(p.total_amount).label('invoice_received')).filter(
            and_(p.client_id == client_id, p.state.in_([1, 9]), type == 0,
                 p.invoice_time == d)).first().invoice_received)
        print('invoice_received {}'.format(invoice_received))
        invoice_sent = z(p.session().query(func.sum(p.total_amount).label('invoice_sent')).filter(
            and_(p.client_id == client_id, p.state.in_([1, 9]), type == 1,
                 p.invoice_time == d)).first().invoice_sent)
        print('invoice_sent {}'.format(invoice_sent))
        p = BalanceHistory
        begin_balance = z(p.session().query(func.sum(p.mutual_balance).label('begin_balance')).filter(
            and_(p.client_id == client_id,
                 p.date == dp)).first().begin_balance)
        print('begin_balance {}'.format(begin_balance))
        mutual_balance = begin_balance + payment_received_total + debit_note_received - debit_note_sent - \
                         payment_sent_total - credit_note_sent - invoice_received + invoice_sent

        client_name = Client.get(client_id).name

        return dict(client_id=client_id, client_name=client_name, date=str(d.date()),
                    payment_received=payment_received_total, payment_sent=payment_sent_total,
                    debit_received=debit_note_received, debit_sent=debit_note_sent,
                    credit_received=credit_note_received, credit_sent=credit_note_sent,
                    invoice_received=invoice_received, invoice_sent=invoice_sent,
                    begin_balance=begin_balance, mutual_balance=mutual_balance)


# region +++JurisdictionPrefix
class JurisdictionPrefix(BaseModel):
    __tablename__ = 'jurisdiction_prefix'

    id = Column(Integer, primary_key=True)
    alias = Column(String(40))
    prefix = Column(PrefixRange, nullable=False, index=True, unique=True)
    jurisdiction_id = Column(Integer, index=True)  # The charging area
    jurisdiction_country_id = Column(Integer, index=True)  # country
    jurisdiction_name = Column(String(100))  # state
    jurisdiction_country_name = Column(String(100))  # , unique=True???)
    ocn = Column(String(10))
    lata = Column(String(10))
    block_id = Column(String(1))
    effective_date = Column(DateTime(True))

    # code = synonym('prefix')
    country = synonym('jurisdiction_country_name')
    state = synonym('jurisdiction_name')


# endregion  ---JurisdictionPrefix
# +++ DidBillingPlan
class DidBillingPlan(BaseModel):
    __tablename__ = 'did_billing_plan'
    RATE_TYPE = {1: 'fixed', 2: 'variable'}
    PAY_TYPE = {0: 'weekly', 1: 'monthly'}
    PRICE_TYPE = {0: 'weekly', 1: 'monthly'}

    id = Column(Integer, primary_key=True, server_default=text("nextval('did_billing_plan_id_seq'::regclass)"))
    did_price = Column(Float, default=0.0, server_default=text("0"))
    min_price = Column(Float, default=0.0, server_default=text("0"))
    name = Column(String(255), unique=True)
    payphone_subcharge = Column(String)
    monthly_charge = Column(Numeric, default=0.0)
    rate_table_id = Column(ForeignKey('rate_table.rate_table_id', ondelete='SET NULL'), nullable=True)
    rate_type = Column(ChoiceType(RATE_TYPE), server_default='1')
    pay_type = Column(ChoiceType(PAY_TYPE), server_default='0')
    fee_per_port = Column(Numeric, default=0.0)
    price_type = Column(ChoiceType(PRICE_TYPE), server_default='0')

    update_at = Column(DateTime(timezone=True), default=datetime.now(UTC))
    update_by = Column(String)
    min_time = Column(Integer)
    interval = Column(Integer)
    setup_fee = synonym('did_price')
    monthly_fee = synonym('monthly_charge')
    rate_per_min = synonym('min_price')

    rate_table = relationship('RateTable',
                              primaryjoin='foreign(DidBillingPlan.rate_table_id) == RateTable.rate_table_id',
                              uselist=False)

    # sell_dids = relationship('DidBillingPlan',
    #                           primaryjoin='foreign(DidBillingRel.sell_billing_plan_id) == DidBillingPlan.id',
    #                           uselist=True, back_populates='sell_billing_plan')
    # buy_dids = relationship('DidBillingPlan',
    #                         primaryjoin='foreign(DidBillingRel.buy_billing_plan_id) == DidBillingPlan.id',
    #                         uselist=True, back_populates='buy_billing_plan')

    def rate_table_name(self):
        return self.rate_table.name


# --- DidBillingPlan
class DidBillingRel(BaseModel):
    __tablename__ = 'did_billing_brief'
    id = Column(Integer, primary_key=True)
    did_number = Column(String(50))
    vendor_billing_plan_id = Column(Integer)
    client_billing_plan_id = Column(Integer)
    vendor_trunk_id = Column(Integer)
    client_trunk_id = Column(Integer)
    start_date = Column(DateTime(timezone=True))
    end_date = Column(DateTime(timezone=True))
    enable_for_clients = Column(Boolean, server_default='false')
    last_charged_monthly_fees = Column(Integer, server_default='-1')
    next_mrc_billing_date = Column(Date(), index=True)
    nrc_billing_date = Column(Date(), index=True)
    did = synonym('did_number')
    buy_billing_plan_id = synonym('client_billing_plan_id')
    client_res_id = synonym('client_trunk_id')
    vendor_res_id = synonym('vendor_trunk_id')
    sell_billing_plan_id = synonym('vendor_billing_plan_id')


class DidTransaction(BaseModel):
    __tablename__ = 'did_transaction'

    id = Column(Integer, primary_key=True)
    did_billing_id = Column(ForeignKey('did_billing_rel.id', ondelete='CASCADE'), index=True,
                            doc='Reference to  did_billing_rel(id) (API version >=6)')
    date = Column(DateTime(True), nullable=False, server_default=func.now())
    did_number = Column(String(50), nullable=False)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'), nullable=False)
    mrc = Column(Numeric(20, 5), nullable=True)
    nrc = Column(Numeric(20, 5), nullable=True)
    port_fee = Column(Numeric(20, 5), nullable=True)
    created_on = Column(DateTime(True), nullable=False, server_default=func.now())


# region +++ DidBillingRel
class DidBillingRelv6(BaseModel):
    __tablename__ = 'did_billing_rel'
    TOLL_FREE_DID_PREFIXES = ['1800', '1888', '1877', '1866', '1855', '1844']
    STATUS = {0: 'PNDNG', 1: 'INSVC', 2: 'PDISC'}
    id = Column(Integer, primary_key=True, server_default=text("nextval('did_billing_rel_id_seq'::regclass)"))
    did = Column(String(50))
    sell_billing_plan_id = Column(Integer)
    buy_billing_plan_id = Column(Integer)
    did_billing_id = Column(Integer)
    ingress_res_id = Column(Integer)
    egress_res_id = Column(Integer)
    start_date = Column(Date)
    end_date = Column(Date)
    is_available = Column(Boolean, default=False)
    status = Column(ChoiceType(STATUS), server_default='0')
    is_sms = Column(Boolean, default=False)

    did_plus = column_property(case([(func.length(did) == 10, func.concat('+1', did))], else_=func.concat('+', did)))

    state = column_property(select([JurisdictionPrefix.jurisdiction_name]).where(
        JurisdictionPrefix.prefix == func.substr(did, 1, 7)).correlate_except(JurisdictionPrefix))
    ocn = column_property(select([JurisdictionPrefix.ocn]).where(
        JurisdictionPrefix.prefix == func.substr(did, 1, 7)).correlate_except(JurisdictionPrefix))
    lata = column_property(select([JurisdictionPrefix.lata]).where(
        JurisdictionPrefix.prefix == func.substr(did, 1, 7)).correlate_except(JurisdictionPrefix))
    # mrc = column_property(select([DidVendor.mrc]).where(
    #     and_(egress_res_id == Resource.resource_id, DidVendor.id == Resource.client_id)).correlate_except(DidVendor,
    #                                                                                                       Resource))
    # nrc = column_property(select([DidVendor.nrc]).where(
    #     and_(egress_res_id == Resource.resource_id, DidVendor.id == Resource.client_id)).correlate_except(DidVendor,
    #                                                                                                      Resource))

    # created_time = synonym('start_date')???
    assigned_date = column_property(case([(ingress_res_id.is_(None), None)], else_=start_date))

    # client_res_id = synonym('ingress_res_id')
    # vendor_res_id = synonym('egress_res_id')
    client_res_id = synonym('egress_res_id')
    vendor_res_id = synonym('ingress_res_id')

    vendor_billing_rule_id = synonym('sell_billing_plan_id')

    client_billing_rule_id = synonym('buy_billing_plan_id')

    vendor_billing_rule = relationship('DidBillingPlan',
                                       primaryjoin='foreign(DidBillingRelv6.sell_billing_plan_id) == DidBillingPlan.id',
                                       uselist=False)  # , back_populates='sell_dids')
    sell_billing_plan = synonym('vendor_billing_rule')

    buy_billing_plan = relationship('DidBillingPlan',
                                    primaryjoin='foreign(DidBillingRelv6.buy_billing_plan_id) == DidBillingPlan.id',
                                    uselist=False)  # , back_populates='buy_dids')
    client_billing_rule = synonym('buy_billing_plan')

    vendor_res = relationship('Resource',
                              primaryjoin='foreign(DidBillingRelv6.vendor_res_id) == Resource.resource_id',
                              uselist=False)

    client_res = relationship('Resource',
                              primaryjoin='foreign(DidBillingRelv6.client_res_id) == Resource.resource_id',
                              uselist=False)
    rate = relationship('Rate', primaryjoin='Rate.code==foreign(DidBillingRelv6.did)')
    # item = relationship('ProductItems', primaryjoin='ProductItems.digits==foreign(DidBillingRel.did)')

    vendor_billing_rule_name = column_property(
        select([DidBillingPlan.name]).where(DidBillingPlan.id == sell_billing_plan_id).correlate_except(DidBillingPlan))
    client_billing_rule_name = column_property(
        select([DidBillingPlan.name]).where(DidBillingPlan.id == buy_billing_plan_id).correlate_except(DidBillingPlan))

    is_effective = column_property(or_(end_date.is_(None), end_date > datetime.now(UTC).date()))
    is_assigned = column_property(egress_res_id.isnot(None))
    is_toll_free = column_property(func.substr(did, 1, 4).in_(TOLL_FREE_DID_PREFIXES))
    did_type = column_property(case([(func.substr(did, 1, 4).in_(TOLL_FREE_DID_PREFIXES), 'toll_free')], else_='local'))

    # is_unassigned = column_property(and_(ingress_res_id.is_(None), or_(end_date.is_(None), end_date > datetime.now(UTC))))

    def before_save(self):
        # if self.is_available and self.client_res_id:
        #    self.is_available=False
        pass

    @classmethod
    def on_import_delete(cls, rec, mode):
        if mode == 'exact':
            res_ids = [r.resource_id for r in Resource.get(rec.egress_res_id).client.egress_trunks]
            old = cls.filter(and_(cls.did == rec.did, cls.client_res_id.in_(res_ids))).filter(
                cls.end_date.is_(None)).all()
        else:
            old = cls.filter(and_(cls.did == rec.did)).filter(
                cls.end_date.is_(None)).all()
        for o in old:
            o.end_date = datetime.utcnow().date()
            o.is_available = False
            o.save()

    @classmethod
    def did_billing_rel_disconnect(cls, did, client_id, vendor_id, buy_billing_plan_id=None, user_name='unknown'):
        old = cls.filter(or_(cls.did == did, cls.did == did[1:])).filter(cls.end_date.is_(None)).all()
        for o in old:
            o.end_date = datetime.utcnow().date()
            o.is_available = False
            o.save()

    @classmethod
    def did_billing_rel_create(cls, did, client_id, vendor_id, buy_billing_plan_id=None, user_name='unknown'):
        # obj.created_time = datetime.now(UTC)
        if len(did) == 10:
            did = '1' + did
        log.debug(
            'did_billing_rel_create(did={},client_id={},'
            'vendor_id={},buy_billing_plan_id={},user_name={})'.format(did, client_id, vendor_id, buy_billing_plan_id,
                                                                       user_name))
        vendor = Client.get(vendor_id)
        if not vendor:
            raise ValidationError('Bad did vendor {} '.format(client_id))
        vendor_api = None  # DidVendor.get(vendor_id) #todo DidVendor?
        log.debug('vendor api {}'.format(vendor_api))
        client = Client.get(client_id)
        if not client:
            raise ValidationError('Bad did client {} '.format(client_id))

        obj = cls(did=did, vendor_res_id=vendor.resource.resource_id, client_res_id=client.resource.resource_id)
        did = cls.filter(and_(cls.did == obj.did, cls.end_date.is_(None), cls.start_date.isnot(None),
                              cls.vendor_res_id == vendor.resource.resource_id)).first()
        if not obj.vendor_billing_rule_id:
            if vendor_api:
                obj.vendor_billing_rule_id = vendor_api.vendor_billing_rule_id
                obj.did_billing_id = obj.vendor_billing_rule_id
            elif did:
                obj.vendor_billing_rule_id = did.vendor_billing_rule_id
                obj.did_billing_id = did.did_billing_id
        if not obj.buy_billing_plan_id:
            if vendor_api:
                obj.buy_billing_plan_id = vendor_api.client_billing_rule_id
            elif did:
                obj.buy_billing_plan_id = did.buy_billing_plan_id
        if not obj.buy_billing_plan_id:
            res = client.resource
        else:
            rule = DidBillingPlan.get(obj.buy_billing_plan_id)
            res = client.get_assigned_billing_plan_resource(rule)

        old = cls.filter(cls.did == obj.did).filter(cls.end_date.is_(None)).all()

        for o in old:
            if o.id == obj.id:
                continue
            o.end_date = datetime.utcnow().date()
            o.is_available = False
            obj.ingress_res_id = o.ingress_res_id
            obj.buy_billing_plan_id = o.buy_billing_plan_id
            obj.sell_billing_plan_id = o.sell_billing_plan_id
            o.save()

        res.update_by = user_name
        ProductItems.assign_client_did(res, obj.did, user_name)
        obj.client_res = res
        obj.did_billing_id = obj.buy_billing_plan_id
        if obj.vendor_res_id:
            if obj.vendor_billing_rule_id:
                vendor_billing_rule = DidBillingPlan.get(obj.vendor_billing_rule_id)
                rate_table = vendor_billing_rule.rate_table
                vendor_res = Resource.get(obj.vendor_res_id)
                prefix = vendor_res.assign_vendor_did_prefix(obj.did, rate_table)
                prefix.save()
        obj.start_date = datetime.utcnow().date()
        try:
            if obj.buy_billing_plan_id and obj.client_res_id:
                plan = DidBillingPlan.get(obj.buy_billing_plan_id)
                client_id = Resource.get(obj.client_res_id).client_id
                if plan.setup_fee:
                    ClientBalanceOperationAction(client_id=client_id, balance=plan.setup_fee, action=3,
                                                 create_by='did_setup_fee').save()
        except Exception as e:
            log.debug('did_billing_rel created, but deducting setupo fee errors: {}'.format(e))
        log.debug('did_billing_rel created {}'.format(obj))
        return obj

    def create_time(self):
        if self.rate:
            return self.rate.create_time
        else:
            return None

    def assign_time(self):
        if self.item:
            return self.item.update_at
        else:
            return None

    @hybrid_property
    def actions(self):
        try:
            DIRECTION = {0: 'ALL', 1: 'ingress', 2: 'egress'}
            ACT = {'plus prefix': 'Add Prefix', 'minus prefix': 'Remove Prefix'}
            AF = {'plus prefix': 'digits', 'minus prefix': 'num_digits'}
            DIR = {'modify the caller': 'ani', 'modify the called': 'dnis'}
            TYPE = {0: 'modify the caller', 1: 'modify the called'}
            ret = []
            ac = None
            if self.client_res and self.client_res.directions:
                ac = {'ani': {}, 'dnis': {}}
                for a in self.client_res.directions:
                    if not a.action in ['plus prefix', 'minus prefix']:
                        continue
                    if not a.direction == 'egress':
                        continue
                    ac[DIR[a.type]]['type'] = ACT[a.action]
                    ac[DIR[a.type]][AF[a.action]] = a.digits

            if self.client_res and self.client_res.replace_actions:
                if not ac:
                    ac = {'ani': {}, 'dnis': {}}
                for a in self.client_res.replace_actions:
                    TYPE = {0: 'ani only replace the caller', 1: 'dnis only replace the called',
                            2: 'both replace the calling and called'}
                    if a.type in ['ani only replace the caller', 'both replace the calling and called']:
                        ac['ani']['type'] = 'Replace'
                        ac['ani']['new_number'] = a.ani
                        ac['ani']['digits'] = a.ani_prefix
                    if a.type in ['dins only replace the called', 'both replace the calling and called']:
                        ac['dnis']['type'] = 'Replace'
                        ac['dnis']['new_number'] = a.dnis
                        ac['dnis']['digits'] = a.dnis_prefix
            return ac
        except Exception as e:
            log.warning('Cannot get actions for did repository {}'.format(str(e)))

    @actions.setter
    def actions(self, value):
        try:
            ACT = {'Add Prefix': 'plus prefix', 'Remove Prefix': 'minus prefix',
                   'Replace': 'both replace the calling and called'}
            for d in self.client_res.directions:
                d.delete()
            for d in self.client_res.replace_actions:
                d.delete()
            if value:
                a = value
                ani = ani_pref = ani_len = None
                dnis = dnis_pref = dnis_len = None
                digits = None
                for t in ['ani', 'dnis']:
                    if t in a:
                        ac = a[t]
                        if not 'num_digits' in ac:
                            ac['num_digits'] = 0
                        if not 'digits' in ac:
                            ac['digits'] = ''
                        num_digits = None
                        if 'type' in ac:
                            actp = ac['type']
                            if actp == 'Add Prefix' and 'digits' in ac:
                                digits = ac['digits']
                            if actp == 'Remove Prefix' and 'num_digits' in ac:
                                digits = ac['num_digits']
                            if actp == 'Replace' and t == 'ani' and 'new_number' in ac:
                                ani = ac['new_number']
                                ani_pref = ac['digits']
                                ani_len = ac['num_digits']
                            if actp == 'Replace' and t == 'dnis' and 'new_number' in ac:
                                dnis = ac['new_number']
                                dnis_pref = ac['digits']
                                dnis_len = ac['num_digits']
                        rd = 'modify the caller' if t == 'ani' else 'modify the called'
                    if digits:
                        self.client_res.directions.append(
                            ResourceDirection(action=ACT[actp], type=rd, digits=digits, direction='egress'))

                if ani or dnis:
                    tp = None
                    if ani:
                        tp = 'ani only replace the caller'
                    if dnis:
                        tp = 'dnis only replace the caller'
                    if ani and dnis:
                        tp = 'both replace the calling and called'
                    if tp:
                        rep = ResourceReplaceAction(type=tp, ani_prefix=ani_pref, ani=ani, dnis_prefix=dnis_pref,
                                                    dnis=dnis,
                                                    ani_min_length=ani_len, ani_max_length=ani_len,
                                                    dnis_min_length=dnis_len, dnis_max_length=dnis_len)
                        self.client_res.replace_actions.append(rep)
        except Exception as e:
            log.warning('Cannot get actions for did repository {}'.format(str(e)))


DidBillingRel.client_id = column_property(select([Resource.client_id]). \
    where(Resource.resource_id == DidBillingRel.client_res_id).correlate_except(
    Resource))
DidBillingRel.did_client_res_name = column_property(select([Resource.alias]). \
                                                    where(
    Resource.resource_id == DidBillingRel.client_res_id).correlate_except(Resource))
DidBillingRel.did_client_name = column_property(select([Resource.client_name]). \
                                                where(
    Resource.resource_id == DidBillingRel.client_res_id).correlate_except(Resource))
DidBillingRel.did_client_is_online = column_property(select([Resource.client_is_online]). \
                                                     where(
    Resource.resource_id == DidBillingRel.client_res_id).correlate_except(Resource))
DidBillingRel.vendor_id = column_property(select([Resource.client_id]). \
    where(Resource.resource_id == DidBillingRel.vendor_res_id).correlate_except(
    Resource))
DidBillingRel.did_vendor_res_name = column_property(select([Resource.alias]). \
                                                    where(
    Resource.resource_id == DidBillingRel.vendor_res_id).correlate_except(Resource))
DidBillingRel.did_vendor_name = column_property(select([Resource.client_name]). \
                                                where(
    Resource.resource_id == DidBillingRel.vendor_res_id).correlate_except(Resource))


# endregion--- DidBillingRel


class DidReport(BaseModel):
    __tablename__ = 'did_report'

    report_time = Column(DateTime(True))
    duration = Column(Integer)
    ingress_bill_time = Column(Integer)
    ingress_call_cost = Column(Numeric(15, 6))
    ingress_total_calls = Column(Integer)
    not_zero_calls = Column(Integer)
    ingress_success_calls = Column(Integer)
    ingress_client_id = Column(Integer)
    ingress_id = Column(Integer)
    egress_client_id = Column(Integer)
    egress_id = Column(Integer)
    egress_bill_time = Column(Integer)
    egress_call_cost = Column(Numeric(15, 6))
    egress_total_calls = Column(Integer)
    egress_success_calls = Column(Integer)
    did = Column(String(100))
    ingress_ip = Column(String(20))
    egress_ip = Column(String(20))
    pdd = Column(Integer)
    ingress_busy_calls = Column(Integer)
    egress_busy_calls = Column(Integer)
    ingress_cancel_calls = Column(Integer)
    egress_cancel_calls = Column(Integer)

    __table_args__ = (PrimaryKeyConstraint(
        'ingress_client_id', 'ingress_id', 'egress_client_id', 'egress_id', 'report_time'
    ),)


class DidInvoiceChargeDetail(BaseModel):
    __tablename__ = 'did_invoice_charge_detail'
    id = Column(Integer, primary_key=True)
    invoice_number = Column(ForeignKey('invoice_summary.invoice_number', ondelete='CASCADE'))
    did_number = Column(String(64), nullable=False, index=True)
    client_billing_plan_id = Column(ForeignKey('did_billing_plan.id', ondelete='CASCADE'))
    client_trunk_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'))
    nrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    mrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    minutes = Column(Integer, server_default='0')
    attempts = Column(Integer, server_default='0')
    call_charge = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    start_date = Column(Date)
    end_date = Column(Date)
    mrc_number = Column(Integer, server_default='0')
    pay_type = Column(Integer, server_default='0')
    did_mrc = column_property(mrc * mrc_number)

    total = column_property((mrc * mrc_number).op('+')(nrc).op('+')(call_charge))
    billing_plan_name = column_property(select([DidBillingPlan.name]).where(
        DidBillingPlan.id == client_billing_plan_id).correlate_except(DidBillingPlan))
    billing_plan_mrc_rate = column_property(select([DidBillingPlan.monthly_charge]).where(
        DidBillingPlan.id == client_billing_plan_id).correlate_except(DidBillingPlan))
    billing_plan_nrc_rate = column_property(select([DidBillingPlan.did_price]).where(
        DidBillingPlan.id == client_billing_plan_id).correlate_except(DidBillingPlan))


class DidInvoiceChargeBreakdown(BaseModel):
    __tablename__ = 'did_invoice_charge_breakdown'
    id = Column(Integer, primary_key=True)
    invoice_number = Column(ForeignKey('invoice_summary.invoice_number', ondelete='CASCADE'))
    client_billing_plan_id = Column(ForeignKey('did_billing_plan.id', ondelete='CASCADE'))
    client_trunk_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'))
    nrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    mrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    minutes = Column(Integer, server_default='0')
    attempts = Column(Integer, server_default='0')
    call_charge = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    did_count = Column(Integer, server_default='0')


class DidInvoicePortCharge(BaseModel):
    __tablename__ = 'did_invoice_port_charge'
    id = Column(Integer, primary_key=True)
    invoice_number = Column(ForeignKey('invoice_summary.invoice_number', ondelete='CASCADE'))
    client_trunk_id = Column(ForeignKey('resource.resource_id', ondelete='CASCADE'))
    billing_port_type = Column(ChoiceType(Resource.BILLING_PORT_TYPE), server_default='0')
    price_per_max_channel = Column(Numeric(32, 6), server_default=text('0.00'))
    max_channels = Column(Integer, server_default='0')
    price_per_actual_channel = Column(Numeric(32, 6), server_default=text('0.00'))
    did_count = Column(Integer, server_default='0')
    cost_per_port = Column(Numeric(32, 6), server_default=text('0.00'))
    capacity = Column(Integer, server_default='0')
    dnis_cap_limit = Column(Integer, server_default='0')
    did_count = Column(Integer, server_default='0')
    channels = Column(Integer, server_default='0')
    port_charge = Column(Numeric(32, 6), server_default=text('0.00'))


# region +++ Invoice
class Invoice(BaseModel):
    __tablename__ = 'invoice'
    STATE = {-1: 'void', 0: 'normal', 1: 'verify', 9: 'send'}
    TYPE = {0: 'sent(out--buy)', 1: 'received(in--sell)', 2: 'sent(all)', 3: 'incoming'}
    DISPUTED = {0: 'Non-Disputed', 1: 'Disputed', 2: 'Dispute Resolved'}
    STATUS = {0: 'creating', 1: 'zero cdr', 2: 'done', 3: 'only support buy/sell'}
    OUTPUT_TYPE = {0: 'pdf', 1: 'excel', 2: 'html'}
    CREATE_TYPE = {0: 'auto', 1: 'manual'}

    invoice_id = Column(Integer, primary_key=True, server_default=text("nextval('invoice_invoice_id_seq'::regclass)"))
    api_params = Column(Text)
    api_requests = Column(Integer, server_default=text("0"))
    buy_minutes = Column(Float)
    buy_service_charge = Column(Numeric(30, 10))
    buy_total = Column(Numeric(30, 10))
    cdr_link = Column(Text)
    cdr_path = Column(String(300))
    cdr_token = Column(Text)
    client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'), index=True)
    create_type = Column(ChoiceType(CREATE_TYPE), default=0)  # 0-Auto-generated Invoice; 1-Manual Invoice
    credit_amount = Column(Numeric(30, 5), server_default=text("0"))
    credit_note = Column(Numeric(30, 10))
    credit_remaining = Column(Numeric(20, 5))
    current_balance = Column(Numeric(30, 10), nullable=False)
    current_charge = Column(Numeric(30, 10))
    decimal_place = Column(Integer, server_default=text("5"))
    decimals_num = Column(Integer)
    decurring_charge = Column(Numeric(30, 10))
    disputed = Column(ChoiceType(DISPUTED), default=0)  # 0-Non-Disputed; 1-Disputed; 2-Dispute Resolved
    disputed_amount = Column(Numeric(30, 10))
    due_date = Column(Date)
    egress_cdr_file = Column(String(255))
    finance_charge = Column(Numeric(30, 10))
    generate_copy_time = Column(DateTime(True))
    generate_end_time = Column(DateTime(True))
    generate_start_time = Column(DateTime(True))
    generate_stats_time = Column(DateTime(True))
    include_detail = Column(Boolean, server_default=text("true"))
    ingress_cdr_file = Column(String(255))
    invoice_balance_time = Column(DateTime(True))
    invoice_end = Column(DateTime(True), nullable=False)
    invoice_include_payment = Column(Boolean)
    invoice_jurisdictional_detail = Column(Boolean, server_default=text("true"))
    invoice_log_id = Column(Integer)
    invoice_number = Column(String(20), nullable=False, unique=True)
    invoice_start = Column(DateTime(True), nullable=False)
    invoice_time = Column(DateTime(True), nullable=False,
                          server_default=text("('now'::text)::timestamp(0) with time zone"))
    invoice_use_balance_type = Column(Integer, server_default=text("0"))
    invoice_zone = Column(String(10))
    is_invoice_account_summary = Column(Boolean, server_default=text("false"))
    is_send_as_link = Column(Boolean, server_default=text("false"))
    is_short_duration_call_surcharge_detail = Column(Boolean, server_default=text("false"))
    is_show_by_date = Column(Boolean, server_default=text("false"))
    is_show_code_100 = Column(Boolean, server_default=text("false"))
    is_show_code_name = Column(Boolean, server_default=text("false"))
    is_show_country = Column(Boolean, server_default=text("false"))
    is_show_daily_usage = Column(Boolean, server_default=text("false"))
    is_show_detail_trunk = Column(Boolean, server_default=text("false"))
    is_show_total_trunk = Column(Boolean, server_default=text("false"))
    link_cdr = Column(String(50))
    lrn_cost = Column(Numeric(30, 10))
    lrn_numbers = Column(Integer)
    lrn_rate = Column(Numeric(30, 10))
    new_balance = Column(Numeric(30, 10))
    non_recurring_charge = Column(Numeric(30, 10))
    output_type = Column(Integer, server_default=text("0"))
    paid = Column(Boolean, nullable=False, server_default=text("false"))
    pay_amount = Column(Numeric(30, 5), nullable=False, server_default=text("0"))
    payment_credit = Column(Numeric(30, 10))
    payment_term = Column(String)
    pdf_path = Column(String(300))
    please_pay = Column(Numeric(20, 5))
    previous_balance = Column(Numeric(30, 10))
    rate_value = Column(Integer, server_default=text("0"))
    reconcile_file_path = Column(String)
    reconcile_state = Column(Integer, server_default=text("0"))
    scc_calls = Column(Integer)
    scc_cost = Column(Numeric(30, 5))
    scc_per = Column(Float)
    scc_sec = Column(Integer)
    sell_minutes = Column(Float)
    sell_service_charge = Column(Numeric(30, 10))
    sell_total = Column(Numeric(30, 10))
    send_time = Column(DateTime(True))
    state = Column(ChoiceType(STATE), nullable=False, default=0)  # 1:void;0:normal;1:verify;9:send
    status = Column(ChoiceType(STATUS))
    tax = Column(Numeric(30, 10))
    total_amount = Column(Numeric(30, 5), nullable=False, server_default=text("0"))
    total_calls = Column(Integer, server_default=text("0"))
    total_minutes = Column(Float)
    type = Column(Integer)
    unlimited_credit_unlimited = Column(Boolean, server_default=text("false"))
    usage_detail_fields = Column(Text)

    amount = synonym('total_amount')
    invoice_date = synonym('invoice_time')
    invoice_period_from = synonym('invoice_start')
    invoice_period_to = synonym('invoice_end')
    start_date = synonym('invoice_start')
    end_date = synonym('invoice_end')
    invoice_amount = synonym('total_amount')
    # url=synonym('link_cdr')

    # log = relationship('InvoiceLog', back_populates='invoices', uselist=False)
    client = relationship(Client, back_populates='invoices', uselist=False)
    payments = relationship('PaymentInvoice', uselist=True, back_populates='invoice')

    invoice_period = column_property(
        cast(cast(invoice_start, Date), String) + ' ' + cast(cast(invoice_end, Date), String))

    client_name = column_property(select([Client.name]).where(Client.client_id == client_id).correlate_except(Client))
    carrier_name = synonym('client_name')
    auto_send_invoice = column_property(
        select([Client.auto_send_invoice]).where(Client.client_id == client_id).correlate_except(Client))
    payment_term_name = column_property(
        select([PaymentTerm.name]).where(and_(Client.payment_term_id == PaymentTerm.payment_term_id,
                                              Client.client_id == client_id)).correlate_except(PaymentTerm, Client)
    )
    company_type = column_property(
        select([Client.company_type]).where(Client.client_id == client_id).correlate_except(Client))

    invoice_payment = column_property(func.coalesce(
        select([func.sum(ClientPayment.amount)]).where(
            and_(ClientPayment.invoice_number == invoice_number,
                 ClientPayment.payment_type == 'invoice payment received')).correlate_except(ClientPayment).as_scalar(),
        0.0))
    client_credit = column_property(func.coalesce(
        select([func.sum(ClientPayment.amount)]).where(
            and_(ClientPayment.invoice_number == invoice_number,
                 ClientPayment.payment_type == 2)).correlate_except(ClientPayment).as_scalar(), 0.0))
    client_offset = column_property(func.coalesce(
        select([func.sum(ClientPayment.amount)]).where(
            and_(ClientPayment.invoice_number == invoice_number,
                 ClientPayment.payment_type == 3)).correlate_except(ClientPayment).as_scalar(), 0.0))
    credit_note = column_property(func.coalesce(
        select([func.sum(ClientPayment.amount)]).where(
            and_(ClientPayment.invoice_number == invoice_number,
                 ClientPayment.payment_type == 'credit note sent')).correlate_except(ClientPayment).as_scalar(), 0.0))

    def payment_term(self):
        return self.client.payment_term

    @property
    def cdr_url(self):
        # if self.ingress_cdr_file:  # self.status == 'done':
        #     return '{}{}/files/{}'.format(settings.API_SCHEME, settings.API_HOST,
        #                                   self.ingress_cdr_file)
        # else:
        #     return 'cdr file was not created'
        return self.cdr_link

    @property
    def fmt_cdr_url(self):
        if self.cdr_link:
            return "<a href='{}'>CDR LINK</a>".format(self.cdr_link)
        else:
            return 'cdr file was not created'

    @property
    def cdr_data(self):
        if self.ingress_cdr_file:  # self.status == 'done':
            # return open(self.cdr_path+'/'+self.ingress_cdr_file,'rb').read()
            return open(settings.FILES['upload_to'] + '/' + self.ingress_cdr_file, 'rb').read()
        else:
            return None

    @property
    def url(self):
        if self.pdf_path:  # self.status == 'done':
            return '{}{}{}/invoice/{}.pdf'.format(settings.API_SCHEME, settings.API_HOST,
                                                  settings.API_BASE_PATH, self.invoice_num)
        else:
            return None

    @property
    def data_url(self):
        if self.pdf_path:  # self.status == 'done':
            return '{}{}{}/invoice/{}.json'.format(settings.API_SCHEME, settings.API_HOST,
                                                   settings.API_BASE_PATH, self.invoice_num)
        else:
            return None

    @property
    def invoice_link(self):
        import base64
        base_url = self.session().query(SystemParameter.base_url).first().base_url
        b64id = base64.encodebytes(str(self.invoice_id).encode()).decode().replace('\n', '')
        # return '{}/pr/pr_invoices/createpdf_invoice/{}'.format(base_url,self.invoice_number)
        return '{}/homes/getInvoicePdf/{}'.format(base_url, b64id)

    @property
    def fmt_invoice_link(self):
        return "<a href='{}'>Download Invoice here</a>".format(self.invoice_link)

    @property
    def download_link(self):
        return self.url

    @property
    def file_data(self):
        try:
            f = open('{}/{}'.format(settings.FILES['pdf_path'], self.pdf_path.replace('invoice/', '')), 'rb')
            return f.read()
        except:
            return None

    @classmethod
    def _max_number(cls):
        ret = cls.query().session.query(func.max(func.cast(cls.invoice_number, Integer)).label("max_number")).first()
        if ret and ret.max_number:
            return ret.max_number
        else:
            return 0

    @classmethod
    def max_number(cls):
        sql = "SELECT nextval('class4_seq_invoice_no') AS last_invoice_number"
        last_invoice_number = cls.session.execute(sql).fetchall()[0][0]
        while cls.filter(cls.invoice_number == str(last_invoice_number)).first():
            last_invoice_number = cls.session.execute(sql).fetchall()[0][0]
        return last_invoice_number

    # @classmethod
    # def create_invoice(cls, client_id, start_date, end_date, gmt, show_cdr, show_trunk_breakdown):
    #     # start_date=start_date+gmt;
    #     # end_date=end_date+gmt;
    #     obj = cls(client_id=client_id, invoice_start=start_date, invoice_end=end_date, invoice_zone=gmt,
    #               is_short_duration_call_surcharge_detail=show_cdr, is_show_detail_trunk=show_trunk_breakdown)
    #     obj.state = 0
    #
    #     return obj

    @property
    def fmt_start_date(self):
        return str(self.invoice_start)

    @property
    def fmt_end_date(self):
        return str(self.invoice_end)

    @property
    def fmt_invoice_amount(self):
        return dec_prn(self.amount, 2)
        tz=get_time_zone(self.invoice_zone)
        create_time = datetime.now(tz)
INVOICE_SEND_STATE = 'send'
Client.over_due = column_property(select([func.sum(Invoice.total_amount - Invoice.pay_amount).label('over_due')]). \
                                  where(
    and_(Invoice.client_id == Client.client_id, Invoice.paid.isnot(True), Invoice.state == literal_column("9"),
         Invoice.due_date < datetime.now(UTC))). \
                                  correlate_except(Invoice))
ClientPayment.invoice_total_amount = column_property(
    select([Invoice.total_amount]).where(ClientPayment.invoice_number == Invoice.invoice_number).correlate_except(
        Invoice))
ClientPayment.invoice_paid_amount = column_property(
    select([ClientPayment.amount.op('+')(Invoice.total_amount).op('-')(ClientPayment.egress_amount)]).where(
        ClientPayment.invoice_number == Invoice.invoice_number).correlate_except(Invoice))
# ClientPayment.invoice_unpaid_amount = column_property(select([Invoice.total_amount-Invoice.pay_amount]).where(ClientPayment.invoice_number==Invoice.invoice_number).correlate_except(Invoice))
ClientPayment.invoice_due_date = column_property(
    select([Invoice.due_date]).where(ClientPayment.invoice_number == Invoice.invoice_number).correlate_except(Invoice))
ClientPayment.invoice_period = column_property(
    select([Invoice.invoice_period]).where(ClientPayment.invoice_number == Invoice.invoice_number).correlate_except(
        Invoice))
ClientPayment.invoice_id = column_property(
    select([Invoice.invoice_id]).where(Invoice.invoice_number == ClientPayment.invoice_number).correlate_except(
        Invoice))
Client.mutual_balance_old = column_property(select([BalanceHistory.mutual_balance]). \
                                            where(Client.client_id == BalanceHistory.client_id). \
                                            order_by(BalanceHistory.date.desc()).limit(1). \
                                            correlate_except(BalanceHistory).label('mutual_balance'))


class PaymentInvoice(BaseModel):
    __tablename__ = 'payment_invoice'

    id = Column(Integer, primary_key=True, server_default=text("nextval('payment_invoice_id_seq'::regclass)"))
    payment_id = Column(ForeignKey('client_payment.client_payment_id'))
    invoice_id = Column(ForeignKey('invoice.invoice_id'))
    amount = Column(Numeric(20, 5))
    client_id = column_property(
        select([Invoice.client_id.label('client_client_id')]).where(Invoice.invoice_id == invoice_id).correlate_except(
            Invoice))

    invoice = relationship('Invoice', uselist=False, back_populates='payments')
    payment = relationship('ClientPayment', uselist=False, back_populates='invoices')


# endregion --- Invoice

#
class InvoiceHistory(BaseModel):
    __tablename__ = 'invoice_history'

    id = Column(Integer, primary_key=True, server_default=text("nextval('invoice_history_id_seq'::regclass)"))
    client_id = Column(Integer, nullable=False, index=True)
    last_invoice_for = Column(Date, index=True, server_default=text("('now'::text)::date"))
    reminders_count = Column(Integer, server_default=text("0"))
    next_reminder_timestamp = Column(Integer, server_default=text("0"))
    invoice_filename = Column(String(500))
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    next_invoice_date = Column(Date)
    last_invoice_amount = Column(Numeric(20, 5))
    last_invoice_period = Column(String(32))

    client = relationship(Client, uselist=False, primaryjoin="foreign(InvoiceHistory.client_id)==Client.client_id",
                          back_populates='last_invoice_history')


Invoice.last_invoice_for = column_property(select([InvoiceHistory.last_invoice_for]). \
                                           where(Invoice.client_id == InvoiceHistory.client_id).order_by(
    InvoiceHistory.last_invoice_for.desc()).limit(1).correlate_except(InvoiceHistory))
Client.last_invoice_date = column_property(select([InvoiceHistory.last_invoice_for]). \
                                           where(Client.client_id == InvoiceHistory.client_id).order_by(
    InvoiceHistory.last_invoice_for.desc()).limit(1).correlate_except(InvoiceHistory))

Client.last_invoice_amount = column_property(select([InvoiceHistory.last_invoice_amount]). \
                                             where(Client.client_id == InvoiceHistory.client_id).order_by(
    InvoiceHistory.last_invoice_for.desc()).limit(1).correlate_except(InvoiceHistory))

Client.last_invoice_period = column_property(select([InvoiceHistory.last_invoice_period]). \
                                             where(Client.client_id == InvoiceHistory.client_id).order_by(
    InvoiceHistory.last_invoice_for.desc()).limit(1).correlate_except(InvoiceHistory))

Client.next_invoice_date = column_property(select([InvoiceHistory.next_invoice_date]). \
                                           where(Client.client_id == InvoiceHistory.client_id).order_by(
    InvoiceHistory.last_invoice_for.desc()).limit(1).correlate_except(InvoiceHistory))


class InvoiceDebug(BaseModel):
    __tablename__ = 'invoice_debug'
    id = Column(Integer, primary_key=True)
    time = Column(DateTime(True), index=True, nullable=False, server_default=func.now())
    client_id = Column(Integer, index=True, nullable=True)
    text = Column(Text)


class InvoiceTask(BaseModel):
    __tablename__ = 'invoice_task'
    STATUS = {0: 'Initial', 1: 'In Process', 2: 'Successful', 3: 'Failed'}
    id = Column(Integer, primary_key=True)
    client_id = Column(Integer)
    is_auto = Column(Boolean, nullable=False, server_default=text("false"))
    created_by_user_id = Column(Integer)
    created_on = Column(DateTime(True), server_default=func.now())
    bill_period_start = Column(Date)
    bill_period_end = Column(Date)
    show_acc_summary = Column(Boolean, nullable=False, default=False)
    show_summary_of_charges = Column(Boolean, nullable=False, default=False)
    show_payment_applied = Column(Boolean, nullable=False, server_default=text("false"))
    show_rec_charge = Column(Boolean, nullable=False, server_default=text("false"))
    show_non_rec_charge = Column(Boolean, nullable=False, server_default=text("false"))
    show_usage_charge = Column(Boolean, nullable=False, server_default=text("false"))
    show_trunk_detail = Column(Boolean, nullable=False, server_default=text("false"))
    show_daily_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_jd_daily_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_jd_trunk_detail = Column(Boolean, nullable=False, server_default=text("false"))
    show_prefix_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_code_name_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_country_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_inbound_summary = Column(Boolean, nullable=False, server_default=text("false"))
    show_top_100tn_summary = Column(Boolean, nullable=False, server_default=text("false"))

    show_did_charge_breakdown = Column(Boolean, nullable=False, server_default=text("false"))
    show_did_charge_summary = Column(Boolean, nullable=False, server_default=text("false"))

    enable_email = Column(Boolean, nullable=False, server_default=text("false"))
    enable_email_with_cdr = Column(Boolean, nullable=False, server_default=text("false"))
    # addons
    include_tax = Column(Boolean, nullable=False, server_default=text("false"))
    include_scc_charges = Column(Boolean, nullable=False, server_default=text("false"))
    payment_dur_date = Column(Date)
    invoice_zone = Column(String(20),default='+00:00')
    # processing
    status = Column(ChoiceType(STATUS), nullable=False, server_default=text("0"))
    progress = Column(String(1024))
    proc_start_time = Column(DateTime(True))
    proc_end_time = Column(DateTime(True))

    summary = relationship('InvoiceSummary', uselist=False, back_populates='task')

    client_name = column_property(
        select([Client.name]).where(Client.client_id == foreign(client_id)).correlate_except(Client))
    client_type = column_property(
        select([Client.client_type]).where(Client.client_id == foreign(client_id)).correlate_except(Client))
    show_scc_charges = synonym('include_scc_charges')
    @property
    def show_client_data(self):
        return True

    @classmethod
    def create_auto_invoice_task(cls, client_id, start_time, end_time, user_id=None):
        client = Client.get(client_id)
        if client:
            did_invoice_include = ''
            if not client.did_invoice_include is None:
                did_invoice_include = client.did_invoice_include
            orig = client.client_type == 'client'
            obj = cls(client_id=client_id,
                      bill_period_start=start_time,
                      bill_period_end=end_time,
                      show_acc_summary=client.is_invoice_account_summary,
                      show_summary_of_charges='show_summary_of_charges' in did_invoice_include,
                      show_payment_applied=client.invoice_include_payment,
                      show_rec_charge='show_rec_charge' in did_invoice_include,
                      show_non_rec_charge='show_non_rec_charge' in did_invoice_include,
                      show_usage_charge=client.invoice_show_details,
                      show_trunk_detail=client.is_show_detail_trunk,
                      show_daily_summary=client.is_show_daily_usage,
                      show_jd_daily_summary=client.is_show_daily_usage and client.invoice_jurisdictional_detail,
                      show_jd_trunk_detail=client.invoice_jurisdictional_detail,
                      show_prefix_summary=client.is_show_total_trunk,
                      show_code_name_summary=client.is_show_code_name,
                      show_country_summary=client.is_show_country,
                      show_inbound_summary=client.is_show_by_date,
                      show_top_100tn_summary=client.is_show_code_100,
                      show_did_charge_breakdown='show_did_charge_breakdown' in did_invoice_include,
                      show_did_charge_summary='show_did_charge_summary' in did_invoice_include,
                      enable_email=client.email_invoice,
                      enable_email_with_cdr=client.invoice_send_mode_cdr in ('link', 'attachment'),
                      include_tax=client.include_tax,
                      include_scc_charges=client.is_short_duration_call_surcharge_detail,
                      payment_dur_date=client.payment_term.next_date(end_time) + timedelta(days=1),
                      invoice_zone='+00:00' if client.invoice_zone is None else client.invoice_zone,

                      is_auto=True,
                      created_by_user_id=user_id
                      )
            obj.save()
            return obj
        else:
            return None

    def create_summary(self, invoice_number):
        from inv5.invoice_summary import InvoiceSummaryGenerator
        import hashlib
        s = self.bill_period_start
        e = self.bill_period_end
        tz=get_time_zone(self.invoice_zone)
        create_time = datetime.now(tz)
        obj = InvoiceSummaryGenerator(
            invoice_number=invoice_number,
            client_id=self.client_id,
            invoice_start_time=datetime(s.year, s.month, s.day, 0, 0, 0),
            invoice_end_time=datetime(e.year, e.month, e.day, 23, 59, 59),
            task_id=self.id,
            invoice_type=0 if self.is_auto else 1,
            invoice_zone = self.invoice_zone,
            invoice_generation_time=create_time,
            create_time=create_time
        )
        invoice_number = obj.save()
        old = Invoice.filter(Invoice.invoice_number == str(invoice_number)).first()
        while old:
            invoice_number += 1
            old = Invoice.filter(Invoice.invoice_number == str(invoice_number)).first()
            if old:
                continue
            gen = InvoiceSummaryGenerator.filter(InvoiceSummaryGenerator.invoice_number == invoice_number).first()
            if gen:
                continue
            obj.invoice_number = invoice_number
            invoice_number = obj.save()
        obj.invoice_num = hashlib.md5(str(invoice_number).encode()).hexdigest()
        obj.save()
        return obj

    def run(self):
        self.status = 'In Process'
        self.proc_start_time = datetime.now(UTC)
        try:
            self.save()
            old = InvoiceSummary.filter(InvoiceSummary.task_id == self.id).first()
            invoice_number = None
            if old:
                invoice_number = old.invoice_number
                old.delete()
                self.session.commit()
            obj = self.create_summary(invoice_number)
            obj.update()
            obj.save()
            if self.is_auto and not obj.client.invoice_zero and obj.total_amount==0.0:
                self.progress = 'Summary Successful,Zero invoice not to generate'
                self.save()
                raise Exception('Zero invoice not to generate')
            else:
                self.progress = 'Summary Successful'
                self.save()
                obj.create_invoice()
                self.status = 'Successful'
                self.progress = 'Summary Successful,File successful'
                self.save()
        except Exception as e:
            self.progress = 'error:' + str(e)[:1000]
            self.status = 'Failed'
            raise e
        finally:
            self.proc_end_time = datetime.now(UTC)
            self.save()


class InvoiceSummary(BaseModel):
    __tablename__ = 'invoice_summary'
    TYPE = {0: 'auto invoice', 1: 'manual invoice', 2: 'DID auto invoice', 3: 'DID manual invoice'}
    invoice_number = Column(Integer, primary_key=True,
                            server_default=text("nextval('invoice_summary_invoice_number_seq'::regclass)"))
    invoice_num = Column(String(256), index=True)
    client_id = Column(Integer, nullable=False, index=True)
    invoice_generation_time = Column(DateTime, index=True)
    invoice_start_time = Column(DateTime)
    invoice_end_time = Column(DateTime)
    payment_dur_date = Column(Date)
    invoice_zone = Column(String(20))
    payment_id = Column(Integer)
    current_balance = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    previous_balance = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    payments_and_credits = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    balance_forward = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    long_distance_charges = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    taxes = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    scc_calls = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    scc_amount = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    current_charges = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    total_calls = Column(Integer, nullable=False, server_default=text("0"))
    total_minutes = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    total_amount = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    unlimited_credit = Column(Boolean, nullable=False, server_default=text("false"))
    include_tax = Column(Boolean, nullable=False, server_default=text("false"))
    include_scc_charges = Column(Boolean, nullable=False, server_default=text("false"))
    include_daily_details = Column(Boolean, nullable=False, server_default=text("false"))
    include_country_details = Column(Boolean, nullable=False, server_default=text("false"))
    include_code_name_details = Column(Boolean, nullable=False, server_default=text("false"))
    include_trunk_details = Column(Boolean, nullable=False, server_default=text("false"))
    include_trunk_prefix_details = Column(Boolean, nullable=False, server_default=text("false"))
    invoice_type = Column(ChoiceType(TYPE), nullable=False, server_default=text("0"))
    did_mrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    did_nrc = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    create_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    json_content = Column(MutableDict.as_mutable(JSON))  # Column(JSON)
    task_id = Column(ForeignKey('invoice_task.id', ondelete='CASCADE'))

    port_charges = Column(Numeric(32, 6), nullable=False, server_default=text('0.00'))
    task = relationship(InvoiceTask, uselist=False, back_populates='summary')
    did_invoice_charge_detail = relationship(DidInvoiceChargeDetail,
                                             uselist=True, single_parent=True, cascade="all, delete-orphan")
    did_invoice_charge_breakdown = relationship(DidInvoiceChargeBreakdown,
                                                uselist=True, single_parent=True, cascade="all, delete-orphan")
    client = relationship(Client, primaryjoin=Client.client_id == foreign(client_id),
                          uselist=False)
    client_name = column_property(
        select([Client.name]).where(Client.client_id == foreign(client_id)).correlate_except(Client))
    client_address = column_property(
        select([Client.address]).where(Client.client_id == foreign(client_id)).correlate_except(Client))
    client_invoice_zone = column_property(
        select([Client.invoice_zone]).where(Client.client_id == foreign(client_id)).correlate_except(Client))
    @property
    def include_mrc(self):
        return bool(self.did_mrc) or bool(self.did_nrc)

    @property
    def invoice_start_time_tz(self):
        return self.invoice_start_time.replace(tzinfo=get_time_zone(self.invoice_zone))

    @property
    def invoice_end_time_tz(self):
        return self.invoice_end_time.replace(tzinfo=get_time_zone(self.invoice_zone))

    @property
    def start_time(self):
        return int(self.invoice_start_time.replace(tzinfo=get_time_zone(self.invoice_zone)).timestamp())

    @property
    def end_time(self):
        return int(self.invoice_end_time.replace(tzinfo=get_time_zone(self.invoice_zone)).timestamp())

    @property
    def billing_period(self):
        return self.invoice_start_time.strftime("%Y-%m-%d") + ' ' + self.invoice_end_time.strftime("%Y-%m-%d")

    @property
    def config(self):
        if not hasattr(self, '_config_') or self._config_ is None:
            if self.client.invoice_setting and self.client.invoice_setting.setting:
                self._config_ = self.client.invoice_setting.setting
            else:
                self._config_ = SystemParameter.get(1)
        return self._config_

    @property
    def pdf_tpl(self):
        return self.config.pdf_tpl

    @property
    def tpl_number(self):
        return self.config.tpl_number

    @property
    def logo(self):
        import os, base64
        if hasattr(settings, 'INVOICE_LOGO_URL') and settings.INVOICE_LOGO_URL:
            path = '{}/logo_client_{}.png'.format(settings.FILES['upload_to'], self.client_id)
            if os.path.exists(path):
                os.unlink(path)
            url = settings.INVOICE_LOGO_URL.replace('{client_id}', str(self.client_id))
            resp = requests.get(url)
            if resp.status_code == 200:
                with open(path, 'wb') as f:
                    f.write(resp.content)
                    return path

        path = '{}/logo_client_{}.png'.format(settings.FILES['upload_to'], self.client_id)
        conf = SystemParameter.get(1)
        url = None
        try:
            purl = parse_url(conf.base_url)
            ret = requests.get(
                '{}://{}/homes/getInvoiceLogoUrl?client_id={}'.format(purl.scheme, purl.netloc, self.client_id), verify=False)
            if ret.status_code in (200, 201, 202):
                url = ret.text
            else:
                url = '{}://{}/upload/images/ilogo.png'.format(purl.scheme, purl.netloc)
        except:
            pass
        try:
            resp = requests.get(url, verify=False)
            if resp.status_code in (200, 202, 202):
                if 'text/html' in resp.headers['Content-type']:
                    raise Exception('Logo url got not image!')
                with open(path, 'wb') as f:
                    f.write(resp.content)
                from PIL import Image
                img=Image.open(path)  # just try to identify
                img.thumbnail((300,300), Image.ANTIALIAS)
                img.save(path)
                return path
        except Exception as e:
            log.warning('InvoiceSummary Cannot get logo, check system_parameter.base_url {}'.format(e))
            path = '{}/empty_logo.png'.format(settings.FILES['upload_to'])
            pass
        log.warning('InvoiceSummary No logo, use empty one')
        s = b'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mP8z8BQDwAEhQGAhKmMIQAAAABJRU5ErkJggg=='
        with open(path, 'wb') as f:
            f.write(base64.b64decode(s))
        return path

    @property
    def logo_url(self):
        # return self.logo.replace(settings.FILES['upload_to'], settings.API_URL)
        try:
            return self.config.ilogo_url
        except:
            return None

    @property
    def company_info(self):
        return self.config.company_info

    @property
    def company_info_location(self):
        return self.config.company_info_location

    @property
    def billing_info(self):
        return self.config.billing_info

    @property
    def billing_info_location(self):
        return self.config.billing_info_location

    @property
    def filename(self):
        return '{}_{}_{}.pdf'.format(self.invoice_number,str(self.client_name),
                                  datetime.now().strftime("%Y%m%d%H%M"))

    def create_invoice(self):
        from inv5.invoice_pdf import InvoicePdf
        from inv5.invoice_summary import InvoiceSummaryGenerator
        filename = self.filename
        log.warning('payment_dur_date set to {}'.format(str(self.payment_dur_date)))
        try:
            # if self.invoice_type in ('auto invoice', 'manual invoice'):
            old = Invoice.filter(Invoice.invoice_number == str(self.invoice_number)).all()
            if len(old):
                # log.warning('Delete old invoice {}'.format(str(self.invoice_number)))
                # old.delete()
                log.warning('Void old invoices {}'.format(str(self.invoice_number)))
                for o in old:
                    # o.state='void'
                    # o.save()
                    o.delete()
            invoice_type = 0 if self.invoice_type == 'auto invoice' else 1
            gen = InvoiceSummaryGenerator.get(self.invoice_number)
            pdf = InvoicePdf(self, settings.FILES['pdf_path'] + '/' + filename)
            pdf.generate()
            obj = Invoice(
                invoice_number=str(self.invoice_number),
                client_id=self.client_id,
                create_type=invoice_type,
                invoice_start=self.invoice_start_time,
                invoice_end=self.invoice_end_time,
                invoice_time=self.create_time
            )
            obj.due_date = self.payment_dur_date,
            obj.current_balance = self.current_balance,
            obj.previous_balance = self.previous_balance,
            obj.payment_credit = self.payments_and_credits,
            obj.current_charge = self.current_charges,
            obj.total_amount = self.total_amount,
            obj.decimals_num = self.config.invoice_decimal_digits,
            obj.pdf_path = 'invoice/' + filename,
            obj.company_type = self.client.company_type,
            obj.tax = self.taxes,
            obj.scc_calls = self.scc_calls,
            obj.scc_cost = self.scc_amount,
            obj.invoice_zone = self.invoice_zone,
            obj.is_invoice_account_summary = self.task.show_acc_summary,
            obj.is_send_as_link = self.config.invoice_send_mode == 'link',
            obj.is_short_duration_call_surcharge_detail = self.task.include_scc_charges,
            obj.is_show_by_date = self.task.show_daily_summary,
            obj.is_show_code_100 = self.task.show_top_100tn_summary,
            obj.is_show_code_name = self.task.show_code_name_summary,
            obj.is_show_country = self.task.show_country_summary,
            obj.is_show_daily_usage = self.task.show_daily_summary,
            obj.is_show_detail_trunk = self.task.show_jd_trunk_detail,
            obj.is_show_total_trunk = self.task.show_trunk_detail,
            obj.non_recurring_charge = self.did_nrc,
            # obj.payment_term = self.client.payment_term_name,
            obj.status = 2,  # 'done',
            obj.type = 0,  # 'sent(out--buy)'
            obj.save()
        except Exception as e:
            log.error('create invoice error {}'.format(traceback.format_exc()))
            raise e


InvoiceTask.invoice_number = column_property(
    select([InvoiceSummary.invoice_number]).where(
        InvoiceSummary.task_id == foreign(InvoiceTask.id)).correlate_except(InvoiceSummary))

InvoiceTask.invoice_num = column_property(
    select([InvoiceSummary.invoice_num]).where(
        InvoiceSummary.task_id == foreign(InvoiceTask.id)).correlate_except(InvoiceSummary))

Invoice.invoice_num = column_property(
    select([InvoiceSummary.invoice_num]).where(
        InvoiceSummary.invoice_number == cast(Invoice.invoice_number, Integer)).correlate_except(InvoiceSummary))


# region CdrReport

def cdr_reports_start_date():
    tablename = 'cdr_report_detail2%'
    try:
        r = get_db().session.execute(
            text("select min(relname) as tablename from pg_class where reltuples > 0 and relname like '{}';".format(
                tablename)))
        ret = r.fetchone()
        if not ret or not ret.tablename:
            return None
        tn = ret.tablename
        return datetime(int(tn[17:21]), int(tn[21:23]), int(tn[23:25])).date()
    except:
        return None

def cdr_reports_iterator(start_date_time,end_date_time):
    start_day = start_date_time.date()
    database_start_day = cdr_reports_start_date()
    if database_start_day is None:
        raise ValidationError('database has no cdr report data!')
    if database_start_day > start_day:
        start_day = database_start_day
    end_day = end_date_time.date()
    if end_day < start_day:
        end_day = start_day
    days = (end_day - start_day).days
    for day_idx in range(days + 1):
        day_cdr = start_day + timedelta(days=day_idx)
        cdr_date = "%s%02d%02d" % (day_cdr.year, day_cdr.month, day_cdr.day)
        cdr = cdr_report_detail(cdr_date)
        yield cdr


def client_cdr_iterator(start_date_time,end_date_time):
    start_day = start_date_time.date()
    database_start_day = client_cdr_start_date()
    if database_start_day is None:
        raise ValidationError('database has no cdr report data!')
    if database_start_day > start_day:
        start_day = database_start_day
    end_day = end_date_time.date()
    if end_day < start_day:
        end_day = start_day
    days = (end_day - start_day).days
    for day_idx in range(days + 1):
        day_cdr = start_day + timedelta(days=day_idx)
        cdr_date = "%s%02d%02d" % (day_cdr.year, day_cdr.month, day_cdr.day)
        cdr = client_cdr(cdr_date)
        yield cdr

class HostBasedReport(BaseModel):
    __tablename__ = 'host_based_report'
    report_time = Column(DateTime(True))
    duration = Column(Integer)
    ingress_bill_time = Column(Integer)
    ingress_call_cost = Column(Numeric(15, 6))
    ingress_total_calls = Column(Integer)
    not_zero_calls = Column(Integer)
    ingress_success_calls = Column(Integer)
    ingress_busy_calls = Column(Integer)
    pdd = Column(BigInteger)
    ingress_cancel_calls = Column(Integer)
    ingress_client_id = Column(Integer)
    egress_client_id = Column(Integer)
    egress_bill_time = Column(Integer)
    egress_call_cost = Column(Numeric(15, 6))
    egress_total_calls = Column(Integer)
    egress_success_calls = Column(Integer)
    egress_busy_calls = Column(Integer)
    egress_cancel_calls = Column(Integer)
    ingress_ip = Column(String(50))
    egress_ip = Column(String(50))
    ingress_avg_rate = Column(Float(53))
    egress_avg_rate = Column(Float(53))
    cdr_date = Column(String(24))
    ingress_duration = Column(Integer)
    ingress_not_zero_calls = Column(Integer)
    __table_args__ = (PrimaryKeyConstraint(
        'ingress_client_id', 'ingress_ip', 'egress_client_id', 'egress_ip', 'report_time'
    ),)


def host_based_report(date):
    tablename = 'host_based_report{}'.format(date)
    try:
        r = get_db().session.execute(text("select tablename from pg_tables where tablename='{}';".format(tablename)))
        if not r.fetchall()[0][0] == tablename:
            return None
    except:
        return None
    return Table(
        tablename, BaseModel.metadata,
        Column('report_time', DateTime(True)),
        Column('duration', Integer),
        Column('ingress_bill_time', Integer),
        Column('ingress_call_cost', Numeric(15, 6)),
        Column('ingress_total_calls', Integer),
        Column('not_zero_calls', Integer),
        Column('ingress_success_calls', Integer),
        Column('ingress_busy_calls', Integer),
        Column('pdd', BigInteger),
        Column('ingress_cancel_calls', Integer),
        Column('ingress_client_id', Integer),
        Column('egress_client_id', Integer),
        Column('egress_bill_time', Integer),
        Column('egress_call_cost', Numeric(15, 6)),
        Column('egress_total_calls', Integer),
        Column('egress_success_calls', Integer),
        Column('egress_busy_calls', Integer),
        Column('egress_cancel_calls', Integer),
        Column('ingress_ip', String(50)),
        Column('egress_ip', String(50)),
        Column('ingress_avg_rate', Float(53)),
        Column('egress_avg_rate', Float(53)),
        Column('cdr_date', String(24)),
        Column('ingress_duration', Integer),
        Column('ingress_not_zero_calls', Integer),
        extend_existing=True,
    )


def did_report(date):
    tablename = 'did_report{}'.format(date)
    try:
        r = get_db().session.execute(text("select tablename from pg_tables where tablename='{}';".format(tablename)))
        if not r.fetchall()[0][0] == tablename:
            return None
    except:
        return None
    return Table(
        tablename, BaseModel.metadata,
        Column('report_time', DateTime(True)),
        Column('duration', Integer),
        Column('ingress_bill_time', Integer),
        Column('ingress_call_cost', Numeric(15, 6)),
        Column('ingress_total_calls', Integer),
        Column('not_zero_calls', Integer),
        Column('ingress_success_calls', Integer),
        Column('ingress_client_id', Integer),
        Column('ingress_id', Integer),
        Column('egress_client_id', Integer),
        Column('egress_id', Integer),
        Column('egress_bill_time', Integer),
        Column('egress_call_cost', Numeric(15, 6)),
        Column('egress_total_calls', Integer),
        Column('egress_success_calls', Integer),
        Column('did', String(100)),
        Column('ingress_ip', Text()),
        Column('egress_ip', Text()),
        Column('cdr_date', String(24)),
        Column('pdd', Integer),
        Column('ingress_busy_calls', Integer),
        Column('egress_busy_calls', Integer),
        Column('ingress_cancel_calls', Integer),
        Column('egress_cancel_calls', Integer),
        Column('cdr_date', String(24)),
        Column('mrc', Numeric(15, 6)),
        Column('nrc', Numeric(15, 6)),
        Column('surcharge', Numeric(15, 6)),
        extend_existing=True,
    )


def cdr_report_detail(date):
    tablename = 'cdr_report_detail{}'.format(date)
    try:
        sql = "select relname from pg_class where relname='{}';".format(tablename)
        r = get_db().session.execute(text(sql)).fetchall()
        if not r or not r[0][0] == tablename:
            return None
    except Exception as e:
        return None
    return Table(
        tablename, BaseModel.metadata,
        Column('report_time', DateTime(True)),
        Column('duration', Integer),
        Column('ingress_bill_time', Integer),
        Column('ingress_call_cost', Numeric(15, 6)),
        Column('lnp_cost', Numeric(15, 8)),
        Column('ingress_total_calls', Integer),
        Column('not_zero_calls', Integer),
        Column('ingress_success_calls', Integer),
        Column('ingress_busy_calls', Integer),
        Column('lrn_calls', Integer),
        Column('pdd', BigInteger),
        Column('ingress_cancel_calls', Integer),
        Column('ingress_client_id', Integer),
        Column('ingress_id', Integer),
        Column('ingress_country', String(100)),
        Column('ingress_code_name', String(100)),
        Column('egress_client_id', Integer),
        Column('egress_id', Integer),
        Column('egress_country', String(100)),
        Column('egress_code_name', String(100)),
        Column('egress_code', String(100)),
        Column('ingress_prefix', String(100)),
        Column('egress_bill_time', Integer),
        Column('egress_call_cost', Numeric(15, 6)),
        Column('egress_total_calls', Integer),
        Column('egress_success_calls', Integer),
        Column('egress_busy_calls', Integer),
        Column('egress_cancel_calls', Integer),
        Column('not_zero_calls_30', Integer),
        Column('duration_30', Integer),
        Column('not_zero_calls_6', Integer),
        Column('duration_6', Integer),
        Column('call_18s', Integer),
        Column('call_24s', Integer),
        Column('call_2h', Integer),
        Column('call_3h', Integer),
        Column('call_4h', Integer),
        Column('call_12s', Integer),
        Column('ingress_rate', Float(53)),
        Column('egress_rate', Float(53)),
        Column('product_rout_id', Integer),
        Column('ingress_rate_date', BigInteger),
        Column('egress_rate_date', BigInteger),
        Column('incoming_bandwidth', Integer),
        Column('outgoing_bandwidth', Integer),
        Column('ingress_call_cost_intra', Numeric(15, 6)),
        Column('ingress_call_cost_inter', Numeric(15, 6)),
        Column('egress_call_cost_intra', Numeric(15, 6)),
        Column('egress_call_cost_inter', Numeric(15, 6)),
        Column('ingress_bill_time_intra', Integer),
        Column('ingress_bill_time_inter', Integer),
        Column('egress_bill_time_intra', Integer),
        Column('egress_bill_time_inter', Integer),
        Column('agent_id', Integer),
        Column('agent_rate', Float(53)),
        Column('agent_cost', Numeric(12, 6)),
        Column('ingress_rate_table_id', Integer),
        Column('route_plan_id', Integer),
        Column('orig_jur_type', Integer),
        Column('term_jur_type', Integer),
        Column('par_id', Integer),
        Column('origination_destination_host_name', String(255)),
        Column('termination_source_host_name', String(255)),
        Column('binary_value_of_release_cause_from_protocol_stack', String(100)),
        Column('release_cause', SmallInteger),
        Column('release_cause_from_protocol_stack', String(100)),
        Column('inter_ingress_total_calls', Integer),
        Column('intra_ingress_total_calls', Integer),
        Column('inter_duration', Integer),
        Column('intra_duration', Integer),
        Column('inter_not_zero_calls', Integer),
        Column('intra_not_zero_calls', Integer),
        Column('q850_cause_count', Integer),
        Column('npr_count', Integer),
        Column('ingress_call_cost_local', Numeric(15, 6)),
        Column('ingress_call_cost_ij', Numeric(15, 6)),
        Column('egress_call_cost_local', Numeric(15, 6)),
        Column('egress_call_cost_ij', Numeric(15, 6)),
        Column('cdr_date', String(24)),
        Column('ingress_code', String(100)),
        Column('ring_pdd', Integer),
        Column('egress_no_ring', Integer),
        Column('ingress_duration', Integer),
        Column('ingress_inter_duration', Integer),
        Column('ingress_intra_duration', Integer),
        Column('ingress_inter_not_zero_calls', Integer),
        Column('ingress_intra_not_zero_calls', Integer),
        Column('ingress_not_zero_calls', Integer),
        Column('ingress_not_zero_calls_30', Integer),
        Column('ingress_duration_30', Integer),
        Column('ingress_not_zero_calls_6', Integer),
        Column('ingress_duration_6', Integer),
        Column('ingress_call_18s', Integer),
        Column('ingress_call_24s', Integer),
        Column('ingress_call_2h', Integer),
        Column('ingress_call_3h', Integer),
        Column('ingress_call_4h', Integer),
        Column('ingress_call_12s', Integer),
        Column('src_ocn', String(100)),
        Column('src_lata', Integer),
        Column('dst_ocn', String(100)),
        Column('dst_lata', Integer),
        Column('lrn_ocn', String(100)),
        Column('lrn_lata', Integer),
        Column('origination_source_host_name', String(255)),
        Column('termination_destination_host_name', String(255)),
        extend_existing=True,
    )


class CrdReportDetailTable(object):
    report_time = Column(DateTime(timezone=True))
    duration = Column(Integer)
    ingress_bill_time = Column(Integer)
    ingress_call_cost = Column(Numeric(15, 6))
    lnp_cost = Column(Numeric(15, 8))
    ingress_total_calls = Column(Integer)
    not_zero_calls = Column(Integer)
    ingress_success_calls = Column(Integer)
    ingress_busy_calls = Column(Integer)
    lrn_calls = Column(Integer)
    pdd = Column(BigInteger)
    ingress_cancel_calls = Column(Integer)
    ingress_client_id = Column(Integer)
    ingress_id = Column(Integer)
    ingress_country = Column(String(100))
    ingress_code_name = Column(String(100))
    egress_client_id = Column(Integer)
    egress_id = Column(Integer)
    egress_country = Column(String(100))
    egress_code_name = Column(String(100))
    egress_code = Column(String(100))
    ingress_prefix = Column(String(100))
    egress_bill_time = Column(Integer)
    egress_call_cost = Column(Numeric(15, 6))
    egress_total_calls = Column(Integer)
    egress_success_calls = Column(Integer)
    egress_busy_calls = Column(Integer)
    egress_cancel_calls = Column(Integer)
    not_zero_calls_30 = Column(Integer)
    duration_30 = Column(Integer)
    not_zero_calls_6 = Column(Integer)
    duration_6 = Column(Integer)
    call_18s = Column(Integer)
    call_24s = Column(Integer)
    call_2h = Column(Integer)
    call_3h = Column(Integer)
    call_4h = Column(Integer)
    call_12s = Column(Integer)
    ingress_rate = Column(Float(53))
    egress_rate = Column(Float(53))
    product_rout_id = Column(Integer)
    ingress_rate_date = Column(BigInteger)
    egress_rate_date = Column(BigInteger)
    incoming_bandwidth = Column(Integer)
    outgoing_bandwidth = Column(Integer)
    ingress_call_cost_intra = Column(Numeric(15, 6))
    ingress_call_cost_inter = Column(Numeric(15, 6))
    egress_call_cost_intra = Column(Numeric(15, 6))
    egress_call_cost_inter = Column(Numeric(15, 6))
    ingress_bill_time_intra = Column(Integer)
    ingress_bill_time_inter = Column(Integer)
    egress_bill_time_intra = Column(Integer)
    egress_bill_time_inter = Column(Integer)
    agent_id = Column(Integer)
    agent_rate = Column(Float(53))
    agent_cost = Column(Numeric(12, 6))
    ingress_rate_table_id = Column(Integer)
    route_plan_id = Column(Integer)
    orig_jur_type = Column(Integer)
    term_jur_type = Column(Integer)
    par_id = Column(Integer)
    origination_destination_host_name = Column(String(255))
    termination_source_host_name = Column(String(255))
    binary_value_of_release_cause_from_protocol_stack = Column(String(100))
    release_cause = Column(SmallInteger)
    release_cause_from_protocol_stack = Column(String(100))
    inter_ingress_total_calls = Column(Integer)
    intra_ingress_total_calls = Column(Integer)
    inter_duration = Column(Integer)
    intra_duration = Column(Integer)
    inter_not_zero_calls = Column(Integer)
    intra_not_zero_calls = Column(Integer)
    q850_cause_count = Column(Integer)
    npr_count = Column(Integer)
    ingress_call_cost_local = Column(Numeric(15, 6))
    ingress_call_cost_ij = Column(Numeric(15, 6))
    egress_call_cost_local = Column(Numeric(15, 6))
    egress_call_cost_ij = Column(Numeric(15, 6))
    cdr_date = Column(String(24))
    ingress_code = Column(String(100))


class CdrReportDetail(BaseModel):
    __tablename__ = 'cdr_report_detail'

    report_time = Column(DateTime(timezone=True))
    duration = Column(Integer)
    ingress_bill_time = Column(Integer)
    ingress_call_cost = Column(Numeric(15, 6))
    lnp_cost = Column(Numeric(15, 8))
    ingress_total_calls = Column(Integer)
    not_zero_calls = Column(Integer)
    ingress_success_calls = Column(Integer)
    ingress_busy_calls = Column(Integer)
    lrn_calls = Column(Integer)
    pdd = Column(BigInteger)
    ingress_cancel_calls = Column(Integer)
    ingress_client_id = Column(Integer)
    ingress_id = Column(Integer)
    ingress_country = Column(String(100))
    ingress_code_name = Column(String(100))
    egress_client_id = Column(Integer)
    egress_id = Column(Integer)
    egress_country = Column(String(100))
    egress_code_name = Column(String(100))
    egress_code = Column(String(100))
    ingress_prefix = Column(String(100))
    egress_bill_time = Column(Integer)
    egress_call_cost = Column(Numeric(15, 6))
    egress_total_calls = Column(Integer)
    egress_success_calls = Column(Integer)
    egress_busy_calls = Column(Integer)
    egress_cancel_calls = Column(Integer)
    not_zero_calls_30 = Column(Integer)
    duration_30 = Column(Integer)
    not_zero_calls_6 = Column(Integer)
    duration_6 = Column(Integer)
    call_18s = Column(Integer)
    call_24s = Column(Integer)
    call_2h = Column(Integer)
    call_3h = Column(Integer)
    call_4h = Column(Integer)
    call_12s = Column(Integer)
    ingress_rate = Column(Float(53))
    egress_rate = Column(Float(53))
    product_rout_id = Column(Integer)
    ingress_rate_date = Column(BigInteger)
    egress_rate_date = Column(BigInteger)
    incoming_bandwidth = Column(Integer)
    outgoing_bandwidth = Column(Integer)
    ingress_call_cost_intra = Column(Numeric(15, 6))
    ingress_call_cost_inter = Column(Numeric(15, 6))
    egress_call_cost_intra = Column(Numeric(15, 6))
    egress_call_cost_inter = Column(Numeric(15, 6))
    ingress_bill_time_intra = Column(Integer)
    ingress_bill_time_inter = Column(Integer)
    egress_bill_time_intra = Column(Integer)
    egress_bill_time_inter = Column(Integer)
    agent_id = Column(Integer)
    agent_rate = Column(Float(53))
    agent_cost = Column(Numeric(12, 6))
    ingress_rate_table_id = Column(Integer)
    route_plan_id = Column(Integer)
    orig_jur_type = Column(Integer)
    term_jur_type = Column(Integer)
    par_id = Column(Integer)
    origination_destination_host_name = Column(String(255))
    termination_source_host_name = Column(String(255))
    binary_value_of_release_cause_from_protocol_stack = Column(String(100))
    release_cause = Column(SmallInteger)
    release_cause_from_protocol_stack = Column(String(100))
    inter_ingress_total_calls = Column(Integer)
    intra_ingress_total_calls = Column(Integer)
    inter_duration = Column(Integer)
    intra_duration = Column(Integer)
    inter_not_zero_calls = Column(Integer)
    intra_not_zero_calls = Column(Integer)
    q850_cause_count = Column(Integer)
    npr_count = Column(Integer)
    ingress_call_cost_local = Column(Numeric(15, 6))
    ingress_call_cost_ij = Column(Numeric(15, 6))
    egress_call_cost_local = Column(Numeric(15, 6))
    egress_call_cost_ij = Column(Numeric(15, 6))
    cdr_date = Column(String(24))
    ingress_code = Column(String(100))
    ring_pdd = Column(Integer)
    egress_no_ring = Column(Integer)
    ingress_duration = Column(Integer)
    ingress_inter_duration = Column(Integer)
    ingress_intra_duration = Column(Integer)
    ingress_inter_not_zero_calls = Column(Integer)
    ingress_intra_not_zero_calls = Column(Integer)
    ingress_not_zero_calls = Column(Integer)
    ingress_not_zero_calls_30 = Column(Integer)
    ingress_duration_30 = Column(Integer)
    ingress_not_zero_calls_6 = Column(Integer)
    ingress_duration_6 = Column(Integer)
    ingress_call_18s = Column(Integer)
    ingress_call_24s = Column(Integer)
    ingress_call_2h = Column(Integer)
    ingress_call_3h = Column(Integer)
    ingress_call_4h = Column(Integer)
    ingress_call_12s = Column(Integer)
    src_ocn = Column(String(100))
    src_lata = Column(Integer)
    dst_ocn = Column(String(100))
    dst_lata = Column(Integer)
    lrn_ocn = Column(String(100))
    lrn_lata = Column(Integer)
    origination_source_host_name = Column(String(255))
    termination_destination_host_name = Column(String(255))
    __table_args__ = (PrimaryKeyConstraint(
        'ingress_client_id', 'ingress_id', 'egress_client_id', 'egress_id', 'product_rout_id', 'agent_id',
        'ingress_rate_table_id', 'route_plan_id', 'par_id', 'report_time'
    ),)


# +++
def client_cdr_start_date():
    tablename = 'client_cdr2%'
    try:
        r = get_db().session.execute(
            text("select min(relname) as tablename from pg_class where reltuples > 0 and relname like '{}';".format(
                tablename)))
        ret = r.fetchone()
        if not ret or not ret.tablename:
            return None
        tn = ret.tablename
        return datetime(int(tn[10:14]), int(tn[14:16]), int(tn[16:18])).date()
    except:
        return None


def client_cdr(date):
    tablename = 'client_cdr{}'.format(date)
    try:
        r = get_db().session.execute(text("select tablename from pg_tables where tablename='{}';".format(tablename)))
        if not r.fetchall()[0][0] == tablename:
            return None
    except:
        return None
    return Table(
        tablename, BaseModel.metadata,
        Column('id', BigInteger),
        Column('record_sequence_number', String(100)),
        Column('version_number', String(100)),
        Column('record_type', String(100)),
        Column('connection_type', String(100)),
        Column('session_id', String(100)),
        Column('release_cause', SmallInteger),
        Column('start_time_of_date', BigInteger),
        Column('answer_time_of_date', BigInteger),
        Column('release_tod', BigInteger),
        Column('minutes_west_of_greenwich_mean_time', SmallInteger),
        Column('release_cause_from_protocol_stack', String(100)),
        Column('binary_value_of_release_cause_from_protocol_stack', String(100)),
        Column('first_release_dialogue', String(100)),
        Column('trunk_id_origination', String(100)),
        Column('voip_protocol_origination', String(100)),
        Column('origination_source_number', String(100), index=True),
        Column('origination_source_host_name', String(100)),
        Column('origination_destination_number', String(100), index=True),
        Column('origination_destination_host_name', String(100)),
        Column('origination_call_id', String(150)),
        Column('origination_remote_payload_ip_address', String(100)),
        Column('origination_remote_payload_udp_address', Integer),
        Column('origination_local_payload_ip_address', String(100)),
        Column('origination_local_payload_udp_address', Integer),
        Column('origination_codec_list', String(100)),
        Column('origination_ingress_packets', Integer),
        Column('origination_egress_packets', Integer),
        Column('origination_ingress_octets', Integer),
        Column('origination_egress_octets', Integer),
        Column('origination_ingress_packet_loss', Integer),
        Column('origination_ingress_delay', Integer),
        Column('origination_ingress_packet_jitter', Integer),
        Column('trunk_id_termination', String(100)),
        Column('voip_protocol_termination', String(100)),
        Column('termination_source_number', String(100)),
        Column('termination_source_host_name', String(100)),
        Column('termination_destination_number', String(100)),
        Column('termination_destination_host_name', String(100)),
        Column('termination_call_id', String(150)),
        Column('termination_remote_payload_ip_address', String(100)),
        Column('termination_remote_payload_udp_address', Integer),
        Column('termination_local_payload_ip_address', String(100)),
        Column('termination_local_payload_udp_address', Integer),
        Column('termination_codec_list', String(100)),
        Column('termination_ingress_packets', Integer),
        Column('termination_egress_packets', Integer),
        Column('termination_ingress_octets', Integer),
        Column('termination_egress_octets', Integer),
        Column('termination_ingress_packet_loss', Integer),
        Column('termination_ingress_delay', Integer),
        Column('termination_ingress_packet_jitter', Integer),
        Column('final_route_indication', String(100)),
        Column('routing_digits', String(100), index=True),
        Column('call_duration', Integer, index=True),
        Column('pdd', Integer),
        Column('ring_time', BigInteger),
        Column('callduration_in_ms', Integer),
        Column('conf_id', String(100)),
        Column('call_type', SmallInteger),
        Column('ingress_id', Integer),
        Column('ingress_client_id', Integer),
        Column('ingress_client_rate_table_id', Integer),
        Column('ingress_client_currency_id', Integer),
        Column('ingress_client_rate', Float(53)),
        Column('ingress_client_currency', String(100)),
        Column('ingress_client_bill_time', Integer),
        Column('ingress_client_bill_result', SmallInteger),
        Column('ingress_client_cost', Numeric(12, 6)),
        Column('time', DateTime(True), index=True, server_default=text("('now'::text)::timestamp(0) with time zone")),
        Column('egress_id', Integer),
        Column('egress_rate_table_id', Integer),
        Column('egress_rate', Float(53)),
        Column('egress_cost', Numeric(12, 6)),
        Column('egress_bill_time', Integer),
        Column('egress_client_id', Integer),
        Column('egress_client_currency_id', Integer),
        Column('egress_client_currency', String(100)),
        Column('egress_six_seconds', Integer),
        Column('egress_bill_minutes', Float),
        Column('egress_bill_result', SmallInteger),
        Column('ingress_bill_minutes', Float),
        Column('ingress_dnis_type', SmallInteger),
        Column('ingress_rate_type', SmallInteger),
        Column('lrn_dnis', String(100)),
        Column('egress_dnis_type', SmallInteger),
        Column('egress_rate_type', SmallInteger),
        Column('translation_ani', String(100)),
        Column('ani_code_id', Integer),
        Column('dnis_code_id', Integer),
        Column('item_id', Integer),
        Column('ingress_rate_id', Integer),
        Column('egress_rate_id', Integer),
        Column('rerate_time', DateTime(True)),
        Column('orig_code', String(100)),
        Column('orig_code_name', String(100)),
        Column('orig_country', String(100)),
        Column('term_code', String(100)),
        Column('term_code_name', String(100)),
        Column('term_country', String(100)),
        Column('ingress_rate_effective_date', BigInteger),
        Column('egress_rate_effective_date', BigInteger),
        Column('egress_erro_string', Text),
        Column('route_plan', Integer),
        Column('dynamic_route', Integer),
        Column('static_route', Integer),
        Column('contract_id', String(100)),
        Column('order_id', String(100)),
        Column('order_type', String(100)),
        Column('lrn_number_vendor', SmallInteger),
        Column('lnp_dipping_cost', Numeric(10, 8)),
        Column('is_final_call', SmallInteger, index=True),
        Column('egress_code_asr', Float),
        Column('egress_code_acd', Float),
        Column('route_prefix', String(100)),
        Column('is_manual_kill', Boolean, server_default=text("false")),
        Column('orig_call_duration', Integer),
        Column('orig_delay_second', Integer),
        Column('term_delay_second', Integer),
        Column('trunk_type', SmallInteger),
        Column('origination_profile_port', Integer),
        Column('termination_profile_port', Integer),
        Column('o_trunk_type2', SmallInteger),
        Column('o_billing_method', SmallInteger),
        Column('t_trunk_type2', SmallInteger),
        Column('t_billing_method', SmallInteger),
        Column('campaign_id', Integer),
        Column('tax', Integer),
        Column('agent_id', Integer),
        Column('agent_rate', Float(53)),
        Column('agent_cost', Numeric(12, 6)),
        Column('orig_jur_type', Integer),
        Column('term_jur_type', Integer),
        Column('ring_epoch', BigInteger),
        Column('end_epoch', BigInteger),
        Column('par_id', Integer),
        Column('paid_user', String(100)),
        Column('rpid_user', String(100)),
        Column('timeout_type', SmallInteger),
        Column('q850_cause', Integer),
        Column('q850_cause_string', String(100)),
        Column('src_lata', Integer),
        Column('src_ocn', String(100)),
        Column('src_rc', String(100)),
        Column('dst_lata', Integer),
        Column('dst_ocn', String(100)),
        Column('dst_rc', String(100)),
        Column('lrn_lata', Integer),
        Column('lrn_ocn', String(100)),
        Column('lrn_rc', String(100)),
        extend_existing=True,
    )


class ClientCdr(BaseModel):
    __tablename__ = 'client_cdr'

    id = Column(Integer, primary_key=True, server_default=text("nextval('client_cdr_id_seq'::regclass)"))
    record_sequence_number = Column(String(100))
    version_number = Column(String(100))
    record_type = Column(String(100))
    connection_type = Column(String(100))
    session_id = Column(String(100))
    release_cause = Column(SmallInteger)
    start_time_of_date = Column(BigInteger)
    answer_time_of_date = Column(BigInteger)
    release_tod = Column(BigInteger)
    minutes_west_of_greenwich_mean_time = Column(SmallInteger)
    release_cause_from_protocol_stack = Column(String(100))
    binary_value_of_release_cause_from_protocol_stack = Column(String(100))
    first_release_dialogue = Column(String(100))
    trunk_id_origination = Column(String(100))
    voip_protocol_origination = Column(String(100))
    origination_source_number = Column(String(100))
    origination_source_host_name = Column(String(100))
    origination_destination_number = Column(String(100))
    origination_destination_host_name = Column(String(100))
    origination_call_id = Column(String(150))
    origination_remote_payload_ip_address = Column(String(100))
    origination_remote_payload_udp_address = Column(Integer)
    origination_local_payload_ip_address = Column(String(100))
    origination_local_payload_udp_address = Column(Integer)
    origination_codec_list = Column(String(100))
    origination_ingress_packets = Column(Integer)
    origination_egress_packets = Column(Integer)
    origination_ingress_octets = Column(Integer)
    origination_egress_octets = Column(Integer)
    origination_ingress_packet_loss = Column(Integer)
    origination_ingress_delay = Column(Integer)
    origination_ingress_packet_jitter = Column(Integer)
    trunk_id_termination = Column(String(100))
    voip_protocol_termination = Column(String(100))
    termination_source_number = Column(String(100))
    termination_source_host_name = Column(String(100))
    termination_destination_number = Column(String(100))
    termination_destination_host_name = Column(String(100))
    termination_call_id = Column(String(150))
    termination_remote_payload_ip_address = Column(String(100))
    termination_remote_payload_udp_address = Column(Integer)
    termination_local_payload_ip_address = Column(String(100))
    termination_local_payload_udp_address = Column(Integer)
    termination_codec_list = Column(String(100))
    termination_ingress_packets = Column(Integer)
    termination_egress_packets = Column(Integer)
    termination_ingress_octets = Column(Integer)
    termination_egress_octets = Column(Integer)
    termination_ingress_packet_loss = Column(Integer)
    termination_ingress_delay = Column(Integer)
    termination_ingress_packet_jitter = Column(Integer)
    final_route_indication = Column(String(100))
    routing_digits = Column(String(100))
    call_duration = Column(Integer)
    pdd = Column(Integer)
    ring_time = Column(BigInteger)
    callduration_in_ms = Column(Integer)
    conf_id = Column(String(100))
    call_type = Column(SmallInteger)
    ingress_id = Column(Integer)
    ingress_client_id = Column(Integer)
    ingress_client_rate_table_id = Column(Integer)
    ingress_client_currency_id = Column(Integer)
    ingress_client_rate = Column(Float(53))
    ingress_client_currency = Column(String(100))
    ingress_client_bill_time = Column(Integer)
    ingress_client_bill_result = Column(SmallInteger)
    ingress_client_cost = Column(Numeric(12, 6))
    time = Column(DateTime(timezone=True), index=True, default=datetime.now(UTC))
    egress_id = Column(Integer)
    egress_rate_table_id = Column(Integer)
    egress_rate = Column(Float(53))
    egress_cost = Column(Numeric(12, 6))
    egress_bill_time = Column(Integer)
    egress_client_id = Column(Integer)
    egress_client_currency_id = Column(Integer)
    egress_client_currency = Column(String(100))
    egress_six_seconds = Column(Integer)
    egress_bill_minutes = Column(Float)
    egress_bill_result = Column(SmallInteger)
    ingress_bill_minutes = Column(Float)
    ingress_dnis_type = Column(SmallInteger)
    ingress_rate_type = Column(SmallInteger)
    lrn_dnis = Column(String(100))
    egress_dnis_type = Column(SmallInteger)
    egress_rate_type = Column(SmallInteger)
    translation_ani = Column(String(100))
    ani_code_id = Column(Integer)
    dnis_code_id = Column(Integer)
    item_id = Column(Integer)
    ingress_rate_id = Column(Integer)
    egress_rate_id = Column(Integer)
    rerate_time = Column(DateTime(timezone=True))
    orig_code = Column(String(100))
    orig_code_name = Column(String(100))
    orig_country = Column(String(100))
    term_code = Column(String(100))
    term_code_name = Column(String(100))
    term_country = Column(String(100))
    ingress_rate_effective_date = Column(BigInteger)
    egress_rate_effective_date = Column(BigInteger)
    egress_erro_string = Column(Text)
    route_plan = Column(Integer)
    dynamic_route = Column(Integer)
    static_route = Column(Integer)
    contract_id = Column(String(100))
    order_id = Column(String(100))
    order_type = Column(String(100))
    lrn_number_vendor = Column(SmallInteger)
    lnp_dipping_cost = Column(Numeric(10, 8))
    is_final_call = Column(SmallInteger)
    egress_code_asr = Column(Float)
    egress_code_acd = Column(Float)
    route_prefix = Column(String(100))
    is_manual_kill = Column(Boolean, default=False)
    orig_call_duration = Column(Integer)
    orig_delay_second = Column(Integer)
    term_delay_second = Column(Integer)
    trunk_type = Column(SmallInteger)
    origination_profile_port = Column(Integer)
    termination_profile_port = Column(Integer)
    o_trunk_type2 = Column(SmallInteger)
    o_billing_method = Column(SmallInteger)
    t_trunk_type2 = Column(SmallInteger)
    t_billing_method = Column(SmallInteger)
    campaign_id = Column(Integer)
    tax = Column(Integer)
    agent_id = Column(Integer)
    agent_rate = Column(Float(53))
    agent_cost = Column(Numeric(12, 6))
    orig_jur_type = Column(Integer)
    term_jur_type = Column(Integer)
    ring_epoch = Column(BigInteger)
    end_epoch = Column(BigInteger)
    par_id = Column(Integer)
    paid_user = Column(String(100))
    rpid_user = Column(String(100))
    timeout_type = Column(SmallInteger)
    q850_cause = Column(Integer)
    q850_cause_string = Column(String(100))
    src_lata = Column(Integer)
    src_ocn = Column(String(100))
    src_rc = Column(String(100))
    dst_lata = Column(Integer)
    dst_ocn = Column(String(100))
    dst_rc = Column(String(100))
    lrn_lata = Column(Integer)
    lrn_ocn = Column(String(100))
    lrn_rc = Column(String(100))
    """
    server = synonym('egress_id')
    ingress_trunk = synonym('ingress_id')
    ingress_host = synonym('origination_destination_host_name')
    egress_host = synonym('termination_source_host_name')
    ani = synonym('translation_ani')
    dnis = synonym('lrn_dnis')
    time = synonym('call_duration')
    """


def generate_request_id():
    def gen():
        import random
        letters = [chr(i) for i in range(97, 123)] + [chr(i) for i in range(65, 91)]
        digits = [chr(i) for i in range(48, 58)]
        ret = ''
        for c in range(0, 16):
            ret = ret + random.choice(letters + digits)
        return ret.capitalize()

    return gen


class CdrAsyncTask(BaseModel):
    __tablename__ = 'cdr_async_task'
    TYPE = {0: 'manual', 1: 'auto', 2: 'download'}
    STATUS = {0: 'initial', 1: 'finished', 2: 'running', 3: 'success', 4: 'fail'}
    METHOD = {0: 'create', 1: 'upload', 2: 'email'}
    request_id = Column(String(16), primary_key=True, default=generate_request_id())
    request_client_id = Column(ForeignKey('client.client_id', ondelete='CASCADE'))
    user_id = Column(ForeignKey('users.user_id', ondelete='CASCADE'))
    fields = Column(String(1024))
    filter = Column(JSON)
    download_link = Column(String(512))
    expiration_time = Column(Integer)
    job_start_time = Column(Integer)
    job_end_time = Column(Integer)
    job_id_seq = Sequence('cdr_async_task_job_id_seq')
    # job_id = Column(Integer,server_default=text("nextval('cdr_async_task_job_id_seq'::regclass)"))
    job_id = Column(Integer, server_default=job_id_seq.next_value())
    msg = Column(String(512))
    progress = Column(String(4))
    count = Column(Integer)
    size = Column(Integer)
    status = Column(String(32))
    email = Column(String(512))
    orig_file = Column(String(512))
    mail_status = Column(String(512))
    method = Column(String(32))

    sender_id = Column(ForeignKey('mail_sender.id', ondelete='CASCADE'))
    subject = Column(String(512))
    content = Column(Text)
    mail_cc = Column(String(512))
    type = Column(ChoiceType(TYPE), server_default='1')

    ftp_url = Column(String(512))  # (s)ftp server url to upload (ftp=ftp://myftpserver.com)
    ftp_port = Column(SmallInteger)  # server port (default: 21 -> ftp; 22 -> sftp)
    ftp_user = Column(String(512))  # server user name (default: Anonymous)
    ftp_password = Column(String(512))  # server password (default: None)
    ftp_dir = Column(String(
        512))  # server path to upload: for ftp path from access point; for sftp -> path from root. (default: "/")
    ftp_file_name = Column(String(
        512))  # filename prefix (70 chars max). Engine will generate name of the file using the following template:
    # [prefix]_[yyyy-mm-dd__hh_mm_ss]_[yyyy-mm-dd__hh_mm_ss]_[breakdown].[ext]
    ftp_max_lines = Column(Integer)  # max lines per file
    ftp_compress = Column(String(512))  # compress file. Allowed formats: gz, tar.gz, tar.bz2
    ftp_include_headers = Column(
        SmallInteger)  # include headers line in csv file: 0 -> do not include; 1 -> include. Default: 1 -> with headers
    ftp_file_breakdown = Column(
        SmallInteger)  # breakdown results: 0","descripton":"as one file; 1 -> as hourly file; 2 -> as daily file. Default: 0 -> single file

    client = relationship(Client, uselist=False,  # back_populates='low_balance_config',
                          enable_typechecks=False)
    user = relationship(User, uselist=False)
    client_id = synonym('request_client_id')

    # direct = relationship('SendCdrDirect', uselist=False, single_parent=True, cascade="all, delete-orphan")

    # @property
    # def email(self):
    #    return self.email_to
    @hybrid_property
    def start_time(self):
        try:
            return cast(self.filter['start_time'], String)
        except:
            return cast('', String)

    @hybrid_property
    def end_time(self):
        try:
            return cast(self.filter['end_time'], String)
        except:
            return cast('', String)

    @property
    def _download_link(self):
        if self.orig_file:
            return '{}{}/files/{}'.format(settings.API_SCHEME, settings.API_HOST,
                                          self.orig_file)
        return None

    @property
    def file_name(self):
        if self.orig_file:
            return '{}/{}'.format(settings.FILES['upload_to'], self.orig_file)
        return None

    @property
    def file_name_xls(self):
        return '{}/cdr_{}.xls'.format(settings.FILES['upload_to'], self.request_id)

    @property
    def file_data(self):
        if self.orig_file:
            return open(self.file_name, 'rb').read()
        return None

    @property
    def fmt_start_time(self):
        return datetime.fromtimestamp(self.start_time, UTC)

    @property
    def fmt_end_time(self):
        if 'end_time' in self.filter:
            return datetime.fromtimestamp(self.end_time, UTC)
        return None

    @property
    def fmt_download_link(self):
        return '{}/files/cdr_{}.xls'.format(settings.FILE_URL,
                                            self.request_id)


# endregion

class CdrExportLog(BaseModel):
    __tablename__ = 'cdr_export_log'

    id = Column(Integer, primary_key=True, server_default=text("nextval('cdr_export_log_id_seq'::regclass)"))
    export_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    status = Column(Integer, server_default=text("0"))
    file_name = Column(String(500))
    user_id = Column(Integer)
    is_dipp = Column(Boolean)
    backend_pid = Column(Integer, server_default=text("0"))
    cdr_start_time = Column(DateTime(True))
    cdr_end_time = Column(DateTime(True))
    sql = Column(String)
    file_rows = Column(Integer)
    error_msg = Column(Text)
    send_mail = Column(String(200))
    where_sql = Column(String)
    show_fields_sql = Column(String)
    total_hours = Column(Integer, server_default=text("0"))
    finished_hours = Column(Integer, server_default=text("0"))
    stop_time = Column(DateTime(True))
    download_cdr_from = Column(String(200))
    download_cdr_subject = Column(Text)
    download_cdr_content = Column(Text)
    download_cdr_cc = Column(String(200))
    send_type = Column(Integer)
    total_date = Column(Integer, server_default=text("0"))
    finished_date = Column(Integer, server_default=text("0"))
    finished_time = Column(DateTime(True))
    job_id = Column(Integer)
    pid = Column(Integer)
    file_dir = Column(Text)
    file_path = Column(Text)
    total_days = Column(Integer)
    completed_days = Column(Integer)


class CdrExportTask(BaseModel):
    __tablename__ = 'cdr_export_task'
    STATUS = {0: 'Initial', 1: 'In Process', 2: 'Successful', 3: 'Failed'}
    id = Column(Integer, primary_key=True)
    operator_user = Column(String(40))
    export_cdr_file = Column(String(1024))
    cdr_counts = Column(Integer, server_default='0')
    cdr_start_time = Column(DateTime(True))
    cdr_end_time = Column(DateTime(True))
    csv_file_headers = Column(String(2048))
    cdr_headers = Column(String(2048))
    cdr_filter = Column(String(2048))
    status = Column(ChoiceType(STATUS), server_default='0')
    progress = Column(String(1024))
    start_time = Column(DateTime(True))
    end_time = Column(DateTime(True))
    create_time = Column(DateTime(True), server_default=func.now())
    email_status = column_property(
        select([CdrExportLog.send_mail]).where(CdrExportLog.job_id == id).correlate_except(
            CdrExportLog))
    email_finished_on = column_property(
        select([CdrExportLog.finished_time]).where(CdrExportLog.job_id == id).correlate_except(
            CdrExportLog))

    # cdr_count = synonym(cdr_counts)
    # begin_time = synonym(cdr_start_time)
    @property
    def cdr_count(self):
        return self.cdr_counts

    @property
    def begin_time(self):
        return self.cdr_start_time

    @property
    def fmt_end_time(self):
        return str(self.cdr_end_time)

    @property
    def file_name(self):
        return self.export_cdr_file

    @property
    def file_name_xls(self):
        return '{}/cdr_{}.xls'.format(settings.FILES['upload_to'], self.id)

    # @property
    # def cdr_link(self):
    #     import base64
    #     base_url = self.session().query(SystemParameter.base_url).first().base_url
    #     b64id = base64.encodebytes(str(self.id).encode()).decode().replace('\n', '')
    #     return '{}/homes/cdr_export_download/{}'.format(base_url, b64id)
    @property
    def cdr_link(self):
        import jwt
        conf = self.session().query(SystemParameter.base_url, SystemParameter.cdr_token).first()
        base_url = conf.base_url
        exp = (datetime.now(UTC) + timedelta(days=1)).timestamp()
        token_data = {'id': self.id, 'exp': exp}
        token = jwt.encode(token_data, conf.cdr_token, algorithm='HS256')
        if type(token) != type(''):
            token = token.decode()
        return '{}/homes/cdr_export_download?token={}'.format(base_url, token)

    @property
    def download_link(self):
        return self.cdr_link


# +++ RateSendLog
class RateSendLog(BaseModel):
    __tablename__ = 'rate_send_log'
    STATUS = {
        0: 'waiting',
        2: 'generated',
        3: 'sent',
        4: 'sent failed',
        5: 'sent success'
    }
    FORMAT = {0: 'csv', 1: 'json', 2: 'xls'}
    SEND_TYPE = {0: 'Send to carriers using this rate table', 1: 'Specify my own recipients',
                 2: 'Send to specified trunks'}
    DOWNLOAD_METHOD = {0: 'link', 1: 'attachment'}
    ZIP = {0: 'false', 1: 'true'}
    SENT_AREA = {0: 'product', 1: 'rate'}
    DATE_FORMAT = {0: 'Date Only', 1: 'Date and Time'}

    rate_table_id = Column(Integer, nullable=False)
    status = Column(ChoiceType(STATUS), nullable=False)
    zip = Column(SmallInteger)  # Column(ChoiceType(ZIP))
    file = Column(String(500))
    error = Column(String(500))
    format = Column(ChoiceType(FORMAT), nullable=False)
    id = Column(Integer, primary_key=True, server_default=text("nextval('rate_send_log_id_seq'::regclass)"))
    email_template_id = Column(Integer)
    create_time = Column(DateTime(True))
    effective_date = Column(String(50))
    start_effective_date = Column(String(50))
    download_deadline = Column(Date)
    is_email_alert = Column(Boolean, nullable=False, server_default=text("false"))
    is_disable = Column(Boolean, nullable=False, server_default=text("false"))
    is_temp = Column(Boolean, server_default=text("false"))
    headers = Column(String(500))
    send_type = Column(ChoiceType(SEND_TYPE), default=0, server_default=text("0"))
    send_specify_email = Column(String)
    resource_ids = Column(String(500))
    download_method = Column(ChoiceType(DOWNLOAD_METHOD), default=1, server_default=text("1"))
    total_records = Column(Integer)
    completed_records = Column(Integer)

    sent_area = Column(ChoiceType(SENT_AREA), default=1, server_default=text("1"))
    date_format = Column(ChoiceType(DATE_FORMAT), default=0, server_default=text("0"))

    # indicate_inc_dec = Column(Boolean, server_default=text("false"))

    rate_table = relationship(RateTable, primaryjoin='foreign(RateSendLog.rate_table_id)==RateTable.rate_table_id')
    # email_template = relationship(SendRateTemplate,
    #                               primaryjoin='foreign(RateSendLog.email_template_id)==SendRateTemplate.id',
    #                               uselist=False)
    # email_direct = relationship(SendRateDirect, uselist=False, single_parent=True, cascade="all, delete-orphan")
    details = relationship('RateSendLogDetail', back_populates='parent_job', uselist=True,
                           single_parent=True, cascade="all, delete-orphan")

    job_id = synonym('id')

    process = column_property(case([(total_records > 0, (1.0 * completed_records) / (1.0 * total_records))], else_=0.0))

    rate_table_name = column_property(select([RateTable.name]). \
                                      where(RateTable.rate_table_id == rate_table_id).correlate_except(RateTable))

    @property
    def download_token(self):
        try:
            return self.details[0].download_token
        except:
            return None

    @hybrid_property
    def trunks(self):
        try:
            return self.resource_ids.split(',')
        except:
            return []

    @trunks.setter
    def trunks(self, value):
        try:
            self.resource_ids = ','.join([str(x) for x in value])
        except:
            pass

    @hybrid_property
    def err(self):
        return self.error

    @err.setter
    def err(self, value):
        try:
            if len(value) < 500:
                self.error = value
            else:
                self.error = value[0:500]
        except:
            self.error = value


class RateSendLogDetail(BaseModel):
    __tablename__ = 'rate_send_log_detail'
    STATUS = {
        None: 'Waiting',
        1: 'In Progress',
        2: 'completed',
        3: 'failed',
        4: 'Downloaded',
        5: 'Not Yet Downloaded'
    }
    id = Column(Integer, primary_key=True, server_default=text("nextval('rate_send_log_detail_id_seq'::regclass)"))
    log_id = Column(ForeignKey('rate_send_log.id', ondelete='CASCADE'))
    resource_id = Column(Integer)
    send_to = Column(String(100))
    status = Column(ChoiceType(STATUS))
    error = Column(Text)
    salt = Column(String)
    download_date = Column(Date)
    sent_on = Column(DateTime(True))

    parent_job = relationship('RateSendLog', back_populates='details', uselist=False)
    resource = relationship('Resource', primaryjoin='foreign(RateSendLogDetail.resource_id)==Resource.resource_id')
    download_log = relationship('RateDownloadLog',
                                primaryjoin='foreign(RateSendLogDetail.id) == RateDownloadLog.log_detail_id',
                                back_populates='send_log', uselist=True)

    trunk_id = synonym('resource_id')
    job_id = synonym('log_id')

    result = column_property(
        case([(download_date > datetime.now(UTC), 'expired'), (status == 'failed', 'canceled')], else_='normal'))

    process = column_property(select([RateSendLog.process]). \
                              where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    create_time = column_property(select([RateSendLog.create_time]). \
                                  where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    download_deadline = column_property(select([RateSendLog.download_deadline]). \
                                        where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    effective_date = column_property(select([RateSendLog.effective_date]). \
                                     where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    # file = column_property(select([RateSendLog.file]). \
    #                        where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    send_type = column_property(select([RateSendLog.send_type]). \
                                where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    sent_area = column_property(select([RateSendLog.sent_area]). \
                                where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    rate_table_name = column_property(select([RateSendLog.rate_table_name]). \
                                      where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    rate_table_id = column_property(select([RateSendLog.rate_table_id]). \
                                    where(RateSendLog.id == log_id).correlate_except(RateSendLog))

    trunk_name = column_property(select([Resource.alias]). \
                                 where(Resource.resource_id == resource_id).correlate_except(Resource))
    client_name = column_property(select([Client.name]). \
                                  where(
        and_(Resource.resource_id == resource_id, Client.client_id == Resource.client_id)).correlate_except(Resource,
                                                                                                            Client))

    client_id = column_property(select([Client.client_id]). \
                                where(
        and_(Resource.resource_id == resource_id, Client.client_id == Resource.client_id)).correlate_except(Resource,
                                                                                                            Client))

    @property
    def file(self):
        return self.parent_job.file.split('/')[-1]

    @hybrid_property
    def download_token(self):
        import jwt
        exp = datetime.now(UTC) + timedelta(days=settings.JWT_TTL_DAYS)
        token_data = {'id': self.id, 'exp': exp}
        token = jwt.encode(token_data, settings.JWT_SIGNATURE).decode('utf-8')
        return token

    @hybrid_property
    def download_url(self):
        import base64
        flg = base64.encodebytes(str(self.job_id).encode()).decode().replace('\n', '')
        base_url = self.session().query(SystemParameter.base_url).first().base_url
        # return '{}/rates/download_send_rate_file?flg={}'.format(base_url, flg)
        return '{}/download_rate/index?salt={}'.format(base_url, self.salt)

    @property
    def fmt_rate_update_file_name(self):
        return self.file


RateSendLog.email_count = column_property(
    select([func.sum(func.array_length(func.regexp_split_to_array(RateSendLogDetail.send_to, ';'), 1))]). \
        where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(RateSendLogDetail))

RateSendLog.recipient_count = column_property(select([func.count(RateSendLogDetail.id)]). \
    where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(
    RateSendLogDetail))

# RateSendLogDetail.email_count = column_property(select([RateSendLog.email_count]). \
#     where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(
#     RateSendLog))
# RateSendLogDetail.recipient_count = column_property(select([RateSendLog.recipient_count]). \
#     where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(
#     RateSendLog))
RateSendLogDetail.email_count = column_property(
    func.array_length(func.regexp_split_to_array(RateSendLogDetail.send_to, ';'), 1))

RateSendLogDetail.recipient_count = column_property(select([func.count(RateSendLog.id)]). \
    where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(
    RateSendLog))

RateSendLogDetail.process = column_property(select([RateSendLog.process]). \
    where(RateSendLog.id == RateSendLogDetail.log_id).correlate_except(
    RateSendLog))


class RateDownloadLog(BaseModel):
    __tablename__ = 'rate_download_log'

    id = Column(Integer, primary_key=True, server_default=text("nextval('rate_download_log_id_seq'::regclass)"))
    resource_id = Column(Integer)
    file_path = Column(String)
    download_time = Column(DateTime(True))
    download_ip = Column(String)
    log_detail_id = Column(Integer)
    download_username = Column(String(100))

    resource = relationship('Resource', primaryjoin='foreign(RateDownloadLog.resource_id) == Resource.resource_id',
                            uselist=False)
    send_log = relationship('RateSendLogDetail',
                            primaryjoin='foreign(RateSendLogDetail.id) == RateDownloadLog.log_detail_id',
                            back_populates='download_log', uselist=False)


class FrundDetection(BaseModel):
    __tablename__ = 'frund_detection'
    SEND_EMAIL_TYPE = {0: 'None', 1: 'Your Own NOC', 2: 'Partner NOC', 3: 'Both NOC'}
    id = Column(Integer, primary_key=True)  # , server_default=text("nextval('frund_detection_id_seq'::regclass)"))
    rule_name = Column(String(100), index=True)
    trunk_id_list = Column(String(100))
    criteria_1hour_min = Column(Numeric(15, 6))
    criteria_1hour_revenue = Column(Numeric(15, 6))
    criteria_24hour_min = Column(Numeric(15, 6))
    criteria_24hour_revenue = Column(Numeric(15, 6))
    is_block = Column(Boolean)
    send_email_type = Column(ChoiceType(SEND_EMAIL_TYPE))
    email_from = Column(Integer)
    email_ticket_subject = Column(String(200))
    email_ticket_content = Column(Text)
    trigger_criteria_1hour_time = Column(DateTime(True))
    trigger_criteria_24hour_time = Column(DateTime(True))
    create_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))

    # hourly_asr_of_trunk_is_below = Column(Integer, nullable=True)
    # hourly_acd_of_trunk_is_below = Column(Integer, nullable=True)
    # hourly_minute_of_trunk_is_above = Column(Integer, nullable=True)
    # hourly_cost_of_trunk_is_above = Column(Integer, nullable=True)
    # daily_asr_of_trunk_is_below = Column(Integer, nullable=True)
    # daily_acd_of_trunk_is_below = Column(Integer, nullable=True)
    # daily_minute_of_trunk_is_above = Column(Integer, nullable=True)
    # daily_cost_of_trunk_is_above = Column(Integer, nullable=True)

    @property
    def subject(self):
        return self.email_ticket_subject
        # return 'Fraud detected'

    @property
    def html_content(self):
        return self.email_ticket_content

    @property
    def cc_mail(self):
        return ''

    @property
    def from_mail_id(self):
        return self.email_from

    @property
    def to_mail(self):
        return ''

    def get_attachment(self, env):
        return []


class FrundDetectionHistory(BaseModel):
    __tablename__ = 'frund_detection_history'
    BLOCK_TYPE = {0: '1h minute', 1: '1h revenue', 2: '24h minute', 3: '24 revenue'}
    id = Column(Integer,
                primary_key=True)  # , server_default=text("nextval('frund_detection_history_id_seq'::regclass)"))
    frund_detection_id = Column(Integer, index=True)
    ingress_id = Column(Integer)
    block_type = Column(ChoiceType(BLOCK_TYPE))
    criteria_1hour_min = Column(Numeric(15, 6))
    criteria_1hour_revenue = Column(Numeric(15, 6))
    criteria_24hour_min = Column(Numeric(15, 6))
    criteria_24hour_revenue = Column(Numeric(15, 6))
    actual_1h_minute = Column(Numeric(15, 6))
    actual_24h_minute = Column(Numeric(15, 6))
    actual_1h_revenue = Column(Numeric(15, 6))
    actual_24h_revenue = Column(Numeric(15, 6))
    is_block = Column(Boolean, server_default=text("false"))
    send_email_type = Column(Integer)
    is_send_email = Column(Boolean, server_default=text("false"))
    partner_email_msg = Column(Text)
    partner_email_status = Column(Boolean)
    partner_email = Column(String(200))
    system_email_msg = Column(Text)
    system_email_status = Column(Boolean)
    system_email = Column(String(200))
    create_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    api_email_send_time = Column(DateTime(True), index=True)
    api_email_result = Column(String(1024))

    trunk = relationship('Resource', primaryjoin='foreign(FrundDetectionHistory.ingress_id) == Resource.resource_id')
    rule = relationship(FrundDetection, primaryjoin=foreign(frund_detection_id) == FrundDetection.id)
    rule_name = column_property(
        select([FrundDetection.rule_name]).where(FrundDetection.id == frund_detection_id).correlate_except(
            FrundDetection))

    @property
    def block_value(self):
        if self.block_type == '1h minute':
            return self.actual_1h_minute
        elif self.block_type == '1h revenue':
            return self.actual_1h_revenue
        elif self.block_type == '24h minute':
            return self.actual_24h_minute
        elif self.block_type == '24 revenue':
            return self.actual_24h_revenue
        else:
            return None

    @property
    def limit_value(self):
        if self.block_type == '1h minute':
            return self.criteria_1hour_min
        elif self.block_type == '1h revenue':
            return self.criteria_1hour_revenue
        elif self.block_type == '24h minute':
            return self.criteria_24hour_min
        elif self.block_type == '24 revenue':
            return self.criteria_24hour_revenue
        else:
            return None


class MonitoredRule(BaseModel):
    __tablename__ = 'monitored_rule'
    """COMMENT ON COLUMN monitored_rule.trunk_id_list is 'Monitored trunk id list, separated by comma';
COMMENT ON COLUMN monitored_rule.include_codes is 'Monitored code list,separated by comma. if empty, monitor all code';
COMMENT ON COLUMN monitored_rule.exclude_codes is 'Unmonitored code list,separated by comma';
COMMENT ON COLUMN monitored_rule.monitored_type is '0: trunk; 1: trunk and dnis; 2: trunk and ani; 3: dnis; 4: ani; 5: trunk and destination; 6: trunk and country; 7: trunk and code';
COMMENT ON COLUMN monitored_rule.acd_action is 'xxx_ation   0: ignore; 1: >; 2: <; 3: =';
COMMENT ON COLUMN monitored_rule.sdp_duation is '0: none; 1: 6sec; 2: 12sec; 3: 18sec; 4: 24sec; 5: 30sec';
COMMENT ON COLUMN monitored_rule.exe_schedule_type is '0: never; 1: every specific minutes; 2: daily; 3: weekly';
COMMENT ON COLUMN monitored_rule.run_day_of_week is '0: Sunday; 1: Monday; 2: Tuesday; 3: Wednesday; 4: Thursday; 5: Friday; 6: Saturday';
COMMENT ON COLUMN monitored_rule.send_email_type is '0 - None; 1 - Your Own NOC; 2 - Partner NOC; 3 - Both NOC';
"""
    MONITORED_TYPE = {0: 'trunk', 1: 'trunk and dnis', 2: 'trunk and ani', 3: 'dnis', 4: 'ani',
                      5: 'trunk and destination', 6: 'trunk and country', 7: 'trunk and code'}
    ACTION = {0: 'ignore', 1: '>', 2: '<', 3: '='}
    SDP_DUATION = {0: 'none', 1: '6sec', 2: '12sec', 3: '18sec', 4: '24sec', 5: '30sec'}
    EXE_SCHEDULE_TYPE = {0: 'never', 1: 'every specific minuses', 2: 'daily', 3: 'weekly'}
    RUN_DAY_OF_WEEK = {0: 'Sunday', 1: 'Monday', 2: 'Tuesday', 3: 'Wednesday', 4: 'Thursday', 5: 'Friday',
                       6: 'Saturday'}
    SEND_EMAIL_TYPE = {0: 'None', 1: 'Your Own NOC', 2: 'Partner NOC', 3: 'Both NOC'}

    id = Column(Integer, primary_key=True)  # , server_default=text("nextval('monitored_rule_id_seq'::regclass)"))
    rule_name = Column(String(100))
    is_active = Column(Boolean, nullable=False, server_default=text("true"))
    is_ingress_trunk = Column(Boolean, server_default=text("true"))
    is_all_trunk = Column(Boolean, nullable=False, server_default=text("false"))
    trunk_id_list = Column(String(200))
    include_codes = Column(String(500))
    exclude_codes = Column(String(500))
    monitored_type = Column(ChoiceType(MONITORED_TYPE), server_default=text("0"))
    acd_action = Column(ChoiceType(ACTION))
    acd_value = Column(Integer)
    asr_action = Column(ChoiceType(ACTION))
    asr_value = Column(Numeric(15, 6))
    sdp_action = Column(ChoiceType(ACTION))
    sdp_value = Column(Numeric(15, 6))
    sdp_duation = Column(ChoiceType(SDP_DUATION))
    pdd_action = Column(ChoiceType(ACTION))
    pdd_value = Column(Integer)
    profit_action = Column(ChoiceType(ACTION))
    profit_value = Column(Numeric(15, 6))
    revenue_action = Column(ChoiceType(ACTION))
    revenue_value = Column(Numeric(15, 6))
    min_call_attempt = Column(Integer)
    exe_schedule_type = Column(ChoiceType(EXE_SCHEDULE_TYPE))
    specific_minutes = Column(Integer)
    daily_run_time = Column(Time)
    weekly_run_time = Column(Time)
    run_day_of_week = Column(ChoiceType(RUN_DAY_OF_WEEK))
    sample_size = Column(Integer, nullable=False, server_default=text("5"))
    is_block = Column(Boolean, server_default=text("false"))
    is_auto_unblock = Column(Boolean, server_default=text("false"))
    unblock_after_min = Column(Integer)
    last_run_time = Column(DateTime(True))
    next_run_time = Column(DateTime(True))
    send_email_type = Column(ChoiceType(SEND_EMAIL_TYPE), server_default=text("3"))
    email_from = Column(Integer)
    email_ticket_subject = Column(String(200))
    email_ticket_content = Column(Text)
    create_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    updated_at = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    updated_by = Column(String(200))

    @property
    def from_mail_id(self):
        return self.email_from

    @property
    def subject(self):
        return self.email_ticket_subject

    @property
    def html_content(self):
        return self.email_ticket_content

    @property
    def cc_mail(self):
        return ''

    @property
    def to_mail(self):
        return 'email'

    def get_attachment(self, env):
        return []


class MonitoredRuleHistory(BaseModel):
    __tablename__ = 'monitored_rule_history'

    id = Column(Integer,
                primary_key=True)  # , server_default=text("nextval('monitored_rule_history_id_seq'::regclass)"))
    monitored_rule_id = Column(Integer, index=True)
    start_time = Column(DateTime(True))
    end_time = Column(DateTime(True))
    is_ingress_trunk = Column(Boolean)
    trunk_id = Column(Integer)
    dnis = Column(String(64))
    ani = Column(String(64))
    code_name = Column(String(100))
    country = Column(String(100))
    code = Column(String(64))
    total_call = Column(Integer)
    not_zero_call = Column(Integer)
    sdp_call = Column(Integer)
    call_duration = Column(Integer)
    pdd = Column(Integer)
    ingress_cost = Column(Numeric(30, 6))
    egress_cost = Column(Numeric(30, 6))
    block_id_list = Column(Text)
    unblock_time = Column(DateTime(True), index=True)
    unblock_exe_time = Column(DateTime(True), index=True)
    create_time = Column(DateTime(True), server_default=text("('now'::text)::timestamp(0) with time zone"))
    api_email_send_time = Column(DateTime(True), index=True)
    api_email_result = Column(String(1024))
    monitored_rule = relationship(MonitoredRule, primaryjoin=foreign(monitored_rule_id) == MonitoredRule.id)

    rule_name = column_property(
        select([MonitoredRule.rule_name]).where(MonitoredRule.id == monitored_rule_id).correlate_except(MonitoredRule))
    block_list = synonym('block_id_list')
    sample_start_time = synonym('start_time')
    sample_end_time = synonym('end_time')
    total_calls = synonym('total_call')

    @property
    def limit_table(self):
        rule = self.monitored_rule
        return """
    <table class="limit_table" name="limit_table">
    <tr><th></th><th>acd</th><th>asr</th><th>sdp</th><th>pdd</th><th>profitability</th><th>revenue</th></tr>
    <tr><td>limit value</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>
    <tr><td>operator</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>
    </table>
    """.format(rule.acd_value, rule.asr_value, rule.sdp_value, rule.pdd_value,
               rule.profit_value, rule.revenue_value,
               rule.acd_action, rule.asr_action, rule.sdp_action, rule.pdd_action, rule.profit_action,
               rule.revenue_action)

    @property
    def rule_setup(self):
        rule = self.monitored_rule
        all = 'all' if rule.is_all_trunk else 'selected'
        return """
            <table class="rule_setup" name="rule_setup">
            <tr><th>Name</th><th>Value</th></tr>
            <tr><td>scope</td><td>{} {}</td></tr>
            <tr><td>include</td><td>{}</td></tr>
            <tr><td>exclude</td><td>{}</td></tr>
            <tr><td>monitoring on</td><td>{}</td></tr>
            <tr><td>trunks</td><td>{}</td></tr>
            <tr><td>execution schedule</td><td>{}</td></tr>
            <tr><td>do block</td><td>{}</td></tr>
            </table>
            """.format(all, 'ingress' if rule.is_ingress_trunk else 'egress', rule.include_codes, rule.exclude_codes,
                       rule.monitored_type,
                       rule.trunk_id_list, rule.exe_schedule_type, rule.is_block)

    @property
    def acd(self):
        if self.not_zero_call:
            return self.call_duration / self.not_zero_call
        else:
            return None

    @property
    def asr(self):
        if self.total_call:
            return self.not_zero_call / self.total_call * 100
        else:
            return None

    def monitoring_result(self):
        head = ''
        row = ''
        for key in (
                'total_call', 'not_zero_call', 'sdp_call', 'call_duration', 'pdd', 'ingress_cost', 'egress_cost', 'acd',
                'asr'):
            head = head + '<th>{}</th>'.format(key)
            # if key in self.type_name:
            #     row = row + '<td style="background-color: rgba(234, 174, 165, 1)">{}</td>'.format(getattr(self, key))
            # else:
            row = row + '<td>{}</td>'.format(round(getattr(self, key, 0.0), 2))
        return '<table class="monitoring_result" name="monitoring_result"><tr>{}</tr><tr>{}</tr></table>'.format(head,
                                                                                                                 row)
